// %A%
//----------------------------------------------------------------------
// Source: vcline.cpp
//
// Copyright (c) 1999-2002 Brooktrout Technology, Needham, MA
// All rights reserved
//
// Edit History:
//
// 09-Mar-00  FJN  Created
// 05-Aug-02  FJN  Implement command line -a# and -b#
// 04-Sep-02  FJN  Implement DTMF "3#" to enter conference listen only
// 04-Sep-02  FJN  Invoke BfvConferenceEnd when final party exits conference
// 04-Sep-02  FJN  Terminate all relevant node connections when final party exits conference
// 04-Sep-02  FJN  Create node connections for only one conference
// 17-Jun-03  FJN  Use BfvLineOriginateCall to "busy out"
// 18-Jun-03  FJN  Do not report errors from BfvLineOriginateCall when "busy out"
//----------------------------------------------------------------------

#ifdef WIN32
#include "stdafx.h"
#include "vcdemo.h"
#include "MainFrm.h"
#include "vcdemoDoc.h"
#include "vcdemoView.h"
#include <io.h>
#else
#define O_BINARY 0
#define	UNIT_MAX	128
#endif

#include <fcntl.h>
#include <errno.h>

#ifdef WIN32
extern "C"
{
#endif
#include "btlib.h"
#include "acclib.h"
#include "pthread.h"
#include "semaphore.h"
#ifdef WIN32
void sleep(int);
}
#endif

#define	SEM_SHARE_LOCAL		0
#define	SEM_SHARE_GLOBAL	1

#define	READ_BUF_SIZE		1024
#define	UNIT_BASE			2
#define	SLOTS_PER_T1		24
#define	NODE_MAX			16
#define	UNIT_MAX	        128

typedef enum
{
	CallStateAwaitingIncomingCall,
	CallStateAnsweringIncomingCall,
	CallStateMainMenu,
	CallStateGetConfID,
	CallStateInConf
} CallStateEnum;

typedef struct
{
	BTLINE *lp;
	unsigned role;
	unsigned span;
	unsigned timeSlot;
	unsigned confID;
	unsigned callRef;
	int keySave;
	CallStateEnum callState;
} UnitStruct;

typedef struct
{
	unsigned confID;
	BOOL fBusy;
} NodeStruct;

void XWndMsgBox(char *);
void *XWndProc(void *);
void XWndStatusUpdate(unsigned, char *);
void CallInit(void);

static sem_t semDebug;
static UnitStruct UnitData[UNIT_MAX];
static NodeStruct NodeData[NODE_MAX];

unsigned unitNum = 30;
unsigned chanPerDsp = 5;
unsigned unitBaseA = 0;
unsigned unitBaseB = 0;
unsigned verbose = 0;
BOOL fNoCall = FALSE;

//-----------------------------------------------------------------------
// Procedure ConferenceErrorName
//-----------------------------------------------------------------------
char *
ConferenceErrorName(
	int codeError)
{
	switch (codeError)
	{
	case 0 :
		return("");
	case CONF_ERROR_BAD_PARAMETERS :
		return("bad parameters");
	case CONF_ERROR_BAD_ADMIN_INIT :
		return("bad admin init");
	case CONF_ERROR_ADMIN_EXISTS :
		return("admin exists");
	case CONF_ERROR_ADMIN_MISSING :
		return("admin missing");
	case CONF_ERROR_ID_MISSING_RESERVED :
		return("ID missing reserved");
	case CONF_ERROR_ID_MISSING_NOT_RESERVED :
		return("ID missing not reserved");
	case CONF_ERROR_INSUFFICIENT_RESOURCES :
		return("insufficient resources");
	case CONF_ERROR_ID_IN_USE :
		return("ID in use");
	case CONF_ERROR_RESOURCE_HANDLER_ERROR :
		return("resource handler error");
	case CONF_ERROR_INVALID_CONFERENCE :
		return("invalid conference");
	case CONF_ERROR_CALLREF_IN_USE :
		return("call reference in use");
	case CONF_ERROR_RELEASING_RESOURCES :
		return("releasing resources");
	case CONF_ERROR_INVALID_CALLREF :
		return("invalid call reference");
	case CONF_ERROR_SYSTEM :
		return("system");
	case CONF_ERROR_TALKER_LIMIT :
		return("talker limit");
	case CONF_ERROR_ID_NOT_BEGUN :
		return("ID not begun");
	default :
		return("unknown");
	}
}

//-----------------------------------------------------------------------
// Procedure DebugFunc
//-----------------------------------------------------------------------

void DebugFunc(char *p)
{
	FILE *fp;
	sem_wait(&semDebug);
	if (fp = fopen("vcdemo.log", "a+"))
	{
		fseek(fp, 0, SEEK_END);
		fprintf(fp, "%s\n", p);
		fclose(fp);
	}
	else
		XWndMsgBox("DebugFunc fopen vcdemo.log failure");
	sem_post(&semDebug);
}

//-----------------------------------------------------------------------
// Procedure TerminationCheck
//-----------------------------------------------------------------------
int
TerminationCheck(
	BTLINE *lp,
	char *termParms)
{
	int key;
	struct args_tone toneArgs;
	unsigned unit = (unsigned) termParms;
	char szText[256];
	BTERR btErr;

	BT_ZERO(toneArgs);
	key = BfvToneGet(lp, &toneArgs);
	if (key >= 0)
	{
		UnitData[unit].keySave = key;
		return(1);
	}
	else
	{
		if (toneArgs.res.status == BT_STATUS_TIMEOUT)
			return(0);

		BfvErrorMessage(lp, &toneArgs.res, &btErr);
		sprintf(szText,
			"TerminationCheck%u BfvToneGet failure %s",
			unit,
			btErr.long_msg);
		XWndMsgBox(szText);

		UnitData[unit].keySave = 0;
		return(1);
	}
}

//-----------------------------------------------------------------------
// Procedure VoicePlay
//-----------------------------------------------------------------------

void VoicePlay(
	BTLINE *lp,
	unsigned unit,
	char *szFName)
{
	int fd;
	int bytesRead;
	char szText[256];
	unsigned char readBuf[READ_BUF_SIZE];
	struct args_speech speechArgs;
	BTERR btErr;

	if ((fd = open(szFName, O_RDONLY | O_BINARY)) < 0)
	{
		sprintf(szText,
			"VoicePlay%u open %s failure %d",
			unit,
			szFName,
			errno);
		XWndMsgBox(szText);
		return;
	}

	while ((bytesRead = read(fd, readBuf, READ_BUF_SIZE)) > 0)
	{
		BT_ZERO(speechArgs);
		speechArgs.buf = readBuf;
		speechArgs.size = bytesRead;
		speechArgs.coding_fmt = CODE_ADPCM;
		speechArgs.rate = RATE_8000;
		speechArgs.bits_per_samp = BITS_4;
		speechArgs.func = TerminationCheck;
		speechArgs.arg = (char *) unit;
		BfvSpeechPlayData(lp, &speechArgs);

		if (speechArgs.res.status != BT_STATUS_OK)
		{
			if (speechArgs.res.status != BT_STATUS_USER_TERMINATED)
			{
				BfvErrorMessage(lp, &speechArgs.res, &btErr);
				sprintf(szText,
					"VoicePlay%u BfvSpeechPlayData failure %s",
					unit,
					btErr.long_msg);
				XWndMsgBox(szText);
			}
			close(fd);
			return;
		}
	}

	BT_ZERO(speechArgs);
	speechArgs.buf = readBuf;
	speechArgs.size = 0;
	speechArgs.coding_fmt = CODE_ADPCM;
	speechArgs.rate = RATE_STOP;
	speechArgs.bits_per_samp = BITS_4;
	speechArgs.func = TerminationCheck;
	speechArgs.arg = (char *) unit;
	BfvSpeechPlayData(lp, &speechArgs);

	if (speechArgs.res.status != BT_STATUS_OK &&
		speechArgs.res.status != BT_STATUS_USER_TERMINATED)
	{
		BfvErrorMessage(lp, &speechArgs.res, &btErr);
		sprintf(szText,
			"VoicePlay%u BfvSpeechPlayData failure %s",
			unit,
			btErr.long_msg);
		XWndMsgBox(szText);
	}
	close(fd);
}

//-----------------------------------------------------------------------
// Procedure IncomingChannelThread
//-----------------------------------------------------------------------
void *
IncomingChannelThread(
	void *arg)
{
	int iKey;
	int iStatus;
	int fNoLoopCurrent;
	unsigned unit = (unsigned) arg;
	unsigned unitNode;
	unsigned dsp;
	char szText[256];
	char szKeys[128];
	char *pKeys;
	BTLINE *lp;
	BTERR btErr;
	struct args_line_admin lineArgs;
	struct args_telephone telArgs;
	struct args_tone toneArgs;
	struct args_conference conferenceArgs;
	struct args_conf_party partyArgs;
	struct args_tel_ctrl_call_sw switchArgs;
	struct args_dsp_role dspRoleArgs;
	CONF_PARTY_MODES confMode;
	UnitStruct* pUnitData;
	NodeStruct* pNodeData;

#ifdef DEBUG_VERBOSE
	sprintf(szText, "IncomingChannelThread%u starting", arg);
	XWndMsgBox(szText);
#endif

	//-----------------------------------------------------------------
	// attach channel
	//-----------------------------------------------------------------
	XWndStatusUpdate(unit, "attaching");
	BT_ZERO(lineArgs);
	lineArgs.unit = unit;
	if (!(lp = BfvLineAttach(&lineArgs)))
	{
		BfvErrorMessage(NULL, &lineArgs.res, &btErr);
		sprintf(szText,
			"IncomingChannelThread%u BfvLineAttach failure %s",
			unit,
			btErr.long_msg);
		XWndMsgBox(szText);
		return(NULL);
	}
	UnitData[unit].lp = lp;

	//-----------------------------------------------------------------
	// reset channel
	//-----------------------------------------------------------------
	XWndStatusUpdate(unit, "reseting");
	BT_ZERO(lineArgs);
	lineArgs.config_file_name = "btcall.cfg";
	if (BfvLineReset(lp, &lineArgs) < 0)
	{
		BfvErrorMessage(lp, &lineArgs.res, &btErr);
		sprintf(szText,
			"IncomingChannelThread%u BfvLineReset failure %s",
			unit,
			btErr.long_msg);
		XWndMsgBox(szText);
		return(NULL);
	}

	//-----------------------------------------------------------------
	// connect channel
	//-----------------------------------------------------------------
	// compute DSP number
	dsp = unit / chanPerDsp;

	// if DSP bridge channel (last channel on each DSP)
	if (UnitData[unit].role == SPEECH_CONF_PARTY_ROLE_NODE_DEF)
	{
		// if DSP bridge channel for group-A
		if (unit / chanPerDsp == unitBaseA / chanPerDsp)
		{
			XWndStatusUpdate(unit, "node group A");

			// make connection from this DSP (channel) to next DSP (channel)
			BT_ZERO(switchArgs);
			switchArgs.conn_mode = CALL_SW_DUPLEX_AND_SIGNALING_DEF;
			switchArgs.src_port_class = CALL_SW_PORT_CHANNEL_DEF;
			switchArgs.src_port_unit = UNIT_BASE + unit;
			switchArgs.src_stream = 0;
			switchArgs.src_slot = 0;
			switchArgs.dest_port_class = CALL_SW_PORT_CHANNEL_DEF;
			switchArgs.dest_port_unit = UNIT_BASE + unitBaseB + chanPerDsp - 1;
			switchArgs.dest_stream = 0;
			switchArgs.dest_slot = 0;

			sprintf(szText,
				"IncomingChannelThread%u BfvCallSWConnect node conn_mode %u src_port_class %u src_port_unit %u src_stream %u src_slot %u dest_port_class %u dest_port_unit %u dest_stream %u dest_slot %u",
				unit,
				switchArgs.conn_mode,
				switchArgs.src_port_class,
				switchArgs.src_port_unit,
				switchArgs.src_stream,
				switchArgs.src_slot,
				switchArgs.dest_port_class,
				switchArgs.dest_port_unit,
				switchArgs.dest_stream,
				switchArgs.dest_slot);
			DebugFunc(szText);

			if (BfvCallSWConnect(lp, &switchArgs) < 0)
			{
				BfvErrorMessage(lp, &switchArgs.res, &btErr);
				sprintf(szText,
					"IncomingChannelThread%u BfvCallSWConnect node failure %s",
					unit,
					btErr.long_msg);
				XWndMsgBox(szText);
				return(NULL);
			}
		}
		// if DSP bridge channel for group-B
		else if (unit / chanPerDsp == unitBaseB / chanPerDsp)
			XWndStatusUpdate(unit, "node group B");

		// wait forever
		while (1)
			sleep(60);
	}

	// if not DSP bridge channel
	else
	{
		// make connection from DSP (channel) to T1 time slot (network)
		BT_ZERO(switchArgs);
		switchArgs.conn_mode = CALL_SW_DUPLEX_AND_SIGNALING_DEF;
		switchArgs.src_port_class = CALL_SW_PORT_CHANNEL_DEF;
		switchArgs.src_port_unit = UNIT_BASE + unit;
		switchArgs.src_stream = 0;
		switchArgs.src_slot = 0;
		switchArgs.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
		switchArgs.dest_port_unit = UnitData[unit].span;
		switchArgs.dest_stream = 0;
		switchArgs.dest_slot = UnitData[unit].timeSlot;

		sprintf(szText,
			"IncomingChannelThread%u BfvCallSWConnect terminal conn_mode %u src_port_class %u src_port_unit %u src_stream %u src_slot %u dest_port_class %u dest_port_unit %u dest_stream %u dest_slot %u",
			unit,
			switchArgs.conn_mode,
			switchArgs.src_port_class,
			switchArgs.src_port_unit,
			switchArgs.src_stream,
			switchArgs.src_slot,
			switchArgs.dest_port_class,
			switchArgs.dest_port_unit,
			switchArgs.dest_stream,
			switchArgs.dest_slot);
		DebugFunc(szText);

		if (BfvCallSWConnect(lp, &switchArgs) < 0)
		{
			BfvErrorMessage(lp, &switchArgs.res, &btErr);
			sprintf(szText,
				"IncomingChannelThread%u BfvCallSWConnect terminal failure %s",
				unit,
				btErr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

	}

	//-----------------------------------------------------------------
	// busy out some ports
	//-----------------------------------------------------------------
	if (!(unit >= unitBaseA && unit < unitBaseA + chanPerDsp) &&
		!(unit >= unitBaseB && unit < unitBaseB + chanPerDsp))
	{
		XWndStatusUpdate(unit, "busy out");
		UnitData[unit].callState = CallStateAnsweringIncomingCall;
		BT_ZERO(telArgs);
		telArgs.phonenum = "";
		telArgs.call_protocol_code = CALL_PROTOCOL_VOICE;
		BfvLineOriginateCall(lp, &telArgs);
#if 0
		if (telArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &telArgs.res, &btErr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineOriginateCall failure %s",
				unit,
				btErr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}
#endif
		while (1)
			sleep(60);
	}

	//-----------------------------------------------------------------
	// forever loop
	//-----------------------------------------------------------------
	while (1)
	{
		//-----------------------------------------------------------------
		// wait for incoming call
		//-----------------------------------------------------------------
		UnitData[unit].callState = CallStateAwaitingIncomingCall;
		XWndStatusUpdate(unit, "waiting for incoming call");
		BT_ZERO(telArgs);
		BfvLineWaitForCall(lp, &telArgs);
		if (telArgs.res.line_status != WAIT_FOR_CALL_OK)
		{
			BfvErrorMessage(lp, &telArgs.res, &btErr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineWaitForCall failure %s",
				unit,
				btErr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		//-----------------------------------------------------------------
		// answer incoming call
		//-----------------------------------------------------------------
		UnitData[unit].callState = CallStateAnsweringIncomingCall;
		XWndStatusUpdate(unit, "answering incoming call");
		BT_ZERO(telArgs);
		BfvLineAnswer(lp, &telArgs);

		if (telArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &telArgs.res, &btErr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineAnswer failure %s",
				unit,
				btErr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		fNoLoopCurrent = FALSE;

		//-----------------------------------------------------------------
		// flush buffered DTMF
		//-----------------------------------------------------------------
		BT_ZERO(toneArgs);
		BfvToneFlush(lp, &toneArgs);

		if (toneArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &toneArgs.res, &btErr);
			sprintf(szText,
				"IncomingChannelThread%u BfvToneFlush failure %s",
				unit,
				btErr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		//-----------------------------------------------------------------
		// setup DTMF acquisition parameters
		//-----------------------------------------------------------------
		BT_ZERO(toneArgs);
		toneArgs.decode_flag = DTMF_16TONE;
		BfvToneDetectEnable(lp, &toneArgs);

		if (toneArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &toneArgs.res, &btErr);
			sprintf(szText,
				"IncomingChannelThread%u BfvToneDetectEnable failure %s",
				unit,
				btErr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		//-----------------------------------------------------------------
		// play welcome
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "playing welcome.vox");
		VoicePlay(lp, unit, "welcome.vox");

		if (UnitData[unit].keySave == 0)
		{
			XWndStatusUpdate(unit, "playing menu1.vox");
			VoicePlay(lp, unit, "menu1.vox");
		}

		//-----------------------------------------------------------------
		// set and display state
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "main menu; acquiring DTMF");
		UnitData[unit].callState = CallStateMainMenu;

		//-----------------------------------------------------------------
		// reset tone pointers
		//-----------------------------------------------------------------
		pKeys = szKeys;
		szKeys[0] = 0;

		//-----------------------------------------------------------------
		// in call main processing loop
		//-----------------------------------------------------------------
		while (1)
		{
			//-------------------------------------------------------------
			// acquire DTMF digit
			//-------------------------------------------------------------
			if (UnitData[unit].keySave > 0)
			{
				iKey = UnitData[unit].keySave;
				UnitData[unit].keySave = 0;
			}
			else
			{
				BT_ZERO(toneArgs);
				toneArgs.millisecs = 10 * 60 * 1000;
				iKey = BfvToneGet(lp, &toneArgs);
			}

			//-------------------------------------------------------------
			// if DTMF digit available
			//-------------------------------------------------------------
			if (iKey > 0)
			{
				sprintf(szText, "processing key %c", iKey);
				XWndStatusUpdate(unit, szText);

				if (iKey == '*')
					break;

				if (iKey == '#')
				{
					sprintf(szText, "processing tone command %s", szKeys);
					XWndStatusUpdate(unit, szText);

					switch (UnitData[unit].callState)
					{
					//-----------------------------------------------------
					// main menu command
					//-----------------------------------------------------
					case CallStateMainMenu :
						if (!strcmp(szKeys, "1"))
						{
							XWndStatusUpdate(unit, "playing menu2.vox");
							VoicePlay(lp, unit, "menu2.vox");
							XWndStatusUpdate(unit, "get confID; acquiring DTMF");
							UnitData[unit].callState = CallStateGetConfID;
							confMode = CONF_MODE_FULL_DUPLEX;
						}
						else if (!strcmp(szKeys, "3"))
						{
							XWndStatusUpdate(unit, "playing menu2.vox");
							VoicePlay(lp, unit, "menu2.vox");
							XWndStatusUpdate(unit, "get confID; acquiring DTMF");
							UnitData[unit].callState = CallStateGetConfID;
							confMode = CONF_MODE_LISTEN_ONLY;
						}
						else
						{
							XWndStatusUpdate(unit, "playing sorry.vox");
							VoicePlay(lp, unit, "sorry.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
						}
						break;

					//-----------------------------------------------------
					// conference ID command
					//-----------------------------------------------------
					case CallStateGetConfID :
						sscanf(szKeys, "%u", &UnitData[unit].confID);

						BT_ZERO(partyArgs);
						partyArgs.confID = UnitData[unit].confID;
						partyArgs.callRef = 0;
						partyArgs.type = CONF_TYPE_DEFAULT;
						partyArgs.mode = CONF_MODE_DEFAULT;
						partyArgs.mute = 0;
						partyArgs.noiseLimit = CONF_NOISE_LIMIT_DEFAULT;
						partyArgs.volume = 1;
						partyArgs.agcMax = 0;
						iStatus = BfvConferencePartyAdd(lp, &partyArgs);

						if (iStatus == BT_STATUS_OK)
						{
							UnitData[unit].callRef = partyArgs.callRef;
							sprintf(szText, "conference %u party %u; acquiring DTMF", UnitData[unit].confID, UnitData[unit].callRef);
							XWndStatusUpdate(unit, szText);
							UnitData[unit].callState = CallStateInConf;

							// see if node connections exist for different conference
							for (pNodeData = NodeData; pNodeData < &NodeData[NODE_MAX]; pNodeData++)
								if (pNodeData->fBusy &&
									pNodeData->confID != UnitData[unit].confID)
									break;

							// if no node connections exist for difference conference
							if (pNodeData >= &NodeData[NODE_MAX])
							{
								dsp = unit / chanPerDsp;
								if (!NodeData[dsp].fBusy)
								{
									BT_ZERO(dspRoleArgs);
									dspRoleArgs.confID = UnitData[unit].confID;
									dspRoleArgs.operation = ALLOCATE;
									dspRoleArgs.role = CONF_CHANNEL_ROLE_NODE;
									dspRoleArgs.mode = CONF_MODE_FULL_DUPLEX;
									dspRoleArgs.volume = 1;
									dspRoleArgs.mute = 0;
									dspRoleArgs.clampKeys = 0;

									iStatus = BfvConferenceDspRole(
										UnitData[unit / chanPerDsp * chanPerDsp + chanPerDsp - 1].lp,
										&dspRoleArgs);

									if (iStatus != BT_STATUS_OK)
									{
										sprintf(szText,
											"IncomingChannelThread%u BfvConferenceDspRole failure %s (%d)",
											unit,
											ConferenceErrorName(dspRoleArgs.res.line_status),
											dspRoleArgs.res.line_status);
										XWndMsgBox(szText);
									}

									NodeData[dsp].confID = UnitData[unit].confID;
									NodeData[dsp].fBusy = TRUE;
								}
							}
							break;
						}

						// if party not added successfully, and for reason other than invalid conference
						if (partyArgs.res.line_status != CONF_ERROR_INVALID_CONFERENCE)
						{
							sprintf(szText,
								"IncomingChannelThread%u BfvConferencePartyAdd failure %s (%d)",
								unit,
								ConferenceErrorName(partyArgs.res.line_status),
								partyArgs.res.line_status);
							XWndMsgBox(szText);

							XWndStatusUpdate(unit, "playing menu1.vox");
							VoicePlay(lp, unit, "menu1.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
							UnitData[unit].callState = CallStateMainMenu;
							break;
						}

						// here with invalid conference. attempt to create argument conference
						BT_ZERO(conferenceArgs);
						conferenceArgs.rendezvousKey = 1;
						conferenceArgs.initialAgcMax = 0;
						conferenceArgs.initialMode = CONF_MODE_DEFAULT;
						conferenceArgs.initialMute = 0;
						conferenceArgs.initialNoiseLimit = CONF_NOISE_LIMIT_DEFAULT;
						conferenceArgs.initialType = CONF_TYPE_DEFAULT;
						conferenceArgs.initialVolume = 1;
						conferenceArgs.maxTalkers = unitNum;
						conferenceArgs.minReserveSize = unitNum;
						conferenceArgs.clampKeys = 1;
						conferenceArgs.confID = UnitData[unit].confID;
						iStatus = BfvConferenceBegin(&conferenceArgs);

						// if conference not created successfully
						if (iStatus != BT_STATUS_OK)
						{
							sprintf(szText,
								"IncomingChannelThread%u BfvConferenceBegin failure %s (%d)",
								unit,
								ConferenceErrorName(conferenceArgs.res.line_status),
								conferenceArgs.res.line_status);
							XWndMsgBox(szText);

							XWndStatusUpdate(unit, "playing menu1.vox");
							VoicePlay(lp, unit, "menu1.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
							UnitData[unit].callState = CallStateMainMenu;
							break;
						}

						// attempt again to become party to argument conference
						BT_ZERO(partyArgs);
						partyArgs.confID = UnitData[unit].confID;
						partyArgs.callRef = 0;
						partyArgs.type = CONF_TYPE_DEFAULT;
						partyArgs.mode = CONF_MODE_DEFAULT;
						partyArgs.mute = 0;
						partyArgs.noiseLimit = CONF_NOISE_LIMIT_DEFAULT;
						partyArgs.volume = 1;
						partyArgs.agcMax = 0;
						iStatus = BfvConferencePartyAdd(lp, &partyArgs);

						// if party not added successfully. this time don't care why failure occurred
						if (iStatus != BT_STATUS_OK)
						{
							sprintf(szText,
								"IncomingChannelThread%u BfvConferencePartyAdd failure %s (%d)",
								unit,
								ConferenceErrorName(partyArgs.res.line_status),
								partyArgs.res.line_status);
							XWndMsgBox(szText);

							XWndStatusUpdate(unit, "playing menu1.vox");
							VoicePlay(lp, unit, "menu1.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
							UnitData[unit].callState = CallStateMainMenu;
							break;
						}

						// here with party added successfully
						UnitData[unit].callRef = partyArgs.callRef;
						sprintf(szText,
							"conference %u party %u; acquiring DTMF",
							UnitData[unit].confID,
							UnitData[unit].callRef);
						XWndStatusUpdate(unit, szText);
						UnitData[unit].callState = CallStateInConf;

						// see if node connections exist for different conference
						for (pNodeData = NodeData; pNodeData < &NodeData[NODE_MAX]; pNodeData++)
							if (pNodeData->fBusy &&
								pNodeData->confID != UnitData[unit].confID)
								break;

						// if no node connections exist for difference conference
						if (pNodeData >= &NodeData[NODE_MAX])
						{
							dsp = unit / chanPerDsp;
							if (!NodeData[dsp].fBusy)
							{
								BT_ZERO(dspRoleArgs);
								dspRoleArgs.confID = UnitData[unit].confID;
								dspRoleArgs.operation = ALLOCATE;
								dspRoleArgs.role = CONF_CHANNEL_ROLE_NODE;
								dspRoleArgs.mode = CONF_MODE_FULL_DUPLEX;
								dspRoleArgs.volume = 1;
								dspRoleArgs.mute = 0;
								dspRoleArgs.clampKeys = 0;

								iStatus = BfvConferenceDspRole(
									UnitData[unit / chanPerDsp * chanPerDsp + chanPerDsp - 1].lp,
									&dspRoleArgs);

								if (iStatus != BT_STATUS_OK)
								{
									sprintf(szText,
										"IncomingChannelThread%u BfvConferenceDspRole failure %s (%d)",
										unit,
										ConferenceErrorName(dspRoleArgs.res.line_status),
										dspRoleArgs.res.line_status);
									XWndMsgBox(szText);
								}

								NodeData[dsp].confID = UnitData[unit].confID;
								NodeData[dsp].fBusy = TRUE;
							}
						}
						break;

					//-----------------------------------------------------
					// in conference command
					//-----------------------------------------------------
					case CallStateInConf :
						// command "2#" means exit conference, return to main menu
						if (!strcmp(szKeys, "2"))
						{
							// drop party from conference
							BT_ZERO(partyArgs);
							partyArgs.confID = UnitData[unit].confID;
							partyArgs.callRef = UnitData[unit].callRef;
							iStatus = BfvConferencePartyDrop(lp, &partyArgs);
							UnitData[unit].callRef = 0;

							if (iStatus == BT_STATUS_OK)
							{
								// see if any parties remaining in conference
								for (pUnitData = UnitData; pUnitData < &UnitData[unitNum]; pUnitData++)
									if (pUnitData->callRef &&
										pUnitData->confID == UnitData[unit].confID)
										break;

								// if no parties remaining in conference
								if (pUnitData >= &UnitData[unitNum])
								{
									// terminate node connections, if present
									for (pNodeData = NodeData, unitNode = chanPerDsp - 1;
										pNodeData < &NodeData[NODE_MAX];
										pNodeData++, unitNode += chanPerDsp)

										if (pNodeData->fBusy &&
											pNodeData->confID == UnitData[unit].confID)
										{
											BT_ZERO(dspRoleArgs);
											dspRoleArgs.operation = FREE;
											dspRoleArgs.role = CONF_CHANNEL_ROLE_NODE;

											iStatus = BfvConferenceDspRole(
												UnitData[unitNode].lp,
												&dspRoleArgs);

											if (iStatus != BT_STATUS_OK)
											{
												sprintf(szText,
													"IncomingChannelThread%u BfvConferenceDspRole failure %s (%d)",
													unit,
													ConferenceErrorName(dspRoleArgs.res.line_status),
													dspRoleArgs.res.line_status);
												XWndMsgBox(szText);
											}

											pNodeData->fBusy = FALSE;
										}

									// terminate conference
									BT_ZERO(conferenceArgs);
									conferenceArgs.rendezvousKey = 1;
									conferenceArgs.confID = UnitData[unit].confID;
									iStatus = BfvConferenceEnd(&conferenceArgs);

									if (iStatus != BT_STATUS_OK)
									{
										sprintf(szText,
											"IncomingChannelThread%u BfvConferenceEnd failure %s (%d)",
											unit,
											ConferenceErrorName(conferenceArgs.res.line_status),
											conferenceArgs.res.line_status);
										XWndMsgBox(szText);
									}
								}

								// return to main menu
								XWndStatusUpdate(unit, "playing menu1.vox");
								VoicePlay(lp, unit, "menu1.vox");
								XWndStatusUpdate(unit, "main menu; acquiring DTMF");
								UnitData[unit].callState = CallStateMainMenu;
							}
							else
							{
								sprintf(szText,
									"IncomingChannelThread%u BfvConferencePartyDrop failure %s (%d)",
									unit,
									ConferenceErrorName(partyArgs.res.line_status),
									partyArgs.res.line_status);
								XWndMsgBox(szText);
							}
						}
						else
						{
#if 0 // cannot play voice while party in conference
							XWndStatusUpdate(unit, "playing sorry.vox");
							VoicePlay(lp, unit, "sorry.vox");
#endif
							sprintf(szText, "conference %u party %u; acquiring DTMF", UnitData[unit].confID, UnitData[unit].callRef);
							XWndStatusUpdate(unit, szText);
						}
						break;
					}

					pKeys = szKeys;
					szKeys[0] = 0;
				}
				else
				{
					*pKeys++ = iKey;
					*pKeys = 0;
				}
			}
			else
			{
				BfvErrorMessage(lp, &toneArgs.res, &btErr);
				sprintf(szText,
					"IncomingChannelThread%u BfvToneGet failure %s",
					unit,
					btErr.long_msg);
				XWndMsgBox(szText);

				if (toneArgs.res.status == BT_STATUS_ERROR_HANGUP &&
					toneArgs.res.line_status == HNG_NO_LOOP_CURRENT)
				{
					fNoLoopCurrent = TRUE;
					break;
				}
			}
		}

		if (UnitData[unit].callRef)
		{
			BT_ZERO(partyArgs);
			partyArgs.confID = UnitData[unit].confID;
			partyArgs.callRef = UnitData[unit].callRef;
			iStatus = BfvConferencePartyDrop(lp, &partyArgs);
			UnitData[unit].callRef = 0;

			if (iStatus == BT_STATUS_OK)
			{
				// see if any parties remaining in conference
				for (pUnitData = UnitData; pUnitData < &UnitData[unitNum]; pUnitData++)
					if (pUnitData->callRef &&
						pUnitData->confID == UnitData[unit].confID)
						break;

				// if no parties remaining in conference
				if (pUnitData >= &UnitData[unitNum])
				{
					// terminate node connections, if present
					for (pNodeData = NodeData, unitNode = chanPerDsp - 1;
						pNodeData < &NodeData[NODE_MAX];
						pNodeData++, unitNode += chanPerDsp)

						if (pNodeData->fBusy &&
							pNodeData->confID == UnitData[unit].confID)
						{
							BT_ZERO(dspRoleArgs);
							dspRoleArgs.operation = FREE;
							dspRoleArgs.role = CONF_CHANNEL_ROLE_NODE;

							iStatus = BfvConferenceDspRole(
								UnitData[unitNode].lp,
								&dspRoleArgs);

							if (iStatus != BT_STATUS_OK)
							{
								sprintf(szText,
									"IncomingChannelThread%u BfvConferenceDspRole failure %s (%d)",
									unit,
									ConferenceErrorName(dspRoleArgs.res.line_status),
									dspRoleArgs.res.line_status);
								XWndMsgBox(szText);
							}

							pNodeData->fBusy = FALSE;
						}

					// terminate conference
					BT_ZERO(conferenceArgs);
					conferenceArgs.rendezvousKey = 1;
					conferenceArgs.confID = UnitData[unit].confID;
					iStatus = BfvConferenceEnd(&conferenceArgs);

					if (iStatus != BT_STATUS_OK)
					{
						sprintf(szText,
							"IncomingChannelThread%u BfvConferenceEnd failure %s (%d)",
							unit,
							ConferenceErrorName(conferenceArgs.res.line_status),
							conferenceArgs.res.line_status);
						XWndMsgBox(szText);
					}
				}
			}
			else
			{
				sprintf(szText,
					"IncomingChannelThread%u BfvConferencePartyDrop failure %s (%d)",
					unit,
					ConferenceErrorName(partyArgs.res.line_status),
					partyArgs.res.line_status);
				XWndMsgBox(szText);
			}
		}

		if (!fNoLoopCurrent)
		{
			XWndStatusUpdate(unit, "playing bye.vox");
			VoicePlay(lp, unit, "bye.vox");

			//-----------------------------------------------------------------
			// terminate incoming call
			//-----------------------------------------------------------------
			XWndStatusUpdate(unit, "terminating call");
			BT_ZERO(telArgs);
			BfvLineTerminateCall(lp, &telArgs);

			if (telArgs.res.status != BT_STATUS_OK)
			{
				BfvErrorMessage(lp, &telArgs.res, &btErr);
				sprintf(szText,
					"IncomingChannelThread%u BfvLineTerminateCall failure %s",
					unit,
					btErr.long_msg);
				XWndMsgBox(szText);
			}
		}

		//-----------------------------------------------------------------
		// reset channel
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "reseting");
		BT_ZERO(lineArgs);
		lineArgs.config_file_name = "btcall.cfg";
		lineArgs.skip_teleph = 1;
		if (BfvLineReset(lp, &lineArgs) < 0)
		{
			BfvErrorMessage(lp, &lineArgs.res, &btErr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineReset failure %s",
				unit,
				btErr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}
	}
}

//-----------------------------------------------------------------------
// Procedure VoiceConferenceInit
//-----------------------------------------------------------------------
#ifdef WIN32
void VoiceConferenceInit(void)
#else
int main(int argc, char **argv)
#endif
{
	int ndx;
	int iStatus;
	unsigned unit;
	unsigned timeSlot;
	pthread_t threadID;
#ifdef UNIXWARE
	pthread_attr_t threadAttr;
#endif
	struct args_conf_admin confAdmin;

#ifdef WIN32
	int argc = 0;
	char* argv[32];
	char* cmd = GetCommandLine();
	char* token = strtok(cmd, " \t");
	while (token)
	{
		argv[argc++] = token;
		token = strtok(NULL, " \t");
	}
#endif
	for (ndx = 1; ndx < argc; ndx++)

		// -c<#> set number of channels reserved for conferencing
		if (!strnicmp(argv[ndx], "-c", 2))
			sscanf(&argv[ndx][2], "%u", &unitNum);
		// -p<#> set channels per DSP
		else if (!strnicmp(argv[ndx], "-p", 2))
			sscanf(&argv[ndx][2], "%u", &chanPerDsp);
		// -a<#> set group-a base
		else if (!strnicmp(argv[ndx], "-a", 2))
			sscanf(&argv[ndx][2], "%u", &unitBaseA);
		// -b<#> set group-b base
		else if (!strnicmp(argv[ndx], "-b", 2))
			sscanf(&argv[ndx][2], "%u", &unitBaseB);
		// -v<#> set verbose level
		else if (!strnicmp(argv[ndx], "-v", 2))
			sscanf(&argv[ndx][2], "%u", &verbose);
		// -nocall
		else if (!stricmp(argv[ndx], "-nocall"))
			fNoCall = TRUE;
		// none of the above
		else
		{
			XWndMsgBox("Usage: bridge [-a#] [-b#] [-c#] [-p#] [-v#] [-nocall]");
			exit(1);
		}

	// if group-b base not specified
	if (unitBaseB == unitBaseA)
		unitBaseB = unitBaseA + chanPerDsp;

	// create debug semaphore
	if (sem_init(&semDebug, SEM_SHARE_LOCAL, 1) == -1)
	{
		XWndMsgBox("VoiceConferenceInit sem_init failure");
		exit(1);
	}

	// enable debug display
	if (verbose >= 2)
	{
		BfvDebugModeSet(DEBUG_ALL);
		BfvDebugFuncSet(DebugFunc);
	}
	if (verbose >= 1)
	{
		BfvConferenceDebugModeSet(CONF_TRACE_LEVEL_HIGH, 0xFFFFFFFF);
		BfvConferenceDebugFuncSet(DebugFunc);
	}

	// startup conference administrator
	BT_ZERO(confAdmin);
	confAdmin.rendezvousKey = 1;
	confAdmin.minID = 1;
	confAdmin.maxID = 100;
	iStatus = BfvConferenceAdminInitialize(&confAdmin);

	if (iStatus != BT_STATUS_OK)
	{
		XWndMsgBox("VoiceConferenceInit BfvConferenceAdminInitialize failure");
		exit(1);
	}

#ifndef WIN32
	{
		sem_t semXWndInit;
		if (sem_init(&semXWndInit, SEM_SHARE_LOCAL, 0) == -1)
		{
			printf("VoiceConferenceInit sem_init failure\n");
			exit(1);
		}
#ifdef UNIXWARE
		memset(&threadAttr, 0, sizeof(pthread_attr_t));
		threadAttr.__pt_attr_stacksize = 1024 * 1024;
		pthread_create(&threadID, &threadAttr, XWndProc, (void *) &semXWndInit);
#else
		pthread_create(&threadID, NULL, XWndProc, (void *) &semXWndInit);
#endif
		sem_wait(&semXWndInit);
		sem_destroy(&semXWndInit);
	}
#endif

	// startup incoming channel threads
	timeSlot = 0;
	for (unit = 0; unit < unitNum; unit++)
	{
		if (unit % chanPerDsp == chanPerDsp - 1)
			UnitData[unit].role = SPEECH_CONF_PARTY_ROLE_NODE_DEF;
		else
		{
			UnitData[unit].role = SPEECH_CONF_PARTY_ROLE_TERMINAL_DEF;
			UnitData[unit].span = timeSlot / SLOTS_PER_T1;
			UnitData[unit].timeSlot = timeSlot % SLOTS_PER_T1;
			timeSlot++;
		}

#ifdef UNIXWARE
		memset(&threadAttr, 0, sizeof(pthread_attr_t));
		threadAttr.__pt_attr_stacksize = 1024 * 1024;
		iStatus = pthread_create(&threadID, &threadAttr, IncomingChannelThread, (void *) unit);
#else
		iStatus = pthread_create(&threadID, NULL, IncomingChannelThread, (void *) unit);
#endif
		printf("iStatus %d threadID %d\n", iStatus, threadID);
	}

	// startup outgoing channels
	if (!fNoCall)
		CallInit();

#ifndef WIN32
	// wait forever
	while (1)
		sleep(60);
#endif
}
