// %A%
//----------------------------------------------------------------------
// Source: callproc.cpp
//
// Copyright (c) 1999-2000 Brooktrout Technology, Needham, MA
// All rights reserved
//
// General Purpose:
//
// The functions in this module are invoked by GUI functions in
// CallOriginateDialog.cpp, which are in turn invoked by user
// manipulation of the pop-up non-modal "Call Originate" dialog
// box. To facilitate testing of the conferencing software, provision
// is made for one T1 controller on a TR1000 to be connected via a
// loopback cable to another T1 controller on the same TR1000. Spans
// and time slots are set up accordingly, so a call originating on
// the second T1 controller rings (and is answered by) an analogous
// time slot on the first T1 controller. DTMF can then be sent, thus
// simulating the action of a caller's touch tone keypad. Conferences
// can be entered and exited. Finally, a call can be terminated.
//
// The source below assumes define constant unitNum indicates the
// number of answering channels on the first T1 controller and the
// number of originating channels on the second T1 controller. Therefore,
// with unitNum=12, channel 12 is looped back to channnel 0, channel
// 13 to channel 1, etc..
//
// Typical sequencing of the pop-up window is as follows:
// Port=12 DTMF=w click Originate (port 0 answers, plays welcome.vox)
// Port=12 DTMF=1# click Tone (port 0 recognizes enter conference command)
// Port=12 DTMF=52# click Tone (port 0 attempts to enter conference 52)
// ..
// Port=12 DTMF=2# click Tone (port 0 recognizes exit conference command)
// Port=12 DTMF=* click Tone (port 0 recognizes hangup command)
// Port=12 DTMF=N/A click HangUp
//
// Edit History:
//
// 06-Dec-99  FJN  Created
//----------------------------------------------------------------------

#ifdef WIN32
#include "stdafx.h"
#include "vcdemo.h"
#include "MainFrm.h"
#include "vcdemoDoc.h"
#include "vcdemoView.h"
#endif
#include <string.h>

#ifdef WIN32
extern "C"
{
#endif
#include "btlib.h"
#ifdef WIN32
}
#endif

#define	UNIT_BASE	2
#define	UNIT_MAX	128

typedef struct
{
	BTLINE *lp;
	int fBusy;
} CallStruct;

static CallStruct CallData[UNIT_MAX];

extern BOOL fConfVerbose;
extern unsigned unitNum;
void DebugFunc(char *p);
void XWndMsgBox(char *p);

//----------------------------------------------------------------------
// Procedure CallInit
//----------------------------------------------------------------------
void CallInit(void)
{
	unsigned unit;
	struct args_line_admin lineArgs;
	struct args_tel_ctrl_call_sw switchArgs;
	BTERR bterr;
	char szText[256];

	for (unit = unitNum; unit < unitNum * 2; unit++)
	{
		//----------------------------------------------------------------------
		// attach channel
		//----------------------------------------------------------------------
		BT_ZERO(lineArgs);
		lineArgs.unit = unit;
		if (!(CallData[unit-unitNum].lp = BfvLineAttach(&lineArgs)))
		{
			BfvErrorMessage(NULL, &lineArgs.res, &bterr);
			sprintf(szText,
				"CallInit%u : BfvLineAttach failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return;
		}

		//----------------------------------------------------------------------
		// reset channel
		//----------------------------------------------------------------------
		BT_ZERO(lineArgs);
		lineArgs.config_file_name = "btcall.cfg";
		if (BfvLineReset(CallData[unit-unitNum].lp, &lineArgs) < 0)
		{
			BfvErrorMessage(CallData[unit-unitNum].lp, &lineArgs.res, &bterr);
			sprintf(szText,
				"CallInit%u : BfvLineReset failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return;
		}

		//----------------------------------------------------------------------
		// connect channel
		//----------------------------------------------------------------------
		// make connection from DSP (channel) to T1 time slot (network)
		BT_ZERO(switchArgs);
		switchArgs.conn_mode = CALL_SW_DUPLEX_AND_SIGNALING_DEF;
		switchArgs.src_port_class = CALL_SW_PORT_CHANNEL_DEF;
		switchArgs.src_port_unit = unit + UNIT_BASE;
		switchArgs.src_stream = 0;
		switchArgs.src_slot = 0;
		switchArgs.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
		switchArgs.dest_port_unit = 1;			// t1 controller span
		switchArgs.dest_stream = 0;				// unused
		switchArgs.dest_slot = unit - unitNum;	// t1 time slot
		if (BfvCallSWConnect(CallData[unit-unitNum].lp, &switchArgs) < 0)
		{
			BfvErrorMessage(CallData[unit-unitNum].lp, &switchArgs.res, &bterr);
			sprintf(szText,
				"CallInit%u BfvCallSWConnect terminal failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return;
		}

		if (fConfVerbose)
		{
			sprintf(szText,
				"CallInit%u BfvCallSWConnect conn_mode %u src_port_class %u src_port_unit %u src_stream %u src_slot %u dest_port_class %u dest_port_unit %u dest_stream %u dest_slot %u",
				unit,
				switchArgs.conn_mode,
				switchArgs.src_port_class,
				switchArgs.src_port_unit,
				switchArgs.src_stream,
				switchArgs.src_slot,
				switchArgs.dest_port_class,
				switchArgs.dest_port_unit,
				switchArgs.dest_stream,
				switchArgs.dest_slot);
			DebugFunc(szText);
		}
	}
}

//----------------------------------------------------------------------
// Procedure CallOriginate
//----------------------------------------------------------------------
void CallOriginate(unsigned unit, char *pDial)
{
	struct args_telephone telArgs;
	BTERR bterr;
	char szText[256];

	printf("CallOriginate%u : dialing %s\n", unit, pDial);
	if (unit < unitNum ||
		unit >= unitNum * 2 ||
		!CallData[unit-unitNum].lp ||
		CallData[unit-unitNum].fBusy)
		return;

	BT_ZERO(telArgs);
	telArgs.phonenum = pDial;
	telArgs.call_protocol_code = CALL_PROTOCOL_VOICE;
	BfvLineOriginateCall(CallData[unit-unitNum].lp, &telArgs);
	if (telArgs.res.status == BT_STATUS_OK)
	{
		CallData[unit-unitNum].fBusy = TRUE;
#ifdef DEBUG_VERBOSE
		sprintf(szText, "CallOriginate%u : BfvLineOriginateCall success\n", unit);
		XWndMsgBox(szText);
#endif
	}
	else
	{
		BfvErrorMessage(CallData[unit-unitNum].lp, &telArgs.res, &bterr);
		sprintf(szText, "CallOriginate%u : BfvLineOriginateCall failure %d %s",
			unit,
			telArgs.res.line_status,
			bterr.long_msg);
		XWndMsgBox(szText);
	}
}

//----------------------------------------------------------------------
// Procedure CallTerminate
//----------------------------------------------------------------------
void CallTerminate(unsigned unit)
{
	struct args_telephone telArgs;
	BTERR bterr;
	char szText[256];

	printf("CallTerminate%u : terminating\n", unit);
	if (unit < unitNum ||
		unit >= unitNum + unitNum ||
		!CallData[unit-unitNum].lp ||
		!CallData[unit-unitNum].fBusy)
		return;

	BT_ZERO(telArgs);
	BfvLineTerminateCall(CallData[unit-unitNum].lp, &telArgs);
	if (telArgs.res.status == BT_STATUS_OK)
	{
		CallData[unit-unitNum].fBusy = FALSE;
#ifdef DEBUG_VERBOSE
		sprintf(szText, "CallTerminate%u : BfvLineTerminateCall success\n", unit);
		XWndMsgBox(szText);
#endif
	}
	else
	{
		BfvErrorMessage(CallData[unit-unitNum].lp, &telArgs.res, &bterr);
		sprintf(szText,
			"CallTerminate%u : BfvLineTerminateCall failure %s",
			unit,
			bterr.long_msg);
		XWndMsgBox(szText);
	}
}

//----------------------------------------------------------------------
// Procedure CallPlayTones
//----------------------------------------------------------------------
void CallPlayTones(unsigned unit, char *pDial)
{
	struct args_tone toneArgs;
	BTERR bterr;
	char szText[256];

	printf("CallPlayTones%u : keying %s\n", unit, pDial);
	if (unit < unitNum ||
		unit >= unitNum + unitNum ||
		!CallData[unit-unitNum].lp ||
		!CallData[unit-unitNum].fBusy)
		return;

	while (*pDial)
	{
		BT_ZERO(toneArgs);
		toneArgs.digit = *pDial++;
		toneArgs.millisecs = 100;
		toneArgs.tone_type = DTMF_TONE;
		BfvTonePlay(CallData[unit-unitNum].lp, &toneArgs);
		if (toneArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(CallData[unit-unitNum].lp, &toneArgs.res, &bterr);
			sprintf(szText,
				"CallPlayTones%u : BfvTonePlay failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return;
		}
#ifdef DEBUG_VERBOSE
		sprintf(szText, "CallPlayTones%u : BfvTonePlay success\n", unit);
		XWndMsgBox(szText);
#endif
	}
}
