// %A%
//----------------------------------------------------------------------
// Source: vcline.cpp
//
// Copyright (c) 1999-2000 Brooktrout Technology, Needham, MA
// All rights reserved
//
// General Purpose:
//
// The functions in this module serve to spawn one thread per incoming
// channel. Each incoming channel thread initializes and awaits an
// incoming call. Once answered, it plays welcome.vox and awaits DTMF
// interaction with the caller.
//
// All DTMF commands are terminated with "#", and the following DTMF
// commands are recognized:
//
// (from main menu) "1#" enter conference menu
// (from conference menu) "nnn#" enter conference nnn
// (from conference) "2#" exit conference, enter main menu
// (from anywhere) "*" terminate call
// 
// Edit History:
//
// 06-Dec-99  FJN  Created
// 03-Sep-02  FJN  Enhance to support DTMF "3#" to enter conference listen only
// 30-Sep-02  FJN  Enhance to support DTMF "4#" when in conference to change mode to full duplex
// 30-Sep-02  FJN  Enhance to support DTMF "5#" when in conference to change mode to listen only
// 17-Jun-03  FJN  Use BfvLineOriginateCall to "busy out"
// 18-Jun-03  FJN  Do not report errors from BfvLineOriginateCall when "busy out"
//----------------------------------------------------------------------

#ifdef WIN32
#include "stdafx.h"
#include "vcdemo.h"
#include "MainFrm.h"
#include "vcdemoDoc.h"
#include "vcdemoView.h"
#include <io.h>
#else
#define O_BINARY 0
#define	UNIT_MAX	128
#endif

#include <fcntl.h>
#include <errno.h>

#ifdef WIN32
extern "C"
{
#endif

#include "btlib.h"
#include "acclib.h"
#include "pthread.h"
#include "semaphore.h"

#ifdef WIN32
void sleep(int);
}
#endif

#define	SEM_SHARE_LOCAL		0
#define	SEM_SHARE_GLOBAL	1

#define	UNIT_BASE			2
#define	READ_BUF_SIZE		1024
#define	SLOTS_PER_T1		24

typedef enum
{
	CallStateAwaitingIncomingCall,
	CallStateAnsweringIncomingCall,
	CallStateMainMenu,
	CallStateGetConfID,
	CallStateInConf
} CallStateEnum;

typedef struct
{
	BTLINE *lp;
	unsigned span;
	unsigned timeSlot;
	unsigned confID;
	unsigned callRef;
	int keySave;
	CallStateEnum callState;
} UnitStruct;

void XWndMsgBox(char *);
void XWndStatusUpdate(unsigned, char *);
void *XWndProc(void *);
void CallInit(void);

static sem_t semDebug;
static UnitStruct UnitData[UNIT_MAX];

BOOL fConfVerbose = FALSE;
unsigned unitNum = 30;
unsigned busyOutMax = 5;

static char szUsage[] = "Usage : simpleconf [-v] [-c#] [-b#]";

//-----------------------------------------------------------------------
// Procedure ConferenceErrorName
//-----------------------------------------------------------------------
char *
ConferenceErrorName(
	int codeError)
{
	switch (codeError)
	{
	case 0 :
		return("");
	case CONF_ERROR_BAD_PARAMETERS :
		return("bad parameters");
	case CONF_ERROR_BAD_ADMIN_INIT :
		return("bad admin init");
	case CONF_ERROR_ADMIN_EXISTS :
		return("admin exists");
	case CONF_ERROR_ADMIN_MISSING :
		return("admin missing");
	case CONF_ERROR_ID_MISSING_RESERVED :
		return("ID missing reserved");
	case CONF_ERROR_ID_MISSING_NOT_RESERVED :
		return("ID missing not reserved");
	case CONF_ERROR_INSUFFICIENT_RESOURCES :
		return("insufficient resources");
	case CONF_ERROR_ID_IN_USE :
		return("ID in use");
	case CONF_ERROR_RESOURCE_HANDLER_ERROR :
		return("resource handler error");
	case CONF_ERROR_INVALID_CONFERENCE :
		return("invalid conference");
	case CONF_ERROR_CALLREF_IN_USE :
		return("call reference in use");
	case CONF_ERROR_RELEASING_RESOURCES :
		return("releasing resources");
	case CONF_ERROR_INVALID_CALLREF :
		return("invalid call reference");
	case CONF_ERROR_SYSTEM :
		return("system");
	case CONF_ERROR_TALKER_LIMIT :
		return("talker limit");
	case CONF_ERROR_ID_NOT_BEGUN :
		return("ID not begun");
	default :
		return("unknown");
	}
}

//-----------------------------------------------------------------------
// Procedure DebugFunc
//-----------------------------------------------------------------------

void DebugFunc(char *p)
{
	FILE *fp;
	sem_wait(&semDebug);
	if (fp = fopen("vcdemo.log", "a+"))
	{
		fseek(fp, 0, SEEK_END);
		fprintf(fp, "%s\n", p);
		fclose(fp);
	}
	else
		XWndMsgBox("DebugFunc fopen vcdemo.log failure");
	sem_post(&semDebug);
}

//-----------------------------------------------------------------------
// Procedure TerminationCheck
//-----------------------------------------------------------------------
int
TerminationCheck(
	BTLINE *lp,
	char *termParms)
{
	int key;
	struct args_tone toneArgs;
	unsigned unit = (unsigned) termParms;
	char szText[256];
	BTERR btErr;

	BT_ZERO(toneArgs);
	key = BfvToneGet(lp, &toneArgs);
	if (key >= 0)
	{
		UnitData[unit].keySave = key;
		return(1);
	}
	else
	{
		if (toneArgs.res.status == BT_STATUS_TIMEOUT)
			return(0);

		BfvErrorMessage(lp, &toneArgs.res, &btErr);
		sprintf(szText,
			"TerminationCheck%u BfvToneGet failure %s",
			unit,
			btErr.long_msg);
		XWndMsgBox(szText);

		UnitData[unit].keySave = 0;
		return(1);
	}
}

//-----------------------------------------------------------------------
// Procedure VoicePlay
//-----------------------------------------------------------------------

void VoicePlay(
	BTLINE *lp,
	unsigned unit,
	char *szFName)
{
	int fd;
	int bytesRead;
	char szText[256];
	unsigned char readBuf[READ_BUF_SIZE];
	struct args_speech speechArgs;
	BTERR btErr;

	if ((fd = open(szFName, O_RDONLY | O_BINARY)) < 0)
	{
		sprintf(szText,
			"VoicePlay%u open %s failure %d",
			unit,
			szFName,
			errno);
		XWndMsgBox(szText);
		return;
	}

	while ((bytesRead = read(fd, readBuf, READ_BUF_SIZE)) > 0)
	{
		BT_ZERO(speechArgs);
		speechArgs.buf = readBuf;
		speechArgs.size = bytesRead;
		speechArgs.coding_fmt = CODE_ADPCM;
		speechArgs.rate = RATE_8000;
		speechArgs.bits_per_samp = BITS_4;
		speechArgs.func = TerminationCheck;
		speechArgs.arg = (char *) unit;
		BfvSpeechPlayData(lp, &speechArgs);
		if (speechArgs.res.status != BT_STATUS_OK)
		{
			if (speechArgs.res.status != BT_STATUS_USER_TERMINATED)
			{
				BfvErrorMessage(lp, &speechArgs.res, &btErr);
				sprintf(szText,
					"VoicePlay%u BfvSpeechPlayData failure %s",
					unit,
					btErr.long_msg);
				XWndMsgBox(szText);
			}
			close(fd);
			return;
		}
	}

	BT_ZERO(speechArgs);
	speechArgs.buf = readBuf;
	speechArgs.size = 0;
	speechArgs.coding_fmt = CODE_ADPCM;
	speechArgs.rate = RATE_STOP;
	speechArgs.bits_per_samp = BITS_4;
	speechArgs.func = TerminationCheck;
	speechArgs.arg = (char *) unit;
	BfvSpeechPlayData(lp, &speechArgs);
	if (speechArgs.res.status != BT_STATUS_OK &&
		speechArgs.res.status != BT_STATUS_USER_TERMINATED)
	{
		BfvErrorMessage(lp, &speechArgs.res, &btErr);
		sprintf(szText,
			"VoicePlay%u BfvSpeechPlayData failure %s",
			unit,
			btErr.long_msg);
		XWndMsgBox(szText);
	}
	close(fd);
}

//-----------------------------------------------------------------------
// Procedure IncomingChannelThread
//
// Calling argument is unit number, zero based.
//-----------------------------------------------------------------------
void *
IncomingChannelThread(
	void *arg)
{
	int fNoLoopCurrent;
	int iKey;
	int iStatus;
	unsigned unit = (unsigned) arg;
	char szText[256];
	char szKeys[128];
	char *pKeys;
	BTLINE *lp;
	BTERR bterr;
	struct args_line_admin lineArgs;
	struct args_telephone telArgs;
	struct args_tone toneArgs;
	struct args_conference conferenceArgs;
	struct args_conf_party partyArgs;
	struct args_tel_ctrl_call_sw switchArgs;
	CONF_PARTY_MODES confMode;

#ifdef DEBUG_VERBOSE
	sprintf(szText, "IncomingChannelThread%u starting", arg);
	XWndMsgBox(szText);
#endif

	//-----------------------------------------------------------------
	// attach channel
	//-----------------------------------------------------------------
	XWndStatusUpdate(unit, "attaching");
	BT_ZERO(lineArgs);
	lineArgs.unit = unit;
	if (!(lp = BfvLineAttach(&lineArgs)))
	{
		BfvErrorMessage(lp, &lineArgs.res, &bterr);
		sprintf(szText,
			"IncomingChannelThread%u BfvLineAttach failure %s",
			unit,
			bterr.long_msg);
		XWndMsgBox(szText);
		return(NULL);
	}
	UnitData[unit].lp = lp;

	//-----------------------------------------------------------------
	// reset channel
	//-----------------------------------------------------------------
	XWndStatusUpdate(unit, "reseting");
	BT_ZERO(lineArgs);
	lineArgs.config_file_name = "btcall.cfg";
	if (BfvLineReset(lp, &lineArgs) < 0)
	{
		BfvErrorMessage(lp, &lineArgs.res, &bterr);
		sprintf(szText,
			"IncomingChannelThread%u BfvLineReset failure %s",
			unit,
			bterr.long_msg);
		XWndMsgBox(szText);
		return(NULL);
	}

	//-----------------------------------------------------------------
	// connect channel
	// make connection from DSP (channel) to T1 time slot (network)
	//-----------------------------------------------------------------
	XWndStatusUpdate(unit, "connecting");
	BT_ZERO(switchArgs);
	switchArgs.conn_mode = CALL_SW_DUPLEX_AND_SIGNALING_DEF;
	switchArgs.src_port_class = CALL_SW_PORT_CHANNEL_DEF;
	switchArgs.src_port_unit = UNIT_BASE + unit;
	switchArgs.src_stream = 0;
	switchArgs.src_slot = 0;
	switchArgs.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
	switchArgs.dest_port_unit = UnitData[unit].span;
	switchArgs.dest_stream = 0;
	switchArgs.dest_slot = UnitData[unit].timeSlot;

	if (fConfVerbose)
	{
		sprintf(szText,
			"IncomingChannelThread%u BfvCallSWConnect conn_mode %u src_port_class %u src_port_unit %u src_stream %u src_slot %u dest_port_class %u dest_port_unit %u dest_stream %u dest_slot %u",
			unit,
			switchArgs.conn_mode,
			switchArgs.src_port_class,
			switchArgs.src_port_unit,
			switchArgs.src_stream,
			switchArgs.src_slot,
			switchArgs.dest_port_class,
			switchArgs.dest_port_unit,
			switchArgs.dest_stream,
			switchArgs.dest_slot);
		DebugFunc(szText);
	}

	if (BfvCallSWConnect(lp, &switchArgs) < 0)
	{
		BfvErrorMessage(lp, &switchArgs.res, &bterr);
		sprintf(szText,
			"IncomingChannelThread%u BfvCallSWConnect terminal failure %s",
			unit,
			bterr.long_msg);
		XWndMsgBox(szText);
		return(NULL);
	}

	//-----------------------------------------------------------------
	// busy out some ports
	//-----------------------------------------------------------------
	if (busyOutMax &&
		unit >= busyOutMax)
	{
		XWndStatusUpdate(unit, "busy out");
		UnitData[unit].callState = CallStateAnsweringIncomingCall;
		BT_ZERO(telArgs);
		telArgs.phonenum = "";
		telArgs.call_protocol_code = CALL_PROTOCOL_VOICE;
		BfvLineOriginateCall(lp, &telArgs);
#if 0
		if (telArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &telArgs.res, &bterr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineOriginateCall failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}
#endif
		while (1)
			sleep(60);
	}

	//-----------------------------------------------------------------
	// forever loop
	//-----------------------------------------------------------------
	while (1)
	{
		//-----------------------------------------------------------------
		// wait for incoming call
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "waiting for incoming call");
		UnitData[unit].callState = CallStateAwaitingIncomingCall;
		BT_ZERO(telArgs);
		BfvLineWaitForCall(lp, &telArgs);
		if (telArgs.res.status != BT_STATUS_OK ||
			telArgs.res.line_status != WAIT_FOR_CALL_OK)
		{
			BfvErrorMessage(lp, &telArgs.res, &bterr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineWaitForCall failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		//-----------------------------------------------------------------
		// answer incoming call
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "answering incoming call");
		UnitData[unit].callState = CallStateAnsweringIncomingCall;
		BT_ZERO(telArgs);
		BfvLineAnswer(lp, &telArgs);
		if (telArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &telArgs.res, &bterr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineAnswer failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		fNoLoopCurrent = FALSE;

		//-----------------------------------------------------------------
		// flush buffered DTMF
		//-----------------------------------------------------------------
		BT_ZERO(toneArgs);
		BfvToneFlush(lp, &toneArgs);
		if (toneArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &toneArgs.res, &bterr);
			sprintf(szText,
				"IncomingChannelThread%u BfvToneFlush failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		//-----------------------------------------------------------------
		// setup DTMF acquisition parameters
		//-----------------------------------------------------------------
		BT_ZERO(toneArgs);
		toneArgs.decode_flag = DTMF_16TONE;
		BfvToneDetectEnable(lp, &toneArgs);
		if (toneArgs.res.status != BT_STATUS_OK)
		{
			BfvErrorMessage(lp, &toneArgs.res, &bterr);
			sprintf(szText,
				"IncomingChannelThread%u BfvToneDetectEnable failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}

		//-----------------------------------------------------------------
		// play welcome
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "playing welcome.vox");
		VoicePlay(lp, unit, "welcome.vox");

		if (UnitData[unit].keySave == 0)
		{
			XWndStatusUpdate(unit, "playing menu1.vox");
			VoicePlay(lp, unit, "menu1.vox");
		}

		//-----------------------------------------------------------------
		// set and display state
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "main menu; acquiring DTMF");
		UnitData[unit].callState = CallStateMainMenu;

		//-----------------------------------------------------------------
		// reset tone pointers
		//-----------------------------------------------------------------
		pKeys = szKeys;
		szKeys[0] = 0;

		//-----------------------------------------------------------------
		// in call main processing loop
		//-----------------------------------------------------------------
		while (1)
		{
			//-------------------------------------------------------------
			// acquire DTMF digit
			//-------------------------------------------------------------
			if (UnitData[unit].keySave > 0)
			{
				iKey = UnitData[unit].keySave;
				UnitData[unit].keySave = 0;
			}
			else
			{
				BT_ZERO(toneArgs);
				toneArgs.millisecs = 10 * 60 * 1000;
				iKey = BfvToneGet(lp, &toneArgs);
			}

			//-------------------------------------------------------------
			// if DTMF digit available
			//-------------------------------------------------------------
			if (iKey > 0)
			{
				sprintf(szText, "processing key %c", iKey);
				XWndStatusUpdate(unit, szText);

				// star means terminate call
				if (iKey == '*')
					break;

				// pound means terminate command
				if (iKey == '#')
				{
					sprintf(szText, "processing tone command %s", szKeys);
					XWndStatusUpdate(unit, szText);

					switch (UnitData[unit].callState)
					{
					//-----------------------------------------------------
					// main menu command
					//-----------------------------------------------------
					case CallStateMainMenu :
						// command "1#" means goto conference ID acquire state, duplex
						if (!strcmp(szKeys, "1"))
						{
							XWndStatusUpdate(unit, "playing menu2.vox");
							VoicePlay(lp, unit, "menu2.vox");
							XWndStatusUpdate(unit, "get confID; acquiring DTMF");
							UnitData[unit].callState = CallStateGetConfID;
							confMode = CONF_MODE_FULL_DUPLEX;
						}
						// command "3#" means goto conference ID acquire state, listen only
						else if (!strcmp(szKeys, "3"))
						{
							XWndStatusUpdate(unit, "playing menu2.vox");
							VoicePlay(lp, unit, "menu2.vox");
							XWndStatusUpdate(unit, "get confID; acquiring DTMF");
							UnitData[unit].callState = CallStateGetConfID;
							confMode = CONF_MODE_LISTEN_ONLY;
						}
						else
						{
							XWndStatusUpdate(unit, "playing sorry.vox");
							VoicePlay(lp, unit, "sorry.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
						}
						break;

					//-----------------------------------------------------
					// conference ID command
					//-----------------------------------------------------
					case CallStateGetConfID :
						sscanf(szKeys, "%u", &UnitData[unit].confID);

						// attempt to become party to argument conference
						BT_ZERO(partyArgs);
						partyArgs.confID = UnitData[unit].confID;
						partyArgs.callRef = 0;
						partyArgs.type = CONF_TYPE_DEFAULT;
						partyArgs.mode = confMode;
						partyArgs.mute = 0;
						partyArgs.noiseLimit = CONF_NOISE_LIMIT_DEFAULT;
						partyArgs.volume = 1;
						partyArgs.agcMax = 0;
						iStatus = BfvConferencePartyAdd(lp, &partyArgs);

						// if party added successfully
						if (iStatus == BT_STATUS_OK)
						{
							UnitData[unit].callRef = partyArgs.callRef;
							sprintf(szText,
								"conference %u party %u; acquiring DTMF",
								UnitData[unit].confID,
								UnitData[unit].callRef);
							XWndStatusUpdate(unit, szText);
							UnitData[unit].callState = CallStateInConf;
							break;
						}

						// if party not added successfully, and for reason other than invalid conference
						if (partyArgs.res.line_status != CONF_ERROR_INVALID_CONFERENCE)
						{
							sprintf(szText,
								"IncomingChannelThread%u BfvConferencePartyAdd failure %s (%d)",
								unit,
								ConferenceErrorName(partyArgs.res.line_status),
								partyArgs.res.line_status);
							XWndMsgBox(szText);

							XWndStatusUpdate(unit, "playing menu1.vox");
							VoicePlay(lp, unit, "menu1.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
							UnitData[unit].callState = CallStateMainMenu;
							break;
						}

						// here with invalid conference. attempt to create argument conference
						BT_ZERO(conferenceArgs);
						conferenceArgs.rendezvousKey = 1;
						conferenceArgs.initialAgcMax = 0;
						conferenceArgs.initialMode = CONF_MODE_DEFAULT;
						conferenceArgs.initialMute = 0;
						conferenceArgs.initialNoiseLimit = CONF_NOISE_LIMIT_DEFAULT;
						conferenceArgs.initialType = CONF_TYPE_DEFAULT;
						conferenceArgs.initialVolume = 0;
						conferenceArgs.comfortNoise = CONF_COMFORT_NOISE_NORMAL;
						conferenceArgs.maxTalkers = unitNum;
						conferenceArgs.minReserveSize = unitNum;
						conferenceArgs.clampKeys = 1;
						conferenceArgs.confID = UnitData[unit].confID;
						iStatus = BfvConferenceBegin(&conferenceArgs);

						// if conference not created successfully
						if (iStatus != BT_STATUS_OK)
						{
							sprintf(szText,
								"IncomingChannelThread%u BfvConferenceBegin failure %s (%d)",
								unit,
								ConferenceErrorName(conferenceArgs.res.line_status),
								conferenceArgs.res.line_status);
							XWndMsgBox(szText);

							XWndStatusUpdate(unit, "playing menu1.vox");
							VoicePlay(lp, unit, "menu1.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
							UnitData[unit].callState = CallStateMainMenu;
							break;
						}

						// attempt again to become party to argument conference
						BT_ZERO(partyArgs);
						partyArgs.confID = UnitData[unit].confID;
						partyArgs.callRef = 0;
						partyArgs.type = CONF_TYPE_DEFAULT;
						partyArgs.mode = confMode;
						partyArgs.mute = 0;
						partyArgs.noiseLimit = CONF_NOISE_LIMIT_DEFAULT;
						partyArgs.volume = 1;
						partyArgs.agcMax = 0;
						iStatus = BfvConferencePartyAdd(lp, &partyArgs);

						// if party not added successfully. this time don't care why failure occurred
						if (iStatus != BT_STATUS_OK)
						{
							sprintf(szText,
								"IncomingChannelThread%u BfvConferencePartyAdd failure %s (%d)",
								unit,
								ConferenceErrorName(partyArgs.res.line_status),
								partyArgs.res.line_status);
							XWndMsgBox(szText);

							XWndStatusUpdate(unit, "playing menu1.vox");
							VoicePlay(lp, unit, "menu1.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
							UnitData[unit].callState = CallStateMainMenu;
							break;
						}

						// here with party added successfully
						UnitData[unit].callRef = partyArgs.callRef;
						sprintf(szText,
							"conference %u party %u; acquiring DTMF",
							UnitData[unit].confID,
							UnitData[unit].callRef);
						XWndStatusUpdate(unit, szText);
						UnitData[unit].callState = CallStateInConf;
						break;

					//-----------------------------------------------------
					// in conference command
					//-----------------------------------------------------
					case CallStateInConf :
						// command "2#" means exit conference, return to main menu
						if (!strcmp(szKeys, "2"))
						{
							// drop party from conference
							BT_ZERO(partyArgs);
							partyArgs.confID = UnitData[unit].confID;
							partyArgs.callRef = UnitData[unit].callRef;
							iStatus = BfvConferencePartyDrop(lp, &partyArgs);
							UnitData[unit].callRef = 0;
							if (iStatus != BT_STATUS_OK)
							{
								sprintf(szText,
									"IncomingChannelThread%u BfvConferencePartyDrop failure %s (%d)",
									unit,
									ConferenceErrorName(partyArgs.res.line_status),
									partyArgs.res.line_status);
								XWndMsgBox(szText);

								XWndStatusUpdate(unit, "playing menu1.vox");
								VoicePlay(lp, unit, "menu1.vox");
								XWndStatusUpdate(unit, "main menu; acquiring DTMF");
								UnitData[unit].callState = CallStateMainMenu;
								break;
							}

							// return to main menu
							XWndStatusUpdate(unit, "playing menu1.vox");
							VoicePlay(lp, unit, "menu1.vox");
							XWndStatusUpdate(unit, "main menu; acquiring DTMF");
							UnitData[unit].callState = CallStateMainMenu;
						}
						// command "4#" means set mode to full duplex
						else if (!strcmp(szKeys, "4"))
						{
							BT_ZERO(partyArgs);
							partyArgs.confID = UnitData[unit].confID;
							partyArgs.callRef = UnitData[unit].callRef;
							iStatus = BfvConferencePartyQuery(lp, &partyArgs);
							if (iStatus != BT_STATUS_OK)
							{
								sprintf(szText,
									"IncomingChannelThread%u BfvConferencePartyQuery failure %s (%d)",
									unit,
									ConferenceErrorName(partyArgs.res.line_status),
									partyArgs.res.line_status);
								XWndMsgBox(szText);
								break;
							}
							// do NOT re-zero the argument structure--retain info from query
							partyArgs.mode = CONF_MODE_FULL_DUPLEX;
							iStatus = BfvConferencePartyModify(lp, &partyArgs);
							if (iStatus != BT_STATUS_OK)
							{
								sprintf(szText,
									"IncomingChannelThread%u BfvConferencePartyModify failure %s (%d)",
									unit,
									ConferenceErrorName(partyArgs.res.line_status),
									partyArgs.res.line_status);
								XWndMsgBox(szText);
							}
						}
						// command "5#" means set mode to listen only
						else if (!strcmp(szKeys, "5"))
						{
							BT_ZERO(partyArgs);
							partyArgs.confID = UnitData[unit].confID;
							partyArgs.callRef = UnitData[unit].callRef;
							iStatus = BfvConferencePartyQuery(lp, &partyArgs);
							if (iStatus != BT_STATUS_OK)
							{
								sprintf(szText,
									"IncomingChannelThread%u BfvConferencePartyQuery failure %s (%d)",
									unit,
									ConferenceErrorName(partyArgs.res.line_status),
									partyArgs.res.line_status);
								XWndMsgBox(szText);
								break;
							}
							// do NOT re-zero the argument structure--retain info from query
							partyArgs.mode = CONF_MODE_LISTEN_ONLY;
							iStatus = BfvConferencePartyModify(lp, &partyArgs);
							if (iStatus != BT_STATUS_OK)
							{
								sprintf(szText,
									"IncomingChannelThread%u BfvConferencePartyModify failure %s (%d)",
									unit,
									ConferenceErrorName(partyArgs.res.line_status),
									partyArgs.res.line_status);
								XWndMsgBox(szText);
							}
						}
						// bad command
						else
						{
#if 0 // cannot play voice while party in conference
							XWndStatusUpdate(unit, "playing sorry.vox");
							VoicePlay(lp, unit, "sorry.vox");
#endif
							sprintf(szText,
								"in conference %u party %u; acquiring DTMF",
								UnitData[unit].confID,
								UnitData[unit].callRef);
							XWndStatusUpdate(unit, szText);
						}
						break;
					}

					// reset command string
					pKeys = szKeys;
					szKeys[0] = 0;
				}
				else
				{
					// append key to command string
					*pKeys++ = iKey;
					*pKeys = 0;
				}
			}
			else
			{
				// here if BfvToneGet failed. complain
				BfvErrorMessage(lp, &toneArgs.res, &bterr);
				sprintf(szText,
					"IncomingChannelThread%u BfvToneGet failure %s",
					unit,
					bterr.long_msg);
				XWndMsgBox(szText);

				// if BfvToneGet failed for reason of no loop current (caller disconnected)
				if (toneArgs.res.status == BT_STATUS_ERROR_HANGUP &&
					toneArgs.res.line_status == HNG_NO_LOOP_CURRENT)
				{
					fNoLoopCurrent = TRUE;
					break;
				}
			}
		}

		//-----------------------------------------------------------------
		// if still a party to conference, drop
		//-----------------------------------------------------------------
		if (UnitData[unit].callRef)
		{
			BT_ZERO(partyArgs);
			partyArgs.confID = UnitData[unit].confID;
			partyArgs.callRef = UnitData[unit].callRef;
			iStatus = BfvConferencePartyDrop(lp, &partyArgs);
			UnitData[unit].callRef = 0;
			if (iStatus != BT_STATUS_OK)
			{
				sprintf(szText,
					"IncomingChannelThread%u BfvConferencePartyDrop failure %s (%d)",
					unit,
					ConferenceErrorName(partyArgs.res.line_status),
					partyArgs.res.line_status);
				XWndMsgBox(szText);
			}
		}

		//-----------------------------------------------------------------
		// if not terminating call because of no loop current
		//-----------------------------------------------------------------
		if (!fNoLoopCurrent)
		{
			XWndStatusUpdate(unit, "playing bye.vox");
			VoicePlay(lp, unit, "bye.vox");

			//-----------------------------------------------------------------
			// terminate incoming call
			//-----------------------------------------------------------------
			XWndStatusUpdate(unit, "terminating call");
			BT_ZERO(telArgs);
			BfvLineTerminateCall(lp, &telArgs);
			if (telArgs.res.status != BT_STATUS_OK)
			{
				BfvErrorMessage(lp, &telArgs.res, &bterr);
				sprintf(szText,
					"IncomingChannelThread%u BfvLineTerminateCall failure %s",
					unit,
					bterr.long_msg);
				XWndMsgBox(szText);
			}
		}

		//-----------------------------------------------------------------
		// reset channel
		//-----------------------------------------------------------------
		XWndStatusUpdate(unit, "reseting");
		BT_ZERO(lineArgs);
		lineArgs.config_file_name = "btcall.cfg";
		lineArgs.skip_teleph = 1;
		if (BfvLineReset(lp, &lineArgs) < 0)
		{
			BfvErrorMessage(lp, &lineArgs.res, &bterr);
			sprintf(szText,
				"IncomingChannelThread%u BfvLineReset failure %s",
				unit,
				bterr.long_msg);
			XWndMsgBox(szText);
			return(NULL);
		}
	}
}

//-----------------------------------------------------------------------
// Procedure VoiceConferenceInit
//
// Enable debug trace.
// Initialize conference administrator.
// Spawn one thread for each incoming channel.
// Initialize loopback call originate handler.
//-----------------------------------------------------------------------
#ifdef WIN32
void VoiceConferenceInit(void)
#else
int main(int argc, char **argv)
#endif
{
	register i;
	int iStatus;
	unsigned unit;
	pthread_t threadID;
#ifdef UNIXWARE
	pthread_attr_t threadAttr;
#endif
	struct args_conf_admin confAdmin;
#ifdef WIN32
	int argc;
	char argv[8][32];
	argc = sscanf(GetCommandLine(),
		"%s %s %s %s %s %s %s %s",
		argv[0],
		argv[1],
		argv[2],
		argv[3],
		argv[4],
		argv[5],
		argv[6],
		argv[7]);
#endif

	for (i = 1; i < argc; i++)
		if (!stricmp(argv[i], "-v"))
			fConfVerbose = TRUE;
		else if (!strnicmp(argv[i], "-b", 2))
			sscanf(&argv[i][2], "%u", &busyOutMax);
		else if (!strnicmp(argv[i], "-c", 2))
			sscanf(&argv[i][2], "%u", &unitNum);
		else
		{
#ifdef WIN32
			AfxMessageBox(szUsage);
#else
			printf("%s\n", szUsage);
#endif
			exit(1);
		}

	// create debug log file semaphore
	if (sem_init(&semDebug, SEM_SHARE_LOCAL, 1) == -1)
	{
		XWndMsgBox("VoiceConferenceInit sem_init failure");
		exit(1);
	}

	// enable debug display
	if (fConfVerbose)
	{
		FILE* fp;
		if (fp = fopen("vcdemo.log", "w"))
		{
			fprintf(fp, "simpleconf startup, BusyOutMax %u, ACC API %s\n", busyOutMax, ACC_API_VERSION);
			fclose(fp);
		}
		BfvDebugModeSet(DEBUG_ALL);
		BfvDebugFuncSet(DebugFunc);
		BfvConferenceDebugModeSet(CONF_TRACE_LEVEL_HIGH, 0xFFFFFFFF);
		BfvConferenceDebugFuncSet(DebugFunc);
	}

	// startup conference administrator
	BT_ZERO(confAdmin);
	confAdmin.rendezvousKey = 1;
	confAdmin.minID = 1;
	confAdmin.maxID = 100;
	iStatus = BfvConferenceAdminInitialize(&confAdmin);
	if (iStatus != BT_STATUS_OK)
	{
		XWndMsgBox("VoiceConferenceInit BfvConferenceAdminInitialize failure");
		exit(1);
	}

#ifndef WIN32
	{
		sem_t semXWndInit;
		if (sem_init(&semXWndInit, SEM_SHARE_LOCAL, 0) == -1)
		{
			printf("VoiceConferenceInit sem_init failure\n");
			exit(1);
		}
#ifdef UNIXWARE
		memset(&threadAttr, 0, sizeof(pthread_attr_t));
		threadAttr.__pt_attr_stacksize = 1024 * 1024;
		pthread_create(&threadID, &threadAttr, XWndProc, (void *) &semXWndInit);
#else
		pthread_create(&threadID, NULL, XWndProc, (void *) &semXWndInit);
#endif
		sem_wait(&semXWndInit);
		sem_destroy(&semXWndInit);
	}
#endif

	// startup incoming channel threads
	for (unit = 0; unit < unitNum; unit++)
	{
		UnitData[unit].span = unit / SLOTS_PER_T1;
		UnitData[unit].timeSlot = unit % SLOTS_PER_T1;

#ifdef UNIXWARE
		memset(&threadAttr, 0, sizeof(pthread_attr_t));
		threadAttr.__pt_attr_stacksize = 1024 * 1024;
		iStatus = pthread_create(&threadID, &threadAttr, IncomingChannelThread, (void *) unit);
#else
		iStatus = pthread_create(&threadID, NULL, IncomingChannelThread, (void *) unit);
#endif
		printf("iStatus %d threadID %d\n", iStatus, threadID);
	}

	// startup outgoing channel
	CallInit();

#ifndef WIN32
	// wait forever
	while (1)
		sleep(60);
#endif
}
