/*
 * bftdump.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/* Used to list the contents of a T.434 BFT file. */

/* For Windows only:
   If an application uses a DLL version of the Bfv API library, under some
   conditions the application and the Bfv API library may link to different
   versions or instances of the compiler-supplied C runtime (CRT) libraries.
   This creates incompatibilities between file pointers or allocated memory
   among the application and the Bfv API library.

   Under these conditions, the application must use the Bfv API supplied _dll_
   versions of the file I/O and memory allocation functions if the pointer or
   memory is passed to/from the application and the library.

   The Bfv API DLL libraries links with MSVCRT.DLL. When using the Visual
   Studio mechanism to compile for debugging, applications may be linked with
   MSVCRTD.DLL. When using Visual Studio.NET, applications link with
   MSVCR71.DLL. When using the DLLS version of the Bfv API library, it also
   links with MVCRT.DLL, but it is a different instance of the CRT library.
*/

#include <string.h>
#ifndef T434_STANDALONE
#include "btlib.h"
#else
#include <stdio.h>
#include "t434.h"
#endif

int print_func BT_ARGS ((unsigned tag, unsigned subtag, long length,
                         long data,char *arg));

void usage()
{
    fprintf (stderr, "usage: bftdump <t434file>\n");
    exit(1);
}

int main (int argc, char **argv)
{
    FILE *fp;
    struct args_t434 args;
    struct args_debug_mode args_debug;
#ifndef T434_STANDALONE
    BTERR bterr;
    
    BT_API_SET_VER();
#endif
    
    if (argc != 2)
        usage();

#ifndef USES_FAT_FILESYSTEM
#ifdef MILL_DLL
    fp = _dll_fopen(argv[1],"r");
#else
    fp = fopen(argv[1],"r");
#endif
#else
#ifdef MILL_DLL
    fp = _dll_fopen(argv[1],"rb");
#else
    fp = fopen(argv[1],"rb");
#endif
#endif
    if (fp == NULL)
    {
        fprintf(stderr,"Error opening %s\n",argv[1]);
        exit(1);
    }
    
#ifndef T434_STANDALONE
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
#endif
    
#ifndef T434_STANDALONE
    BT_ZERO(args);
#else
    memset(&args,0,sizeof(args));
#endif
    args.fp = fp;
    args.parse_func = print_func;
    if (BfvT434ParseFile (&args))
    {
#ifndef T434_STANDALONE
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr, "BfvT434ParseFile: %s\n",bterr.long_msg);
#else
        fprintf (stderr, "BfvT434ParseFile error\n");
#endif
    }
#if 0
    printf("File parsed was of version %d\n",args.t434_version);
#endif
    
#ifdef MILL_DLL
    _dll_fclose(fp);
#else
    fclose(fp);
#endif
    
    return(0);
}

print_func (unsigned tag, unsigned subtag,long length,long data,char *arg)
{
    char *ctmp;
    int tmp;
    int i,j;
    struct gentm *gt;
    struct args_t434 args;
    
#ifndef T434_STANDALONE
    BT_ZERO(args);
#else
    memset(&args,0,sizeof(args));
#endif
    args.tag = tag;
    /* T434 tag value and name */
    ctmp = BfvT434TagName(&args);
    printf("Tag %02X %s ",tag,ctmp);
    i = strlen(ctmp);
    args.tag = subtag;
    /* Subtag (ASN1 base tag) name */
    ctmp = BfvT434TagName(&args);
    if (i + (int)strlen(ctmp) > 60)
        printf("\n\t");
    printf("%s ",ctmp);
    
    if (length == LENGTH_INDEFINITE)
    {
        printf ("Length indefinite.\n");
        printf ("Cannot handle, aborting.\n");
        return(1);
    }
    
    switch (subtag)
    {
      case ASN1_Integer:
        printf ("%ld %08lx\n", data, data);
        break;
      case ASN1_GeneralizedTime:
        gt = (struct gentm *)data;
        /* Get formatted date & time, but skip day of week, which will
           be wrong, since the tm_mday field is not set. */
        ctmp = asctime (&gt->tm) + 4;
        tmp = strlen(ctmp);
        if (ctmp[tmp - 1] == '\n')
            ctmp[tmp - 1] = '\0';
        printf ("\n\t%s ", ctmp);
        
        if (gt->time_diff_mins == GEN_TIME_DIFF_LOCAL)
            printf("Local\n");
        else if (gt->time_diff_mins == 0)
            printf("UTC\n");
        else
        {
            i = gt->time_diff_mins % 60;
            if (i < 0)
                i = -i;
            printf("Local %+03d:%02d\n",gt->time_diff_mins / 60,i);
        }
        break;
      case ASN1_GraphicString:
        if (length <= 60)
            printf ("\"%s\"\n", (char *)data);
        else
            printf ("Length %ld\n", length);
        break;
      case ASN1_BitString:
        if (length > 60)
        {
            printf ("Length %ld\n", length);
            break;
        }
        ctmp = (char *)data;
        for (i = 0; i < length; i++)
        {
            for (j = 7; j >= 0; j--)
                putchar((ctmp[i] & (1 << j)) ? '1' : '0');
        }
        putchar('\n');
        break;
      case ASN1_ObjectIdentifier:
      case ASN1_OctetString:
        if (length > 60)
        {
            printf ("Length %ld\n", length);
            break;
        }
        printf ("bytes ");
        ctmp = (char *)data;
        for (i = 0; i < length; i++)
            printf("%02X ",ctmp[i]);
        putchar('\n');
        break;
      case ASN1_OctetString | ASN1_IDENTIFIER_CONSTRUCTED:
      case ASN1_GraphicString | ASN1_IDENTIFIER_CONSTRUCTED:
      case ASN1_BitString | ASN1_IDENTIFIER_CONSTRUCTED:
      case ASN1_USER_DEFINED:
        printf ("Length %ld\n", length);
        break;
      case ASN1_END_OF_SEQ:
        printf("\n");
        break;
      default:
        printf ("Unknown Subtag type %X, Length %ld\n", subtag, length);
        break;
    }
    return(0);
}
