/*
 * bftstrip.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/* Used to extract a data file from a T.434 BFT file. */

/* For Windows only:
   If an application uses a DLL version of the Bfv API library, under some
   conditions the application and the Bfv API library may link to different
   versions or instances of the compiler-supplied C runtime (CRT) libraries.
   This creates incompatibilities between file pointers or allocated memory
   among the application and the Bfv API library.

   Under these conditions, the application must use the Bfv API supplied _dll_
   versions of the file I/O and memory allocation functions if the pointer or
   memory is passed to/from the application and the library.

   The Bfv API DLL libraries links with MSVCRT.DLL. When using the Visual
   Studio mechanism to compile for debugging, applications may be linked with
   MSVCRTD.DLL. When using Visual Studio.NET, applications link with
   MSVCR71.DLL. When using the DLLS version of the Bfv API library, it also
   links with MVCRT.DLL, but it is a different instance of the CRT library.
*/

#include <string.h>
#ifndef T434_STANDALONE
#include "btlib.h"
#else
#include <stdio.h>
#include "t434.h"
#endif

int extract_func BT_ARGS((unsigned tag, unsigned subtag, long length,
                          long data, char *arg));

char filename[512];
int file_number;

/* Define PARSE_BT_OLD_COMPAT to assume the data_file_content is in the format
   used by the old Brooktrout code, and many other implementations.
   In that format, data_file_content contains the actual data.
   Otherwise, the actual data is contained within tagged fields and must
   be parsed. */
#if 0
#define PARSE_BT_OLD_COMPAT
#endif

void usage()
{
    fprintf (stderr, "usage: bftstrip <t434file>\n");
    exit(1);
}

int main (int argc, char **argv)
{
    FILE *fp;
    char *infname;
    struct args_t434 args;
    struct args_debug_mode args_debug;
#ifndef T434_STANDALONE
    BTERR bterr;
    
    BT_API_SET_VER();
#endif
    
    if (argc != 2)
        usage();

    infname = argv[1];
    
#ifndef USES_FAT_FILESYSTEM
#ifdef MILL_DLL
    fp = _dll_fopen(infname,"r");
#else
    fp = fopen(infname,"r");
#endif
#else
#ifdef MILL_DLL
    fp = _dll_fopen(infname,"rb");
#else
    fp = fopen(infname,"rb");
#endif
#endif
    if (fp == NULL)
    {
        fprintf(stderr,"Error opening %s\n",infname);
        exit(1);
    }
    
#ifndef T434_STANDALONE
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
#endif
    
#ifndef T434_STANDALONE
    BT_ZERO(args);
#else
    memset(&args,0,sizeof(args));
#endif
    args.fp = fp;
    args.parse_func = extract_func;
#ifdef PARSE_BT_OLD_COMPAT
    args.raw_content = 1;
#endif
    if (BfvT434ParseFile (&args))
    {
#ifndef T434_STANDALONE
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr, "BfvT434ParseFile: %s\n",bterr.long_msg);
#else
        fprintf (stderr, "BfvT434ParseFile error\n");
#endif
    }
    
#ifdef MILL_DLL
    _dll_fclose(fp);
#else
    fclose(fp);
#endif
    
    return(0);
}

extract_func(unsigned tag,unsigned subtag,long length,long data,char *arg)
{
    FILE *infp, *outfp;
    char buf[512];
    int n;
    char *ctmp,*fname;
    struct args_t434 args;
    
    if (length == LENGTH_INDEFINITE)
    {
        printf ("Indefinite length found for tag %02X.\n",tag);
        printf ("Cannot handle, aborting.\n");
        return(1);
    }
    
    switch (tag)
    {
      case T434_filename:
        if (subtag != ASN1_GraphicString)
            break;
        /* We look at only the first first filename component in sequence */
        if (filename[0])
            break;
        /* Reduce the filename to the last path component, so that we
           will only write to the current directory. Only the last
           string in the sequence will be available in filename. */
        /* Relies on BfvT434ParseFile 0-terminating the strings before
           calling this user-supplied function. */
        fname = (char *)data;
        if (ctmp = strrchr(fname,'/'))
            fname = ctmp + 1;
#ifdef USES_FAT_FILESYSTEM
        if (ctmp = strrchr(fname,'\\'))
            fname = ctmp + 1;
        if (fname[1] == ':')
            fname += 2;
#endif
        strcpy(filename,fname);
        break;
      case T434_data_file_content:
        infp = (FILE *)data;
        if (!filename[0])
            sprintf(filename,"tmp%d.bft",file_number);
        file_number++;
#ifndef USES_FAT_FILESYSTEM
#ifdef MILL_DLL
        outfp = _dll_fopen(filename,"w");
#else
        outfp = fopen(filename,"w");
#endif
#else
#ifdef MILL_DLL
        outfp = _dll_fopen(filename,"wb");
#else
        outfp = fopen(filename,"wb");
#endif
#endif
        if (outfp == NULL)
        {
            fprintf(stderr,"Error opening %s\n",filename);
            return(1);
        }
        
#ifndef T434_STANDALONE
        BT_ZERO(args);
#else
        memset(&args,0,sizeof(args));
#endif
        args.tag = subtag;
        printf("Writing file %s of length %ld. Data type is %s.\n",
               filename,length,BfvT434TagName(&args));
        
        n = sizeof(buf);
        while (length > 0)
        {
            if (n > length)
                n = length;
#ifdef MILL_DLL
            if (_dll_fread(buf,n,1,infp) != 1)
#else
            if (fread(buf,n,1,infp) != 1)
#endif
            {
                fprintf(stderr,"File read error\n");
#ifdef MILL_DLL
                _dll_fclose(outfp);
#else
                fclose(outfp);
#endif
                return(1);
            }
#ifdef MILL_DLL
            if (_dll_fwrite(buf,n,1,outfp) != 1)
#else
            if (fwrite(buf,n,1,outfp) != 1)
#endif
            {
                fprintf(stderr,"File write error\n");
#ifdef MILL_DLL
                _dll_fclose(outfp);
#else
                fclose(outfp);
#endif
                return(1);
            }
            length -= n;
        }
#ifdef MILL_DLL
        _dll_fclose(outfp);
#else
        fclose(outfp);
#endif
        filename[0] = '\0';
        break;
    }
    return(0);
}
