/*
 * dlfax.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The dlfax program uses the highest level infopkt sending routines to
send facsimiles and the dialing database functions to implement
dialing restrictions. If you include the -c argument, the application
uses the dialing restrictions of the specified country. If you include
the -l argument, the application only lists the contents of the dialing
database; it does not dial.
***************************************************************************/

#include "btlib.h"
#include <string.h>

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

#define DIAL_DATABASE "dialdb"

void process_results BT_ARGS((BTLINE *lp,FAX_RES *fres,long duration));
void process_aoc BT_ARGS((BTLINE *lp));
int get_dl_restrictions BT_ARGS((BTLINE *lp));

void
usage ()
{
    fprintf (stderr, "usage: dlfax -l  or  dlfax [options] infopktfile\n");
    fprintf (stderr, "  -l          list contents of dialing database\n");
    fprintf (stderr, "  -s num      send to given phone number\n");
    fprintf (stderr, "  -u unitnum  use specified unit number\n");
    fprintf (stderr, "  -c ccode    use country code for dialing restrictions\n");
    fprintf (stderr, "One of -l or -s is required.\n");
    exit (1);
}

char *local_id = "Brooktrout Fax";
char *snum;
int unit;
int done;
unsigned ccode;

int list_func(struct dialdb *dbentry, char *arg)
{
    static int first = 1;
    char *ttmp;
    time_t time_tmp;
    
    if (first)
    {
        first = 0;
        printf("Contents of Dial Database:\n");
        printf("==========================\n");
    }
    
    if (dbentry->phone_num[0] == 0)
        printf("(blank entry)\n");
    else
    {
        printf("Phone: %s, Flags: %X\n",
               dbentry->phone_num,dbentry->flags);
        printf("\tWrong count: %d, Unsuccessful count: %d, Total count: %d\n",
               dbentry->wrong_count,dbentry->unsucc_count,
               dbentry->total_count);
        time_tmp = dbentry->next_dial_time;
        printf("\tNext valid dial time: %s",
               (dbentry->next_dial_time != (unsigned)-1) ?
               ctime(&time_tmp) : "Never\n");
        if (dbentry->reference_time)
        {
            ttmp = ctime(&time_tmp);
            if (ttmp[strlen(ttmp) - 1] == '\n')
                ttmp[strlen(ttmp) - 1] = '\0';
        }
        else
            ttmp = "N/A";
        printf("\tTime of entry: %s\n",ttmp);
    }
    
    return(0);
}

void list_dialdb()
{
    struct args_telephone args;
    
    BT_ZERO(args);
    args.db_file = DIAL_DATABASE;
    args.db_func = list_func;
    if (BfvDialDBList(&args))
    {
        printf("Dialing database file not found.\n");
        exit(1);
    }
}


int
main (int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    BTLINE *lp;
    char *name;
    void dowork();
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    while ((c = getopt (argc, argv, "ls:u:c:")) != EOF)
    {
        switch (c)
        {
          case 'l':
            list_dialdb();
            exit(0);
          case 'u':
            unit = atoi (optarg);
            break;
          case 's':
            snum = optarg;
            break;
          case 'c':
            if (sscanf(optarg,"%x",&ccode) != 1)
                usage();
            if (ccode == 0)
            {
                fprintf(stderr,
                        "Illegal country code. Use number from ccode.h\n");
                exit(1);
            }
            break;
          default:
            usage ();
        }
    }
    
    if (optind == argc || snum == NULL)
        usage ();

    name = argv[optind++];
    
    if (optind != argc)
        usage ();

    BT_ZERO(args);
    args.unit = unit;
    if ((lp = BfvLineAttach (&args)) == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
        exit (1);
    }
    
    dowork(lp,name);
    
    return(0);
}

void
dowork(BTLINE *lp, char *name)
{
    struct infopkt_stream *ips;
    BTERR bterr;
    long start_time;
    static char *reason_str[] = {
        "No reason (bug?).",
        "Delay after 'wrong' result.",
        "Delay after 'unsuccessful' result.",
        "Blacklisted in advance.",
        "Blacklisted manually.",
        "Blacklisted due to excessive 'wrong' results.",
        "Blacklisted due to excessive 'unsuccessful' results.",
        "Blacklisted due to excessive total bad results.",
        "Delay after too many 'wrong' results per time period.",
        "Delay after too many 'unsuccessful' results per time period.",
        "Delay after too many total bad results per time period.",
        "Delay after 'unsuccessful' result on trunk line.",
    };
#define NUM_REASONS (sizeof(reason_str) / sizeof(*reason_str))
    char *reduced_num, *ctmp;
    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_fax args_fax;
    struct args_infopkt args_infopkt;
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (lp, &args_admin) < 0)
    {
        BfvErrorMessage(lp,&args_admin.res,&bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                 bterr.long_msg,args_admin.reset_status);
        exit(1);
    }
    
    /* Check for digital configuration error */
    if (args_admin.reset_status & RST_DIG_CONFIG_UNREADABLE)
    {
        fprintf (stderr, "Telephony/Digital configuration file not found.\n");
        exit(1);
    }
    
    if (!LINE_HAS_CAP(lp,CAP_FAX_XMIT))
    {
        fprintf (stderr, "unit %d does not support fax\n", LINE_UNIT_NUM(lp));
        exit(1);
    }
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = name;
    args_infopkt.fmode = "r";
    if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr, "BfvInfopktOpen: %s\n",bterr.long_msg);
        exit(1);
    }
    
    /* Allow use of dl restrictions for another country for testing. */
    if (ccode && get_dl_restrictions(lp))
    {
        fprintf(stderr,"Error looking up country in config file.\n");
        exit(1);
    }
    
    /* Reduce number by removing everything up to the last w. */
    reduced_num = snum;
    if (ctmp = strchr(reduced_num,'w'))
        reduced_num = ctmp + 1;
    if (ctmp = strchr(reduced_num,'W'))
        reduced_num = ctmp + 1;

#if 1

    /* Use high level call to do dialing & call progress and dialing
       database operations. */
    BT_ZERO(args_tel);
    args_tel.phonenum = snum;
    args_tel.call_protocol_code = CALL_PROTOCOL_FAX;
    args_tel.func = NULL;
    args_tel.db_file = DIAL_DATABASE;
    args_tel.raw_number = reduced_num;
    BfvLineOrigCallDB (lp, &args_tel);
    if (args_tel.res.status == BT_STATUS_ERROR &&
        args_tel.res.line_status == APIERR_TOOSOON)
    {
        printf("Can't call phone number for another %ld seconds.\n",
               args_tel.wait_time);
        if (args_tel.reason >= 0 && args_tel.reason < NUM_REASONS)
            printf("Reason: %s\n",reason_str[args_tel.reason]);
        else
            printf("Reason: Delay or blacklist for an unknown reason (bug?).\n");
        goto got_err_1;
    }

#else
    
    /* Use low level calls to do dialing & call progress and dialing
       database operations. */
    
    /* Check to see if phone number can be (re)dialed at this time. */
    BT_ZERO(args_tel);
    args_tel.db_file = DIAL_DATABASE;
    args_tel.raw_number = reduced_num;
    if (BfvDialDBCheck (lp, &args_tel))
    {
        fprintf(stderr,"Error checking dialing database.\n");
        exit(1);
    }
    if (args_tel.wait_time != 0)
    {
        printf("Can't call phone number for another %ld seconds.\n",
               args_tel.wait_time);
        if (args_tel.reason >= 0 && args_tel.reason < NUM_REASONS)
            printf("Reason: %s\n",reason_str[args_tel.reason]);
        else
            printf("Reason: Delay or blacklist for an unknown reason (bug?).\n");
        goto got_err_1;
    }
    
    args_tel.phonenum = snum;
    args_tel.call_protocol_code = CALL_PROTOCOL_FAX;
    BfvLineOriginateCall (lp, &args_tel);
    
    /* Update dialing database information for this phone number */
    args_tel.raw_number = reduced_num;
    args_tel.db_res = &args_tel.res;
    if (BfvDialDBUpdate(lp, DIAL_DATABASE, reduced_num, &cres.res))
    {
        fprintf(stderr,"Error updating dialing database.\n");
        exit(1);
    }
    
#endif
    
    if (args_tel.res.status != BT_STATUS_OK || 
        LINE_STATE(lp) != LINE_STATE_CONNECTED)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("BfvLineOriginateCall: %s\n",bterr.long_msg);
        goto got_err_2;
    }
    
    /* Start timing after we detect Fax answer tone */
    start_time = time(NULL);
    
    BT_ZERO(args_fax);
    args_fax.s_ips = ips;
    args_fax.local_id = local_id;
    BfvFaxSend (lp, &args_fax);
    if (args_fax.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_fax.res,&bterr);
        printf ("BfvFaxSend: %s\n",bterr.long_msg);
        goto got_err;
    }
    
got_err:
    process_results(lp,&args_fax.fax_res,time(NULL) - start_time);
    
got_err_2:
    BT_ZERO(args_tel);
    BfvLineTerminateCall (lp, &args_tel);

    process_aoc(lp);

got_err_1:
    BT_ZERO(args_infopkt);
    args_infopkt.ips = ips;
    BfvInfopktClose (&args_infopkt);
    
    BT_ZERO(args_admin);
    BfvLineDetach (lp, &args_admin);
    
    exit(0);
}

/* Uses internal API knowledge. For testing purposes only. No developer
   should modify the country parameters in this way. */
/* Read the country information for the country specified by ccode, and
   replace the the dialing restriction parameters in effect with those of
   the new country. */
int
get_dl_restrictions(BTLINE *lp)
{
    struct country_phone_info country, *phone;
    FILE *fp;
    unsigned short phone_size, country_code;
    short ncountries;
#if 0
    unsigned char fifo[128],*p;
#endif
    
#ifndef USES_FAT_FILESYSTEM
    fp = fopen(LINE_CONFIG_STRUCT(lp)->bt_cparm,"r");
#else
    fp = fopen(LINE_CONFIG_STRUCT(lp)->bt_cparm,"rb");
#endif
    if (fp == NULL)
    {
        fprintf(stderr,"File open error.\n");
        return(1);
    }
    
    if (_bfv_fread_short((char *)&phone_size,
                         sizeof(phone_size), 1, fp) != 1 ||
        _bfv_fread_short((char *)&ncountries,
                         sizeof(ncountries), 1, fp) != 1)
    {
        fprintf(stderr,"File read error.\n");
        fclose(fp);
        return(1);
    }
    
    if (phone_size != sizeof(struct country_phone_info))
    {
        fprintf(stderr,"Wrong phone structure size.\n");
        fclose(fp);
        return(1);
    }
    
    while (ncountries--)
    {
        if (_bfv_fread_short((char *)&country_code,
                             sizeof(country_code), 1, fp) != 1)
        {
            fprintf(stderr,"File read error.\n");
            fclose(fp);
            return(1);
        }
        if (country_code != ccode)
        {
            if (fseek(fp, (long)sizeof(struct country_phone_info),
                      SEEK_CUR) < 0)
            {
                fprintf(stderr,"File read error.\n");
                fclose(fp);
                return(1);
            }
            continue;
        }
        if (fread((char *)&country, sizeof(struct country_phone_info), 1, fp)
            != 1)
        {
            fprintf(stderr,"File read error.\n");
            fclose(fp);
            return(1);
        }
        break;
    }
    if (ncountries < 0)
    {
        fprintf(stderr,"Country not found. Use number from ccode.h\n");
        fclose(fp);
        return(1);
    }
    
    fclose(fp);
    
#ifdef BT_BIG_ENDIAN
    BYTE_SWAP_SHORT(country.dl_wr_lmt_time);
    BYTE_SWAP_SHORT(country.dl_uns_lmt_time);
    BYTE_SWAP_SHORT(country.dl_min_wrong_first);
    BYTE_SWAP_SHORT(country.dl_min_wrong);
    BYTE_SWAP_SHORT(country.dl_min_unsucc_first);
    BYTE_SWAP_SHORT(country.dl_min_unsucc);
#endif
    
    phone = LINE_PHONE_STRUCT(lp);
    
    /* Set country code so any country-specific restrictions not implemented
       by parameters will take effect, usually pre-blacklisted numbers. */
    phone->country_code = country.country_code;
    /* Dialing restriction parameters */
    phone->dl_configuration = country.dl_configuration;
    phone->dl_wrong_limit = country.dl_wrong_limit;
    phone->dl_unsucc_limit = country.dl_unsucc_limit;
    phone->dl_sum_limit = country.dl_sum_limit;
    phone->dl_max_limit = country.dl_max_limit;
    phone->dl_wr_lmt_time = country.dl_wr_lmt_time;
    phone->dl_uns_lmt_time = country.dl_uns_lmt_time;
    phone->dl_min_wrong_first = country.dl_min_wrong_first;
    phone->dl_min_wrong = country.dl_min_wrong;
    phone->dl_min_unsucc_first = country.dl_min_unsucc_first;
    phone->dl_min_unsucc = country.dl_min_unsucc;
    
#if 0
    /* Min_on_hook is not a dialing restriction parameter per se, but
       may be used as such. Must actually send the parameter to the channel
       in order to take effect. May not truly be effective though, since
       the parameter is set back to the value for the country in the
       btcall.cfg after reset. */
    p = fifo;
    *p++ = phone->dtone_len;
    *p++ = phone->dtone_timeout_lowbyte;
    *p++ = phone->loopcur_len;
    *p++ = phone->loopcur_timeout;
    *p++ = phone->loopcur_debounce;
    *p++ = phone->dtone_timeout_highbyte;
    *p++ = phone->dial_tone_min;
    *p++ = phone->drop_out_time;
    *p++ = country.min_on_hook;
    BfvLineSendCmd (lp, TR_DIAL_PARAM, p-fifo, p-fifo, fifo);
#endif
    
    printf("Using dialing restrictions for country code %X.\n",ccode);
    return(0);
}

char *
conf_vals(int conf)
{
    switch (conf)
    {
      case FSK_MCF_1:
      case FSK_MCF_2:
        return("MCF");
      case FSK_RTN_1:
      case FSK_RTN_2:
        return("RTN");
      case FSK_RTP_1:
      case FSK_RTP_2:
        return("RTP");
      case FSK_PIN_1:
      case FSK_PIN_2:
        return("PIN");
      case FSK_PIP_1:
      case FSK_PIP_2:
        return("PIP");
      default:
        return("?");
    }
}

void
process_results(BTLINE *lp, FAX_RES *fres, long duration)
{
    static char *bft_names[] =
        {NULL,"T.434","BTM",NULL,"DTM",NULL,NULL,NULL,"EDIFACT"};
    static int bit_rates[] =
        {2400,4800,7200,9600,7200,9600,12000,14400,12000,14400,
         2400,4800,7200,9600,12000,14400,16800,19200,
         21600,24000,26400,28800,31200,33600};
    static int ecm_vals[] = {0,256,64};
    static char *fmt_names[2][2] = {{"MH","MR"},{"MMR","MMR"}};
    static char *res_names[] =
        {"200Hx100V(Norm)",
         "200Hx200V(Fine)",
         "200Hx400V",
         "300Hx300V",
         "400Hx400V",
         "600Hx600V", 
         "1200Hx1200V",
         "300Hx600V",
         "400Hx800V",
         "600Hx1200V",
         "100Hx100V",
        };
    static char *width_names[] = {"A4","B4","A3"};
    PAGE_RES *pageres;
    PAGE_RES *temp;
    int jj;
    
    /* loop down the page res list */
    pageres = fres->reslist_head;
    jj = 1;
    while (pageres)
    {
        printf("Page %2d:  ", jj);
        if (pageres->bft_type)
            printf("BFT type %s  ",bft_names[pageres->bft_type]);
        else
        {
            if (pageres->eff_page_type)
            {
                printf("EFF Page: ");
                if (pageres->eff_page_type & BT_EFF_JPEG_ENABLE)
                {
                    printf("JPEG(");
                    if (pageres->eff_page_type & BT_EFF_JPEG_FULLCOLOR)
                        printf("full color, ");
                    else
                        printf("grayscale, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_DEFAULT_TABLES)
                        printf("default tbls, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_12BIT)
                        printf("12bit, ");
                    else
                        printf("8bit, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_NOSUBSAMPLING)
                        printf("1:1:1, ");
                    else
                        printf("4:1:1, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_CUSTOM_ILLUMINANT)
                        printf("cust illum, ");
                    else
                        printf("std illum, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_CUSTOM_GAMUT)
                        printf("cust gamut");
                    else
                        printf("std gamut");
                    printf(")\n          ");
                }
                if (pageres->eff_page_type & BT_EFF_JBIG)
                {
                    printf("JBIG(");
                    if (pageres->eff_page_type & BT_EFF_JBIG_L0)
                        printf("L0");
                    else
                        printf("std");
                    printf(") ");
                }
                if (pageres->eff_page_type & BT_EFF_LOSSLESS)
                    printf("LOSSLESS COLOR ");
                if (pageres->eff_page_type & BT_EFF_PLANE_INTERLEAVE)
                    printf("PLANE_INTERLEAVE ");
                if (pageres->eff_page_type & BT_EFF_MRC_MASK)
                    printf("MRC(%X) ",
                           (pageres->eff_page_type & BT_EFF_MRC_MASK) >> 4);
                if (pageres->eff_page_type & BT_EFF_PAGE_LENGTH_STRIP)
                    printf("PAGE_LEN_STRIP ");
            }
            else
            {
                if (pageres->direction)
                    printf("ASCII %d  ",pageres->ascii_bytes);
                printf("BdLn %d  TotLn %d  ",
                       pageres->bad_lines, pageres->total_lines);
            }
            printf("Res %s  ",res_names[pageres->resolution]);
            printf("Width %s  ",width_names[pageres->width]);
        }
        printf("Bps %u\n",bit_rates[pageres->bit_rate]);
        
        printf("          ConfVal %s  SigLev %04X  LineNoise %04X",
               conf_vals(pageres->confirm_value),
               pageres->sig_level,pageres->line_noise);
        if (!pageres->direction)
            printf("  SigQual %04X",pageres->sig_quality);
        printf("\n");
        
        temp = pageres;
        pageres = pageres->next;
#ifdef MILL_DLL
        _dll_free(temp);
#else
        free(temp);
#endif
        jj++;
    }
    printf("Duration %ld secs  ",duration);
    printf("ECM %d  ",ecm_vals[LINE_DCS(lp)->Error_Correction_Mode]);
    printf("Line Format %s",
           fmt_names[LINE_DCS(lp)->MMR][LINE_DCS(lp)->TwoD_Coding]);
    printf("\n");
}

void
process_aoc(BTLINE *lp)
{
    AOC_INFO *aoc = LINE_AOC_INFO(lp);
        
    if (aoc == NULL || aoc->status == BT_AOC_UNAVAILABLE)
        printf("Charge information unavailable.\n");
    else if (aoc->status == BT_AOC_FREE)
        printf("Charge information: FREE.\n");
    else if (aoc->status == BT_AOC_UNITS)
        printf("Charge information: Unit type %d, value %ld\n",
               aoc->charge_info,aoc->charge);
    else if (aoc->status == BT_AOC_CURRENCY)
    {
        static char *currency_names[] = {
            "1/1000","1/100","1/10","1","10","100","1000" };
#define NUM_CURRENCIES (sizeof(currency_names) / sizeof(*currency_names))
        char *cur_name;
            
        cur_name = (0 <= aoc->charge_info &&
                    aoc->charge_info < NUM_CURRENCIES) ?
            currency_names[aoc->charge_info] :
            "Invalid currency";
        printf("Charge information: Currency multiplier %s (%d), value %ld\n",
               cur_name,aoc->charge_info,aoc->charge);
    }
    else
        printf("Charge information: Bad status value %d\n",
               aoc->status);
}
