/*
 * eccllvoice.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 *
 * ECC Low-Level Voice sample application.
 *
 * This program is a derivative of voice.c.  Perform a diff to see how an
 * application can be modified to use low-level ECC API calls.
 *
 * The eccllvoice program uses the speech-infopkt-file recording and playing
 * routines to record and play speech. Recording begins when you start
 * speaking. Recording continues for a maximum of ten seconds or the
 * time specified in the -n option.
 */

#include <string.h>
#include <ctype.h>
#include "btlib.h"

int play_func BT_ARGS((BTLINE *lp,char *arg));
void process_aoc BT_ARGS((BTLINE *lp));
void process_call_info BT_ARGS((CALL_RES *call_res));
void clear_call BT_ARGS((BTLINE *lp));
void check_line_state BT_ARGS((BTLINE *lp));

void
usage ()
{
    fprintf (stderr, "usage: eccllvoice [options] infopktfile\n");
    fprintf (stderr, "  -p          play\n");
    fprintf (stderr, "  -r          record (default 10 secs)\n");
    fprintf (stderr, " one of -p or -r is required\n");
    fprintf (stderr, "  -c num      call given number, else await ring\n");
    fprintf (stderr, "  -u unitnum  use specified unit number\n");
    fprintf (stderr, "  -n secs     specify recording time\n");
    fprintf (stderr, "  -l          loop forever sending or receiving\n");
    fprintf (stderr, "  -v          enable API debug mode\n");
    fprintf (stderr, "  -f          specify record coding format (dflt:pcm_ulaw)\n");
    fprintf (stderr, "use the number of the format, or one of:\n");
    fprintf (stderr, "   adpcm[32,24],pcm_ulaw[64,48,88],pcm_alaw[64,48,88],\n");
    fprintf (stderr, "   g723-1[-53,-63],g729-a,sx7300,sx9600,gsm_610,rawdata,\n");
    fprintf (stderr, "   g726[-40,-32,-24,-16],linear[128]\n");
    fprintf (stderr, "'#' causes immediate termination of play/record.\n");
    fprintf (stderr, "On playback:\n");
    fprintf (stderr, "    '1' = gain up, '2' = gain down, '3' = speed up, '4' = speed down.\n");
    exit (1);
}

int cflag;
char *cnum;
int unit = 2;
int pflag;
int rflag;
int lflag;
int done;
int format_specified = 0;
long the_timeout = 10000L;

unsigned coding_fmt;
struct coding_info {
    char name[30];
    unsigned fmt;
    unsigned rate;
    unsigned bits_per_samp;
} fmt_list[] = {
    { {"pcm_ulaw"}, CODE_PCM_ULAW, RATE_8000, BITS_8 },
    { {"pcm_ulaw64"}, CODE_PCM_ULAW, RATE_8000, BITS_8 },
    { {"pcm_ulaw48"}, CODE_PCM_ULAW, RATE_6000, BITS_8 },
    { {"pcm_ulaw88"}, CODE_PCM_ULAW, RATE_11000, BITS_8 },
    { {"pcm_alaw"}, CODE_PCM_ALAW, RATE_8000, BITS_8 },
    { {"pcm_alaw64"}, CODE_PCM_ALAW, RATE_8000, BITS_8 },
    { {"pcm_alaw48"}, CODE_PCM_ALAW, RATE_6000, BITS_8 },
    { {"pcm_alaw88"}, CODE_PCM_ALAW, RATE_11000, BITS_8 },
    { {"adpcm"}, CODE_ADPCM, RATE_8000, BITS_4 },
    { {"adpcm32"}, CODE_ADPCM, RATE_8000, BITS_4 },
    { {"adpcm24"}, CODE_ADPCM, RATE_6000, BITS_4 },
    { {"g723-1"}, CODE_G723_1, RATE_5300, BITS_4 },
    { {"g723-1-53"}, CODE_G723_1, RATE_5300, BITS_4 },
    { {"g723-1-63"}, CODE_G723_1, RATE_6300, BITS_4 },
    { {"g729-a"}, CODE_G729_A, RATE_8000, BITS_4 },
    { {"sx7300"}, CODE_SX7300, RATE_8000, BITS_4 },
    { {"sx9600"}, CODE_SX9600, RATE_8000, BITS_4 },
    { {"gsm_610"}, CODE_GSM_610, RATE_13000, BITS_8 },
    { {"rawdata"}, CODE_RAWDATA, RATE_8000, BITS_8 },
    { {"g726"}, CODE_G726, RATE_8000, BITS_5 },
    { {"g726-40"}, CODE_G726, RATE_8000, BITS_5 },
    { {"g726-32"}, CODE_G726, RATE_8000, BITS_4 },
    { {"g726-24"}, CODE_G726, RATE_8000, BITS_3 },
    { {"g726-16"}, CODE_G726, RATE_8000, BITS_2 },
    { {"linear"}, CODE_LINEAR, RATE_8000, BITS_16 },
    { {"linear128"}, CODE_LINEAR, RATE_8000, BITS_16 },
    /* Not supported or compatibility only */
    { {"cvsd"}, CODE_CVSD, RATE_20000, BITS_1 },
    { {"subband"}, CODE_SUBBAND, RATE_8000, BITS_3 },
    { {"oki_adpcm"}, CODE_OKI_ADPCM, RATE_8000, BITS_4 },
    { {"rhet_adpcm"}, CODE_RHET_ADPCM, RATE_8000, BITS_4 },
    { {"rhet_adpcm32"}, CODE_RHET_ADPCM, RATE_8000, BITS_4 },
    { {"rhet_adpcm24"}, CODE_RHET_ADPCM, RATE_6000, BITS_4 },
    { {"dvi_adpcm"}, CODE_DVI_ADPCM, RATE_8000, BITS_4 },
    { {"dvi_adpcm32"}, CODE_DVI_ADPCM, RATE_8000, BITS_4 },
    { {"dvi_adpcm24"}, CODE_DVI_ADPCM, RATE_6000, BITS_4 },
    { {"ms_gsm"}, CODE_MS_GSM, RATE_8000, BITS_4 },
    { {"gsm_660"}, CODE_GSM_660, RATE_12300, BITS_8 },
};
#define NUM_FMTS (sizeof(fmt_list) / sizeof(*fmt_list))
/* Extra braces around strings above required to avoid QNX 2 compiler bug */


void dowork BT_ARGS ((BTLINE *, char *));
void dowork1 BT_ARGS ((BTLINE *, char *));

unsigned
format_string_lookup(char *fmtstr)
{
    unsigned i;
    
    for (i = 0; i < NUM_FMTS; i++)
    {
        if (!strcmp(fmtstr,fmt_list[i].name))
            break;
    }
    if (i >= NUM_FMTS)
    {
        fprintf(stderr,"Unknown format '%s'\n",fmtstr);
        exit(1);
    }
    return(i);
}


int
main (int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    BTLINE *lp;
    char *name;
    int i;
    unsigned tmp;
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();

    while ((c = getopt (argc, argv, "c:pru:lf:n:Lv")) != EOF)
    {
        switch (c)
        {
          case 'l':
            lflag = 1;
            break;
          case 'u':
            unit = atoi (optarg);
            break;
          case 'c':
            cflag = 1;
            cnum = optarg;
            break;
          case 'p':
            pflag = 1;
            break;
          case 'r':
            rflag = 1;
            break;
          case 'f':
            if (isdigit(optarg[0]))
            {
                tmp = atoi(optarg);
                for (i = 0; i < NUM_FMTS; i++)
                {
                    if (fmt_list[i].fmt == tmp)
                    {
                        coding_fmt = i;
                        break;
                    }
                }
                if (i >= NUM_FMTS)
                {
                    fprintf(stderr,"Invalid format %d\n",tmp);
                    exit(1);
                }
            }
            else
                coding_fmt = format_string_lookup(optarg);
            format_specified = 1;
            break;
          case 'n':
            the_timeout = 1000L * atoi(optarg);
            break;
          case 'v':
            BT_ZERO(args_debug);
            args_debug.mode = DEBUG_ALL;
            args_debug.options = API_DBG_MODE_SET;
            BfvDebugModeSetAdv(&args_debug);
            break;
          default:
            usage ();
        }
    }
    
    if (optind >= argc)
        usage ();

    name = argv[optind++];
    
    if (optind != argc)
        usage ();

    if (pflag + rflag != 1)
    {
        fprintf (stderr, "one of -p or -r is required\n");
        exit (1);
    }
    
    BT_ZERO(args);
    args.unit = unit;
    if ((lp = BfvLineAttach (&args)) == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
        exit (1);
    }
    
    dowork1 (lp, name);
    
    return(0);
}

void
dowork1 (BTLINE *lp, char *name)
{
    struct args_line_admin args_admin;
    
    if (!format_specified)
    {
        coding_fmt = format_string_lookup("pcm_ulaw");
    }
    
    if (lflag)
    {
        for (;;)
            dowork (lp, name);
    }
    else
        dowork (lp, name);

    BT_ZERO(args_admin);
    BfvLineDetach (lp, &args_admin);
    
    done = 1;
}

void
dowork (BTLINE *lp, char *name)
{
    struct infopkt_stream *ips;
    BTERR bterr;
    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_speech args_speech;
    struct args_infopkt args_infopkt;
    struct args_tone args_tone;
    struct args_cc args_cc;
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (lp, &args_admin) < 0)
    {
        BfvErrorMessage(lp,&args_admin.res,&bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                 bterr.long_msg,args_admin.reset_status);
        exit(1);
    }
    
    /* Check for digital configuration error */
    if (args_admin.reset_status & RST_DIG_CONFIG_UNREADABLE)
    {
        fprintf (stderr, "Telephony/Digital configuration file not found.\n");
        exit(1);
    }
    
    if (rflag && !LINE_HAS_CAP(lp,CAP_SPEECH_RECORD))
    {
        fprintf (stderr, "unit %d does not support speech recording\n",
                 LINE_UNIT_NUM(lp));
        exit(1);
    }
    
    if (pflag && !LINE_HAS_CAP(lp,CAP_SPEECH_PLAY))
    {
        fprintf (stderr, "unit %d does not support speech playback\n",
                 LINE_UNIT_NUM(lp));
        exit(1);
    }

    check_line_state(lp);
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = name;
    args_infopkt.fmode = rflag ? "w" : "r";
    if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        printf ("BfvInfopktOpen: %s\n",bterr.long_msg);
        exit(1);
    }
    
    if (cflag)
    {
        /* start the outgoing call */
        BT_ZERO(args_cc);
        args_cc.phonenum = cnum;
        args_cc.call_protocol_code = CALL_PROTOCOL_VOICE;
        BfvCallSetup(lp, &args_cc);
        if (args_cc.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_cc.res,&bterr);
            printf ("BfvCallSetup: %s\n",bterr.long_msg);
            clear_call(lp);
            exit (1);
        }

        /* wait for the outgoing call setup to complete */
        BT_ZERO(args_cc);
        args_cc.timeout = 60000L;
        args_cc.call_protocol_code = CALL_PROTOCOL_VOICE;
        BfvCallWaitForComplete(lp, &args_cc);
        if (args_cc.res.status != BT_STATUS_OK || 
            LINE_STATE(lp) != LINE_STATE_CONNECTED)
        {
            BfvErrorMessage(lp,&args_cc.res,&bterr);
            printf ("BfvCallWaitForComplete: %s\n",bterr.long_msg);
            clear_call(lp);
            exit (1);
        }
    }
    else
    {
        /* enable ring detection */
        BT_ZERO(args_cc);
        args_cc.mode = 1;
        BfvCallRingDetect(lp, &args_cc);
        if (args_cc.res.line_status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_cc.res,&bterr);
            printf ("BfvCallRingDetect: %s\n",bterr.long_msg);
            exit (1);
        }

        /* wait forever for a call to come in */
        BT_ZERO(args_cc);
        BfvCallWaitForSetup (lp, &args_cc);
        if (args_cc.res.status != BT_STATUS_OK || 
            args_cc.res.line_status != WAIT_FOR_CALL_OK)
        {
            BfvErrorMessage(lp,&args_cc.res,&bterr);
            printf ("BfvCallWaitForSetup: %s\n",bterr.long_msg);
            exit (1);
        }
        process_call_info(&args_cc.cres);

        /* start answering the call */
        BT_ZERO(args_cc);
        BfvCallAccept (lp, &args_cc);
        if (args_cc.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_cc.res,&bterr);
            printf ("BfvCallAccept: %s\n",bterr.long_msg);
            exit (1);
        }

        /* wait for the call to be answered */
        BT_ZERO(args_cc);
        BfvCallWaitForAccept (lp, &args_cc);
        if (args_cc.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_cc.res,&bterr);
            printf ("BfvCallWaitForAccept: %s\n",bterr.long_msg);
            exit (1);
        }
    }

    BT_ZERO(args_tone);
    args_tone.decode_flag = DTMF_12TONE; 
    BfvToneDetectEnable(lp, &args_tone);
    
    /* Clear previous CP data */
    BT_ZERO(args_tel);
    while (BfvDataCP(lp,&args_tel) == 0);
    
    /* Enable call progress, and allow BfvToneGet to check for CP. 
       Can be done only in a restricted mode. */
    BT_ZERO(args_tel);
    args_tel.call_protocol_code = CALL_PROTOCOL_VOICE_NO_RAW;
    args_tel.call_mode = BT_ORIGINATE;
    if (BfvLineCallProgressEnable(lp,&args_tel) < 0)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("BfvLineCallProgressEnable: %s\n",bterr.long_msg);

        /* hang up the call */
        clear_call(lp);

        exit(1);
    }
    
    if (rflag)
    {
        /* record for 10 seconds. Uses specified format, bps value, bit per
           sample value, with AFE 8000samp/sec, MSB data. Will do silence
           removal, no silence timeout. */
        BT_ZERO(args_speech);
        args_speech.ips = ips;
        args_speech.coding_fmt = fmt_list[coding_fmt].fmt;
        args_speech.rate = fmt_list[coding_fmt].rate;
        args_speech.bits_per_samp = fmt_list[coding_fmt].bits_per_samp;
        args_speech.afe_rate = AFE_8000;
        args_speech.data_fmt = SPCH_MSB;
        args_speech.timeout = the_timeout;
        args_speech.silence_timeout = 0L;
        args_speech.s_compr = 1;
        args_speech.func = play_func;
        BfvSpeechRecord (lp, &args_speech);
    }
    else
    {
        /* play till file is done */
        BT_ZERO(args_speech);
        args_speech.ips = ips;
        args_speech.func = play_func;
        BfvSpeechPlay (lp, &args_speech);
    }
    
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_speech.res,&bterr);
        printf ("BfvSpeechPlay/Record: %s\n",bterr.long_msg);
    }
    
    BT_ZERO(args_tel);
    BfvLineCallProgressDisable(lp,&args_tel);

    BT_ZERO(args_tone);
    BfvToneDetectDisable(lp,&args_tone);

    args_infopkt.ips = ips;
    BfvInfopktClose (&args_infopkt);

    /* hang up the call */
    clear_call(lp);

    /* note: advice of charge information is not available until after 
       BfvCallWaitForRelease or BfvLineTerminteCall has been called */
    process_aoc(lp);
}

void check_line_state (BTLINE *lp)
{
    struct args_cc args_cc;
    BTERR bterr;

    /* check the state of the line to see if it is idle 
       this is in this example code just to demonstrate the
       use of this function, It is not a required function call
       Although the function name suggest ISDN only, this function can
       be used on any line type including analog and T1 robbed bit.*/
    BT_ZERO(args_cc);
    BfvCallStatus(lp, &args_cc);
    if (args_cc.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_cc.res,&bterr);
        printf ("BfvCallStatus: %s\n",bterr.long_msg);
        exit (1);
    }
    switch(args_cc.state) {
      case BST_IDLE:
        printf ("The line is idle\n");
        break;
      case BST_DIAL:
        printf ("The line has started an outbound call and is still dialing the number\n");
        break;
      case BST_DIAL_COMPLETE:
        printf ("The line has started an outbound call and has completed dialing\n");
        break;
      case BST_ALERTING:
        printf ("The line has detected an incoming call and is ready to be answered (DID collection is complete)\n");
        break;
      case BST_WAIT_FOR_CALL:
        printf ("The line is waiting for an incoming call\n");
        break;
      case BST_CALL_DETECTED:
        printf ("The line has detected an incoming call but it is not ready to be answered (DID collection is in progress)\n");
        break;
      case BST_CONNECTED:
        printf ("The line is connected\n");
        break;
      case BST_CLEAR_CALL:
        printf ("The line is clearing a call\n");
        break;
      case BST_DOWN:
        /* Note: this state is not always bad, 
           For inbound call functions on a line that is down, the system
           will wait for the line to come back into service and then wait
           for an incoming call automatically. without reporting an error
           to the application. For an outbound call functions, the system
           will wait for a period of time (set in the call control
           configuration file) for the line to come into service before
           reporting an error. if the line comes into service before the
           timeout, the call will be placed as normal. This is to allow
           ISDN protocols (PRI & BRI) to come 
           into service without causing an error in the application */
        printf ("The line is Out of Service\n");
        break;
    }
}


void clear_call (BTLINE *lp)
{
    struct args_cc args_cc;
    BTERR bterr;

    /* hang up the call */
    BT_ZERO(args_cc);
    BfvCallDisconnect (lp, &args_cc);
    if (args_cc.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_cc.res,&bterr);
        printf ("BfvCallDisconnect: %s\n",bterr.long_msg);
    }

    BT_ZERO(args_cc);
    BfvCallWaitForRelease (lp, &args_cc);
    if (args_cc.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_cc.res,&bterr);
        printf ("BfvCallWaitForRelease: %s\n",bterr.long_msg);
    }
}


/* # = Abort, 1 = Gain Up, 2 = Gain Down, 3 = Speed Up, 4 = Speed Down. */
int
play_func (BTLINE *lp, char *arg)
{
    static int action_table[] = { 
        SPCH_MOD_GAIN_UP, SPCH_MOD_GAIN_DOWN, 
        SPCH_MOD_RATE_UP, SPCH_MOD_RATE_DOWN,
    };
    int tmp;
    struct args_tone args_tone;
    struct args_speech args_speech;
    struct args_telephone args_tel;
    
    BT_ZERO(args_tone);
    args_tone.tone_cp = 1;
    switch (tmp = BfvToneGet(lp,&args_tone))
    {
      case '1':
      case '2':
      case '3':
      case '4':
        if (!rflag)
        {
            /* Take appropriate action */
            BT_ZERO(args_speech);
            args_speech.modification = action_table[tmp - '1'];
            BfvSpeechModify(lp,&args_speech);
        }
        return(0);
      case '#':
        /* Stop playing/recording */
        return(1);
      case 1:
        /* CP available */
        BT_ZERO(args_tel);
        if (BfvDataCP(lp,&args_tel))
        {
            printf("CP data reported available, but is not!\n");
            return(0);
        }
        if (args_tel.type != CP_TYPE_BOARD_DET)
        {
            printf("Unexpected CP type %u found.\n",args_tel.type);
            return(0);
        }
        switch (args_tel.data)
        {
          case CS_BUSY1:
          case CS_DIALTON:
            /* Stop on busy/reorder or dialtone */
            return(1);
          case CS_SILENCE:
          case CS_CNGDETCT:
            return(0);
          default:
            printf("Unexpected CP data value %02X found.\n",args_tel.data);
            return(0);
        }
    }
    if (args_tone.res.status != BT_STATUS_OK &&
        args_tone.res.status != BT_STATUS_TIMEOUT)
    {
        BTERR bterr;
        
        BfvErrorMessage(lp,&args_tone.res,&bterr);
        printf ("BfvToneGet: %s\n",bterr.long_msg);
        return(1);
    }
    return (0); /* Keep playing */
}

void
process_aoc(BTLINE *lp)
{
    AOC_INFO *aoc = LINE_AOC_INFO(lp);
        
    if (aoc == NULL || aoc->status == BT_AOC_UNAVAILABLE)
        printf("Charge information unavailable.\n");
    else if (aoc->status == BT_AOC_FREE)
        printf("Charge information: FREE.\n");
    else if (aoc->status == BT_AOC_UNITS)
        printf("Charge information: Unit type %d, value %ld\n",
               aoc->charge_info,aoc->charge);
    else if (aoc->status == BT_AOC_CURRENCY)
    {
        static char *currency_names[] = {
            "1/1000","1/100","1/10","1","10","100","1000" };
#define NUM_CURRENCIES (sizeof(currency_names) / sizeof(*currency_names))
        char *cur_name;
            
        cur_name = (0 <= aoc->charge_info &&
                    aoc->charge_info < NUM_CURRENCIES) ?
            currency_names[aoc->charge_info] :
            "Invalid currency";
        printf("Charge information: Currency multiplier %s (%d), value %ld\n",
               cur_name,aoc->charge_info,aoc->charge);
    }
    else
        printf("Charge information: Bad status value %d\n",
               aoc->status);
}

void
process_call_info(CALL_RES *call_res)
{
    if (strlen(call_res->called_party_number))
        printf("Called party number: \"%s\"\n",
            call_res->called_party_number);
    if (strlen(call_res->called_party_subaddress))
        printf("Called party subaddress: \"%s\"\n",
           call_res->called_party_subaddress);
    if (strlen(call_res->calling_party_number))
        printf("Calling party number: \"%s\"\n",
           call_res->calling_party_number);
    if (strlen(call_res->calling_party_subaddress))
        printf("Calling party subaddress: \"%s\"\n",
           call_res->calling_party_subaddress);
    if (strlen(call_res->redir_number))
        printf("Redirection number: \"%s\", reason %X\n",
           call_res->redir_number, call_res->redir_reason);
}
