/*
 * faxml.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The faxml program uses the mid-level infopkt file fax sending and
receiving routines to send or receive facsimiles.
***************************************************************************/

#include <memory.h>
#include "btlib.h"

typedef int BOOL;
#define TRUE 1
#define FALSE 0

int done;
BOOL send_mode = FALSE, recv_mode = FALSE;
int expect_another;

void process_results BT_ARGS((BTLINE *lp,FAX_RES *fres,long duration));
void process_aoc BT_ARGS((BTLINE *lp));
void process_call_info BT_ARGS((CALL_RES *call_res));

void
usage()
{
    printf("usage: faxml [-u <unitno>] [-v] -s <phoneno> <infopktfile>\n");
    printf("or:    faxml [-u <unitno>] [-v] -r <infopktfile>\n");
    printf("\tu = unit number (fax card)\n");
    printf("\ts = send mode\n");
    printf("\tr = receive mode\n");
    printf("\tv = turn on API debug mode\n");
    exit( 1 );
}

int
main(int argc, char **argv)
{
    extern int optind;
    extern char *optarg;
    int c;
    BTLINE *lp = NULL;
    int unit;
    char *name;
    char *snum;
    void dowork();
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    
    unit = 0;
    snum = ",1234";
    
    while ( (c = getopt( argc, argv, "u:s:rv" )) != EOF )
    {
        switch ( c )
        {
          case 'u':
            unit = atoi( optarg );
            break;
          case 's':
            send_mode = TRUE;
            snum = optarg;
            break;
          case 'r':
            recv_mode = TRUE;
            break;
          case 'v':
            BT_ZERO(args_debug);
            args_debug.mode = DEBUG_ALL;
            args_debug.options = API_DBG_MODE_SET;
            BfvDebugModeSetAdv(&args_debug);
            break;
          default:
            usage();
        }
    }
    
    if ( optind+1 != argc || (!send_mode && !recv_mode) )
        usage();

    name = argv[ optind ];
    
    BT_ZERO(args);
    args.unit = unit;
    lp = BfvLineAttach(&args);
    if (lp == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        printf ("BfvLineAttach: %s\n",bterr.long_msg);
        exit(1);
    }
    
    dowork(lp,name,snum);
    
    return(0);
}

void
dowork(BTLINE *lp, char *name, char *snum)
{
    struct infopkt_stream *ips = NULL;
    FAX_RES fax_res;
    BTERR bterr;
    int tmp;
    long start_time;
    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_fax args_fax;
    struct args_infopkt args_infopkt;
    
    BT_ZERO(fax_res);
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = name;
    args_infopkt.fmode = send_mode ? "r" : "w";
    ips = BfvInfopktOpen(&args_infopkt);
    if ( ips == NULL )
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        printf ("BfvInfopktOpen: %s\n",bterr.long_msg);
        goto err_1;
    }
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (lp, &args_admin) < 0)
    {
        BfvErrorMessage(lp,&args_admin.res,&bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                 bterr.long_msg,args_admin.reset_status);
        goto err_1;
    }
    
    if (!LINE_HAS_CAP(lp,CAP_FAX_XMIT))
    {
        fprintf (stderr, "unit %d does not support fax\n", LINE_UNIT_NUM(lp));
        goto err_1;
    }
    
    LINE_FAX_RES(lp) = &fax_res;
    
    /*
     *   SSSS  EEEEE  N   N  DDDD
     *  S      E      NN  N  D   D
     *   SSS   EEEE   N N N  D   D
     *      S  E      N  NN  D   D
     *  SSSS   EEEEE  N   N  DDDD
     */
    if ( send_mode )
    {
        /*
         * Send step 1:
         *      Dial out to establish connection
         */
        printf( "BfvLineOriginateCall\n" );
        BT_ZERO(args_tel);
        args_tel.phonenum = snum;
        args_tel.call_protocol_code = CALL_PROTOCOL_FAX;
        BfvLineOriginateCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            LINE_STATE(lp) != LINE_STATE_CONNECTED)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineOriginateCall: %s\n",bterr.long_msg);
            goto err_2;
        }
        
        /* Start timing after we detect Fax answer tone */
        start_time = time(NULL);
        
        /*
         *      Let's do all the individual steps instead of easy BfvFaxSend
         *      This can be informative to just understand how BfvFaxSend works
         *      or you may have other needs (multiple info packet streams,
         *      etc...)
         *              
         *              BT_ZERO(args_fax);
         *              args_fax.s_ips = ips;
         *              args_fax.local_id = local_id;
         *              BfvFaxSend (lp, &args_fax);
         *              if (args_fax.res.status != BT_STATUS_OK )
         *              {
         *                      printf( "problem with send_fax\n" );
         *                      goto err;
         *              }
         */
        
        /*
         * Send step 2:
         *      BfvFaxBeginSend -- peek at a 'doc' info packet in the stream
         *      and setup the fax for transmission (error if there isn't a
         *      'doc' packet in the stream)
         */
        printf( "BfvFaxBeginSend\n" );
        BT_ZERO(args_fax);
        args_fax.s_ips = ips;
        BfvFaxBeginSend( lp, &args_fax);
        if (args_fax.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxBeginSend: %s\n",bterr.long_msg);
            goto err;
        }
        /*
         * Send step 3:
         *      Call BfvFaxGetRemoteInfo and BfvFaxWaitForTraining to talk to
         *      the remote fax and exchange info
         */
        printf( "BfvFaxGetRemoteInfo\n" );
        BT_ZERO(args_fax);
        BfvFaxGetRemoteInfo(lp, &args_fax);
        if (args_fax.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxGetRemoteInfo: %s\n",bterr.long_msg);
            goto err;
        }
        else
            printf ("Remote ID: '%s'\n", args_fax.remote_info.remote_id );
        
        printf( "BfvFaxWaitForTraining\n" );
        BT_ZERO(args_fax);
        BfvFaxWaitForTraining(lp, &args_fax);
        if (args_fax.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxWaitForTraining: %s\n",bterr.long_msg);
            goto err;
        }
        
        /*
         * Send step 4:
         *      Call BfvFaxNextPage to read the 'doc and 'bop' packets and
         *      setup the fax for a page
         *      Stop sending pages if there are no more packets (either EOF or
         *      some other kind of packet)
         */
        for (;;) /* for each page */
        {
            printf( "BfvFaxNextPage\n" );
            BT_ZERO(args_fax);
            args_fax.s_ips = ips;
            if ((tmp = BfvFaxNextPage(lp, &args_fax)) == 0 )
                break;
            if (tmp < 0)
            {
                BfvErrorMessage(lp,&args_fax.res,&bterr);
                printf ("BfvFaxNextPage: %s\n",bterr.long_msg);
                goto err;
            }
            
            /*
             * Send step 5:
             *      Call BfvFaxSendPage to send the 'g3' packets until there
             *      are no more (either the next 'bop', some other kind
             *      of packet, or EOF)
             */
            printf( "BfvFaxSendPage\n" );
            BT_ZERO(args_fax);
            args_fax.s_ips = ips;
            if (BfvFaxSendPage( lp, &args_fax))
            {
                BfvErrorMessage(lp,&args_fax.res,&bterr);
                printf ("BfvFaxSendPage: %s\n",bterr.long_msg);
                goto err;
            }
        }
        
        /*
         * Send step 6:
         *      Call BfvFaxEndOfDocument to finish up
         */
        printf( "BfvFaxEndOfDocument\n" );
        BT_ZERO(args_fax);
        if (BfvFaxEndOfDocument(lp, &args_fax) < 0)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxEndOfDocument: %s\n",bterr.long_msg);
            goto err;
        }
        
        printf( "Done\n" );
    }
    
    /*
     *  RRRR   EEEEE   CCCC  V   V 
     *  R   R  E      C      V   V
     *  RRRR   EEEE   C      V   V
     *  R  R   E      C       V V
     *  R   R  EEEEE   CCCC    V
     */
    if ( recv_mode )
    {
        /*
         * Receive step 1:
         *      Wait for an incoming call
         */
        printf( "Wait for connection (ringing or DID digits)\n" );
        BT_ZERO(args_tel);
        BfvLineWaitForCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            args_tel.res.line_status != WAIT_FOR_CALL_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineWaitForCall: %s\n",bterr.long_msg);
            goto err_1;
        }
        process_call_info(&args_tel.call_res);
        
        /* Start timing after we detect ringing */
        start_time = time(NULL);
        
        /*
         * Receive step 2:
         *      The call is here, pick it up
         */
        printf( "BfvLineAnswer\n" );
        BT_ZERO(args_tel);
        BfvLineAnswer (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineAnswer: %s\n",bterr.long_msg);
            goto err_1;
        }
        
        /*
         *      Let's do all the individual steps instead of easy BfvFaxReceive
         *      This can be informative to just understand how BfvFaxReceive
         *      works or you may have other needs.
         *
         *              BT_ZERO(args_fax);
         *              args_fax.r_ips = ips;
         *              args_fax.local_id = local_id;
         *              BfvFaxReceive (lp, &args_fax);
         *              if (args_fax.res.status != BT_STATUS_OK)
         *              {
         *                      printf( "problem with receive_fax\n" );
         *                      goto err;
         *              }
         */
        
        /*
         * Receive step 3:
         *      Start receiving sequence
         */
        printf( "BfvFaxBeginReceive\n" );
        BT_ZERO(args_fax);
        if (BfvFaxBeginReceive(lp, &args_fax) < 0)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxBeginReceive: %s\n",bterr.long_msg);
            goto err;
        }
        
        /*
         * Receive step 4:
         *      Call BfvFaxGetRemoteInfo and BfvFaxWaitForTraining to talk to
         *      the remote fax and exchange info
         */
        printf( "BfvFaxGetRemoteInfo\n" );
        BT_ZERO(args_fax);
        BfvFaxGetRemoteInfo(lp, &args_fax);
        if (args_fax.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxGetRemoteInfo: %s\n",bterr.long_msg);
            goto err;
        }
        else
            printf ("Remote ID: '%s'\n", args_fax.remote_info.remote_id );

        printf( "BfvFaxWaitForTraining\n" );
        BT_ZERO(args_fax);
        BfvFaxWaitForTraining(lp, &args_fax);
        if (args_fax.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxWaitForTraining: %s\n",bterr.long_msg);
            goto err;
        }
        /*
         * User code could write out infopackets now for header or other
         * information in USER packets.
         */
        
        /*
         * Receive step 5:
         *      Call BfvFaxReceivePages to receive all the pages or
         *      BfvFaxReceivePage to receive them one at a time
         */
        do
        {
            printf( "BfvFaxReceivePage\n" );
            BT_ZERO(args_fax);
            args_fax.r_ips = ips;
            expect_another = BfvFaxReceivePage(lp,&args_fax);
            if (expect_another < 0)
            {
                BfvErrorMessage(lp,&args_fax.res,&bterr);
                printf ("BfvFaxReceivePage: %s\n",bterr.long_msg);
                goto err;
            }
        }
        while (expect_another);
        
        /*
         * Receive step 6:
         *      Wait for completion using BfvFaxEndReception.
         */
        printf( "BfvFaxEndReception\n" );
        BT_ZERO(args_fax);
        if (BfvFaxEndReception(lp, &args_fax) < 0)
        {
            BfvErrorMessage(lp,&args_fax.res,&bterr);
            printf ("BfvFaxEndReception: %s\n",bterr.long_msg);
            goto err;
        }
        
        printf( "Done\n" );
    }
    
err:
    process_results(lp,&fax_res,time(NULL) - start_time);
    
err_2:
    BT_ZERO(args_tel);
    BfvLineTerminateCall (lp, &args_tel);

    process_aoc(lp);

err_1:
    BT_ZERO(args_admin);
    BfvLineDetach (lp, &args_admin);
    
    if (ips != NULL)
    {
        BT_ZERO(args_infopkt);
        args_infopkt.ips = ips;
        BfvInfopktClose (&args_infopkt);
    }
    
    done = 1;
}

char *
conf_vals(int conf)
{
    switch (conf)
    {
      case FSK_MCF_1:
      case FSK_MCF_2:
        return("MCF");
      case FSK_RTN_1:
      case FSK_RTN_2:
        return("RTN");
      case FSK_RTP_1:
      case FSK_RTP_2:
        return("RTP");
      case FSK_PIN_1:
      case FSK_PIN_2:
        return("PIN");
      case FSK_PIP_1:
      case FSK_PIP_2:
        return("PIP");
      default:
        return("?");
    }
}

void
process_results(BTLINE *lp, FAX_RES *fres, long duration)
{
    static char *bft_names[] =
        {NULL,"T.434","BTM",NULL,"DTM",NULL,NULL,NULL,"EDIFACT"};
    static int bit_rates[] =
        {2400,4800,7200,9600,7200,9600,12000,14400,12000,14400,
         2400,4800,7200,9600,12000,14400,16800,19200,
         21600,24000,26400,28800,31200,33600};
    static int ecm_vals[] = {0,256,64};
    static char *fmt_names[2][2] = {{"MH","MR"},{"MMR","MMR"}};
    static char *res_names[] =
        {"200Hx100V(Norm)",
         "200Hx200V(Fine)",
         "200Hx400V",
         "300Hx300V",
         "400Hx400V",
         "600Hx600V", 
         "1200Hx1200V",
         "300Hx600V",
         "400Hx800V",
         "600Hx1200V",
         "100Hx100V",
        };
    static char *width_names[] = {"A4","B4","A3"};
    PAGE_RES *pageres;
    PAGE_RES *temp;
    int jj;
    
    /* loop down the page res list */
    pageres = fres->reslist_head;
    jj = 1;
    while (pageres)
    {
        printf("Page %2d:  ", jj);
        if (pageres->bft_type)
            printf("BFT type %s  ",bft_names[pageres->bft_type]);
        else
        {
            if (pageres->eff_page_type)
            {
                printf("EFF Page: ");
                if (pageres->eff_page_type & BT_EFF_JPEG_ENABLE)
                {
                    printf("JPEG(");
                    if (pageres->eff_page_type & BT_EFF_JPEG_FULLCOLOR)
                        printf("full color, ");
                    else
                        printf("grayscale, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_DEFAULT_TABLES)
                        printf("default tbls, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_12BIT)
                        printf("12bit, ");
                    else
                        printf("8bit, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_NOSUBSAMPLING)
                        printf("1:1:1, ");
                    else
                        printf("4:1:1, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_CUSTOM_ILLUMINANT)
                        printf("cust illum, ");
                    else
                        printf("std illum, ");
                    if (pageres->eff_page_type & BT_EFF_JPEG_CUSTOM_GAMUT)
                        printf("cust gamut");
                    else
                        printf("std gamut");
                    printf(")\n          ");
                }
                if (pageres->eff_page_type & BT_EFF_JBIG)
                {
                    printf("JBIG(");
                    if (pageres->eff_page_type & BT_EFF_JBIG_L0)
                        printf("L0");
                    else
                        printf("std");
                    printf(") ");
                }
                if (pageres->eff_page_type & BT_EFF_LOSSLESS)
                    printf("LOSSLESS COLOR ");
                if (pageres->eff_page_type & BT_EFF_PLANE_INTERLEAVE)
                    printf("PLANE_INTERLEAVE ");
                if (pageres->eff_page_type & BT_EFF_MRC_MASK)
                    printf("MRC(%X) ",
                           (pageres->eff_page_type & BT_EFF_MRC_MASK) >> 4);
                if (pageres->eff_page_type & BT_EFF_PAGE_LENGTH_STRIP)
                    printf("PAGE_LEN_STRIP ");
            }
            else
            {
                if (pageres->direction)
                    printf("ASCII %d  ",pageres->ascii_bytes);
                printf("BdLn %d  TotLn %d  ",
                       pageres->bad_lines, pageres->total_lines);
            }
            printf("Res %s  ",res_names[pageres->resolution]);
            printf("Width %s  ",width_names[pageres->width]);
        }
        printf("Bps %u\n",bit_rates[pageres->bit_rate]);
        
        printf("          ConfVal %s  SigLev %04X  LineNoise %04X",
               conf_vals(pageres->confirm_value),
               pageres->sig_level,pageres->line_noise);
        if (!pageres->direction)
            printf("  SigQual %04X",pageres->sig_quality);
        printf("\n");
        
        temp = pageres;
        pageres = pageres->next;
#ifdef MILL_DLL
        _dll_free(temp);
#else
        free(temp);
#endif
        jj++;
    }
    printf("Duration %ld secs  ",duration);
    printf("ECM %d  ",ecm_vals[LINE_DCS(lp)->Error_Correction_Mode]);
    printf("Line Format %s",
           fmt_names[LINE_DCS(lp)->MMR][LINE_DCS(lp)->TwoD_Coding]);
    printf("\n");
}

void
process_aoc(BTLINE *lp)
{
    AOC_INFO *aoc = LINE_AOC_INFO(lp);
        
    if (aoc == NULL || aoc->status == BT_AOC_UNAVAILABLE)
        printf("Charge information unavailable.\n");
    else if (aoc->status == BT_AOC_FREE)
        printf("Charge information: FREE.\n");
    else if (aoc->status == BT_AOC_UNITS)
        printf("Charge information: Unit type %d, value %ld\n",
               aoc->charge_info,aoc->charge);
    else if (aoc->status == BT_AOC_CURRENCY)
    {
        static char *currency_names[] = {
            "1/1000","1/100","1/10","1","10","100","1000" };
#define NUM_CURRENCIES (sizeof(currency_names) / sizeof(*currency_names))
        char *cur_name;
            
        cur_name = (0 <= aoc->charge_info &&
                    aoc->charge_info < NUM_CURRENCIES) ?
            currency_names[aoc->charge_info] :
            "Invalid currency";
        printf("Charge information: Currency multiplier %s (%d), value %ld\n",
               cur_name,aoc->charge_info,aoc->charge);
    }
    else
        printf("Charge information: Bad status value %d\n",
               aoc->status);
}

void
process_call_info(CALL_RES *call_res)
{
    if (strlen(call_res->called_party_number))
        printf("Called party number: \"%s\"\n",
            call_res->called_party_number);
    if (strlen(call_res->called_party_subaddress))
        printf("Called party subaddress: \"%s\"\n",
           call_res->called_party_subaddress);
    if (strlen(call_res->calling_party_number))
        printf("Calling party number: \"%s\"\n",
           call_res->calling_party_number);
    if (strlen(call_res->calling_party_subaddress))
        printf("Calling party subaddress: \"%s\"\n",
           call_res->calling_party_subaddress);
    if (strlen(call_res->redir_number))
        printf("Redirection number: \"%s\", reason %X\n",
           call_res->redir_number, call_res->redir_reason);
}
