/*
 * ipstrip.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The ipstrip program removes the infopkt header from the G3 or
speech data in a specified infopkt stream file and writes each page of
the converted data to a file, g3data.30x. The first page of converted
fax data is written to the file g3data.301, the second to the file
g3data.302, the third to the file g3data.303, and so on until the entire
infopkt stream file is converted. All speech data are placed in a single
file.
***************************************************************************/

#include <string.h>
#include <ctype.h>
#include "btlib.h"

void usage()
{
    fprintf(stderr, "Usage: ipstrip [-h] [-o <output_base>] <ips_filename>\n");
    fprintf(stderr, "\t-h = add 128-byte \"Brooktrout\" header\n");
    fprintf(stderr, "\t-o = use supplied output filename base\n");
    exit(1);
}

int main(int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    char *ptr;
    char *nami, namo[512];
    int nbyts;
    int pagecnt = 1;
    struct infopkt_stream *ips;
    struct infopkt *ip;
    struct docpkt *docp;
    FILE *outfp;
    int data_received = 1;
    int header_flag = 0;
    int header_written = 0;
    int current_resolution = 0;
    int current_width = 0;
    static char output_base[512] = "g3data";
    int write_header BT_ARGS((FILE *,int,int));
    struct args_infopkt args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    
    while ((c = getopt (argc, argv, "ho:")) != EOF)
    {
        switch (c)
        {
          case 'h':
            header_flag = 1;
            break;
          case 'o':
            strcpy(output_base,optarg);
            break;
          default:
            usage ();
        }
    }
    
    nami = argv[optind++];
    
    if (optind != argc)
        usage ();

    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    BT_ZERO(args);
    args.fname = nami;
    args.fmode = "r";
    if ((ips = BfvInfopktOpen(&args)) == NULL)
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        printf ("BfvInfopktOpen: %s\n",bterr.long_msg);
        exit(1);
    }
    
    sprintf(namo, "%s.3%02d", output_base, pagecnt);
#ifndef USES_FAT_FILESYSTEM
    if ((outfp = fopen(namo, "w")) == NULL)
#else
    if ((outfp = fopen(namo, "wb")) == NULL)
#endif
    {
        printf("error opening output file\n");
        exit(1);
    }
    
    args.ips = ips;
    args.i_mode = INDIR_MODE_FOLLOW;
    while ((ip = BfvInfopktGet(&args)) != NULL)
    {
        switch(ip->type)
        {
          case INFOPKT_G3:
          case INFOPKT_BTG3:
            data_received = 1;
            ptr = (char *)ip + sizeof(struct g3pkt);
            nbyts = ip->length - sizeof(struct g3pkt);
            fwrite(ptr, sizeof(char), nbyts, outfp);
            break;
          case INFOPKT_DOCUMENT_PARAMETERS:
            docp = (struct docpkt *)ip;
            current_resolution = docp->resolution;
            current_width = docp->hor_width;
            /* fall through */
          case INFOPKT_BEGINNING_OF_PAGE:
          case INFOPKT_T30_PARAMETERS:
          case INFOPKT_BFT:
            if (data_received)
            {
                data_received = 0;
                if (pagecnt != 1)
                {
                    fclose(outfp);
                    sprintf(namo, "%s.3%02d", output_base, pagecnt);
#ifndef USES_FAT_FILESYSTEM
                    if ((outfp = fopen(namo, "w")) == NULL)
#else
                    if ((outfp = fopen(namo, "wb")) == NULL)
#endif
                    {
                        printf("error opening output file\n");
                        exit(1);
                    }
                    header_written = 0;
                }
                pagecnt++;
            }
            if (header_flag && !header_written)
            {
                write_header(outfp,current_resolution,current_width);
                header_written = 1;
            }
            break;
          case INFOPKT_SPEECH:
            data_received = 1;
            ptr = (char *)ip + sizeof(struct speechpkt);
            nbyts = ip->length - sizeof(struct speechpkt);
            fwrite(ptr, sizeof(char), nbyts, outfp);
            break;
          case INFOPKT_CVSD:
            data_received = 1;
            ptr = (char *)ip + sizeof(struct cvsdpkt);
            nbyts = ip->length - sizeof(struct cvsdpkt);
            fwrite(ptr, sizeof(char), nbyts, outfp);
            break;
          default:
            break;
        }
    }
    if (args.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        printf ("BfvInfopktGet: %s\n",bterr.long_msg);
    }
    BfvInfopktClose(&args);
    fclose(outfp);
    
    return(0);
}


#define HEADER_SIZE 128
#define HEADER_HORIZ_RES_OFF 4
#define HEADER_VERT_RES_OFF 6
#define HEADER_WIDTH_OFF 9

write_header(FILE *fp, int resolution, int width)
{
    static unsigned char header[HEADER_SIZE] = 
        { 0xbb, 1, 0, 1, 0xc8, 0, 0x64, 0, 1, 0xc0, 6, 1, 1};
    static int line_sizes[] = { 1728, 2048, 2432 };
    int horiz = 200, vert = 100;
    
    width = line_sizes[width];
    switch (resolution)
    {
      case RES_200H_100V:
        horiz = 200;
        vert = 100;
        break;
      case RES_200H_200V:
        horiz = 200;
        vert = 200;
        break;
      case RES_200H_400V:
        horiz = 200;
        vert = 400;
        break;
      case RES_300H_300V:
        horiz = 300;
        vert = 300;
        width = (width / 2) * 3;
        break;
      case RES_400H_400V:
        horiz = 400;
        vert = 400;
        width *= 2;
        break;
    }
    header[HEADER_HORIZ_RES_OFF] = horiz & 0xff;
    header[HEADER_HORIZ_RES_OFF + 1] = (horiz >> 8) & 0xff;
    header[HEADER_VERT_RES_OFF] = vert & 0xff;
    header[HEADER_VERT_RES_OFF + 1] = (vert >> 8) & 0xff;
    header[HEADER_WIDTH_OFF] = width & 0xff;
    header[HEADER_WIDTH_OFF + 1] = width >> 8;
    
    fwrite(header,1,sizeof(header),fp);
    return(0);
}
