/*
 * ivr.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The ivr program is a small, interactive, voice-response system that
permits users to record a message, play a message, receive a fax, and
send a fax. It uses the speech infopkt-file recording and playing
routines and the highest level infopkt-file fax receiving and sending
routines.
***************************************************************************/

/* User must provide the following 5 files, all of which are in the ivr_msg
   subdirectory which must be created. */

#define FILE_GREETING   "ivr_msg/greeting.pkt"
/* Should give a greeting, giving the choices:
   1 to leave a message, 2 to get a message,
   3 to receive a fax, 4 to send a fax, and 9 to rerecord the greeting. */

#define FILE_BYE        "ivr_msg/bye.pkt"
/* Should say goodbye. */

#define FILE_LEAVE      "ivr_msg/leave.pkt"
/* Should ask user to leave a 15 second msg to be recorded.
   Recording will last for 15 seconds or until a touchtone is pressed. */

#define FILE_RECORD     "ivr_msg/record.pkt"
/* Should ask user to begin a new greeting msg to be recorded.
   Recording will last for 15 seconds or until a touchtone is pressed. */

#define FAX_XMIT        "ivr_msg/fax_xmit.pkt"
/* Fax file to transmit */

/* These are created by ivr at runtime */
#define FILE_MSG        "ivr_msg/msg.pkt"
#define FILE_NEW        "ivr_msg/new.pkt"
#define FAX_RCV         "ivr_msg/fax_rcv.pkt"

/* Volume (gain) of beeps played */
#define BEEP_VOLUME -10


#include <string.h>
#include "btlib.h"

void dowork BT_ARGS((BTLINE *)), process_call BT_ARGS((BTLINE *));
int node_top BT_ARGS((BTLINE *)), node_leave_message BT_ARGS ((BTLINE *));
int node_new_greeting BT_ARGS((BTLINE *)),
    node_get_message BT_ARGS((BTLINE *));
int node_rcv_fax BT_ARGS((BTLINE *)), node_xmit_fax BT_ARGS((BTLINE *));
int play_func BT_ARGS((BTLINE *lp,char *arg));
void process_aoc BT_ARGS((BTLINE *lp));
void process_call_info BT_ARGS((CALL_RES *call_res));
#if defined(unix) || defined(M_XENIX)
#undef remove
#define remove unlink
#endif
#if (defined(unix) && !(defined(ATT40) || defined(SCOUNIX) || defined(SOLARIS) || defined(linux))) || defined(M_XENIX)
#define NO_RENAME
int rename BT_ARGS((char *,char *));
#endif

int unit;
int done;
int (*nodefunc) BT_ARGS((BTLINE *));
int lflag;

void
usage ()
{
    fprintf (stderr, "usage: ivr [-u unit] [-L]\n");
    fprintf (stderr, "  -u unitnum  use specified unit number\n");
    fprintf (stderr, "  -L          loop for testing\n");
    exit(1);
}

int
main (int argc, char **argv)
{
    extern char *optarg;
    extern int optind;
    int c;
    BTLINE *lp;
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    while ((c = getopt (argc, argv, "u:L")) != EOF)
    {
        switch (c)
        {
          case 'u':
            unit = atoi (optarg);
            break;
          case 'L':
            lflag = 1;
            break;
          default:
            usage ();
        }
    }
    
    if (optind != argc)
        usage ();

    BT_ZERO(args);
    args.unit = unit;
    if ((lp = BfvLineAttach (&args)) == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
        exit(1);
    }
    
    dowork(lp);
    
    return(0);
}

void dowork(BTLINE *lp)
{
    struct args_line_admin args_admin;
    BTERR bterr;
    
    for (;;)
    {
        BT_ZERO(args_admin);
        args_admin.config_file_name = "btcall.cfg";
        if (BfvLineReset (lp, &args_admin) < 0)
        {
            BfvErrorMessage(lp,&args_admin.res,&bterr);
            fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                     bterr.long_msg,args_admin.reset_status);
            exit(1);
        }
        
        process_call(lp);
        
        if (!lflag)
            exit(0);
    }
}

void
process_call(BTLINE *lp)
{
    int timedout;
    int val;
    int (*oldnodefunc) BT_ARGS((BTLINE *));
    struct infopkt_stream *ips;
    struct args_telephone args_tel;
    struct args_tone args_tone;
    struct args_infopkt args_infopkt;
    struct args_speech args_speech;
    BTERR bterr;
    
    BT_ZERO(args_tel);
    BfvLineWaitForCall (lp, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK || 
        args_tel.res.line_status != WAIT_FOR_CALL_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("BfvLineWaitForCall: %s\n",bterr.long_msg);
        exit(1);
    }
    process_call_info(&args_tel.call_res);
    BT_ZERO(args_tel);
    BfvLineAnswer (lp, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("BfvLineAnswer: %s\n",bterr.long_msg);
        exit(1);
    }

    BT_ZERO(args_tone);
    args_tone.decode_flag = DTMF_16TONE; 
    BfvToneDetectEnable(lp, &args_tone);
    
    nodefunc = node_top;
    
    for (timedout = 0; timedout <= 3; )
    {
        oldnodefunc = nodefunc;
        val = (*nodefunc)(lp);
        if (val)
            break;
        if (oldnodefunc != nodefunc)
            timedout = 0;
        else
            timedout++;
    }
    
    if (val >= 0)
    {
        BT_ZERO(args_infopkt);
        args_infopkt.fname = FILE_BYE;
        args_infopkt.fmode = "r";
        ips = BfvInfopktOpen (&args_infopkt);
        if (ips)
        {
            BT_ZERO(args_speech);
            args_speech.ips = ips;
            BfvSpeechPlay (lp, &args_speech);
            if (args_speech.res.status != BT_STATUS_OK &&
                args_speech.res.status != BT_STATUS_USER_TERMINATED)
            {
                BfvErrorMessage(lp,&args_speech.res,&bterr);
                printf ("BfvSpeechPlay: %s\n",bterr.long_msg);
                exit(1);
            }
            args_infopkt.ips = ips;
            BfvInfopktClose (&args_infopkt);
        }
        else
        {
            BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
            fprintf (stderr,"BfvInfopktOpen %s: %s\n",FILE_BYE,bterr.long_msg);
            exit(1);
        }
    }
    
    BfvLineTerminateCall (lp, &args_tel);
    process_aoc(lp);
}

/* play the speech data in the named file, stopping if a tone comes in.
 */
void
play_msg_stop_on_tone (BTLINE *lp, char *name)
{
    struct infopkt_stream *ips;
    struct args_infopkt args_infopkt;
    struct args_tone args_tone;
    struct args_speech args_speech;
    BTERR bterr;
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = name;
    args_infopkt.fmode = "r";
    ips = BfvInfopktOpen (&args_infopkt);
    
    if (ips)
    {
        BT_ZERO(args_tone);
        BfvToneFlush(lp, &args_tone);
        BT_ZERO(args_speech);
        args_speech.ips = ips;
        args_speech.func = play_func;
        BfvSpeechPlay (lp, &args_speech);
        if (args_speech.res.status != BT_STATUS_OK &&
            args_speech.res.status != BT_STATUS_USER_TERMINATED)
        {
            BfvErrorMessage(lp,&args_speech.res,&bterr);
            printf ("BfvSpeechPlay: %s\n",bterr.long_msg);
            exit(1);
        }
        args_infopkt.ips = ips;
        BfvInfopktClose (&args_infopkt);
    }
    else
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr,"BfvInfopktOpen %s: %s\n",name,bterr.long_msg);
        exit(1);
    }
}

/* play a message, but don't stop if a tone comes in */
void
play_msg (BTLINE *lp, char *name)
{
    struct infopkt_stream *ips;
    struct args_infopkt args_infopkt;
    struct args_speech args_speech;
    BTERR bterr;
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = name;
    args_infopkt.fmode = "r";
    ips = BfvInfopktOpen (&args_infopkt);
    
    if (ips)
    {
        BT_ZERO(args_speech);
        args_speech.ips = ips;
        BfvSpeechPlay (lp, &args_speech);
        if (args_speech.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_speech.res,&bterr);
            printf ("BfvSpeechPlay: %s\n",bterr.long_msg);
            exit(1);
        }
        args_infopkt.ips = ips;
        BfvInfopktClose (&args_infopkt);
    }
    else
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr,"BfvInfopktOpen %s: %s\n",name,bterr.long_msg);
        exit(1);
    }
}

/* each node is represented by a function.  When the function is 
 * called, it plays the node message, then takes node specific
 * action.  For most nodes, the message lists the menu of choices,
 * and the rest of the action is to wait for a tone.  If no
 * tone arrives in the node specific timeout, the function just
 * returns.  The higher level code calls the function again to
 * make it repeat the message and do its action again.  The
 * higher level code is responsible for deciding whether we have
 * stopped making progress and to hang up.  If a tone does arrive,
 * the node uses it to select a new node, and sets the function pointer
 * "nodefunc" to the new function, then returns.  The higher level
 * loop then calls the new nodefunc, which will play the new message,
 * etc. Other actions include starting up the fax board, etc.
 */

/* this is the top level node */
int
node_top (BTLINE *lp)
{
    int digit;
    struct args_tone args_tone;
    BTERR bterr;
    
    /* this message says:
     *
     *  Press 1 to leave a message, press 2 to get the message,
     *  press 3 to send a fax, press 4 to receive a fax,
     *  press 9 to record a new greeting.
     */
    play_msg_stop_on_tone (lp, FILE_GREETING);
    
    /* wait up to 5 seconds for a tone, then repeat message */
    BT_ZERO(args_tone);
    args_tone.millisecs = 5000L;
    digit = BfvToneGet (lp, &args_tone);
    if (digit < 0)
    {
        if (args_tone.res.status != BT_STATUS_TIMEOUT)
        {
            BfvErrorMessage(lp,&args_tone.res,&bterr);
            printf ("BfvToneGet: %s\n",bterr.long_msg);
            exit(1);
        }
        return (0);
    }
    
    switch (digit)
    {
      case '1':
        nodefunc = node_leave_message;
        break;
      case '2':
        nodefunc = node_get_message;
        break;
      case '3':
        /* Receive a fax means ivr program transmits */
        nodefunc = node_xmit_fax;
        break;
      case '4':
        /* Transmit a fax means ivr program receives */
        nodefunc = node_rcv_fax;
        break;
      case '9':
        nodefunc = node_new_greeting;
        break;
      case '*':
        return (1);
    }
    return (0);
}

int
node_new_greeting (BTLINE *lp)
{
    struct infopkt_stream *ips;
    struct args_infopkt args_infopkt;
    struct args_tone args_tone;
    struct args_speech args_speech;
    BTERR bterr;
    
    play_msg_stop_on_tone (lp, FILE_RECORD);
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = FILE_NEW;
    args_infopkt.fmode = "w";
    if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr,"BfvInfopktOpen %s: %s\n",FILE_NEW,bterr.long_msg);
        exit(1);
    }
    
    BT_ZERO(args_tone);
    args_tone.toneid = TONEID_CCITT_1;
    args_tone.millisecs = 100L;
    args_tone.volume = BEEP_VOLUME;
    BfvTonePlayBeep(lp, &args_tone);
    
    printf ("Recording...\n");
    BfvToneFlush(lp, &args_tone);
    
    BT_ZERO(args_speech);
    args_speech.ips = ips;
    args_speech.coding_fmt = CODE_PCM_ULAW;
    args_speech.rate = RATE_8000;
    args_speech.bits_per_samp = BITS_8;
    args_speech.afe_rate = AFE_8000;
    args_speech.data_fmt = SPCH_MSB;
    args_speech.timeout = 15000L;
    args_speech.silence_timeout = 1000L;
    args_speech.s_compr = 1;
    args_speech.func = play_func;
    BfvSpeechRecord (lp, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK &&
        args_speech.res.status != BT_STATUS_USER_TERMINATED &&
        args_speech.res.status != BT_STATUS_TIMEOUT)
    {
        BfvErrorMessage(lp,&args_speech.res,&bterr);
        printf ("BfvSpeechRecord: %s\n",bterr.long_msg);
        exit(1);
    }
    
    BfvToneFlush(lp, &args_tone);
    
    args_infopkt.ips = ips;
    BfvInfopktClose (&args_infopkt);
    
    remove (FILE_GREETING);
    rename (FILE_NEW,FILE_GREETING);
    
    nodefunc = node_top;
    
    return (0);
}


int
node_leave_message (BTLINE *lp)
{
    struct infopkt_stream *ips;
    struct args_infopkt args_infopkt;
    struct args_tone args_tone;
    struct args_speech args_speech;
    BTERR bterr;
    
    play_msg_stop_on_tone (lp, FILE_LEAVE);
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = FILE_NEW;
    args_infopkt.fmode = "w";
    if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr,"BfvInfopktOpen %s: %s\n",FILE_NEW,bterr.long_msg);
        exit(1);
    }
    
    BT_ZERO(args_tone);
    args_tone.toneid = TONEID_CCITT_1;
    args_tone.millisecs = 100L;
    args_tone.volume = BEEP_VOLUME;
    BfvTonePlayBeep(lp, &args_tone);
    
    printf ("Recording...\n");
    BfvToneFlush(lp, &args_tone);
    
    BT_ZERO(args_speech);
    args_speech.ips = ips;
    args_speech.coding_fmt = CODE_PCM_ULAW;
    args_speech.rate = RATE_8000;
    args_speech.bits_per_samp = BITS_8;
    args_speech.afe_rate = AFE_8000;
    args_speech.data_fmt = SPCH_MSB;
    args_speech.timeout = 15000L;
    args_speech.silence_timeout = 1000L;
    args_speech.s_compr = 1;
    args_speech.func = play_func;
    BfvSpeechRecord (lp, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK &&
        args_speech.res.status != BT_STATUS_USER_TERMINATED &&
        args_speech.res.status != BT_STATUS_TIMEOUT)
    {
        BfvErrorMessage(lp,&args_speech.res,&bterr);
        printf ("BfvSpeechRecord: %s\n",bterr.long_msg);
        exit(1);
    }
    
    BfvToneFlush(lp, &args_tone);
    
    args_infopkt.ips = ips;
    BfvInfopktClose (&args_infopkt);
    
    remove (FILE_MSG);
    rename (FILE_NEW,FILE_MSG);
    
    nodefunc = node_top;
    
    return (0);
}

/* tell play to stop as soon as any tone appears in the tone buffer */
int
play_func (BTLINE *lp,char *arg)
{
    struct args_tone args;
    
    BT_ZERO(args);
    if (BfvTonePeek (lp, &args) < 0)
        return (0); /* keep going */
    else
        return (1); /* stop now */
}

int
node_get_message (BTLINE *lp)
{
    play_msg (lp, FILE_MSG);
    nodefunc = node_top;
    return (0);
}

int
node_rcv_fax (BTLINE *lp)
{
    struct infopkt_stream *ips = NULL;
    struct args_fax args_fax;
    struct args_tone args_tone;
    struct args_infopkt args_infopkt;
    BTERR bterr;
    
    nodefunc = node_top;
    
    BT_ZERO(args_tone);
    args_tone.toneid = TONEID_CCITT_1;
    args_tone.millisecs = 100L;
    args_tone.volume = BEEP_VOLUME;
    BfvTonePlayBeep(lp, &args_tone);
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = FAX_RCV;
    args_infopkt.fmode = "w";
    if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr,"BfvInfopktOpen %s: %s\n",FAX_RCV,bterr.long_msg);
        exit(1);
    }
    
    BT_ZERO(args_tone);
    BfvToneDetectDisable (lp,&args_tone);
    
    BT_ZERO(args_fax);
    args_fax.r_ips = ips;
    args_fax.local_id = "this is the local id";
    BfvFaxReceive (lp, &args_fax);
    if (args_fax.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_fax.res,&bterr);
        printf ("BfvFaxReceive: %s\n",bterr.long_msg);
    }
    
    args_infopkt.ips = ips;
    BfvInfopktClose (&args_infopkt);
    
    return (-1);
}

int
node_xmit_fax (BTLINE *lp)
{
    struct infopkt_stream *ips = NULL;
    struct args_fax args_fax;
    struct args_tone args_tone;
    struct args_infopkt args_infopkt;
    struct args_telephone args_tel;
    BTERR bterr;
    
    nodefunc = node_top;
    
    BT_ZERO(args_tone);
    args_tone.toneid = TONEID_CCITT_1;
    args_tone.millisecs = 100L;
    args_tone.volume = BEEP_VOLUME;
    BfvTonePlayBeep(lp, &args_tone);
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = FAX_XMIT;
    args_infopkt.fmode = "r";
    if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr,"BfvInfopktOpen %s: %s\n",FAX_XMIT,bterr.long_msg);
        exit(1);
    }
    
    BT_ZERO(args_tone);
    BfvToneDetectDisable (lp,&args_tone);
    
    BT_ZERO(args_tel);
    args_tel.phonenum = "";
    args_tel.call_protocol_code = CALL_PROTOCOL_FAX;
    BfvLineOriginateCall (lp, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK || 
        LINE_STATE(lp) != LINE_STATE_CONNECTED)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("BfvLineOriginateCall: %s\n",bterr.long_msg);
        goto x_done;
    }
    
    BT_ZERO(args_fax);
    args_fax.s_ips = ips;
    args_fax.local_id = "this is the local id";
    BfvFaxSend (lp, &args_fax);
    if (args_fax.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_fax.res,&bterr);
        printf ("BfvFaxSend: %s\n",bterr.long_msg);
    }
    
x_done:
    
    BT_ZERO(args_tel);
    BfvLineTerminateCall (lp, &args_tel);

    args_infopkt.ips = ips;
    BfvInfopktClose (&args_infopkt);
    return (-1);
}

#ifdef NO_RENAME
/* Some versions of Sys V Unix don't have rename */
rename(char *f1, char *f2)
{
    unlink(f2);
    if (link(f1,f2) < 0)
        return(-1);
    if (unlink(f1) < 0)
        return(-1);
    return(0);
}
#endif

void
process_aoc(BTLINE *lp)
{
    AOC_INFO *aoc = LINE_AOC_INFO(lp);
        
    if (aoc == NULL || aoc->status == BT_AOC_UNAVAILABLE)
        printf("Charge information unavailable.\n");
    else if (aoc->status == BT_AOC_FREE)
        printf("Charge information: FREE.\n");
    else if (aoc->status == BT_AOC_UNITS)
        printf("Charge information: Unit type %d, value %ld\n",
               aoc->charge_info,aoc->charge);
    else if (aoc->status == BT_AOC_CURRENCY)
    {
        static char *currency_names[] = {
            "1/1000","1/100","1/10","1","10","100","1000" };
#define NUM_CURRENCIES (sizeof(currency_names) / sizeof(*currency_names))
        char *cur_name;
            
        cur_name = (0 <= aoc->charge_info &&
                    aoc->charge_info < NUM_CURRENCIES) ?
            currency_names[aoc->charge_info] :
            "Invalid currency";
        printf("Charge information: Currency multiplier %s (%d), value %ld\n",
               cur_name,aoc->charge_info,aoc->charge);
    }
    else
        printf("Charge information: Bad status value %d\n",
               aoc->status);
}

void
process_call_info(CALL_RES *call_res)
{
    if (strlen(call_res->called_party_number))
        printf("Called party number: \"%s\"\n",
            call_res->called_party_number);
    if (strlen(call_res->called_party_subaddress))
        printf("Called party subaddress: \"%s\"\n",
           call_res->called_party_subaddress);
    if (strlen(call_res->calling_party_number))
        printf("Calling party number: \"%s\"\n",
           call_res->calling_party_number);
    if (strlen(call_res->calling_party_subaddress))
        printf("Calling party subaddress: \"%s\"\n",
           call_res->calling_party_subaddress);
    if (strlen(call_res->redir_number))
        printf("Redirection number: \"%s\", reason %X\n",
           call_res->redir_number, call_res->redir_reason);
}
