/*
 * mkbft.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/* Used to create a T.434 BFT file from a data file. */

/* For Windows only:
   If an application uses a DLL version of the Bfv API library, under some
   conditions the application and the Bfv API library may link to different
   versions or instances of the compiler-supplied C runtime (CRT) libraries.
   This creates incompatibilities between file pointers or allocated memory
   among the application and the Bfv API library.

   Under these conditions, the application must use the Bfv API supplied _dll_
   versions of the file I/O and memory allocation functions if the pointer or
   memory is passed to/from the application and the library.

   The Bfv API DLL libraries links with MSVCRT.DLL. When using the Visual
   Studio mechanism to compile for debugging, applications may be linked with
   MSVCRTD.DLL. When using Visual Studio.NET, applications link with
   MSVCR71.DLL. When using the DLLS version of the Bfv API library, it also
   links with MVCRT.DLL, but it is a different instance of the CRT library.
*/

#include <string.h>
#if defined(unix) || defined(M_XENIX) || defined(_AIX)
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#ifndef T434_STANDALONE
#include "btlib.h"
#else
#include <stdio.h>
#include "t434.h"
#endif

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

unsigned char *do_emit BT_ARGS((unsigned tag, unsigned subtag,
                                unsigned char *buf,long *length,
                                char *arg));

char *infname;
struct stat fileinfo;
unsigned filesize;

/* Version to emit. 0 -> latest supported version */
int version = 0;

/* Define EMIT_BT_OLD_COMPAT to emit the data_file_content in the format
   used by the old Brooktrout code, and many other implementations.
   In that format, data_file_content contains the actual data to be
   sent; otherwise the actual data is contained within tagged fields.
   Also in that format, the year in dates is specified with 2 digits;
   otherwise the year is specified with 4 digits. */
#if 0
#define EMIT_BT_OLD_COMPAT
#endif

void usage()
{
    fprintf (stderr, "usage: mkbft -o <t434file> <datafile> ...\n");
    exit(1);
}

int 
main (int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    char *outfname = NULL;
    FILE *infp, *outfp;
    struct args_t434 args;
    struct args_debug_mode args_debug;
#ifndef T434_STANDALONE
    BTERR bterr;
    
    BT_API_SET_VER();
#endif
    
    while ((c = getopt (argc, argv, "o:")) != EOF)
    {
        switch (c)
        {
          case 'o':
            outfname = optarg;
            break;
          default:
            usage ();
        }
    }
    
    if (optind >= argc)
        usage();

    if (outfname == NULL)
        usage();

#ifdef USES_FAT_FILESYSTEM
#ifdef MILL_DLL
    outfp = _dll_fopen(outfname,"wb+");
#else
    outfp = fopen(outfname,"wb+");
#endif
#else
#ifdef MILL_DLL
    outfp = _dll_fopen(outfname,"w+");
#else
    outfp = fopen(outfname,"w+");
#endif
#endif
    if (outfp == NULL)
    {
        fprintf(stderr,"Error opening %s\n",outfname);
        exit(1);
    }
    
#ifndef T434_STANDALONE
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
#endif
    
    while (optind < argc)
    {
        infname = argv[optind++];
        if (stat(infname,&fileinfo) < 0)
        {
            fprintf(stderr,"Error opening %s\n",infname);
            exit(1);
        }
#ifndef USES_FAT_FILESYSTEM
#ifdef MILL_DLL
        infp = _dll_fopen(infname,"r");
#else
        infp = fopen(infname,"r");
#endif
#else
#ifdef MILL_DLL
        infp = _dll_fopen(infname,"rb");
#else
        infp = fopen(infname,"rb");
#endif
#endif
        if (infp == NULL)
        {
            fprintf(stderr,"Error opening %s\n",infname);
            exit(1);
        }
        
#ifndef T434_STANDALONE
        BT_ZERO(args);
#else
        memset(&args,0,sizeof(args));
#endif
        args.fp = outfp;
        args.emit_func = do_emit;
        args.arg = (char *)infp;
        args.t434_version = version;
#ifdef EMIT_BT_OLD_COMPAT
        args.raw_content = 1;
        args.emit_year2 = 1;
#endif
        if (BfvT434EmitFile (&args))
        {
#ifndef T434_STANDALONE
            BfvErrorMessage(NULL,&args.res,&bterr);
            fprintf (stderr, "BfvT434EmitFile: %s\n",bterr.long_msg);
#else
            fprintf (stderr, "BfvT434EmitFile error\n");
#endif
            exit(1);
        }
#ifdef MILL_DLL
        _dll_fclose(infp);
#else
        fclose(infp);
#endif
    }
#ifdef MILL_DLL
    _dll_fclose(outfp);
#else
    fclose(outfp);
#endif
    
    return(0);
}

unsigned char *do_emit (unsigned tag,unsigned subtag,unsigned char *buf,
                        long *length,char *arg)
{
    FILE *infp, *outfp;
    char fbuf[512];
    int n;
    long offset;
    struct gentm *gt;
    
#if 0
{
    struct args_t434 args;
        
#ifndef T434_STANDALONE
    BT_ZERO(args);
#else
    memset(&args,0,sizeof(args));
#endif
    args.tag = tag;
    strcpy(fbuf,BfvT434TagName(&args));
#ifndef T434_STANDALONE
    BT_ZERO(args);
#else
    memset(&args,0,sizeof(args));
#endif
    args.tag = subtag;
    printf("Emit %s type %s\n",fbuf,BfvT434TagName(&args));
}
#endif
    
    switch (tag)
    {
      case T434_filename:
        strcpy((char *)buf,infname);
        *length = strlen(infname);
        return(buf);
      case T434_filesize:
        infp = (FILE *)arg;
#ifdef MILL_DLL
        offset = _dll_ftell(infp);
        if (_dll_fseek(infp,0L,SEEK_END))
#else
        offset = ftell(infp);
        if (fseek(infp,0L,SEEK_END))
#endif
        {
            printf("Error going to eof\n");
            exit(1);
        }
#ifdef MILL_DLL
        filesize = _dll_ftell(infp) - offset;
#else
        filesize = ftell(infp) - offset;
#endif
        buf[0] = filesize & 0xff;
        buf[1] = (filesize >> 8) & 0xff;
        buf[2] = (filesize >> 16) & 0xff;
        buf[3] = (filesize >> 24) & 0xff;
#ifdef MILL_DLL
        if (_dll_fseek(infp,offset,SEEK_SET))
#else
        if (fseek(infp,offset,SEEK_SET))
#endif
        {
            printf("Error going to offset\n");
            exit(1);
        }
        return(buf);
      case T434_dtime_creation:
      case T434_dtime_last_modification:
        gt = (struct gentm *)buf;
        gt->time_diff_mins = GEN_TIME_DIFF_LOCAL;
        /* Although both mtime (modification time) and ctime (creation time)
           are available in many environments, we only use the modification
           time. The creation time retrieved by stat is often not a
           useful quantity. We supply both the modification and creation
           times in the T.434 BFT file because some vendors require one
           of these time fields, some the other. */
        offset = fileinfo.st_mtime;
        memcpy((char *)&gt->tm,(char *)localtime(&offset),sizeof(gt->tm));
        return(buf);
      case T434_data_file_content:
        infp = (FILE *)arg;
        outfp = (FILE *)buf;
#ifdef MILL_DLL
        while ((n = _dll_fread(fbuf,1,sizeof(fbuf),infp)) > 0)
            if (_dll_fwrite(fbuf,1,n,outfp) != (unsigned)n)
#else
        while ((n = fread(fbuf,1,sizeof(fbuf),infp)) > 0)
            if (fwrite(fbuf,1,n,outfp) != (unsigned)n)
#endif
            {
                *length = -1L;
                return(NULL);
            }
        return(buf);
      default:
        return(NULL);
    }
}
