/*
 * mktiff.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/* This program will convert raw G3 files into a valid TIFF Class F file,
   taking the resolution, width, and number of scan lines from the 128-byte
   Brooktrout header into account. If the header does not exist or the
   number of lines is 0, then the actual count of lines from the G3 file
   is computed.

   Files must be in MH/MSB format, with EOLs byte-aligned.
*/

/* For Windows only:
   If an application uses a DLL version of the Bfv API library, under some
   conditions the application and the Bfv API library may link to different
   versions or instances of the compiler-supplied C runtime (CRT) libraries.
   This creates incompatibilities between file pointers or allocated memory
   among the application and the Bfv API library.

   Under these conditions, the application must use the Bfv API supplied _dll_
   versions of the file I/O and memory allocation functions if the pointer or
   memory is passed to/from the application and the library.

   The Bfv API DLL libraries links with MSVCRT.DLL. When using the Visual
   Studio mechanism to compile for debugging, applications may be linked with
   MSVCRTD.DLL. When using Visual Studio.NET, applications link with
   MSVCR71.DLL. When using the DLLS version of the Bfv API library, it also
   links with MVCRT.DLL, but it is a different instance of the CRT library.
*/

#include <string.h>
#if defined(unix) || defined(M_XENIX) || defined(_AIX)
#include <unistd.h>
#endif
#include "btlib.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

int total_pages;

void usage()
{
    fprintf(stderr, "Usage: mktiff -o <tiff_file> <g3filename> ...\n");
    exit(1);
}

int process_file();

int 
main(int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    char *infile, *outfile;
    TFILE *tp;
    FILE *infp;
    int tmp;
    struct args_tiff args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    
    while ((c = getopt (argc, argv, "o:")) != EOF)
    {
        switch (c)
        {
          case 'o':
            outfile = optarg;
            break;
          default:
            usage ();
        }
    }
    
    if ((total_pages = argc - optind) <= 0)
        usage ();

    if (outfile == NULL)
        usage ();

    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    BT_ZERO(args);
    args.fname = outfile;
    args.fmode = "w";
    if ((tp = BfvTiffOpen(&args)) == NULL)
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr, "BfvTiffOpen: %s\n",bterr.long_msg);
        exit (1);
    }
    
    while (optind < argc)
    {
        infile = argv[optind++];
#ifdef USES_FAT_FILESYSTEM
        if ((infp = fopen(infile,"rb")) == NULL)
#else
        if ((infp = fopen(infile,"r")) == NULL)
#endif
        {
            fprintf(stderr,"Can't read %s\n",infile);
            break;
        }
            
        tmp = process_file(infp,tp);
        fclose(infp);
        if (tmp < 0)
        {
            fprintf(stderr,"Error writing output file\n");
            break;
        }
    }
    args.tp = tp;
    BfvTiffClose(&args);
    
    return(0);
}

#define HEADER_SIZE             128
/* Number of initial bytes of header to check */
#define HEADER_COMP_SIZE        1  /* This does not check versions */
#define HEADER_HORIZ_RES_OFF    4
#define HEADER_VERT_RES_OFF     6
#define HEADER_WIDTH_OFF        9
#define HEADER_NUM_LINES_OFF    45

void 
getheader(FILE *infile, int *resolution, int *width, short *num_lines)
{
    static unsigned char std_header[HEADER_SIZE] = 
        { 0xbb, 1, 0, 1, 0xc8, 0, 0x64, 0, 1, 0xc0, 6, 1, 1};
    unsigned char head[HEADER_SIZE];    /* Holds group 3 header info */
    int i;
    int horiz, vert;
    
    i = fread(head,HEADER_SIZE,1,infile);
    
    /* If the header is found, read from it and skip over it. */
    if (i == 1 && !memcmp(head,std_header,HEADER_COMP_SIZE))
    {
        horiz = head[HEADER_HORIZ_RES_OFF] |
            (head[HEADER_HORIZ_RES_OFF + 1] << 8);
        vert = head[HEADER_VERT_RES_OFF] |
            (head[HEADER_VERT_RES_OFF + 1] << 8);
        
        if (horiz == 200 && vert == 100)
            *resolution = RES_200H_100V;
        else if (horiz == 200 && vert == 200)
            *resolution = RES_200H_200V;
        else if (horiz == 200 && vert == 400)
            *resolution = RES_200H_400V;
        else if (horiz == 300 && vert == 300)
            *resolution = RES_300H_300V;
        else if (horiz == 400 && vert == 400)
            *resolution = RES_400H_400V;
        else if (horiz == 600 && vert == 600)
            *resolution = RES_600H_600V;
        else if (horiz == 1200 && vert == 1200)
            *resolution = RES_1200H_1200V;
        else if (horiz == 300 && vert == 600)
            *resolution = RES_300H_600V;
        else if (horiz == 400 && vert == 800)
            *resolution = RES_400H_800V;
        else if (horiz == 600 && vert == 1200)
            *resolution = RES_600H_1200V;
        else if (horiz == 100 && vert == 100)
            *resolution = RES_100H_100V;
        else
            *resolution = RES_200H_100V;

        *width = head[HEADER_WIDTH_OFF] | (head[HEADER_WIDTH_OFF + 1] << 8);
        *num_lines = head[HEADER_NUM_LINES_OFF] |
            (head[HEADER_NUM_LINES_OFF + 1] << 8);
    }
    else
    {
        *resolution = RES_200H_100V;
        *width = 1728;
        *num_lines = 0;
        fseek(infile,0L,SEEK_SET);
    }
}

int zero_count;
int rtc_seen;
int consec_eols;
int eol_seen;

void 
reset_line_count()
{
    zero_count = 0;
    rtc_seen = 0;
    consec_eols = 0;
    eol_seen = 0;
}

/* Count number of MH lines in file. When we see an EOL, we add to the line
   count when we see non-EOL data after the EOL. We will assume an RTC
   at 3 consecutive EOLs, and not process any more data. This could
   be simplified by making use of the assumption of byte-aligned EOLs... */
void 
count_lines(unsigned char *buf, int size, unsigned *num)
{
    int bit_pos;
    unsigned char ctmp;
    
    while (size-- && !rtc_seen)
    {
        ctmp = *buf++;
        for (bit_pos = 7; bit_pos >= 0; bit_pos--)
        {
            if (ctmp & (1 << bit_pos))
            {
                if (zero_count >= 11)
                {
                    if (++consec_eols >= 3)
                    {
                        rtc_seen = 1;
                        break;
                    }
                    eol_seen = 1;
                }
                else
                {
                    consec_eols = 0;
                    if (eol_seen)
                    {
                        (*num)++;
                        eol_seen = 0;
                    }
                }
                zero_count = 0;
            }
            else
                zero_count++;
        }
    }
}

static struct ifd_field our_ifd[] = {
    {TAG_NEWSUBFILETYPE,FT_LONG,1,2},
    {TAG_IMAGEWIDTH,FT_LONG,1,1728},
#define OFF_IM_WIDTH 1
    {TAG_IMAGELENGTH,FT_LONG,1,0},
#define OFF_IMAGELENGTH 2
    {TAG_BITSPERSAMPLE,FT_SHORT,1,1},
    {TAG_COMPRESSION,FT_SHORT,1,3},
    {TAG_PHOTOMETRICINTERP,FT_SHORT,1,0},
    {TAG_FILLORDER,FT_SHORT,1,1},
    {TAG_SAMPLESPERPIXEL,FT_SHORT,1,1},
    {TAG_ROWSPERSTRIP,FT_LONG,1,0},
#define OFF_ROWSPERSTRIP 8
    {TAG_XRESOLUTION,FT_RATIONAL,1,0},
#define OFF_XRES 9
    {TAG_YRESOLUTION,FT_RATIONAL,1,0},
#define OFF_YRES 10
    {TAG_GROUP3OPTIONS,FT_LONG,1,4},
    {TAG_RESOLUTIONUNIT,FT_SHORT,1,TIFF_RES_UNIT_INCHES},
    {TAG_PAGENUMBER,FT_SHORT,2,0},
#define OFF_PAGENUM 13
};
#define NUM_TAGS (sizeof(our_ifd) / sizeof(struct ifd_field))

int
process_file(FILE *fp, TFILE *tp)
{
    static int current_page_num;
    int current_resolution;
    int current_width;
    short num_lines_hdr;
    unsigned num_lines;
    int n;
    unsigned char buffer[1024];
    int i;
    struct args_tiff args;
    BTERR bterr;
    
    getheader(fp,&current_resolution,&current_width,&num_lines_hdr);
    if ((num_lines = num_lines_hdr) == 0)
        reset_line_count();

    while ((n = fread(buffer,1,sizeof(buffer),fp)) > 0)
    {
        BT_ZERO(args);
        args.tp = tp;
        args.buf = buffer;
        args.size = n;
        if (BfvTiffWriteImage (&args))
        {
            BfvErrorMessage(NULL,&args.res,&bterr);
            fprintf (stderr, "BfvTiffWriteImage: %s\n",bterr.long_msg);
            return(-1);
        }
        /* If header does not contain lines, count lines from buffers */
        if (num_lines_hdr == 0)
            count_lines(buffer,n,&num_lines);
    }
    BT_ZERO(args);
    args.tp = tp;
    if (BfvTiffWriteImage (&args))
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr, "BfvTiffWriteImage: %s\n",bterr.long_msg);
        return(-1);
    }
    
#ifdef MILL_DLL
    _dll_fseek(TIFF_FP(tp),0L,SEEK_END);
#else
    fseek(TIFF_FP(tp),0L,SEEK_END);
#endif
    
    args.tp = tp;
    args.resolution = current_resolution;
    BfvTiffWriteRes (&args);
    our_ifd[OFF_XRES].offset = args.horiz_offset;
    our_ifd[OFF_YRES].offset = args.vert_offset;
    
#if 0
    /* Adjust width value for X resolution */
    if (current_resolution == RES_300H_300V)
        current_width = (current_width / 2) * 3;
    else if (current_resolution == RES_400H_400V)
        current_width *= 2;
#endif
    our_ifd[OFF_IM_WIDTH].offset = current_width;
    our_ifd[OFF_PAGENUM].offset =
        (((unsigned)total_pages) << 16) | current_page_num;
    our_ifd[OFF_IMAGELENGTH].offset = our_ifd[OFF_ROWSPERSTRIP].offset =
        num_lines;
    
    for (i = 0; i < NUM_TAGS; i++)
    {
        args.tp = tp;
        args.ifd_field = &our_ifd[i];
        if (BfvTiffWriteIFD (&args))
        {
            BfvErrorMessage(NULL,&args.res,&bterr);
            fprintf (stderr, "BfvTiffWriteIFD: %s\n",bterr.long_msg);
            return(-1);
        }
    }
    args.tp = tp;
    args.ifd_field = NULL;
    if (BfvTiffWriteIFD (&args))
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr, "BfvTiffWriteIFD: %s\n",bterr.long_msg);
        return(-1);
    }
    
    current_page_num++;
    return(0);
}
