/*
 * playp.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The playp program waits for a call, then plays the specified sequence
of phrases from the specified prompt file until the sequence completes
or until the user presses the # key.
***************************************************************************/

#include <string.h>
#include <ctype.h>
#include "btlib.h"

int play_func BT_ARGS((BTLINE *lp,char *arg));
void dowork BT_ARGS((BTLINE *lp,char *name,int ind,int argc,char **argv));
void process_aoc BT_ARGS((BTLINE *lp));
void process_call_info BT_ARGS((CALL_RES *call_res));

int unit;
int done;

void
usage ()
{
    fprintf (stderr, "usage: playp [-u unit] prompt_file phrase_num ...\n");
    exit (1);
}

int
main (int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    BTLINE *lp;
    char *name;
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    while ((c = getopt (argc, argv, "u:")) != EOF)
    {
        switch (c)
        {
          case 'u':
            unit = atoi (optarg);
            break;
          default:
            usage ();
        }
    }
    
    if (optind >= argc)
        usage ();

    name = argv[optind++];
    
    if (optind >= argc)
        usage ();

    BT_ZERO(args);
    args.unit = unit;
    if ((lp = BfvLineAttach (&args)) == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
        exit (1);
    }
    
    dowork(lp,name,optind,argc,argv);
    
    return(0);
}

void
dowork(BTLINE *lp, char *name, int ind, int argc, char **argv)
{
    PROMPT_MAP *prompt_map;
    BTERR bterr;
    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_speech args_speech;
    struct args_infopkt args_infopkt;
    struct args_tone args_tone;
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (lp, &args_admin) < 0)
    {
        BfvErrorMessage(lp,&args_admin.res,&bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                 bterr.long_msg,args_admin.reset_status);
        exit (1);
    }
    
    if (!LINE_HAS_CAP(lp,CAP_SPEECH_PLAY))
    {
        fprintf (stderr, "unit %d does not support speech playback\n",
                 LINE_UNIT_NUM(lp));
        exit (1);
    }
    
    BT_ZERO(args_infopkt);
    args_infopkt.fname = name;
    args_infopkt.fmode = "r";
    if ((prompt_map = BfvPromptOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
        fprintf (stderr,"BfvPromptOpen: %s\n",bterr.long_msg);
        exit (1);
    }
    
    BT_ZERO(args_tel);
    BfvLineWaitForCall (lp, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK || 
        args_tel.res.line_status != WAIT_FOR_CALL_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("BfvLineWaitForCall: %s\n",bterr.long_msg);
        exit (1);
    }
    process_call_info(&args_tel.call_res);
    
    BT_ZERO(args_tel);
    BfvLineAnswer (lp, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("BfvLineAnswer: %s\n",bterr.long_msg);
        exit (1);
    }
    
    BT_ZERO(args_tone);
    args_tone.decode_flag = DTMF_12TONE; 
    BfvToneDetectEnable(lp, &args_tone);
    
    /* play all prompts done */
    while (ind < argc)
    {
        BT_ZERO(args_speech);
        args_speech.prompt_map = prompt_map;
        args_speech.phrase_number = atoi(argv[ind++]);
        args_speech.func = play_func;
        BfvPromptPlay (lp, &args_speech);
        if (args_speech.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_speech.res,&bterr);
            printf ("BfvPromptPlay: %s\n",bterr.long_msg);
            goto got_err;
        }
    }
    BT_ZERO(args_speech);
    args_speech.func = play_func;
    BfvPromptPlay (lp, &args_speech);
    BT_ZERO(args_tone);
    BfvToneDetectDisable (lp,&args_tone);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_speech.res,&bterr);
        printf ("BfvPromptPlay: %s\n",bterr.long_msg);
    }
    
got_err:
    args_infopkt.prompt_map = prompt_map;
    BfvPromptClose (&args_infopkt);
    
    BT_ZERO(args_tel);
    BfvLineTerminateCall (lp, &args_tel);

    process_aoc(lp);
    
    BT_ZERO(args_admin);
    BfvLineDetach (lp, &args_admin);
    
    exit(0);
}

/* Abort on '#' DTMF */
int
play_func (BTLINE *lp, char *arg)
{
    struct args_tone args;
    
    BT_ZERO(args);
    if (BfvToneGet(lp,&args) == '#')
        return(1);      /* Stop playing */
    if (args.res.status != BT_STATUS_OK &&
        args.res.status != BT_STATUS_TIMEOUT)
    {
        BTERR bterr;
        
        BfvErrorMessage(lp,&args.res,&bterr);
        printf ("BfvToneGet: %s\n",bterr.long_msg);
        return(1);
    }
    return (0); /* Keep playing */
}

void
process_aoc(BTLINE *lp)
{
    AOC_INFO *aoc = LINE_AOC_INFO(lp);
        
    if (aoc == NULL || aoc->status == BT_AOC_UNAVAILABLE)
        printf("Charge information unavailable.\n");
    else if (aoc->status == BT_AOC_FREE)
        printf("Charge information: FREE.\n");
    else if (aoc->status == BT_AOC_UNITS)
        printf("Charge information: Unit type %d, value %ld\n",
               aoc->charge_info,aoc->charge);
    else if (aoc->status == BT_AOC_CURRENCY)
    {
        static char *currency_names[] = {
            "1/1000","1/100","1/10","1","10","100","1000" };
#define NUM_CURRENCIES (sizeof(currency_names) / sizeof(*currency_names))
        char *cur_name;
            
        cur_name = (0 <= aoc->charge_info &&
                    aoc->charge_info < NUM_CURRENCIES) ?
            currency_names[aoc->charge_info] :
            "Invalid currency";
        printf("Charge information: Currency multiplier %s (%d), value %ld\n",
               cur_name,aoc->charge_info,aoc->charge);
    }
    else
        printf("Charge information: Bad status value %d\n",
               aoc->status);
}

void
process_call_info(CALL_RES *call_res)
{
    if (strlen(call_res->called_party_number))
        printf("Called party number: \"%s\"\n",
            call_res->called_party_number);
    if (strlen(call_res->called_party_subaddress))
        printf("Called party subaddress: \"%s\"\n",
           call_res->called_party_subaddress);
    if (strlen(call_res->calling_party_number))
        printf("Calling party number: \"%s\"\n",
           call_res->calling_party_number);
    if (strlen(call_res->calling_party_subaddress))
        printf("Calling party subaddress: \"%s\"\n",
           call_res->calling_party_subaddress);
    if (strlen(call_res->redir_number))
        printf("Redirection number: \"%s\", reason %X\n",
           call_res->redir_number, call_res->redir_reason);
}
