/*
 * tones.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The tones program uses DTMF detection routines to detect and
display incoming touchtones and DTMF generation and single
frequency tone generation routines to produce touchtones and other
tones.
***************************************************************************/

#include <string.h>
#include "btlib.h"

#define TRUE 1
#define FALSE 0

int cflag = 0;
int get;
char *cnum;
int done;

void process_aoc BT_ARGS((BTLINE *lp));
void process_call_info BT_ARGS((CALL_RES *call_res));

void
usage()
{
    printf("usage: tones [options]\n");
    printf("\t-u unitnum = unit number (fax or voice card)\n");
    printf("\t-c num = call given number, else await ring\n");
    printf("\t-p = play tones\n");
    printf("\t-g = get tones and display them\n");
    printf("\t-v = turn on API debug mode\n");
    printf("\tOne of -p or -g is required\n");
    exit( 1 );
}

int
main(int argc, char **argv)
{
    extern int optind;
    extern char *optarg;
    int c, unit;
    BTLINE *lp;
    char *tones = "0123456789*#ABCD";
    void dowork();
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    
    unit = 0;
    get = -1;
    
    while ( (c = getopt( argc, argv, "u:c:pgv" )) != EOF )
    {
        switch ( c )
        {
          case 'u':
            unit = atoi( optarg );
            break;
          case 'g':
            get = TRUE;
            break;
          case 'p':
            get = FALSE;
            break;
          case 'v':
            BT_ZERO(args_debug);
            args_debug.mode = DEBUG_ALL;
            args_debug.options = API_DBG_MODE_SET;
            BfvDebugModeSetAdv(&args_debug);
            break;
          case 'c':
            cflag = 1;
            cnum = optarg;
            break;
          default:
            usage();
        }
    }
    
    if (optind != argc)
        usage();

    if (get < 0)
        usage();

    BT_ZERO(args);
    args.unit = unit;
    lp = BfvLineAttach(&args);                  /* Grab a line */
    
    if (lp == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        printf ("BfvLineAttach: %s\n",bterr.long_msg);
        exit( 1 );
    }
    
    dowork (lp, tones);
    
    return(0);
}

void
dowork (BTLINE *lp, char *tones)
{
    BTERR bterr;
    char *cp;
    int tone;
    static int puretone[] = {
        TONEID_CCITT_1,
        TONEID_CCITT_2,
        TONEID_CCITT_3,
        TONEID_CCITT_4,
        TONEID_CCITT_5,
        TONEID_697_HZ,
        TONEID_770_HZ,
        TONEID_852_HZ,
        TONEID_941_HZ,
        TONEID_1209_HZ,
        TONEID_1336_HZ,
        TONEID_1477_HZ,
        TONEID_1633_HZ,
        -1
    };
    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_tone args_tone;
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (lp, &args_admin) < 0)     /* Must always init lines */
    {
        BfvErrorMessage(lp,&args_admin.res,&bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                 bterr.long_msg,args_admin.reset_status);
        exit(1);
    }
    
    if (cflag)
    {
        BT_ZERO(args_tel);
        args_tel.phonenum = cnum;
        args_tel.call_protocol_code = CALL_PROTOCOL_VOICE;
        BfvLineOriginateCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            LINE_STATE(lp) != LINE_STATE_CONNECTED)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineOriginateCall: %s\n",bterr.long_msg);
            exit(1);
        }
    }
    else
    {
        BT_ZERO(args_tel);
        BfvLineWaitForCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            args_tel.res.line_status != WAIT_FOR_CALL_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineWaitForCall: %s\n",bterr.long_msg);
            exit(1);
        }
        process_call_info(&args_tel.call_res);
        
        BT_ZERO(args_tel);
        BfvLineAnswer (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            fprintf (stderr,"BfvLineAnswer: %s\n",bterr.long_msg);
            exit(1);
        }
    }
    
    
    printf( "DTMF tones:\n\t" );
    if (get)
    {
        BT_ZERO(args_tone);
        args_tone.decode_flag = DTMF_16TONE | TONE_REPORT_END;
        BfvToneDetectEnable(lp, &args_tone);
        if (args_tone.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tone.res,&bterr);
            fprintf (stderr,"BfvToneDetectEnable: %s\n",bterr.long_msg);
            goto got_err;
        }
        for (;;)
        {
            /* Get tone (60 sec timeout) */
            BT_ZERO(args_tone);
            args_tone.millisecs = 60000L;
            tone = BfvToneGet(lp, &args_tone);
            if (tone < 0)
            {
                BfvErrorMessage(lp,&args_tone.res,&bterr);
                printf ("BfvToneGet: %s\n",bterr.long_msg);
                break;
            }
            if (tone > 0)
                printf("%c", tone);
            else
                printf("e");
            fflush(stdout);
        }
    }
    else
    {
        cp = tones;
        printf( "%c", *cp );
        fflush(stdout);
        /* Max tone len is 1.5 sec */
        BT_ZERO(args_tone);
        args_tone.digit = *cp++;
        args_tone.millisecs = 1500L;
        args_tone.tone_type = DTMF_TONE;
        BfvTonePlay( lp, &args_tone);
        if (args_tone.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tone.res,&bterr);
            fprintf (stderr,"BfvTonePlay: %s\n",bterr.long_msg);
            goto got_err;
        }
        for (; *cp; cp++ )
        {
            printf( "%c", *cp );
            fflush(stdout);
            /* Play shortest (0.1 sec) */
            BT_ZERO(args_tone);
            args_tone.digit = *cp;
            args_tone.millisecs = 100L;
            args_tone.tone_type = DTMF_TONE;
            BfvTonePlay( lp, &args_tone);
            if (args_tone.res.status != BT_STATUS_OK)
            {
                BfvErrorMessage(lp,&args_tone.res,&bterr);
                fprintf (stderr,"BfvTonePlay: %s\n",bterr.long_msg);
                goto got_err;
            }
        }
        
        printf( "\nVolume test:\n\t" );
        for ( tone = -20; tone <= 3; tone ++)
        {
            printf( "%02d ", tone );
            fflush(stdout);
            BT_ZERO(args_tone);
            args_tone.toneid = TONEID_CCITT_1;
            args_tone.millisecs = 100L;
            args_tone.volume = tone;
            BfvTonePlayBeep(lp, &args_tone);
            if (args_tone.res.status != BT_STATUS_OK)
            {
                BfvErrorMessage(lp,&args_tone.res,&bterr);
                fprintf (stderr,"BfvTonePlayBeep: %s\n",bterr.long_msg);
                goto got_err;
            }
        }
        
        printf( "\nPure tone test:\n\t" );
        for ( tone = 0; puretone[ tone ] >= 0; tone++ )
        {
            printf( "%02d ", puretone[ tone ] );
            fflush(stdout);
            BT_ZERO(args_tone);
            args_tone.toneid = puretone[tone];
            args_tone.millisecs = 100L;
            args_tone.volume = 0;
            BfvTonePlayBeep(lp, &args_tone);
            if (args_tone.res.status != BT_STATUS_OK)
            {
                BfvErrorMessage(lp,&args_tone.res,&bterr);
                fprintf (stderr,"BfvTonePlayBeep: %s\n",bterr.long_msg);
                goto got_err;
            }
        }
    }
    printf( "\n" );
    
got_err:
    BT_ZERO(args_tel);
    BfvLineTerminateCall (lp, &args_tel);

    process_aoc(lp);
    
    BT_ZERO(args_admin);
    BfvLineDetach (lp, &args_admin);    /* Done with the line */
    
    exit( 0 );
}

void
process_aoc(BTLINE *lp)
{
    AOC_INFO *aoc = LINE_AOC_INFO(lp);
        
    if (aoc == NULL || aoc->status == BT_AOC_UNAVAILABLE)
        printf("Charge information unavailable.\n");
    else if (aoc->status == BT_AOC_FREE)
        printf("Charge information: FREE.\n");
    else if (aoc->status == BT_AOC_UNITS)
        printf("Charge information: Unit type %d, value %ld\n",
               aoc->charge_info,aoc->charge);
    else if (aoc->status == BT_AOC_CURRENCY)
    {
        static char *currency_names[] = {
            "1/1000","1/100","1/10","1","10","100","1000" };
#define NUM_CURRENCIES (sizeof(currency_names) / sizeof(*currency_names))
        char *cur_name;
            
        cur_name = (0 <= aoc->charge_info &&
                    aoc->charge_info < NUM_CURRENCIES) ?
            currency_names[aoc->charge_info] :
            "Invalid currency";
        printf("Charge information: Currency multiplier %s (%d), value %ld\n",
               cur_name,aoc->charge_info,aoc->charge);
    }
    else
        printf("Charge information: Bad status value %d\n",
               aoc->status);
}

void
process_call_info(CALL_RES *call_res)
{
    if (strlen(call_res->called_party_number))
        printf("Called party number: \"%s\"\n",
            call_res->called_party_number);
    if (strlen(call_res->called_party_subaddress))
        printf("Called party subaddress: \"%s\"\n",
           call_res->called_party_subaddress);
    if (strlen(call_res->calling_party_number))
        printf("Calling party number: \"%s\"\n",
           call_res->calling_party_number);
    if (strlen(call_res->calling_party_subaddress))
        printf("Calling party subaddress: \"%s\"\n",
           call_res->calling_party_subaddress);
    if (strlen(call_res->redir_number))
        printf("Redirection number: \"%s\", reason %X\n",
           call_res->redir_number, call_res->redir_reason);
}
