/*
 * transfer.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The transfer program shows how to use the high level functions to transfer
a call.  The program will wait for a call on the specified channel and 
then begin the transfer.
***************************************************************************/

#include <string.h>
#include <ctype.h>
#include "btlib.h"
/* Ensure that sleep is defined for all OS's. */
#include "bfax.h"

int play_func BT_ARGS((BTLINE *lp,char *arg));

void
usage ()
{
    fprintf (stderr, "usage: transfer [options] transfer_phone_number\n");
    fprintf (stderr, "  -u unitnum  use specified unit number\n");
    fprintf (stderr, "  -d unitnum  use specified unit number for second LP\n");
    fprintf (stderr, "  -c num      call given number, else await ring\n");
    fprintf (stderr, "  -m <1,2> specify trasfer mode if tranfer capability is LINE_XFER_ALL\n");
    fprintf (stderr, "  -h choose to put the call on hold if 2 b channel transfer\n");
    fprintf (stderr, "  -t line state to complete transfer \n");
    fprintf (stderr, "     <bst_dial_complete, bst_alerting, bst_connected>\n");
    fprintf (stderr, "  -s prompt file to play in supervised transfer\n");
    exit (1);
}

int cflag = 0;
char *cnum;
char *tnum;
char *name;
int unit = 2;
int unit_second = -1; 
int loop_count = 1; 
int specified_transfer_mode = 1;
int sflag = 0;
int transfer_line_state = BST_CONNECTED;
int hold_call = 0;
int supervised_responded = 0;
long the_timeout = 10000L;
BTLINE *lp_second = 0;

struct transfer_line_state_info {
    char name[30];
    unsigned transfer_line_state;
} tl_state_list[] = {
    { {"bst_dial_complete"}, BST_DIAL_COMPLETE },
    { {"bst_alerting"}, BST_ALERTING },
    { {"bst_connected"}, BST_CONNECTED }
};
#define NUM_STATES (sizeof(tl_state_list) / sizeof(*tl_state_list))
/* Extra braces around strings above required to avoid QNX 2 compiler bug */

void dowork BT_ARGS ((BTLINE *));

unsigned
state_string_lookup(char *statestr)
{
    unsigned i;
    
    for (i = 0; i < NUM_STATES; i++)
    {
        if (!strcmp(statestr,tl_state_list[i].name))
            break;
    }
    if (i >= NUM_STATES)
    {
        fprintf(stderr,"Unknown state '%s'\n",statestr);
        exit(1);
    }
    return(tl_state_list[i].transfer_line_state);
}

int
main (int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    BTLINE *lp;
    int i;
    unsigned tmp;
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    while ((c = getopt (argc, argv, "u:d:l:c:m:ht:s:")) != EOF)
    {
        switch (c)
        {
          case 'u':
            unit = atoi (optarg);
            break;
          case 'd':
            unit_second = atoi (optarg);
            break;
          case 'l':
            loop_count = atoi (optarg);
            break;
          case 'c':
            cflag = 1;
            cnum = optarg;
            break;
          case 'm':
            // if specified_transfer_mode is not 2-channel transfer, set to default 1.
            specified_transfer_mode = atoi (optarg);
            if (specified_transfer_mode != 2)
            {
                specified_transfer_mode = 1;
            }
            break;
          case 'h':
            hold_call = 1;
            break;
          case 't':
            if (!sflag)
            {
                if (isdigit(optarg[0]))
                {
                    tmp = atoi(optarg);
                    for (i = 0; i < NUM_STATES; i++)
                    {
                        if (tl_state_list[i].transfer_line_state == tmp)
                        {
                            transfer_line_state = tmp;
                            break;
                        }
                    }
                    if (i >= NUM_STATES)
                    {
                        fprintf(stderr,"Invalid state %d\n",tmp);
                        exit(1);
                    }
                }
                else
                    transfer_line_state = state_string_lookup(optarg);
            }
            break;
          case 's':
            //if supervised, transfer_line_state = BST_CONNECTED only.
            sflag = 1;
            name = optarg;
            transfer_line_state = BST_CONNECTED;
            break;
          default:
            usage ();
        }
    }
    
    if (optind >= argc)
        usage ();

    tnum = argv[optind++];

    BT_ZERO(args);
    args.unit = unit;
    if ((lp = BfvLineAttach (&args)) == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
        exit (1);
    }
    if (unit_second == -1)
    {
        unit_second = unit + 1;
    }
    
    for(i = 1; i <= loop_count; i++) 
    {
	printf("loop %d out of %d\n", i, loop_count);
    	dowork (lp);
    }

    BT_ZERO(args);
    BfvLineDetach (lp, &args);

    return(0);
}

void
dowork (BTLINE *lp)
{
    BTERR bterr;
    struct infopkt_stream *ips;
    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_infopkt args_infopkt;
    struct args_speech args_speech;
    struct args_tone args_tone;
    int tone;
    int transfer_mode;
    BTLINE *get_respond_lp;
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (lp, &args_admin) < 0)
    {
        BfvErrorMessage(lp,&args_admin.res,&bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                 bterr.long_msg,args_admin.reset_status);
        exit(1);
    }
    
    /* Check for digital configuration error */
    if (args_admin.reset_status & RST_DIG_CONFIG_UNREADABLE)
    {
        fprintf (stderr, "Telephony/Digital configuration file not found.\n");
        exit(1);
    }

    /* Check if transfer can be done */
    BT_ZERO(args_tel);
    BfvLineTransferCapabilityQuery(lp, &args_tel);
    if(args_tel.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        fprintf (stderr,"BfvLineTransferCapabilityQuery: %s\n",bterr.long_msg);
        exit (1);
    }
    
    transfer_mode = args_tel.transfer_mode;

    switch(args_tel.transfer_mode)
    {
    case LINE_XFER_SINGLE:
        break;
    case LINE_XFER_ALL:
    case LINE_XFER_TWO_CHAN:
    case LINE_XFER_TWO_CHAN_NEEDS_NAILUP:
        if ((args_tel.transfer_mode == LINE_XFER_ALL) &&
            (specified_transfer_mode == 1))
        {
            // if LINE_XFER_ALL and user picks 1, do as LINE_XFER_SINGLE.
            break;
        }
        else
        {
            struct args_telephone args_tel2;

            BT_ZERO(args_admin);
            args_admin.unit = unit_second;
            if ((lp_second = BfvLineAttach (&args_admin)) == NULL)
            {
                BfvErrorMessage(lp_second,&args_admin.res,&bterr);
                fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
                exit (1);
            }

            BT_ZERO(args_admin);
            args_admin.config_file_name = "btcall.cfg";
            if (BfvLineReset (lp_second, &args_admin) < 0)
            {
                BfvErrorMessage(lp_second,&args_admin.res,&bterr);
                fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                         bterr.long_msg,args_admin.reset_status);
                exit(1);
            }

            BT_ZERO(args_tel2);
            BfvLineTransferCapabilityQuery(lp_second, &args_tel2);
            if(args_tel2.res.status != BT_STATUS_OK)
            {
                BfvErrorMessage(lp_second,&args_tel2.res,&bterr);
                fprintf (stderr,"BfvLineTransferCapabilityQuery: %s\n",bterr.long_msg);
                exit (1);
            }

            if(args_tel2.transfer_group != args_tel.transfer_group)
            {
                fprintf (stderr,"Error: 2nd Line does not belong to the same transfer group\n");
                exit (1);
            }
        }
        break;
    case LINE_XFER_NONE:
        fprintf (stderr,"Error: Transfer is not supported\n");
        exit(1);
    }

    if (sflag)
    {
        BT_ZERO(args_infopkt);
        args_infopkt.fname = name;
        args_infopkt.fmode = "r";
        if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
        {
            BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
            fprintf (stderr, "BfvInfopktOpen: %s\n",bterr.long_msg);
            exit (1);
        }
    }

    if (cflag)
    {
        /* Make an outbound call */
        BT_ZERO(args_tel);
        args_tel.phonenum = cnum;
        args_tel.call_protocol_code = CALL_PROTOCOL_VOICE;
        BfvLineOriginateCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            LINE_STATE(lp) != LINE_STATE_CONNECTED)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            fprintf (stderr, "BfvLineOriginateCall: %s\n",bterr.long_msg);
            goto got_err_1;
        }
    }
    else
    {
        /* Wait for an inbound call */
        BT_ZERO(args_tel);
        BfvLineWaitForCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            args_tel.res.line_status != WAIT_FOR_CALL_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            fprintf (stderr, "BfvLineWaitForCall: %s\n",bterr.long_msg);
            goto got_err_1;
        }

        /* Answer the inbound call */
        BT_ZERO(args_tel);
        BfvLineAnswer (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            fprintf (stderr, "BfvLineAnswer: %s\n",bterr.long_msg);
            goto got_err_1;
        }
    }

    //Some PBXs have a problem if transfer takes place right after accepting the first call.
    //This sleep can be replaced by playing a prompt
    sleep(1);

    // for LINE_XFER_TWO_CHAN_NEEDS_NAILUP, 
    // 2 bchannels need tobe connected by BfvCallSwitchConnect
    // before complete transfer.
    // The high level function BfvLineTransferComplete does the connection
    // automatically in ECC.
    // See transferll.c for low level functions.
    BT_ZERO(args_tel);
    args_tel.phonenum = tnum;
    args_tel.supervised = sflag ? 1 : 0;
    args_tel.transfer_line_state = transfer_line_state;
    args_tel.hold_call = hold_call;
    args_tel.call_protocol_code = CALL_PROTOCOL_VOICE;
    args_tel.lp_second_channel = lp_second;
    BfvLineTransfer(lp, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        fprintf (stderr, "BfvLineTransfer: %s\n",bterr.long_msg);
        goto got_err_2;
    }

    if (sflag)
    {
        //Wait for responds on the second channel, if this is 2 channel transfer
        if(lp_second)
        {
            get_respond_lp = lp_second;
        }
        else
        {
            get_respond_lp = lp;
        }

        BT_ZERO(args_tone);
        args_tone.decode_flag = DTMF_16TONE | TONE_REPORT_END;
        BfvToneDetectEnable(get_respond_lp, &args_tone);
        if (args_tone.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(get_respond_lp,&args_tone.res,&bterr);
            fprintf (stderr,"BfvToneDetectEnable: %s\n",bterr.long_msg);
            goto got_err_2;
        }

        /* play till file is done */
        BT_ZERO(args_speech);
        args_speech.ips = ips;
        args_speech.func = play_func;
        BfvSpeechPlay (get_respond_lp, &args_speech);
    }
    else
    {
        //Some PBXs have a problem if hangup takes place right after transfer command issued.
        //This sleep can be replaced by playing a prompt
        sleep(5);
    }

    // still waiting for digits after prompt played
    if (sflag)
    {
        for (;!supervised_responded;)
        {
            /* Get tone (60 sec timeout) */
            BT_ZERO(args_tone);
            args_tone.millisecs = 60000L;
            tone = BfvToneGet(get_respond_lp, &args_tone);
            if (tone < 0)
            {
                BfvErrorMessage(get_respond_lp,&args_tone.res,&bterr);
                fprintf (stderr, "BfvToneGet: %s\n",bterr.long_msg);
                break;
            }
            if (tone > 0)
            {
                printf("%c", tone);
                switch (tone) 
                {
                case '1':
                    supervised_responded = 1;
                    break;
                case '2':
                    supervised_responded = 2;
                    break;
                }
            }
            else
            {
                printf("e");
            }
            fflush(stdout);
        }
    }
    switch (supervised_responded) 
    {
    case 1:
        BT_ZERO(args_tel);
        args_tel.lp_second_channel = lp_second;
        BfvLineTransferComplete(lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            fprintf (stderr, "BfvLineTransferComplete: %s\n",bterr.long_msg);
        }
        break;
    case 2:
        BT_ZERO(args_tel);
        args_tel.lp_second_channel = lp_second;
        BfvLineTransferCancel(lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            fprintf (stderr, "BfvLineTransferCancel: %s\n",bterr.long_msg);
        }
        break;                    
    }

    if (sflag)
    {
        args_infopkt.ips = ips;
        BfvInfopktClose (&args_infopkt);
    }

got_err_2:
    if(lp_second != 0)
    {
        BT_ZERO(args_admin);
        BfvLineDetach (lp_second, &args_admin);
    }

    BT_ZERO(args_tel);
    BfvLineTerminateCall (lp, &args_tel);

got_err_1:
    return;
}

/* 1 = Accept call, 2 = Reject call. */
int play_func (BTLINE *lp, char *arg)
{
    int tmp;
    struct args_tone args_tone;

    BT_ZERO(args_tone);
    args_tone.tone_cp = 1;

    switch (tmp = BfvToneGet(lp,&args_tone)) 
    {
    case '1':
        supervised_responded = 1;
        return (1);
    case '2':
        supervised_responded = 2;
        return (1);
    }

    if (args_tone.res.status != BT_STATUS_OK &&
    args_tone.res.status != BT_STATUS_TIMEOUT)
    {
        BTERR bterr;

        BfvErrorMessage(lp,&args_tone.res,&bterr);
        fprintf (stderr, "BfvToneGet: %s\n",bterr.long_msg);
        return(1);
    }
    return (0); /* Keep playing */
}
