/*
 * transferll.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The transferll program shows how to use the low level functions to transfer
a call.  The program will wait for a call on the specified channel and 
then begin the transfer.
***************************************************************************/

#include <string.h>
#include <ctype.h>
#include "btlib.h"
/* Ensure that sleep is defined for all OS's. */
#include "bfax.h"

int play_func BT_ARGS((BTLINE *lp,char *arg));
int call_progress_func BT_ARGS ((BTLINE *lp, char *arg));
void HighLevelHold(BTLINE *lp, struct args_telephone *tel_args);
void HighLevelRetrieve(BTLINE *lp, struct args_telephone *tel_args);
void MakeEnquiryCall(BTLINE * lp, int transfer_line_state, struct args_cc *isdn_args);
void HighLevelTransfer(BTLINE *lp, struct args_telephone *tel_args);
void HighLevelTransferComplete(BTLINE *lp, struct args_telephone *tel_args);
void HighLevelTransferCancel(BTLINE *lp, struct args_telephone *tel_args);
void StartSingleChannelTransfer(BTLINE * lp, struct args_telephone * tel_args);
void StartSingleChannelTransferCancel(BTLINE * lp, struct args_telephone * tel_args);
void StartTwoChannelTransfer(BTLINE * lp, struct args_telephone * tel_args);
void NailBchannelsAndCompleteTransfer(BTLINE *lp, struct args_telephone *args_tel);

void
usage ()
{
    fprintf (stderr, "usage: transferll [options] transfer_phone_number\n");
    fprintf (stderr, "  -u unitnum  use specified unit number\n");
    fprintf (stderr, "  -c num      call given number, else await ring\n");
    fprintf (stderr, "  -m <1,2> specify trasfer mode if tranfer capability is LINE_XFER_ALL\n");
    fprintf (stderr, "  -h choose to put the call on hold if 2 b channel transfer\n");
    fprintf (stderr, "  -t line state to complete transfer \n");
    fprintf (stderr, "     <bst_dial_complete, bst_alerting, bst_connected>\n");
    fprintf (stderr, "  -s prompt file to play in supervised transfer\n");
    fprintf (stderr, "  -d to disable auto switch connection if transfer capability is LINE_XFER_TWO_CHAN_NEEDS_NAILUP\n");
    exit (1);
}

int cflag = 0;
char *cnum;
char *tnum;
char *name;
int unit = 2;
int offset = 1; // second unit = unit + offset;
int specified_transfer_mode = 1;
int sflag = 0;
int transfer_line_state = BST_CONNECTED;
int hold_call = 0;
int supervised_responded = 0;
long the_timeout = 10000L;
BTLINE *lp_second = 0;
int disable_auto_sw_connect = 0;

struct transfer_line_state_info {
    char name[30];
    unsigned transfer_line_state;
} tl_state_list[] = {
    { {"bst_dial_complete"}, BST_DIAL_COMPLETE },
    { {"bst_alerting"}, BST_ALERTING },
    { {"bst_connected"}, BST_CONNECTED }
};
#define NUM_STATES (sizeof(tl_state_list) / sizeof(*tl_state_list))
/* Extra braces around strings above required to avoid QNX 2 compiler bug */

void dowork BT_ARGS ((BTLINE *));

unsigned
state_string_lookup(char *statestr)
{
    unsigned i;
    
    for (i = 0; i < NUM_STATES; i++)
    {
        if (!strcmp(statestr,tl_state_list[i].name))
            break;
    }
    if (i >= NUM_STATES)
    {
        fprintf(stderr,"Unknown state '%s'\n",statestr);
        exit(1);
    }
    return(tl_state_list[i].transfer_line_state);
}

int
main (int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    BTLINE *lp;
    int i;
    unsigned tmp;
    struct args_line_admin args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    while ((c = getopt (argc, argv, "u:c:m:ht:s:d")) != EOF)
    {
        switch (c)
        {
          case 'u':
            unit = atoi (optarg);
            break;
          case 'c':
            cflag = 1;
            cnum = optarg;
            break;
          case 'm':
            // if specified_transfer_mode is not 2-channel transfer, set to default 1.
            specified_transfer_mode = atoi (optarg);
            if (specified_transfer_mode != 2)
            {
                specified_transfer_mode = 1;
            }
            break;
          case 'h':
            hold_call = 1;
            break;
          case 't':
            if (!sflag)
            {
                if (isdigit(optarg[0]))
                {
                    tmp = atoi(optarg);
                    for (i = 0; i < NUM_STATES; i++)
                    {
                        if (tl_state_list[i].transfer_line_state == tmp)
                        {
                            transfer_line_state = tmp;
                            break;
                        }
                    }
                    if (i >= NUM_STATES)
                    {
                        fprintf(stderr,"Invalid state %d\n",tmp);
                        exit(1);
                    }
                }
                else
                    transfer_line_state = state_string_lookup(optarg);
            }
            break;
          case 's':
            //if supervised, transfer_line_state = BST_CONNECTED only.
            sflag = 1;
            name = optarg;
            transfer_line_state = BST_CONNECTED;
            break;
          case 'd':
            disable_auto_sw_connect = 1;
            break;
          default:
            usage ();
        }
    }
    
    if (optind >= argc)
        usage ();

    tnum = argv[optind++];

    BT_ZERO(args);
    args.unit = unit;
    if ((lp = BfvLineAttach (&args)) == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
        exit (1);
    }
    
    dowork (lp);

    BT_ZERO(args);
    BfvLineDetach (lp, &args);
    
    return(0);
}

void
dowork (BTLINE *lp)
{
    BTERR bterr;
    struct infopkt_stream *ips;
    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_infopkt args_infopkt;
    struct args_speech args_speech;
    struct args_tone args_tone;
    int tone;
    int transfer_mode;
    BTLINE *get_respond_lp;
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (lp, &args_admin) < 0)
    {
        BfvErrorMessage(lp,&args_admin.res,&bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                 bterr.long_msg,args_admin.reset_status);
        exit(1);
    }
    
    /* Check for digital configuration error */
    if (args_admin.reset_status & RST_DIG_CONFIG_UNREADABLE)
    {
        fprintf (stderr, "Telephony/Digital configuration file not found.\n");
        exit(1);
    }

    /* Check if transfer can be done */
    BT_ZERO(args_tel);
    BfvLineTransferCapabilityQuery(lp, &args_tel);
    if(args_tel.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        fprintf (stderr,"BfvLineTransferCapabilityQuery: %s\n",bterr.long_msg);
        exit (1);
    }
    
    transfer_mode = args_tel.transfer_mode;

    switch(args_tel.transfer_mode)
    {
    case LINE_XFER_SINGLE:
        break;
    case LINE_XFER_ALL:
    case LINE_XFER_TWO_CHAN:
    case LINE_XFER_TWO_CHAN_NEEDS_NAILUP:
        if ((args_tel.transfer_mode == LINE_XFER_ALL) &&
            (specified_transfer_mode == 1))
        {
            // if LINE_XFER_ALL and user picks 1, do as LINE_XFER_SINGLE.
            break;
        }
        else
        {
            struct args_telephone args_tel2;

            BT_ZERO(args_admin);
            args_admin.unit = unit + offset;
            if ((lp_second = BfvLineAttach (&args_admin)) == NULL)
            {
                BfvErrorMessage(lp_second,&args_admin.res,&bterr);
                fprintf (stderr,"BfvLineAttach: %s\n",bterr.long_msg);
                exit (1);
            }

            BT_ZERO(args_admin);
            args_admin.config_file_name = "btcall.cfg";
            if (BfvLineReset (lp_second, &args_admin) < 0)
            {
                BfvErrorMessage(lp_second,&args_admin.res,&bterr);
                fprintf (stderr, "BfvLineReset: %s: status %lX.\n",
                         bterr.long_msg,args_admin.reset_status);
                exit(1);
            }

            BT_ZERO(args_tel2);
            BfvLineTransferCapabilityQuery(lp_second, &args_tel2);
            if(args_tel2.res.status != BT_STATUS_OK)
            {
                BfvErrorMessage(lp_second,&args_tel2.res,&bterr);
                fprintf (stderr,"BfvLineTransferCapabilityQuery: %s\n",bterr.long_msg);
                exit (1);
            }

            if(args_tel2.transfer_group != args_tel.transfer_group)
            {
                fprintf (stderr,"Error: 2nd Line does not belong to the same transfer group\n");
                exit (1);
            }
        }
        break;
    case LINE_XFER_NONE:        
        fprintf (stderr,"Error: Transfer is not supported\n");
        exit(1);
    }

    if (sflag)
    {
        BT_ZERO(args_infopkt);
        args_infopkt.fname = name;
        args_infopkt.fmode = "r";
        if ((ips = BfvInfopktOpen (&args_infopkt)) == NULL)
        {
            BfvErrorMessage(NULL,&args_infopkt.res,&bterr);
            printf ("BfvInfopktOpen: %s\n",bterr.long_msg);
            exit (1);
        }
    }

    if (cflag)
    {
        /* Make an outbound call */
        BT_ZERO(args_tel);
        args_tel.phonenum = cnum;
        args_tel.call_protocol_code = CALL_PROTOCOL_VOICE;
        BfvLineOriginateCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            LINE_STATE(lp) != LINE_STATE_CONNECTED)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineOriginateCall: %s\n",bterr.long_msg);
            goto got_err_1;
        }
    }
    else
    {
        /* Wait for an inbound call */
        BT_ZERO(args_tel);
        BfvLineWaitForCall (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK || 
            args_tel.res.line_status != WAIT_FOR_CALL_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineWaitForCall: %s\n",bterr.long_msg);
            goto got_err_1;
        }

        /* Answer the inbound call */
        BT_ZERO(args_tel);
        BfvLineAnswer (lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("BfvLineAnswer: %s\n",bterr.long_msg);
            goto got_err_1;
        }
    }

    //Some PBXs have a problem if transfer takes place right after accepting the first call.
    //This sleep can be replaced by playing a prompt
    sleep(1);

    BT_ZERO(args_tel);
    args_tel.phonenum = tnum;
    args_tel.supervised = sflag ? 1 : 0;
    args_tel.transfer_line_state = transfer_line_state;
    args_tel.hold_call = hold_call;
    args_tel.call_protocol_code = CALL_PROTOCOL_VOICE;
    args_tel.lp_second_channel = lp_second;
    HighLevelTransfer(lp, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_tel.res,&bterr);
        printf ("LowLevelTransfer: %s\n",bterr.long_msg);
        goto got_err_2;
    }

    if (sflag)
    {
        //Wait for responds on the second channel, if this is 2 channel transfer
        if(lp_second)
        {
            get_respond_lp = lp_second;
        }
        else
        {
            get_respond_lp = lp;
        }

        BT_ZERO(args_tone);
        args_tone.decode_flag = DTMF_16TONE | TONE_REPORT_END;
        BfvToneDetectEnable(get_respond_lp, &args_tone);
        if (args_tone.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(get_respond_lp,&args_tone.res,&bterr);
            fprintf (stderr,"BfvToneDetectEnable: %s\n",bterr.long_msg);
            goto got_err_2;
        }

        /* play till file is done */
        BT_ZERO(args_speech);
        args_speech.ips = ips;
        args_speech.func = play_func;
        BfvSpeechPlay (get_respond_lp, &args_speech);
    }
    else
    {
        //Some PBXs have a problem if hangup takes place right after transfer command issued.
        //This sleep can be replaced by playing a prompt
        sleep(5);
    }

    // still waiting for digits after prompt played
    if (sflag)
    {
        for (;!supervised_responded;)
        {
            /* Get tone (60 sec timeout) */
            BT_ZERO(args_tone);
            args_tone.millisecs = 60000L;
            tone = BfvToneGet(get_respond_lp, &args_tone);
            if (tone < 0)
            {
                BfvErrorMessage(get_respond_lp,&args_tone.res,&bterr);
                printf ("BfvToneGet: %s\n",bterr.long_msg);
                break;
            }
            if (tone > 0)
            {
                printf("%c", tone);
                switch (tone) 
                {
                case '1':
                    supervised_responded = 1;
                    break;
                case '2':
                    supervised_responded = 2;
                    break;
                }
            }
            else
            {
                printf("e");
            }
            fflush(stdout);
        }
    }

    switch (supervised_responded) 
    {
    case 1:
        BT_ZERO(args_tel);
        args_tel.lp_second_channel = lp_second;
        NailBchannelsAndCompleteTransfer(lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("HighLevelTransferComplete: %s\n",bterr.long_msg);
        }
        break;
    case 2:
        BT_ZERO(args_tel);
        args_tel.lp_second_channel = lp_second;
        HighLevelTransferCancel(lp, &args_tel);
        if (args_tel.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_tel.res,&bterr);
            printf ("HighLevelTransferCancel: %s\n",bterr.long_msg);
        }
        break;
    }

    if (sflag)
    {
        args_infopkt.ips = ips;
        BfvInfopktClose (&args_infopkt);
    }

got_err_2:
    if(lp_second != 0)
    {
        BT_ZERO(args_admin);
        BfvLineDetach (lp_second, &args_admin);
    }

    BT_ZERO(args_tel);
    BfvLineTerminateCall (lp, &args_tel);

got_err_1:
    return;
}

/* 1 = Accept call, 2 = Reject call. */
int play_func (BTLINE *lp, char *arg)
{
    int tmp;
    struct args_tone args_tone;

    BT_ZERO(args_tone);
    args_tone.tone_cp = 1;

    switch (tmp = BfvToneGet(lp,&args_tone)) 
    {
    case '1':
        supervised_responded = 1;
        return (1);
    case '2':
        supervised_responded = 2;
        return (1);
    }

    if (args_tone.res.status != BT_STATUS_OK &&
    args_tone.res.status != BT_STATUS_TIMEOUT)
    {
        BTERR bterr;

        BfvErrorMessage(lp,&args_tone.res,&bterr);
        printf ("BfvToneGet: %s\n",bterr.long_msg);
        return(1);
    }
    return (0); /* Keep playing */
}


void HighLevelTransfer(BTLINE *lp, struct args_telephone *tel_args)
{
    //Validate some of the parameters
    if(tel_args->transfer_line_state != BST_DIAL_COMPLETE &&
       tel_args->transfer_line_state != BST_ALERTING &&
       tel_args->transfer_line_state != BST_CONNECTED)
    {
        tel_args->res.status = BT_STATUS_ERROR;
        tel_args->res.line_status = APIERR_BADPARAMETER;
    }

    if(tel_args->res.status == BT_STATUS_OK)
    {
        BfvLineTransferCapabilityQuery(lp, tel_args);
        if(tel_args->res.status == BT_STATUS_OK)
        {
            switch(tel_args->transfer_mode)
            {
            case LINE_XFER_SINGLE:
                //make sure we don't use the secondary channel
                StartSingleChannelTransfer(lp, tel_args);
                break;
            case LINE_XFER_TWO_CHAN:
                // LINE_XFER_TWO_CHAN requires lp_second_channel
                if(tel_args->lp_second_channel)
                {
                    //tel_args->transfer_group must be set with the value for lp
                    //This was done by BfvLineTransferCapabilityQuery
                    StartTwoChannelTransfer(lp, tel_args);
                }
                else
                {
                    tel_args->res.status = BT_STATUS_ERROR;
                    tel_args->res.line_status = APIERR_BADPARAMETER;
                }
                break;
            case LINE_XFER_ALL:
                if(tel_args->lp_second_channel)
                {
                    //tel_args->transfer_group must be set with the value for lp
                    //This was done by BfvLineTransferCapabilityQuery
                    StartTwoChannelTransfer(lp, tel_args);
                }
                else
                {
                    StartSingleChannelTransfer(lp, tel_args);
                }
                break;
            case LINE_XFER_TWO_CHAN_NEEDS_NAILUP:
                // LINE_XFER_TWO_CHAN requires lp_second_channel
                if(tel_args->lp_second_channel)
                {
                    //tel_args->transfer_group must be set with the value for lp
                    //This was done by BfvLineTransferCapabilityQuery
                    StartTwoChannelTransfer(lp, tel_args);
                }
                else
                {
                    tel_args->res.status = BT_STATUS_ERROR;
                    tel_args->res.line_status = APIERR_BADPARAMETER;
                }
                break;
            case LINE_XFER_NONE:
            default:
                tel_args->res.status = BT_STATUS_ERROR;
                tel_args->res.line_status = APIERR_BOARDCAPABILITY;
                break;
            }
        
            if(tel_args->res.status == BT_STATUS_OK && 
               tel_args->supervised == 0)
            {
                NailBchannelsAndCompleteTransfer(lp, tel_args);
            }

            if(tel_args->res.status != BT_STATUS_OK)
            {
                struct args_telephone tel_args2;

                BT_ZERO(tel_args2);
                tel_args2.timeout = tel_args->timeout;
                //Need to reset all
                HighLevelTransferCancel(lp, &tel_args2);

                //give the user the original failure stored in tel_args
            }
        }
    }
}

void HighLevelTransferComplete(BTLINE *lp, struct args_telephone *tel_args)
{
    struct args_cc isdn_args;

    BT_ZERO(isdn_args);

    isdn_args.timeout = tel_args->timeout;
    isdn_args.lp_second_channel = tel_args->lp_second_channel;

    BfvCallTransferComplete(lp, &isdn_args);
    tel_args->res = isdn_args.res;
    if (isdn_args.res.status == BT_STATUS_OK)
    {
        BT_ZERO(isdn_args);
        isdn_args.timeout = tel_args->timeout;
        isdn_args.lp_second_channel = tel_args->lp_second_channel;
        isdn_args.disable_auto_sw_connect = disable_auto_sw_connect;

        BfvCallWaitTransferComplete(lp, &isdn_args);
        tel_args->res = isdn_args.res;
        if (isdn_args.res.status == BT_STATUS_OK)
        {
            BT_ZERO(isdn_args);
            isdn_args.timeout = tel_args->timeout;
            BfvCallWaitForRelease (lp, &isdn_args);
            tel_args->res = isdn_args.res;
        }
    }
}

void HighLevelTransferCancel(BTLINE *lp, struct args_telephone *tel_args)
{
    BfvLineTransferCapabilityQuery(lp, tel_args);
    if(tel_args->res.status == BT_STATUS_OK)
    {
        switch(tel_args->transfer_mode)
        {
        case LINE_XFER_SINGLE:
            //make sure we don't use the secondary channel
            StartSingleChannelTransferCancel(lp, tel_args);
            break;
        case LINE_XFER_TWO_CHAN:
            // LINE_XFER_TWO_CHAN requires lp_second_channel
            if(tel_args->lp_second_channel)
            {
                BfvLineTerminateCallISDN(tel_args->lp_second_channel, tel_args);
                HighLevelRetrieve(lp, tel_args);
            }
            else
            {
                tel_args->res.status = BT_STATUS_ERROR;
                tel_args->res.line_status = APIERR_BADPARAMETER;
            }
            break;
        case LINE_XFER_ALL:
            if(tel_args->lp_second_channel)
            {
                BfvLineTerminateCallISDN(tel_args->lp_second_channel, tel_args);
                HighLevelRetrieve(lp, tel_args);
            }
            else
            {
                StartSingleChannelTransferCancel(lp, tel_args);
            }
            break;
        case LINE_XFER_TWO_CHAN_NEEDS_NAILUP:
            if(tel_args->lp_second_channel)
            {
                BfvLineTerminateCallISDN(tel_args->lp_second_channel, tel_args);
            }
            else
            {
                tel_args->res.status = BT_STATUS_ERROR;
                tel_args->res.line_status = APIERR_BADPARAMETER;
            }
            break;
        case LINE_XFER_NONE:
        default:
            tel_args->res.status = BT_STATUS_ERROR;
            tel_args->res.line_status = APIERR_BOARDCAPABILITY;
            break;
        }
    }
}

void StartSingleChannelTransfer(BTLINE * lp, struct args_telephone * tel_args)
{
    struct args_cc isdn_args;

    HighLevelHold(lp, tel_args);
    if(tel_args->res.status == BT_STATUS_OK)
    {
        //Start enquiry call
        BT_ZERO(isdn_args);
        isdn_args.timeout = tel_args->timeout;
        isdn_args.phonenum = tel_args->phonenum;
        isdn_args.call_protocol_code = tel_args->call_protocol_code;
        isdn_args.arg = tel_args->arg;
        isdn_args.func = tel_args->func;
        isdn_args.call_mode = tel_args->call_mode;

        MakeEnquiryCall(lp, tel_args->transfer_line_state, &isdn_args);
        tel_args->res = isdn_args.res;
        tel_args->cause_code = isdn_args.cause;
    }
}

void StartSingleChannelTransferCancel(BTLINE * lp, struct args_telephone * tel_args)
{
    //If a call is on hold then disconnect and recover held call
    struct args_cc isdn_args;

    do
    {
        BT_ZERO(isdn_args);
        isdn_args.timeout = tel_args->timeout;

        BfvCallStatus(lp, &isdn_args);
        tel_args->res = isdn_args.res;

        if(isdn_args.res.status != BT_STATUS_OK)
            break;

        //If we don't have any calls on hold then we are done
        if(isdn_args.calls_on_hold != 0)
        {
            //If connected then disconnect the enquiry call
            if(LINE_STATE(lp) != LINE_STATE_IDLE)
            {
                BfvLineTerminateCallISDN(lp, tel_args);
            }

            if(tel_args->res.status != BT_STATUS_OK)
                break;

            HighLevelRetrieve(lp, tel_args);
        }
    } while(0);
}

void StartTwoChannelTransfer(BTLINE * lp, struct args_telephone * tel_args)
{
    struct args_telephone tel_args2;
    struct args_cc isdn_args;

    BT_ZERO(tel_args2);
    tel_args2.timeout = tel_args->timeout;

    BfvLineTransferCapabilityQuery(tel_args->lp_second_channel, &tel_args2);
    if(tel_args2.res.status != BT_STATUS_OK)
    {
        tel_args->res = tel_args2.res;
    }
    else
    {
        //Check if both channel belong to the same transfer group
        if(tel_args2.transfer_group != tel_args->transfer_group)
        {
            tel_args->res.status = BT_STATUS_ERROR;
            tel_args->res.line_status = APIERR_BADPARAMETER;
        }
        else
        {
            tel_args->res.status = BT_STATUS_OK;
            tel_args->res.line_status = APIERR_UNCLASSIFIED;
            if(tel_args->hold_call)
            {
                HighLevelHold(lp, tel_args);
            }
            if(tel_args->res.status == BT_STATUS_OK)
            {
                BT_ZERO(isdn_args);
                isdn_args.timeout = tel_args->timeout;
                isdn_args.phonenum = tel_args->phonenum;
                isdn_args.call_protocol_code = tel_args->call_protocol_code;
                isdn_args.arg = tel_args->arg;
                isdn_args.func = tel_args->func;
                isdn_args.call_mode = tel_args->call_mode;

                MakeEnquiryCall(tel_args->lp_second_channel, tel_args->transfer_line_state, &isdn_args);
                tel_args->res = isdn_args.res;
                tel_args->cause_code = isdn_args.cause;
            }
        }
    }
}

void MakeEnquiryCall(BTLINE * lp, int transfer_line_state, struct args_cc *isdn_args)
{
    isdn_args->enquiry_call = 1;

    BfvCallSetup(lp, isdn_args);
    if(isdn_args->res.status == BT_STATUS_OK)
    {
        switch(transfer_line_state)
        {
        case BST_DIAL_COMPLETE:
            // No call progress is needed
            // transfer after initiating the outgoing call

            BfvCallWaitForAlerting(lp, isdn_args);
            break;
        case BST_ALERTING:
            //if call progress is needed check for busy and fail
            //otherwise transfer on any other connecting or connected CP(Silence, Answer, RING, etc)

            /* This user function checks for "ringing". */
            isdn_args->func = call_progress_func;
            BfvCallWaitForComplete(lp, isdn_args);
            if(isdn_args->res.status == BT_STATUS_OK)
            {
                //if in a disconnection state fail
                if(LINE_STATE(lp) == LINE_STATE_IDLE ||
                  (LINE_STATE(lp) == LINE_STATE_OFF_HOOK && 
                  (isdn_args->res.line_status == FCP_DIALTON || //Dial tone
                   isdn_args->res.line_status == FCP_BUSY1   || //Normal busy
                   isdn_args->res.line_status == FCP_BUSY2   || //Distinctive or fast busy
                   isdn_args->res.line_status == FCP_ROBUSY  || //Reorder busy
                   isdn_args->res.line_status == FCP_RECALL  || //Recall dial tone
                   isdn_args->res.line_status == FCP_CONFIRM || //Confirmation dial tone
                   isdn_args->res.line_status == FCP_PULSE   || //Specified pulse dial tone
                   isdn_args->res.line_status == FCP_SITREORD|| //Special Information Reorder tone
                   isdn_args->res.line_status == FCP_SITNOCIR)))//Special Information No Circuit Detected tone
                {
                    isdn_args->res.status = BT_STATUS_ERROR;
                    isdn_args->res.line_status = APIERR_NOTCONNECTED;
                }
            }    
            break;
        case BST_CONNECTED:
            // transfer after if voice is detected on the line

            BfvCallWaitForComplete(lp, isdn_args);
            if (isdn_args->res.status == BT_STATUS_OK)
            {
                //if not in a connected state fail
                if(LINE_STATE(lp) != LINE_STATE_CONNECTED)
                {
                    isdn_args->res.status = BT_STATUS_ERROR;
                    isdn_args->res.line_status = APIERR_NOTCONNECTED;
                }
            }
            break;
        default:
            isdn_args->res.status = BT_STATUS_ERROR;
            isdn_args->res.line_status = APIERR_BADPARAMETER;
            break;
        }
    }
}

void HighLevelHold(BTLINE *lp, struct args_telephone *tel_args)
{
    struct args_cc isdn_args;
    int rc = 0;

    do
    {
        BT_ZERO(isdn_args);
        isdn_args.timeout = tel_args->timeout;

        BfvCallStatus(lp, &isdn_args);
        tel_args->res = isdn_args.res;

        if(isdn_args.res.status != BT_STATUS_OK)
            break;

        if(isdn_args.calls_on_hold == 0)
        {
            if(LINE_STATE(lp) == LINE_STATE_IDLE)
            {
                // No active call present
                tel_args->res.status = BT_STATUS_ERROR;
                tel_args->res.line_status = APIERR_BADSTATE; 
                break;
            }

            BT_ZERO(isdn_args);
            isdn_args.timeout = tel_args->timeout;

            BfvCallHold(lp, &isdn_args);
            tel_args->res = isdn_args.res;

            if(isdn_args.res.status != BT_STATUS_OK)
                break;

            BT_ZERO(isdn_args);
            isdn_args.timeout = tel_args->timeout;

            BfvCallWaitForHold(lp, &isdn_args);
            tel_args->res = isdn_args.res;

            if(isdn_args.res.status != BT_STATUS_OK)
                break;

            isdn_args.calls_on_hold = 1;
        }
        else
        {
            //Already have a call on hold on this channel
            tel_args->res.status = BT_STATUS_ERROR;
            tel_args->res.line_status = APIERR_BADSTATE; 
            break;
        }

        //line is idle and we have a call on hold
    } while(0);
}

void HighLevelRetrieve(BTLINE *lp, struct args_telephone *tel_args)
{
    //If a call is on hold then disconnect and recover held call
    struct args_cc isdn_args;

    do
    {
        BT_ZERO(isdn_args);
        isdn_args.timeout = tel_args->timeout;

        BfvCallStatus(lp, &isdn_args);
        tel_args->res = isdn_args.res;

        if(isdn_args.res.status != BT_STATUS_OK)
            break;

        //If we don't have any calls on hold then we are done
        if(isdn_args.calls_on_hold != 0)
        {
            //Recover the call on hold
            BT_ZERO(isdn_args);
            isdn_args.timeout = tel_args->timeout;

            BfvCallRetrieve(lp, &isdn_args);
            tel_args->res = isdn_args.res;

            if(tel_args->res.status != BT_STATUS_OK)
                break;

            BT_ZERO(isdn_args);
            isdn_args.timeout = tel_args->timeout;

            BfvCallWaitForRetrieve(lp, &isdn_args);
            tel_args->res = isdn_args.res;

            if(tel_args->res.status != BT_STATUS_OK)
                break;          
        }
    } while(0);
}


int
SwitchConnectPrepareBeforeTransfer(MILL_LINE * lp1, MILL_LINE * lp2,
                                   int *port_unit_1, int *stream_1, int *slot_1,
                                   int *port_unit_2, int *stream_2, int *slot_2)
{
    BTERR bterr;
    struct args_tel_ctrl_call_sw args_sw;
    struct args_tel_ctrl_call_sw args_tel_sw;

    if (lp1 == 0 || lp2 == 0)
    {
        return 0;
    }

    //Disconnect two connections from CHANNEL to NETWORK
    BT_ZERO(args_tel_sw);
    while (BfvCallSWGetConns(lp1, &args_tel_sw) > 0)
    {
        if (args_tel_sw.src_port_class == CALL_SW_PORT_CHANNEL_DEF &&
            args_tel_sw.dest_port_class == CALL_SW_PORT_NETWORK_DEF)
        {
            if ((int)args_tel_sw.src_port_unit == unit + 2 ||
                (int)args_tel_sw.src_port_unit == unit + offset + 2)
            {
                BT_ZERO(args_sw);
                args_sw.conn_mode = CALL_SW_DISCONNECT_DEF;
                args_sw.dest_port_class = CALL_SW_PORT_CHANNEL_DEF;
                args_sw.dest_port_unit = args_tel_sw.src_port_unit;
                //args_sw.dest_stream = 0;
                //args_sw.dest_slot = 0;
                args_sw.src_port_class = CALL_SW_PORT_NETWORK_DEF;
                args_sw.src_port_unit = args_tel_sw.dest_port_unit;
                args_sw.src_stream = args_tel_sw.dest_stream;
                args_sw.src_slot = args_tel_sw.dest_slot;

                if ((int)args_tel_sw.src_port_unit == unit + 2)
                {
                    *port_unit_1 = args_tel_sw.dest_port_unit;
                    *stream_1 = args_tel_sw.dest_stream;
                    *slot_1 = args_tel_sw.dest_slot;
                }
                else
                {
                    *port_unit_2 = args_tel_sw.dest_port_unit;
                    *stream_2 = args_tel_sw.dest_stream;
                    *slot_2 = args_tel_sw.dest_slot;
                }

                BfvCallSWConnect(lp1, &args_sw);
                if(args_sw.res.status != BT_STATUS_OK)
                {
                    BfvErrorMessage(lp1,&args_sw.res,&bterr);
                    fprintf (stderr,"BfvCallSwConnect: %s\n",bterr.long_msg);
                    //goto got_err_2;
                }
            }
        }
    }

    //Connect two NETWORK timeslots
    BT_ZERO(args_sw);
    args_sw.conn_mode = CALL_SW_FULL_DUPLEX_DEF;
    args_sw.src_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.src_port_unit = *port_unit_1;
    args_sw.src_stream = *stream_1;
    args_sw.src_slot = *slot_1;
    args_sw.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.dest_port_unit = *port_unit_2;
    args_sw.dest_stream = *stream_2;
    args_sw.dest_slot = *slot_2;

    BfvCallSWConnect(lp1, &args_sw);
    if(args_sw.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp1,&args_sw.res,&bterr);
        fprintf (stderr,"BfvCallSwConnect: %s\n",bterr.long_msg);
        //goto got_err_2;
    }

    return 1;
}

int
SwitchConnectRestoreAfterTransfer(MILL_LINE * lp1, MILL_LINE * lp2,
                                  int port_unit_1, int stream_1, int slot_1,
                                  int port_unit_2, int stream_2, int slot_2)
{
    BTERR bterr;
    struct args_tel_ctrl_call_sw args_sw;

    if (lp1 == 0 || lp2 == 0)
    {
        return 0;
    }

    //Disconnect two NETWORK timeslots
    BT_ZERO(args_sw);
    args_sw.conn_mode = CALL_SW_DISCONNECT_DEF;
    args_sw.src_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.src_port_unit = port_unit_1;
    args_sw.src_stream = stream_1;
    args_sw.src_slot = slot_1;
    args_sw.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.dest_port_unit = port_unit_2;
    args_sw.dest_stream = stream_2;
    args_sw.dest_slot = slot_2;

    BfvCallSWConnect(lp1, &args_sw);
    if(args_sw.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp1,&args_sw.res,&bterr);
        fprintf (stderr,"BfvCallSwConnect: %s\n",bterr.long_msg);
        //goto got_err_2;
    }
    BT_ZERO(args_sw);
    args_sw.conn_mode = CALL_SW_DISCONNECT_DEF;
    args_sw.src_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.src_port_unit = port_unit_2;
    args_sw.src_stream = stream_2;
    args_sw.src_slot = slot_2;
    args_sw.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.dest_port_unit = port_unit_2;
    args_sw.dest_stream = stream_2;
    args_sw.dest_slot = slot_2;

    BfvCallSWConnect(lp1, &args_sw);
    if(args_sw.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp1,&args_sw.res,&bterr);
        fprintf (stderr,"BfvCallSwConnect: %s\n",bterr.long_msg);
        //goto got_err_2;
    }

    //reConnect two connections from CHANNEL to NETWORK
    BT_ZERO(args_sw);
    args_sw.conn_mode = CALL_SW_TRANSMIT_ONLY_DEF;
    args_sw.src_port_class = CALL_SW_PORT_CHANNEL_DEF;
    args_sw.src_port_unit = unit + 2;
    //args_sw.src_stream = 0;
    //args_sw.src_slot = 0;
    args_sw.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.dest_port_unit = port_unit_1;
    args_sw.dest_stream = stream_1;
    args_sw.dest_slot = slot_1;

    BfvCallSWConnect(lp1, &args_sw);
    if(args_sw.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp1,&args_sw.res,&bterr);
        fprintf (stderr,"BfvCallSwConnect: %s\n",bterr.long_msg);
        //goto got_err_2;
    }

    BT_ZERO(args_sw);
    args_sw.conn_mode = CALL_SW_TRANSMIT_ONLY_DEF;
    args_sw.src_port_class = CALL_SW_PORT_CHANNEL_DEF;
    args_sw.src_port_unit = unit + offset + 2;
    //args_sw.src_stream = 0;
    //args_sw.src_slot = 0;
    args_sw.dest_port_class = CALL_SW_PORT_NETWORK_DEF;
    args_sw.dest_port_unit = port_unit_2;
    args_sw.dest_stream = stream_2;
    args_sw.dest_slot = slot_2;

    BfvCallSWConnect(lp1, &args_sw);
    if(args_sw.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp1,&args_sw.res,&bterr);
        fprintf (stderr,"BfvCallSwConnect: %s\n",bterr.long_msg);
        //goto got_err_2;
    }

    return 1;
}


// for LINE_XFER_TWO_CHAN_NEEDS_NAILUP, 
// 2 bchannels need tobe connected by BfvCallSwitchConnect
// before complete transfer.
// The high level function BfvLineTransferComplete does the connection
// automatically in ECC.
// The low level function does it here if auto_sw_connect is false,
// else it is done inside BfvCallWaitTransferComplete automatically.
// The value of auto_sw_connect is passed into BfvCallWaitTransferComplete
void NailBchannelsAndCompleteTransfer(BTLINE *lp, struct args_telephone *args_tel)
{
    struct args_telephone tel_args2;
    int transfer_mode;
    int port_unit_1;
    int stream_1;
    int slot_1;
    int port_unit_2;
    int stream_2;
    int slot_2;

    BT_ZERO(tel_args2);
    MillLineTransferCapabilityQuery(lp, &tel_args2);
    if(tel_args2.res.status == BT_STATUS_OK)
    {
        transfer_mode = tel_args2.transfer_mode;
    }
    if (transfer_mode == ECC_XFER_TWO_CHAN_NEEDS_NAILUP && disable_auto_sw_connect == 1)
    {
        if (lp_second)
        {
            SwitchConnectPrepareBeforeTransfer(lp, lp_second, 
                &port_unit_1, &stream_1, &slot_1,
                &port_unit_2, &stream_2, &slot_2);
        }
    }

    //Complete Transfer
    HighLevelTransferComplete(lp, args_tel);

    if(args_tel->res.status != BT_STATUS_OK)
    {
        struct args_telephone args_tel2;

        BT_ZERO(args_tel2);
        args_tel2.timeout = args_tel->timeout;
        args_tel2.lp_second_channel = lp_second;
        //Need to reset all
        HighLevelTransferCancel(lp, &args_tel2);

        //give the user the original failure stored in tel_args
    }

    if (transfer_mode == ECC_XFER_TWO_CHAN_NEEDS_NAILUP && disable_auto_sw_connect == 1 && lp_second)
    {
        SwitchConnectRestoreAfterTransfer(lp, lp_second,
            port_unit_1, stream_1, slot_1,
            port_unit_2, stream_2, slot_2);
    }
}

/* This function is used by BfvCallWaitForComplete() */
/* to detect ringing. */
int call_progress_func (BTLINE *lp, char *arg)
{
    struct args_telephone args;
    
    BT_ZERO(args);
    while (BfvDataCP (lp, &args) == 0)
    {
        if ((args.data == CS_RING1) || (args.data == CS_RING2))
            return (1);
    }

    return (0);
}
