
/*
 * trombone.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The trombone program demonstrates how to setup and tear down a trombone
call transfer between two parties.
***************************************************************************/

#include "btlib.h"
#include "osilib.h"

#define PROGRAM_VERSION     "1.0.0"
#define MAX_PHONE_NUMBER    64

/* Global Variables */
int  g_iFirstUnit=-1, g_iSecondUnit=-1;
int  g_iErrorState=0;
long g_lTimeout=10000L;
char *g_szWelcomeFileName, *g_szRecordFileName;
char g_szPhoneNumber[MAX_PHONE_NUMBER];
bt_sem g_sWaitForCall;
BTLINE *g_lpP, *g_lpS;

/* Function Prototypes */
int  DigitalTrombone (struct args_tel_ctrl_call_sw *,
                      struct args_tel_ctrl_call_sw *);
int  AnalogTrombone (struct args_tel_ctrl_call_sw *,
                     struct args_tel_ctrl_call_sw *);
void DigitalTromboneCleanup (struct args_tel_ctrl_call_sw *,
                             struct args_tel_ctrl_call_sw *);
void AnalogTromboneCleanup (struct args_tel_ctrl_call_sw *,
                            struct args_tel_ctrl_call_sw *);
void *ThreadFunction (void *);
void DigitalHangup (void);
void AnalogEchoRecord (void);
void AlertThreadFunctionOfError (bt_thread);

/* Function to display program usage. */
void usage ()
{
    fprintf (stderr, "usage: trombone [options]\n");
    fprintf (stderr, "  -p    Primary channel number (Required).\n");
    fprintf (stderr, "  -s    Secondary Channel number (Required).\n");
    fprintf (stderr, "  -w    Welcome prompt file name (Required).\n");
    fprintf (stderr, "  -r    Recorded file name (Required).\n");
    fprintf (stderr, "  -n    Record timeout in seconds [10 secs (Default)].\n");
    fprintf (stderr, "  -d    Phone number to dial [\"1234\" (Default)].\n");
    fprintf (stderr, "  -v    Enable or disable debugging [0-Off, 1-On (Default)].\n");
    fprintf (stderr, "Program version: %s.\n", PROGRAM_VERSION);
    
    exit (1);
}

int main (int argc, char **argv)
{
    int c, iDebugMode=1;
    int iThreadStatus;
    bt_thread tThreadID;
    extern int optind;
    extern char *optarg;
    BTERR bterr;
        
    struct args_tel_ctrl_call_sw args_new_conn;
    struct args_tel_ctrl_call_sw args_lpP_chan_net;
    struct args_tel_ctrl_call_sw args_lpP_net_chan;
    struct args_tel_ctrl_call_sw args_lpS_chan_net;
    struct args_tel_ctrl_call_sw args_lpS_net_chan;

    struct args_line_admin args_admin;
    struct args_telephone args_tel;
    struct args_infopkt args_infopkt;
    struct infopkt_stream *play_ips, *rec_ips;
    struct args_speech args_speech;
    struct args_tone args_tone;
    struct args_debug_mode args_debug;
    
    /* Explictly initialize some global variables. */
    g_szRecordFileName = g_szWelcomeFileName = NULL;
    strcpy ((char *)g_szPhoneNumber, (char *)"1234");
    
    while ((c = getopt (argc, argv, "v:p:s:n:w:d:r:")) != EOF)
    {
        switch (c)
        {
        case 'v':
            iDebugMode = atoi (optarg);
            break;
        
        case 'p':
            g_iFirstUnit = atoi (optarg);
            break;
        
        case 's':
            g_iSecondUnit = atoi (optarg);
            break;
        
        case 'n':
            g_lTimeout = 1000 * atol (optarg);
            break;
        
        case 'w':
            g_szWelcomeFileName = optarg;
            break;
        
        case 'r':
            g_szRecordFileName = optarg;
            break;
        
        case 'd':
            strncpy ((char *)g_szPhoneNumber, (char *)optarg,
              sizeof(char)*MAX_PHONE_NUMBER);
            break;
        
        default:
            usage ();
        }
    }

    if (g_iFirstUnit == -1 || g_iSecondUnit == -1 ||
      g_szWelcomeFileName == NULL || g_szRecordFileName == NULL)
        usage ();
        
    /* Turn on debugging, default. */
    if (iDebugMode != 0)
    {
        BT_ZERO(args_debug);
        args_debug.mode = -1;
        args_debug.options = API_DBG_MODE_SET;
        BfvDebugModeSetAdv(&args_debug);
    }

    /* Create a wait for call semaphore for the secondary channel. */
    if (osiSemaphore_Init (&g_sWaitForCall, 1, S_THREAD))
    {
        fprintf (stdout, "OsiSemaphore_Init() Failed.\n");
                
        exit (1);
    }

    /* Set this semaphore to the non-signaled state. */
    osiSemaphore_Decrement (&g_sWaitForCall, 0);

    /* Attach and Reset first unit. */
    BT_ZERO(args_admin);
    args_admin.unit = g_iFirstUnit;
    if ((g_lpP = BfvLineAttach (&args_admin)) == NULL)
    {
        BfvErrorMessage (g_lpP, &args_admin.res, &bterr);
        fprintf (stderr, "BfvLineAttach: %s\n", bterr.long_msg);
                
        exit (1);
    }

    fprintf (stdout, "Channel %d: Attached.\n", g_iFirstUnit);

    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (g_lpP, &args_admin) < 0)
    {
        BfvErrorMessage (g_lpP, &args_admin.res, &bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX\n", bterr.long_msg,
          args_admin.reset_status);
        goto GOT_ERROR4;
    }

    fprintf (stdout, "Channel %d: Reset.\n", g_iFirstUnit);

    /* Check for digital configuration error */
    if (args_admin.reset_status & RST_DIG_CONFIG_UNREADABLE)
    {
        fprintf (stderr, "Channel %d: Telephony/Digital configuration file not found.\n",
          g_iFirstUnit);
        goto GOT_ERROR4;
    }
    
    /* Check if channel can record speech. */
    if (!LINE_HAS_CAP(g_lpP, CAP_SPEECH_RECORD))
    {
        fprintf (stderr, "Channel %d: Does not support speech recording.\n",
          g_iFirstUnit);
        goto GOT_ERROR4;
    }
    
    /* Check if channel can play speech. */
    if (!LINE_HAS_CAP(g_lpP, CAP_SPEECH_PLAY))
    {
        fprintf (stderr, "Channel %d: Does not support speech playback.\n",
          g_iFirstUnit);
        goto GOT_ERROR4;
    }

    /* Attach and Reset second unit. */
    BT_ZERO(args_admin);
    args_admin.unit = g_iSecondUnit;
    if ((g_lpS = BfvLineAttach (&args_admin)) == NULL)
    {
        BfvErrorMessage (g_lpS, &args_admin.res, &bterr);
        fprintf (stderr, "BfvLineAttach: %s\n", bterr.long_msg);
        goto GOT_ERROR4;
    }

    fprintf (stdout, "Channel %d: Attached.\n", g_iSecondUnit);
    
    BT_ZERO(args_admin);
    args_admin.config_file_name = "btcall.cfg";
    if (BfvLineReset (g_lpS, &args_admin) < 0)
    {
        BfvErrorMessage (g_lpS, &args_admin.res, &bterr);
        fprintf (stderr, "BfvLineReset: %s: status %lX\n", bterr.long_msg,
          args_admin.reset_status);
        goto GOT_ERROR3;
    }

    fprintf (stdout, "Channel %d: Reset.\n", g_iSecondUnit);

    /* Check for digital configuration error */
    if (args_admin.reset_status & RST_DIG_CONFIG_UNREADABLE)
    {
        fprintf (stderr, "Channel %d: Telephony/Digital configuration file not found.\n",
          g_iSecondUnit);
        goto GOT_ERROR3;
    }

    /* Check if channel can record speech. */
    if (!LINE_HAS_CAP(g_lpS, CAP_SPEECH_RECORD))
    {
        fprintf (stderr, "Channel %d: Does not support speech recording.\n",
          g_iFirstUnit);
        goto GOT_ERROR3;
    }

    /*********** Get connection info ****************/

    /* Save the current connections from g_lpP and g_lpS. */
    BT_ZERO(args_lpP_chan_net);
    BT_ZERO(args_lpP_net_chan);

    BT_ZERO(args_lpS_chan_net);
    BT_ZERO(args_lpS_net_chan);
                
    BT_ZERO(args_new_conn);
    while (BfvCallSWGetConns (g_lpP, &args_new_conn) == 1)
    {
        /* The unit number from the firmware is 2 greater */
        /* than the ordinal number that we are using. */
        /* We need to add 2 to our unit variables. */

        if ( (args_new_conn.src_port_unit == (unsigned)(g_iFirstUnit + 2)) &&
             (args_new_conn.src_port_class == CALL_SW_PORT_CHANNEL_DEF) )
        {
            fprintf (stdout,
                     "Channel %d, found src connection info: %d, %d, %d, %d, %d, %d, %d, %d, %d.\n",
                     g_iFirstUnit, args_new_conn.conn_mode,
                     args_new_conn.src_port_class, args_new_conn.src_port_unit,
                     args_new_conn.src_stream, args_new_conn.src_slot,
                     args_new_conn.dest_port_class, args_new_conn.dest_port_unit,
                     args_new_conn.dest_stream, args_new_conn.dest_slot);

            memcpy ((struct args_tel_ctrl_call_sw *)&args_lpP_chan_net,
                    (struct args_tel_ctrl_call_sw *)&args_new_conn,
                    sizeof(struct args_tel_ctrl_call_sw));
        }

        if ( (args_new_conn.dest_port_unit == (unsigned)(g_iFirstUnit + 2)) &&
             (args_new_conn.dest_port_class == CALL_SW_PORT_CHANNEL_DEF) )
        {
            fprintf (stdout,
                     "Channel %d, found dest connection info: %d, %d, %d, %d, %d, %d, %d, %d, %d.\n",
                     g_iFirstUnit, args_new_conn.conn_mode,
                     args_new_conn.src_port_class, args_new_conn.src_port_unit,
                     args_new_conn.src_stream, args_new_conn.src_slot,
                     args_new_conn.dest_port_class, args_new_conn.dest_port_unit,
                     args_new_conn.dest_stream, args_new_conn.dest_slot);

            memcpy ((struct args_tel_ctrl_call_sw *)&args_lpP_net_chan,
                    (struct args_tel_ctrl_call_sw *)&args_new_conn,
                    sizeof(struct args_tel_ctrl_call_sw));
        }

        if ( (args_new_conn.src_port_unit == (unsigned)(g_iSecondUnit + 2)) &&
             (args_new_conn.src_port_class == CALL_SW_PORT_CHANNEL_DEF) )
        {
            fprintf (stdout,
                     "Channel %d, found src connection info: %d, %d, %d, %d, %d, %d, %d, %d, %d.\n",
                     g_iSecondUnit, args_new_conn.conn_mode,
                     args_new_conn.src_port_class, args_new_conn.src_port_unit,
                     args_new_conn.src_stream, args_new_conn.src_slot,
                     args_new_conn.dest_port_class, args_new_conn.dest_port_unit,
                     args_new_conn.dest_stream, args_new_conn.dest_slot);

            memcpy ((struct args_tel_ctrl_call_sw *)&args_lpS_chan_net,
                    (struct args_tel_ctrl_call_sw *)&args_new_conn,
                    sizeof(struct args_tel_ctrl_call_sw));
        }
        
        if ( (args_new_conn.dest_port_unit == (unsigned)(g_iSecondUnit + 2)) &&
             (args_new_conn.dest_port_class == CALL_SW_PORT_CHANNEL_DEF) )
        {
            fprintf (stdout,
                     "Channel %d, found dest connection info: %d, %d, %d, %d, %d, %d, %d, %d, %d.\n",
                     g_iSecondUnit, args_new_conn.conn_mode,
                     args_new_conn.src_port_class, args_new_conn.src_port_unit,
                     args_new_conn.src_stream, args_new_conn.src_slot,
                     args_new_conn.dest_port_class, args_new_conn.dest_port_unit,
                     args_new_conn.dest_stream, args_new_conn.dest_slot);

            memcpy ((struct args_tel_ctrl_call_sw *)&args_lpS_net_chan,
                    (struct args_tel_ctrl_call_sw *)&args_new_conn,
                    sizeof(struct args_tel_ctrl_call_sw));
        }
    }
    
    /************ Wait for first call *************/

    /* Open welcome prompt file. */
    BT_ZERO(args_infopkt);
    args_infopkt.fname = g_szWelcomeFileName;
    args_infopkt.fmode = "r";
    if ((play_ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage (NULL, &args_infopkt.res, &bterr);
        fprintf (stderr, "BfvInfopktOpen: %s\n", bterr.long_msg);
        goto GOT_ERROR3;
    }
    
    /* Create a thread to control the secondary channel. */
    if (osiThread_Create (NULL, 0, ThreadFunction, (void *)NULL, 0,
      (unsigned long *)&tThreadID))
    {
        fprintf (stdout, "osiThread_Create() Failed.\n");
        goto GOT_ERROR3;
    }
    
    fprintf (stdout, "Channel %d: Waiting for a call ...\n", g_iFirstUnit);
    
    /* Wait here for an inbound call. */
    BT_ZERO(args_tel);
    BfvLineWaitForCall (g_lpP, &args_tel);
    if (args_tel.res.line_status != WAIT_FOR_CALL_OK)
    {
        BfvErrorMessage (g_lpP, &args_tel.res, &bterr);
        fprintf (stderr, "BfvLineWaitForCall: %s\n", bterr.long_msg);

        args_infopkt.ips = play_ips;
        BfvInfopktClose (&args_infopkt);

        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR3;
    }
        
    fprintf (stdout, "Channel %d: Off-Hook.\n", g_iFirstUnit);

    BT_ZERO(args_tel);
    BfvLineAnswer (g_lpP, &args_tel);
    
    fprintf (stdout, "Channel %d: Playing welcome prompt %s.\n", g_iFirstUnit,
      g_szWelcomeFileName);
    
    /* Play a welcome prompt to caller. */
    BT_ZERO(args_speech);
    args_speech.ips = play_ips;
    args_speech.func = NULL;
    BfvSpeechPlay (g_lpP, &args_speech);

    /* Close welcome prompt file. */
    args_infopkt.ips = play_ips;
    BfvInfopktClose (&args_infopkt);
    if (args_infopkt.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (NULL, &args_infopkt.res, &bterr);
        fprintf (stderr, "BfvInfopktClose: %s\n",bterr.long_msg);
        fprintf (stdout, "Channel %d: Ignoring BfvInfopktClose() error.\n",
          g_iFirstUnit);
    }

    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechPlay: %s\n",bterr.long_msg);

        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR2;
    }
    
    fprintf (stdout, "Channel %d: Dialing %s ...\n", g_iSecondUnit, g_szPhoneNumber);
        
    /* Dial the second call */
    BT_ZERO(args_tel);
    args_tel.phonenum = g_szPhoneNumber;
    args_tel.call_protocol_code = CALL_PROTOCOL_VOICE;
    BfvLineOriginateCall (g_lpS, &args_tel);
    if (args_tel.res.status != BT_STATUS_OK || LINE_STATE(g_lpS) != LINE_STATE_CONNECTED)
    {
        BfvErrorMessage (g_lpS, &args_tel.res, &bterr);
        fprintf (stderr, "BfvLineOriginateCall: %s\n", bterr.long_msg);

        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR2;
    }
    
    fprintf (stdout, "Channel %d: Off-Hook.\n", g_iSecondUnit);

    /******************************************************************/
    /**************** Switch Connections ******************************/
    /******************************************************************/

    /*************** Disconnect current connections ****************/

    /* Disconnect TRANSMIT on g_lpP from Chan 0 -> Net 0. */
    /* Disconnects are inverted so use net -> chan. */
    BT_ZERO(args_new_conn);
    memcpy ((struct args_tel_ctrl_call_sw *)&args_new_conn,
            (struct args_tel_ctrl_call_sw *)&args_lpP_net_chan,
            sizeof(struct args_tel_ctrl_call_sw));
    args_new_conn.conn_mode = CALL_SW_DISCONNECT_DEF;
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #1: %s\n", bterr.long_msg);

        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR1;
    }

    /* Disconnect TRANSMIT on g_lpS from Chan 1 -> Net 1. */
    /* Disconnects are inverted so use net -> chan. */
    BT_ZERO(args_new_conn);
    memcpy ((struct args_tel_ctrl_call_sw *)&args_new_conn,
            (struct args_tel_ctrl_call_sw *)&args_lpS_net_chan,
            sizeof(struct args_tel_ctrl_call_sw));
    args_new_conn.conn_mode = CALL_SW_DISCONNECT_DEF;
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #2: %s\n", bterr.long_msg);

        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR1;
    }
    
    fprintf (stdout, "Disconnects Done.\n");

    /************** New connections ***************/

    /* New TRANSMIT connection network ts0 -> ts1. */
    BT_ZERO(args_new_conn);
    args_new_conn.conn_mode       = args_lpP_net_chan.conn_mode;
    args_new_conn.src_port_class  = args_lpP_net_chan.src_port_class;
    args_new_conn.src_port_unit   = args_lpP_net_chan.src_port_unit;
    args_new_conn.src_stream      = args_lpP_net_chan.src_stream;
    args_new_conn.src_slot        = args_lpP_net_chan.src_slot;
    args_new_conn.dest_port_class = args_lpS_net_chan.src_port_class;
    args_new_conn.dest_port_unit  = args_lpS_net_chan.src_port_unit;
    args_new_conn.dest_stream     = args_lpS_net_chan.src_stream;
    args_new_conn.dest_slot       = args_lpS_net_chan.src_slot;
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #3: %s\n", bterr.long_msg);
        
        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR1;
    }

    /* New TRANSMIT connection network ts1 -> ts0. */
    BT_ZERO(args_new_conn);
    args_new_conn.conn_mode       = args_lpS_net_chan.conn_mode;
    args_new_conn.src_port_class  = args_lpS_net_chan.src_port_class;
    args_new_conn.src_port_unit   = args_lpS_net_chan.src_port_unit;
    args_new_conn.src_stream      = args_lpS_net_chan.src_stream;
    args_new_conn.src_slot        = args_lpS_net_chan.src_slot;
    args_new_conn.dest_port_class = args_lpP_net_chan.src_port_class;
    args_new_conn.dest_port_unit  = args_lpP_net_chan.src_port_unit;
    args_new_conn.dest_stream     = args_lpP_net_chan.src_stream;
    args_new_conn.dest_slot       = args_lpP_net_chan.src_slot;
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #4: %s\n", bterr.long_msg);

        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR1;
    }

    fprintf (stdout, "New connections Done.\n");

    /************** Setup Trombone Call. ************/

    /* Test for the telephony hardware type. */
    if (LINE_HAS_CAP(g_lpP, CAP_PHIF_ANALOG_LOOP_START))
    {
        fprintf (stdout, "Analog hardware detected.\n");

        if (AnalogTrombone (&args_lpP_chan_net, &args_lpS_chan_net))
        {
            AlertThreadFunctionOfError (tThreadID);
            goto GOT_ERROR1;
        }
    }
    else
    {
        fprintf (stdout, "Digital hardware detected.\n");

        if (DigitalTrombone (&args_lpP_chan_net, &args_lpS_net_chan))
        {
            AlertThreadFunctionOfError (tThreadID);
            goto GOT_ERROR1;
        }
    }
    
    /* Open file and record data */
    BT_ZERO(args_infopkt);
    args_infopkt.fname = g_szRecordFileName;
    args_infopkt.fmode = "w";
    if ((rec_ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage (NULL, &args_infopkt.res, &bterr);
        fprintf (stderr, "BfvInfopktOpen: %s\n", bterr.long_msg);

        AlertThreadFunctionOfError (tThreadID);
        goto GOT_ERROR1;
    }

    /* Signal thread function that a call has arrived. */
    osiSemaphore_Increment (&g_sWaitForCall);

    /* Delay recording by one second. */
    osiWait (1000);

    /* Enable DTMF detection. */
    BT_ZERO(args_tone);
    args_tone.decode_flag = DTMF_12TONE;
    BfvToneDetectEnable (g_lpP, &args_tone);

    fprintf (stdout, "Channel %d: Recording file %s ...\n", g_iFirstUnit,
      g_szRecordFileName);

    /* Record speech from channel 0. */
    BT_ZERO(args_speech);
    args_speech.ips = rec_ips;
    args_speech.coding_fmt = CODE_PCM_ULAW;
    args_speech.rate = RATE_8000;
    args_speech.bits_per_samp = BITS_8;
    args_speech.afe_rate = AFE_8000;
    args_speech.data_fmt = SPCH_MSB;
    args_speech.timeout = g_lTimeout;
    args_speech.vad = 1;
    args_speech.asr_mode = 1;
    BfvSpeechRecord (g_lpP, &args_speech);

    args_infopkt.ips = rec_ips;
    BfvInfopktClose (&args_infopkt);
    if (args_infopkt.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (NULL, &args_infopkt.res, &bterr);
        fprintf (stderr, "BfvInfopktClose: %s\n",bterr.long_msg);
        fprintf (stdout, "Channel %d: Ignoring BfvInfopktClose() error.\n",
          g_iFirstUnit);
    }

    if (args_speech.res.status != BT_STATUS_OK)
    {
        /* If the secondary channel has sent an alert then */
        /* it has detected a hang-up. */
        if (args_speech.res.status == BT_STATUS_ALERT)
        {
            fprintf (stdout, "Channel %d: Alerted!\n", g_iFirstUnit);
        }
        else
        {
            /* If a hang-up or error occurs than send an */
            /* alert to the secondary channel. */
            if (LINE_STATE(g_lpP) == LINE_STATE_IDLE)
            {
                fprintf (stdout, "Channel %d: Hang-up detected!\n", g_iFirstUnit);
            }
            else
            {
                BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
                fprintf (stderr, "BfvSpeechRecord: %s\n", bterr.long_msg);
            }

            fprintf (stdout, "Channel %d: Alerting channel %d.\n", g_iFirstUnit,
              g_iSecondUnit);

            BT_ZERO(args_admin);
            args_admin.unit = g_iSecondUnit;
            BfvLineAlert (&args_admin);
        }
    
        /* Wait for Thread Function to return. */
        osiThread_Wait (tThreadID, &iThreadStatus);
    }
    
    fprintf (stdout, "Channel %d: Recording Done.\n", g_iFirstUnit);

    /******** Clean up *******/

    /* Test for the telephony hardware type. */
    if (LINE_HAS_CAP(g_lpP, CAP_PHIF_ANALOG_LOOP_START))
        AnalogTromboneCleanup (&args_lpP_chan_net, &args_lpS_chan_net);
    else
        DigitalTromboneCleanup (&args_lpP_chan_net, &args_lpS_net_chan);

    /* Disconnect connection network ts0 -> ts1. */
    /* Disconnects are inverted so use ts1 -> ts0. */
    BT_ZERO(args_new_conn);
    args_new_conn.conn_mode       = CALL_SW_DISCONNECT_DEF;
    args_new_conn.src_port_class  = args_lpS_net_chan.src_port_class;
    args_new_conn.src_port_unit   = args_lpS_net_chan.src_port_unit;
    args_new_conn.src_stream      = args_lpS_net_chan.src_stream;
    args_new_conn.src_slot        = args_lpS_net_chan.src_slot;
    args_new_conn.dest_port_class = args_lpP_net_chan.src_port_class;
    args_new_conn.dest_port_unit  = args_lpP_net_chan.src_port_unit;
    args_new_conn.dest_stream     = args_lpP_net_chan.src_stream;
    args_new_conn.dest_slot       = args_lpP_net_chan.src_slot;
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #7: %s\n", bterr.long_msg);
    }

    /* Disconnect connection network ts1 -> ts0. */
    /* Disconnects are inverted so use ts0 -> ts1. */
    BT_ZERO(args_new_conn);
    args_new_conn.conn_mode       = CALL_SW_DISCONNECT_DEF;
    args_new_conn.src_port_class  = args_lpP_net_chan.src_port_class;
    args_new_conn.src_port_unit   = args_lpP_net_chan.src_port_unit;
    args_new_conn.src_stream      = args_lpP_net_chan.src_stream;
    args_new_conn.src_slot        = args_lpP_net_chan.src_slot;
    args_new_conn.dest_port_class = args_lpS_net_chan.src_port_class;
    args_new_conn.dest_port_unit  = args_lpS_net_chan.src_port_unit;
    args_new_conn.dest_stream     = args_lpS_net_chan.src_stream;
    args_new_conn.dest_slot       = args_lpS_net_chan.src_slot;
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #8: %s\n", bterr.long_msg);
    }
    
    fprintf (stdout, "New disconnects Done.\n");
    
    /*********** Replace original connections. **********/
    
    /* Re-Connect g_lpP. */
    BT_ZERO(args_new_conn);
    memcpy ((struct args_tel_ctrl_call_sw *)&args_new_conn,
            (struct args_tel_ctrl_call_sw *)&args_lpP_chan_net,
            sizeof(struct args_tel_ctrl_call_sw));
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #9: %s\n", bterr.long_msg);
    }

    /* Re-Connect g_lpS. */
    BT_ZERO(args_new_conn);
    memcpy ((struct args_tel_ctrl_call_sw *)&args_new_conn,
            (struct args_tel_ctrl_call_sw *)&args_lpS_chan_net,
            sizeof(struct args_tel_ctrl_call_sw));
    if (BfvCallSWConnect (g_lpP, &args_new_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_new_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #10: %s\n", bterr.long_msg);
    }
    
GOT_ERROR1:
    
    BT_ZERO(args_tel);
    BfvLineTerminateCall (g_lpS, &args_tel);

    fprintf (stdout, "Channel %d: On-Hook.\n", g_iSecondUnit);

GOT_ERROR2:

    BT_ZERO(args_tel);
    BfvLineTerminateCall (g_lpP, &args_tel);
    
    fprintf (stdout, "Channel %d: On-Hook.\n", g_iFirstUnit);

GOT_ERROR3:

    BT_ZERO(args_admin);
    BfvLineDetach (g_lpS, &args_admin);
    
GOT_ERROR4:

    BT_ZERO(args_admin);
    BfvLineDetach (g_lpP, &args_admin);
    
    /* Free these system resources. */
    osiSemaphore_Destroy (&g_sWaitForCall);
    osiThread_Destroy (tThreadID);
    
    return (0);
}

int DigitalTrombone (struct args_tel_ctrl_call_sw *args_g_lpP,
             struct args_tel_ctrl_call_sw *args_g_lpS)
{
    struct args_tel_ctrl_call_sw args_conn;
    struct args_speech args_speech;
    BTERR bterr;
    
    /* Reset the echo cancellor on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_RESET;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl RESET: %s\n", bterr.long_msg);
    
        return (1);
    }

    /* Get reference signal from input slot 1 on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_ALT_INPUT_ENABLE;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl ALT_INPUT: %s\n", bterr.long_msg);
    
        return (1);
    }
    
    /* New TRANSMIT connection ts 1 -> chan 0 REF. */
    BT_ZERO(args_conn);
    args_conn.conn_mode       = args_g_lpS->conn_mode;
    args_conn.src_port_class  = args_g_lpS->src_port_class;
    args_conn.src_port_unit   = args_g_lpS->src_port_unit;
    args_conn.src_stream      = args_g_lpS->src_stream;
    args_conn.src_slot        = args_g_lpS->src_slot;
    args_conn.dest_port_class = args_g_lpP->src_port_class;
    args_conn.dest_port_unit  = args_g_lpP->src_port_unit;
    args_conn.dest_stream     = args_g_lpP->src_stream;
    args_conn.dest_slot       = 1;
    if (BfvCallSWConnect (g_lpP, &args_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #5: %s\n", bterr.long_msg);
        
        return (1);
    }
    
    fprintf (stdout, "New echo cancellation connections done.\n");

    return (0);
}

void DigitalTromboneCleanup (struct args_tel_ctrl_call_sw *args_g_lpP,
                 struct args_tel_ctrl_call_sw *args_g_lpS)
{
    struct args_tel_ctrl_call_sw args_conn;
    struct args_speech args_speech;
    BTERR bterr;
    
    /* Disable alternate input on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_ALT_INPUT_DISABLE;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl DISABLE ALT_IN: %s\n",
          bterr.long_msg);
    
        return;
    }
    
    /* Disconnect TRANSMIT connection ts 1 -> chan 0 REF. */
    /* Disconnects are inverted so use ref -> net. */
    BT_ZERO(args_conn);
    args_conn.conn_mode       = CALL_SW_DISCONNECT_DEF;
    args_conn.src_port_class  = args_g_lpP->src_port_class;
    args_conn.src_port_unit   = args_g_lpP->src_port_unit;
    args_conn.src_stream      = args_g_lpP->src_stream;
    args_conn.src_slot        = 1;
    args_conn.dest_port_class = args_g_lpS->src_port_class;
    args_conn.dest_port_unit  = args_g_lpS->src_port_unit;
    args_conn.dest_stream     = args_g_lpS->src_stream;
    args_conn.dest_slot       = args_g_lpP->src_slot;
    if (BfvCallSWConnect (g_lpP, &args_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #6: %s\n", bterr.long_msg);
    
        return;
    }
    
    fprintf (stdout, "Removed echo cancellation connections.\n");

    return;
}

int AnalogTrombone (struct args_tel_ctrl_call_sw *args_g_lpP,
            struct args_tel_ctrl_call_sw *args_g_lpS)
{
    struct args_tel_ctrl_call_sw args_conn;
    struct args_speech args_speech;
    BTERR bterr;
    
    /* Reset the echo cancellor on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_RESET;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl Primary RESET: %s\n",
          bterr.long_msg);
    
        return (1);
    }

    /* Get reference signal from input slot 1 on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_ALT_INPUT_ENABLE;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl Primary ALT_INPUT: %s\n",
          bterr.long_msg);
    
        return (1);
    }

    /* Send echo cancelled data to host and reference slot 1 on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_OUTPUT_SELECT;
    args_speech.output_select = SPEECH_RECORD_TO_HOST_TSLOT_DEF;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl Primary ALT_OUTPUT: %s\n",
          bterr.long_msg);
    
        return (1);
    }

    /* Reset the echo cancellor on secondary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_RESET;
    BfvSpeechEchoCancelControl (g_lpS, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpS, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl Secondary RESET: %s\n",
          bterr.long_msg);
    
        return (1);
    }

    /* Get reference signal from input slot 1 on secondary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_ALT_INPUT_ENABLE;
    BfvSpeechEchoCancelControl (g_lpS, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpS, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl Secondary ALT_INPUT: %s\n",
          bterr.long_msg);
    
        return (1);
    }

    /* Send echo cancelled data to host and reference slot 1 on secondary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_OUTPUT_SELECT;
    args_speech.output_select = SPEECH_RECORD_TO_TSLOT_ONLY_DEF;
    BfvSpeechEchoCancelControl (g_lpS, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpS, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl Secondary ALT_OUTPUT: %s\n",
          bterr.long_msg);
    
        return (1);
    }
    
    /* New TRANSMIT connection chan 1 REF OUT -> chan 0 REF IN. */
    BT_ZERO(args_conn);
    args_conn.conn_mode       = args_g_lpS->conn_mode;
    args_conn.src_port_class  = args_g_lpS->src_port_class;
    args_conn.src_port_unit   = args_g_lpS->src_port_unit;
    args_conn.src_stream      = args_g_lpS->src_stream;
    args_conn.src_slot        = 1;
    args_conn.dest_port_class = args_g_lpP->src_port_class;
    args_conn.dest_port_unit  = args_g_lpP->src_port_unit;
    args_conn.dest_stream     = args_g_lpP->src_stream;
    args_conn.dest_slot       = 1;
    if (BfvCallSWConnect (g_lpP, &args_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #5A: %s\n", bterr.long_msg);
    
        return (1);
    }

    /* New TRANSMIT connection chan 0 REF OUT -> chan 1 REF IN. */
    BT_ZERO(args_conn);
    args_conn.conn_mode       = args_g_lpP->conn_mode;
    args_conn.src_port_class  = args_g_lpP->src_port_class;
    args_conn.src_port_unit   = args_g_lpP->src_port_unit;
    args_conn.src_stream      = args_g_lpP->src_stream;
    args_conn.src_slot        = 1;
    args_conn.dest_port_class = args_g_lpS->src_port_class;
    args_conn.dest_port_unit  = args_g_lpS->src_port_unit;
    args_conn.dest_stream     = args_g_lpS->src_stream;
    args_conn.dest_slot       = 1;
    if (BfvCallSWConnect (g_lpP, &args_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #5B: %s\n", bterr.long_msg);
    
        return (1);
    }
    
    fprintf (stdout, "New echo cancellation connections done.\n");

    return (0);
}

void AnalogTromboneCleanup (struct args_tel_ctrl_call_sw *args_lpP,
               struct args_tel_ctrl_call_sw *args_lpS)
{
    struct args_tel_ctrl_call_sw args_conn;
    struct args_speech args_speech;
    BTERR bterr;
    
    /* Disable alternate input on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_ALT_INPUT_DISABLE;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl DISABLE ALT_IN: %s\n",
          bterr.long_msg);
    
        return;
    }

    /* Route record data to host on primary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_OUTPUT_SELECT;
    args_speech.output_select = SPEECH_RECORD_TO_HOST_ONLY_DEF;
    BfvSpeechEchoCancelControl (g_lpP, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpP, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl DISABLE ALT_OUT: %s\n",
          bterr.long_msg);
    
        return;
    }

    /* Disable alternate input on secondary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_ALT_INPUT_DISABLE;
    BfvSpeechEchoCancelControl (g_lpS, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpS, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl DISABLE ALT_IN: %s\n",
          bterr.long_msg);
    
        return;
    }

    /* Route record data to host on secondary channel. */
    BT_ZERO(args_speech);
    args_speech.echoc_op = ECHOC_OP_OUTPUT_SELECT;
    args_speech.output_select = SPEECH_RECORD_TO_HOST_ONLY_DEF;
    BfvSpeechEchoCancelControl (g_lpS, &args_speech);
    if (args_speech.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage (g_lpS, &args_speech.res, &bterr);
        fprintf (stderr, "BfvSpeechEchoCancelControl DISABLE ALT_OUT: %s\n",
          bterr.long_msg);
    
        return;
    }
    
    /* Disconnect TRANSMIT connection chan 1 REF -> chan 0 REF. */
    /* Disconnects are inverted so use ref 0 -> ref 1. */
    BT_ZERO(args_conn);
    args_conn.conn_mode       = CALL_SW_DISCONNECT_DEF;
    args_conn.src_port_class  = args_lpP->src_port_class;
    args_conn.src_port_unit   = args_lpP->src_port_unit;
    args_conn.src_stream      = args_lpP->src_stream;
    args_conn.src_slot        = 1;
    args_conn.dest_port_class = args_lpS->src_port_class;
    args_conn.dest_port_unit  = args_lpS->src_port_unit;
    args_conn.dest_stream     = args_lpS->src_stream;
    args_conn.dest_slot       = 1;
    if (BfvCallSWConnect (g_lpP, &args_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #6A: %s\n", bterr.long_msg);
    
        return;
    }

    /* Disconnect TRANSMIT connection chan 0 REF -> chan 1 REF. */
    /* Disconnects are inverted so use ref 1 -> ref 0. */
    BT_ZERO(args_conn);
    args_conn.conn_mode       = CALL_SW_DISCONNECT_DEF;
    args_conn.src_port_class  = args_lpS->src_port_class;
    args_conn.src_port_unit   = args_lpS->src_port_unit;
    args_conn.src_stream      = args_lpS->src_stream;
    args_conn.src_slot        = 1;
    args_conn.dest_port_class = args_lpP->src_port_class;
    args_conn.dest_port_unit  = args_lpP->src_port_unit;
    args_conn.dest_stream     = args_lpP->src_stream;
    args_conn.dest_slot       = 1;
    if (BfvCallSWConnect (g_lpP, &args_conn) != 0)
    {
        BfvErrorMessage (NULL, &args_conn.res, &bterr);
        fprintf (stderr, "BfvCallSWConnect #6B: %s\n", bterr.long_msg);
    
        return;
    }

    fprintf (stdout, "Removed echo cancellation connections.\n");

    return;
}


void *ThreadFunction (void *input)
{
    fprintf (stdout, "Channel %d: Thread created and waiting ...\n", g_iSecondUnit);

    osiSemaphore_Decrement (&g_sWaitForCall, INFINITE);

    if (g_iErrorState == 1)
    {
        fprintf (stdout, "Channel %d: Thread exiting due to error ...\n", g_iSecondUnit);
    
        return (NULL);
    }

    /* Test for the telephony hardware type. */
    if (LINE_HAS_CAP(g_lpS, CAP_PHIF_ANALOG_LOOP_START))
        AnalogEchoRecord ();
    else
        DigitalHangup ();

    return (NULL);
}

void DigitalHangup (void)
{
    struct args_packet args_pkt;
    struct args_line_admin args_admin;

    fprintf (stdout, "Channel %d: Watching for a hang-up ...\n", g_iSecondUnit);

    while (1)
    {
        /* Look for a hang-up or alert packet. */
        BT_ZERO(args_pkt);
        args_pkt.timeout = 0L;
        BfvRcvProcessPkt(g_lpS, &args_pkt);
        if (args_pkt.res.status == BT_STATUS_ALERT)
        {
            fprintf (stdout, "Channel %d: Alerted!\n", g_iSecondUnit);
            break;
        }

        if (LINE_STATE(g_lpS) == LINE_STATE_IDLE)
        {
            fprintf (stdout, "Channel %d: Hang-up detected!  Alerting channel %d.\n",
              g_iSecondUnit, g_iFirstUnit);

            BT_ZERO(args_admin);
            args_admin.unit = g_iFirstUnit;
            BfvLineAlert (&args_admin);
            break;
        }
    }
    
    return;
}

void AnalogEchoRecord (void)
{
    struct args_speech args_speech;
    struct infopkt_stream *echo_ips;
    struct args_infopkt args_infopkt;
    struct args_line_admin args_admin;
    BTERR bterr;

    /* Open record echo file. */
    BT_ZERO(args_infopkt);
    args_infopkt.fname = "echo.ips";
    args_infopkt.fmode = "w";
    if ((echo_ips = BfvInfopktOpen (&args_infopkt)) == NULL)
    {
        BfvErrorMessage (NULL, &args_infopkt.res, &bterr);
        fprintf (stderr, "BfvInfopktOpen: %s\n", bterr.long_msg);
    }
    else
    {
        fprintf (stdout, "Channel %d: Recording echo ...\n", g_iSecondUnit);

        /* Record from chan 1 to produce echo cancelled data. */
        BT_ZERO(args_speech);
        args_speech.ips = echo_ips;
        args_speech.coding_fmt = CODE_PCM_ULAW;
        args_speech.rate = RATE_8000;
        args_speech.bits_per_samp = BITS_8;
        args_speech.afe_rate = AFE_8000;
        args_speech.data_fmt = SPCH_MSB;
        args_speech.timeout = 0L;
        args_speech.vad = 1;
        args_speech.asr_mode = 1;
        BfvSpeechRecord (g_lpS, &args_speech);

        args_infopkt.ips = echo_ips;
        BfvInfopktClose (&args_infopkt);
        if (args_infopkt.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage (NULL, &args_infopkt.res, &bterr);
            fprintf (stderr, "BfvInfopktClose: %s\n",bterr.long_msg);
            fprintf (stdout, "Channel %d: Ignoring BfvInfopktClose() error.\n",
              g_iSecondUnit);
        }

        /* This record should never return a status of BT_STATUS_OK. */
        if (args_speech.res.status == BT_STATUS_ALERT)
        {
            fprintf (stdout, "Channel %d: Alerted!\n", g_iSecondUnit);
        }
        else
        {
            if (LINE_STATE(g_lpS) == LINE_STATE_IDLE)
            {
                fprintf (stdout, "Channel %d: Hang-up detected!\n", g_iSecondUnit);
            }
            else
            {
                BfvErrorMessage (g_lpS, &args_speech.res, &bterr);
                fprintf (stderr, "BfvSpeechRecord: %s\n", bterr.long_msg);
            }
            
            fprintf (stdout, "Channel %d: Alerting channel %d.\n", g_iSecondUnit,
              g_iFirstUnit);

            BT_ZERO(args_admin);
            args_admin.unit = g_iFirstUnit;
            BfvLineAlert (&args_admin);
        }
    }
    
    return;
}

void AlertThreadFunctionOfError (bt_thread tID)
{
    int iStatus;

    g_iErrorState = 1;

    /* Signal thread function to stop waiting due to error. */
    osiSemaphore_Increment (&g_sWaitForCall);
    
    osiThread_Wait (tID, &iStatus);

    return;
}

/* End of file. */
