/*
 * tstrip.c
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/**************************************************************************
The tstrip program writes individual G3 pages from a TIFF-F file.
The first page is written to g3data.301, the second to g3data.302, and
so on, until all of the pages are written. This utility converts received
fax data stored in a TIFF-F file to raw G3 format.
***************************************************************************/

#include <string.h>
#include <ctype.h>
#include "btlib.h"

FILE *outfp;
int current_resolution = 0;
int current_width = 0;
short num_lines = 0;
int dont_rev_flag = 0;
int need_to_reverse = 0;
int need_to_rev_rtc = 0;
int skip_rtc = 0;
int pagecnt = 1;
char output_base[512] = "g3data";

void usage()
{
    fprintf(stderr, "Usage: tstrip [-h] [-o <output_base>] [-r] <tiff_filename>\n");
    fprintf(stderr, "\t-h = add 128-byte \"Brooktrout\" header\n");
    fprintf(stderr, "\t-o = use supplied output filename base\n");
    fprintf(stderr, "\t-r = do *not* force output to Brooktrout-standard MSB format\n");
    exit(1);
}

int
main(int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    char *nami, namo[512];
    int header_flag = 0;
    int header_written = 0;
    TFILE *tp;
    int tstrip_ifd_func BT_ARGS((TFILE *tp,struct ifd_field *ifd_field,
                                 char *arg));
    int tmp;
    int n;
    unsigned char buf[1024];
    int write_header BT_ARGS((FILE *fp,int resolution,
                              int width,int num_lines));
    int tstrip_image_func BT_ARGS((unsigned char *buf,int n));
    struct args_tiff args;
    struct args_debug_mode args_debug;
    BTERR bterr;
    
    BT_API_SET_VER();
    BT_ZERO(args_debug);
    args_debug.mode = DEBUG_ALL;
    args_debug.options = API_DBG_MODE_SET;
    BfvDebugModeSetAdv(&args_debug);
    
    while ((c = getopt (argc, argv, "hro:")) != EOF)
    {
        switch (c)
        {
          case 'h':
            header_flag = 1;
            break;
          case 'r':
            dont_rev_flag = 1;
            break;
          case 'o':
            strcpy(output_base,optarg);
            break;
          default:
            usage ();
        }
    }
    
    nami = argv[optind++];
    
    if (optind != argc)
        usage ();
    
    BT_ZERO(args);
    args.fname = nami;
    args.fmode = "r";
    if ((tp = BfvTiffOpen(&args)) == NULL)
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr, "BfvTiffOpen: %s\n",bterr.long_msg);
        exit(1);
    }
    
    sprintf(namo, "%s.3%02d", output_base, pagecnt);
#ifndef USES_FAT_FILESYSTEM
    if ((outfp = fopen(namo, "w")) == NULL)
#else
    if ((outfp = fopen(namo, "wb")) == NULL)
#endif
    {
        printf("error opening output file\n");
        exit(1);
    }
    
    args.tp = tp;
    args.func = tstrip_ifd_func;
    while ((tmp = BfvTiffReadIFD (&args)) > 0)
    {
        if (pagecnt != 1)
        {
            fclose(outfp);
            sprintf(namo, "%s.3%02d", output_base, pagecnt);
#ifndef USES_FAT_FILESYSTEM
            if ((outfp = fopen(namo, "w")) == NULL)
#else
            if ((outfp = fopen(namo, "wb")) == NULL)
#endif
            {
                printf("error opening output file\n");
                exit(1);
            }
            header_written = 0;
        }
#if 0
        if (current_resolution == RES_300H_300V)
            current_width = (current_width / 3) * 2;
        else if (current_resolution == RES_400H_400V)
            current_width /= 2;
#endif
        if (header_flag && !header_written)
        {
            write_header(outfp,current_resolution,current_width,num_lines);
            header_written = 1;
        }
        pagecnt++;
        args.tp = tp;
        args.buf = buf;
        args.size = sizeof(buf);
        while ((n = BfvTiffReadImage(&args)) > 0 &&
               (tmp = tstrip_image_func(buf,n)) == 0);
        if (n < 0 || tmp)
            break;
        tstrip_image_func(NULL,0);
    }
        
    if (args.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr, "BfvTiffReadIFD/Image: %s\n",bterr.long_msg);
    }
    else if (tmp < 0)
        printf("Error occurred\n");

    args.tp = tp;
    BfvTiffClose(&args);
    
    fclose(outfp);
    
    return(0);
}

struct ifd_field xres_field;

int
tstrip_ifd_func(TFILE *tp, struct ifd_field *ifd_field, char *arg)
{
    struct args_tiff args;
    
    if (ifd_field == NULL)
        return(0);
    switch (ifd_field->tag)
    {
      case TAG_XRESOLUTION:
        memcpy(&xres_field,ifd_field,sizeof(*ifd_field));
        break;
      case TAG_YRESOLUTION:
        BT_ZERO(args);
        args.tp = tp;
        args.horiz_ifd_field = &xres_field;
        args.vert_ifd_field = ifd_field;
        if ((current_resolution = BfvTiffReadRes (&args)) < 0)
            return(1);
        break;
      case TAG_IMAGEWIDTH:
        current_width = (short)ifd_field->offset;
        break;
      case TAG_FILLORDER:
        need_to_reverse = dont_rev_flag == 0 && ifd_field->offset != 1;
        need_to_rev_rtc = dont_rev_flag == 1 && ifd_field->offset != 1;
        break;
      case TAG_IMAGELENGTH:
        num_lines = (short)ifd_field->offset;
        break;
      case TAG_COMPRESSION:
        /* MMR does not get RTC added on. */
        if (ifd_field->offset == 4)
            skip_rtc = 1;
        break;
    }
    return(0);
}

int
tstrip_image_func(unsigned char *buf, int n)
{
    static unsigned char RTC[] =
        {0x00,0x01,0x00,0x01,0x00,0x01,0x00,0x01,0x00,0x01,0x00,0x01};
    static unsigned char RTC_rev[] =
        {0x00,0x80,0x00,0x80,0x00,0x80,0x00,0x80,0x00,0x80,0x00,0x80};
    static int need_new_page = 0;
    int check_for_eofb BT_ARGS((unsigned char *buf,int num_bytes));
    
    if (n == 0)
    {
        if (skip_rtc)
            return(0);
        /* Must add RTC back on. */
        if (fwrite(((need_to_rev_rtc) ? RTC_rev : RTC),
                   sizeof(RTC),1,outfp) != 1)
            return(1);
        return(0);
    }
    if (need_to_reverse)
    {
        int i;
        unsigned char ch;
        
        for (i = 0; i < n; i++)
        {
            ch = buf[i];
            buf[i] = ((unsigned)(ch & 0x80) >> 7) |
                ((unsigned)(ch & 0x40) >> 5) |
                ((unsigned)(ch & 0x20) >> 3) |
                ((unsigned)(ch & 0x10) >> 1) |
                ((ch & 0x08) << 1) |
                ((ch & 0x04) << 3) |
                ((ch & 0x02) << 5) |
                ((ch & 0x01) << 7);
        }
    }
    
    /* If using MMR and an EOFB appears, no choice but to go to a new
       file for the next strip of the image. */
    if (need_new_page)
    {
        char namo[512];
        
        fclose(outfp);
        sprintf(namo, "%s.3%02d", output_base, pagecnt);
        pagecnt++;
#ifndef USES_FAT_FILESYSTEM
        if ((outfp = fopen(namo, "w")) == NULL)
#else
        if ((outfp = fopen(namo, "wb")) == NULL)
#endif
        {
            printf("error opening output file\n");
            exit(1);
        }
        need_new_page = 0;
    }
    
    if (fwrite(buf,n,1,outfp) != 1)
        return(1);

    /* If using MMR and an EOFB appears, no choice but to go to a new
       file for the next strip of the image. */
    if (skip_rtc)
        need_new_page = check_for_eofb(buf,n);

    return(0);
}


#define HEADER_SIZE 128
#define HEADER_HORIZ_RES_OFF 4
#define HEADER_VERT_RES_OFF 6
#define HEADER_NUM_LINES_OFF 45
#define HEADER_WIDTH_OFF 9

int
write_header(FILE *fp, int resolution, int width, int num_lines)
{
    static unsigned char header[HEADER_SIZE] = 
        { 0xbb, 1, 0, 1, 0xc8, 0, 0x64, 0, 1, 0xc0, 6, 1, 1};
    int horiz = 200, vert = 100;
    
    switch (resolution)
    {
      case RES_200H_100V:
        horiz = 200;
        vert = 100;
        break;
      case RES_200H_200V:
        horiz = 200;
        vert = 200;
        break;
      case RES_200H_400V:
        horiz = 200;
        vert = 400;
        break;
      case RES_300H_300V:
        horiz = 300;
        vert = 300;
        break;
      case RES_400H_400V:
        horiz = 400;
        vert = 400;
        break;
      case RES_600H_600V:
        horiz = 600;
        vert = 600;
        break;
      case RES_1200H_1200V:
        horiz = 1200;
        vert = 1200;
        break;
      case RES_300H_600V:
        horiz = 300;
        vert = 600;
        break;
      case RES_400H_800V:
        horiz = 400;
        vert = 800;
        break;
      case RES_600H_1200V:
        horiz = 600;
        vert = 1200;
        break;
      case RES_100H_100V:
        horiz = 100;
        vert = 100;
        break;
    }
    header[HEADER_HORIZ_RES_OFF] = horiz & 0xff;
    header[HEADER_HORIZ_RES_OFF + 1] = (horiz >> 8) & 0xff;
    header[HEADER_VERT_RES_OFF] = vert & 0xff;
    header[HEADER_VERT_RES_OFF + 1] = (vert >> 8) & 0xff;
    header[HEADER_WIDTH_OFF] = width & 0xff;
    header[HEADER_WIDTH_OFF + 1] = width >> 8;
    header[HEADER_NUM_LINES_OFF] = num_lines & 0xff;
    header[HEADER_NUM_LINES_OFF + 1] = num_lines >> 8;
    
    fwrite(header,1,sizeof(header),fp);
    return(0);
}

/* If format is MMR, and multiple strips are being used,
   an end of strip and EOFB may occur in this buffer.
   If so, we must start a new strip before sending subsequent data.
   A strip end will always be aligned with the end of data
   in a buffer, but may be followed by bytes of 0's and the EOFB
   may span calls to ReadImage. */
int
check_for_eofb(unsigned char *buf, int num_bytes)
{
    static unsigned char saved_from_last_buf[8];
    static int bytes_saved = 0;
    unsigned char *buf_start;
    int msb_byte;
    unsigned last_4_bytes = 0;
    int tmp_num_bytes = num_bytes;
    int i;
    int ret;
    
    /* Skip 0 bytes at the end. Extra 0 bytes are permitted
       following EOFB */
    while (tmp_num_bytes > 0 && buf[tmp_num_bytes - 1] == 0)
        tmp_num_bytes--;

    /* If number of bytes is < 4, then we may need to use
       saved bytes from the previous buffer. */
    if (tmp_num_bytes < 4)
    {
        buf_start = saved_from_last_buf;
        memcpy(&saved_from_last_buf[bytes_saved],buf,tmp_num_bytes);
        tmp_num_bytes += bytes_saved;
    }
    else
        buf_start = buf;

    /* Turn last 4 bytes into an integer, using MSB form,
       with the last byte the least significant. */
    for (i = 0; i < 4; i++)
    {
        msb_byte = buf_start[tmp_num_bytes - 1 - i];
        if (need_to_rev_rtc)
            msb_byte =
                ((msb_byte & 0x1) << 7) |
                ((msb_byte & 0x2) << 5) |
                ((msb_byte & 0x4) << 3) |
                ((msb_byte & 0x8) << 1) |
                ((msb_byte & 0x10) >> 1) |
                ((msb_byte & 0x20) >> 3) |
                ((msb_byte & 0x40) >> 5) |
                ((msb_byte & 0x80) >> 7);
        last_4_bytes |= msb_byte << (i * 8);
    }
    
    /* The EOFB is 24 bits long, but can be aligned at 8 different
       bit positions. */
    if ((last_4_bytes & 0x00ffffff) == 0x00001001 ||
        ((last_4_bytes >> 1) & 0x00ffffff) == 0x00001001 ||
        ((last_4_bytes >> 2) & 0x00ffffff) == 0x00001001 ||
        ((last_4_bytes >> 3) & 0x00ffffff) == 0x00001001 ||
        ((last_4_bytes >> 4) & 0x00ffffff) == 0x00001001 ||
        ((last_4_bytes >> 5) & 0x00ffffff) == 0x00001001 ||
        ((last_4_bytes >> 6) & 0x00ffffff) == 0x00001001 ||
        ((last_4_bytes >> 7) & 0x00ffffff) == 0x00001001)
    {
        ret = 1;
        bytes_saved = 0;
    }
    else
    {       
        ret = 0;
        if (num_bytes < 3)
            bytes_saved = num_bytes;
        else
            bytes_saved = 3;
        memcpy(saved_from_last_buf,&buf[num_bytes - bytes_saved],bytes_saved);
    }
    return(ret);
}
