/* boardmon.c */
/* Copyright (c) 1989-2005 Brooktrout Technology Inc.  All rights reserved. */

/**************************************************************************
The boardmon program monitors the condition of a board. It displays
the board temperature, the status of the Ethernet port and monitors
one or more telephony spans on the board. It reports the state of the
signaling bits and alarms. It also counts the errors (e.g., framing
errors, CRC errors, slips, BPVs) on each span. It only works with
spans that are configured for robbed-bit signaling.
***************************************************************************/

#include <time.h>
#include "btlib.h"
#include "osilib.h"   /* Operating System Independent */


#define MAX_SPANS       4   /* number of spans on a module */
#define MAX_SLOTS       32  /* number of timeslots on a span */
#define MAX_CHAN (MAX_SLOTS * MAX_SPANS)  /* maximum channels to monitor */
#define MAX_ERRORS      4   /* types of errors reported */
#define ERR_FRM         0   /* index into error types */
#define ERR_CRC         1
#define ERR_BPV         2
#define ERR_SLIP        3

#define MAX_BITS        4  /* number of bits per timeslot */
#define BIT_A           0  /* index into MAX_BITS */
#define BIT_B           1
#define BIT_C           2
#define BIT_D           3

#define BIT_OFF         '.'  /* characters for bit display */
#define BIT_A_ON_IN     'a'
#define BIT_B_ON_IN     'b'
#define BIT_C_ON_IN     'c'
#define BIT_D_ON_IN     'd'
#define BIT_A_ON_OUT    'A'
#define BIT_B_ON_OUT    'B'
#define BIT_C_ON_OUT    'C'
#define BIT_D_ON_OUT    'D'

int debug = 0;            /* program debug mode */
int api_debug = 0;        /* Bfv API debug mode */
int display_mode = 0;     /* enable display */
int user_module = 0x2;    /* default module number */
int user_span = 0;        /* span number, one-based, 0 = all */
unsigned span_type[MAX_SPANS];   /* T1 or E1, etc */
int timeslots[MAX_SPANS][MAX_SLOTS];    /* timeslot is valid */
int any_spans = 0;        /* non-zero if a timeslot exists on any span */
int any_timeslot[MAX_SPANS];    /* non-zero if a timeslot exists on a span */

int network_alarm[MAX_SPANS];   /* current alarm for the span */
unsigned network_error[MAX_SPANS][MAX_ERRORS];  /* counters for each type
                                                   of error on each span */

int link_status;         /* status of Ethernet link */
int link_failover_unit;  /* failover unit for the link */

double temperatureC;      /* board temperature */
double temperatureF;
char time_str[MAX_SPANS][50];  /* display reset timestamp */
int need_display = 0;     /* set to 1 if a display was skipped
                             due to time considerations */

/* characters representing the state of the bits to be displayed
   for each span */
unsigned char bits_in[MAX_SPANS][MAX_BITS][MAX_SLOTS];
unsigned char bits_out[MAX_SPANS][MAX_BITS][MAX_SLOTS];


/* function prototypes */
void sigmonFunc(BTLINE *lp, struct args_signaling *asig);
void notifyFunc(BTLINE *lp, unsigned notify_type,
                struct args_board_notify *args);
void usage(void);
int main(int argc, char **argv);
void dowork(BTLINE *lp);
void setup_notify(BTLINE *lp, int mode);
void display(char *caller);
char *alarm_text(int code);
char *link_text(void);
void reset_error_counts (int span);
int read_temperature(BTLINE *lp);
unsigned long millitime(void);
void console_clear(void);
int key_test(int *key);


/* Signalling bits monitor function */
void
sigmonFunc(BTLINE *lp, struct args_signaling *asig)
{
    int span, slot;
   
    if (debug)
        printf("\n--- Sigmon: span %d  slot %d  in %01X  out %01X\n",
               asig->unit,
               asig->time_slot + 1,
               0x0f & asig->stateIn_1,
               0x0f & asig->stateOut_0);
    span = asig->unit - 1;  /* span number, zero based */
    slot = asig->time_slot;  /* timeslot within the span, zero based */
   
    /* Store the bit values for display */
    bits_in[span][BIT_D][slot] =
        (asig->stateIn_1 & 0x01) ? BIT_D_ON_IN : BIT_OFF;
    bits_in[span][BIT_C][slot] =
        (asig->stateIn_1 & 0x02) ? BIT_C_ON_IN : BIT_OFF;
    bits_in[span][BIT_B][slot] =
        (asig->stateIn_1 & 0x04) ? BIT_B_ON_IN : BIT_OFF;
    bits_in[span][BIT_A][slot] =
        (asig->stateIn_1 & 0x08) ? BIT_A_ON_IN : BIT_OFF;
   
    bits_out[span][BIT_D][slot] =
        (asig->stateOut_0 & 0x01) ? BIT_D_ON_OUT : BIT_OFF;
    bits_out[span][BIT_C][slot] =
        (asig->stateOut_0 & 0x02) ? BIT_C_ON_OUT : BIT_OFF;
    bits_out[span][BIT_B][slot] =
        (asig->stateOut_0 & 0x04) ? BIT_B_ON_OUT : BIT_OFF;
    bits_out[span][BIT_A][slot] =
        (asig->stateOut_0 & 0x08) ? BIT_A_ON_OUT : BIT_OFF;
   
    display("signal");
}


/* Board notification function */
void
notifyFunc(BTLINE *lp, unsigned notify_type, struct args_board_notify *args)
{
    if (debug)
        printf("\n--- Notify: span %d,  type %X  ", args->unit, notify_type);

    if (notify_type == ADMIN_NOTIFY_ENET_LINK_DEF)
    {
        if (debug)
            printf("Link: status %X\n", args->link_status);

        link_status = args->link_status;
        link_failover_unit = args->failover_unit;
        display("notify 6");
    }
   
    /* ignore notifications for telephony spans that are not in service */
    if (any_timeslot[args->unit - 1] == 0)
    {
        if (debug)
            printf("ignored\n");
        return;
    }
   
    /* ignore notifications for telephony spans that we are not monitoring */
    if ((user_span != 0) &&  /* one span only */
        (user_span != (int) args->unit))  /* and not ours */
    {
        if (debug)
            printf("ignored\n");
        return;
    }
   
    if (notify_type == ADMIN_NOTIFY_NETWORK_ALARM_DEF)
    {
        if (debug)
            printf("Alarm: type %X\n", args->network_alarm);

        network_alarm[args->unit-1] = args->network_alarm;
        display("notify 1");
    }
   
    if (notify_type == ADMIN_NOTIFY_NETWORK_ERROR_DEF)
    {
        if (debug)
            printf("Error: type %X\n", args->network_error);

        /* count each type of error */
        switch (args->network_error)
        {
          case ADMIN_UNIT_ERROR_FRAMING_DEF:
            network_error[args->unit-1][ERR_FRM]++;
            display("notify 2");
            break;
           
          case ADMIN_UNIT_ERROR_CRC_DEF:
            network_error[args->unit-1][ERR_CRC]++;
            display("notify 3");
            break;
           
          case ADMIN_UNIT_ERROR_BPV_DEF:
            network_error[args->unit-1][ERR_BPV]++;
            display("notify 4");
            break;
           
          case ADMIN_UNIT_ERROR_SLIP_DEF:
            network_error[args->unit-1][ERR_SLIP]++;
            display("notify 5");
            break;
           
          case ADMIN_UNIT_ERROR_NONE_DEF:
            /* no need to display when error goes away */
          default:
            break;
        }  /* end switch */
    }  /* end errors */
}

void usage(void)
{
    fprintf (stderr, "usage: boardmon [-m mod] [-s span] [-d] [-v] [-h]\n");
    fprintf (stderr, "  -m <mod>    Use specified module (default 2)\n");
    fprintf (stderr, "  -s <span>   Use specified span only [1-%i] (default all)\n",
             MAX_SPANS);
    fprintf (stderr, "  -d          Enable program debug mode\n");
    fprintf (stderr, "  -v          Enable API debug mode\n");
    fprintf (stderr, "  -h          Help\n");
    fprintf (stderr, "\nKey presses while running:\n");
    fprintf (stderr, "  1           Reset error counters for span 1\n");
    fprintf (stderr, "  2           Reset error counters for span 2\n");
    fprintf (stderr, "  3           Reset error counters for span 3\n");
    fprintf (stderr, "  4           Reset error counters for span 4\n");
    fprintf (stderr, "  r or R      Reset error counters for all spans\n");
    fprintf (stderr, "  q or Q      Quit the program\n");
    exit(1);
}

int main(int argc, char **argv)
{
    int c, span, slot, bit;
    extern int optind;
    extern char *optarg;
    BTLINE *lp;
    BTERR bterr;
    struct args_line_admin args;
    struct args_addr_info args_info;
    struct args_tel_ctrl_call_sw args_tel;
    int channels;
   
    BT_API_SET_VER();
   
    while ((c = getopt (argc, argv, "dm:s:hv")) != EOF)
    {
        switch (c)
        {
          case 'd':
            debug = 1;  /* program debug on */
            break;
          case 'm':
            if (sscanf(optarg, "%x", &user_module) != 1)
                usage();
            break;
          case 's':
            if (sscanf(optarg, "%i", &user_span) != 1)
                usage();

            if (user_span < 1 || user_span > MAX_SPANS)
            {
                fprintf(stderr, "Span number %d outside valid range [1-%d], exiting...\n", 
                    user_span, MAX_SPANS);
                exit(1);
            }
            break;
          case 'v':
            api_debug = 1;
            BfvDebugModeSet(DEBUG_ALL);  /* API debug on */
            break;
          case 'h':
          default:
            usage ();
        }
    }
   
    if (optind != argc)
        usage ();

    /* Make sure the selected module is valid */
    BT_ZERO(args_info);
    args_info.mod_num = user_module;
    BfvModuleInfo(NULL, &args_info);
    if (args_info.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(NULL, &args_info.res,&bterr);
        fprintf(stderr, "BfvModuleInfo: %s\n",bterr.long_msg);
        exit(1);
    }
   
    if (!args_info.exists)
    {
        fprintf(stderr,"Specified module doesn't exist.\n");
        exit(1);
    }
   
    if (!args_info.hardware)
    {
        fprintf(stderr, "Specified module is not a hardware module.\n");
        exit(1);
    }
   
    channels = args_info.num_channels - 1;
   
    if (channels == 0)
    {
        printf("\nNo channels found, exiting...");
        exit(1);
    }
   
    /* initialize arrays */
    for (span = 0; span < MAX_SPANS; span++)
    {
        network_alarm[span] = ADMIN_UNIT_ALARM_NONE_DEF;
        reset_error_counts(span);
        any_timeslot[span] = 0;
        span_type [span] = CALL_SW_PORT_CHANNEL_DEF;
       
        for (slot = 0; slot < MAX_SLOTS; slot++)
        {
            timeslots[span][slot] = 0;

            for (bit = 0; bit < MAX_BITS; bit++)
            {
                bits_in [span][bit][slot] = BIT_OFF;
                bits_out[span][bit][slot] = BIT_OFF;
            }
        }
    }
   
   
    /* Get a line pointer to the module */
    BT_ZERO(args);
    args.dest_addr.mm_bModule = user_module;
    args.dest_addr.mm_bChannel = 1;
    args.present = 1;
    args.unique = 0;
    args.reset_on_close = 0;
    if ((lp = BfvSessionAttach(&args)) == NULL)
    {
        BfvErrorMessage(lp,&args.res,&bterr);
        fprintf(stderr, "BfvSessionAttach: %s\n",bterr.long_msg);
        exit (1);
    }
   
    /* Reset the line */
    BT_ZERO(args);
    args.config_file_name = "btcall.cfg";
    args.skip_teleph = 1;    
    args.no_init = 1;
    if (BfvLineReset(lp, &args)  < 0)
    {
        BfvErrorMessage(lp, &args.res,&bterr);
        fprintf(stderr, "BfvLineReset: %s: status %lX.\n", 
                bterr.long_msg, args.reset_status);
        exit(1);
    }
    
    /* Create a map of all the valid timeslots for this module */
    BT_ZERO(args_tel);
    while (BfvCallSWGetInfo(lp, &args_tel) == 1)
    {
        /* Connection is a telephony unit */
        if ((args_tel.src_port_class == CALL_SW_PORT_NETWORK_DEF) &&
            ((args_tel.port_type == CALL_SW_PORT_T1_DEF) ||
             (args_tel.port_type == CALL_SW_PORT_E1_DEF)))
        {
            for (slot = 0; slot < (int) args_tel.src_slot; ++slot)
            {
                timeslots[args_tel.src_port_unit][slot] = 1;
            }
            any_timeslot[args_tel.src_port_unit] = 1;
            any_spans = 1;
        }
    }
   
    /* See if any spans are configured */
    if (any_spans == 0)
    {
        fprintf(stderr, "Could not find any configured spans, exiting...\n");
        exit(1);
    }   
   
    if (user_span != 0)  /* only check if a particular span was specified */
        if (any_timeslot[user_span-1] == 0)
        {
            fprintf(stderr,
                    "Could not find any configured timeslots on span %i, exiting...\n",
                    user_span);
            exit(1);
        }
       
    /* check the port_type */
    BT_ZERO(args_tel);
    while (BfvCallSWGetInfo(lp, &args_tel) > 0)
    {
        if (debug)
            printf("Port class %X, Unit %d, Stream %d, Slot %d, Type %X\n",
                   args_tel.src_port_class,
                   args_tel.src_port_unit,
                   args_tel.src_stream,
                   args_tel.src_slot,
                   args_tel.port_type);

        if ((args_tel.src_port_class == CALL_SW_PORT_NETWORK_DEF) && 
            (args_tel.src_port_unit >= 0) &&
            (args_tel.src_port_unit < MAX_SPANS))
            span_type[args_tel.src_port_unit] = args_tel.port_type;
    }
       
    dowork(lp);
       
    BT_ZERO(args);
    BfvSessionDetach(lp, &args);
       
    return(0);
}

void
dowork(BTLINE *lp)
{
    int span, key, run_flag = 1;
    static unsigned long last_time = 0L;
    unsigned long now;
    BTERR bterr;
    struct args_packet apkt;
    
    
    read_temperature(lp);  /* get initial reading */
    
    setup_notify(lp, 1);  /* enable callback functions */
    
    display_mode = 1;  /* enable display */
    
    display("initial");  /* Do initial display */
    
    
    /* Wait for a packet from the board and process it */
    /* until ^C is hit.  Check for key hit to reset counters. */
    while (run_flag)
    {
        if (key_test(&key))
        {
            switch (key)
            {
              case 'q':
              case 'Q':
                run_flag = 0;
                break;
                
                /* reset individual span */
              case '1':
                reset_error_counts(1);
                display("reset");
                break;
              case '2':
                reset_error_counts(2);
                display("reset");
                break;
              case '3':
                reset_error_counts(3);
                display("reset");
                break;
              case '4':
                reset_error_counts(4);
                display("reset");
                break;
              case 'r':
              case 'R':
                for (span = 0; span < MAX_SPANS; span++)
                    reset_error_counts(span + 1);  /* reset all spans */
                display("reset");
                break;
              default:
                ;
            }
        }  /* end keyboard hit */
        
        
        /* read the board temperature every 10 seconds */
        now = millitime();
        
        if ((now - last_time) >= 10000)
        {
            last_time = now;
            if (!read_temperature(lp))
                display("temperature");
        }
        
        
        /* Wait for a Bfv packet, with timeout */
        BT_ZERO(apkt);
        apkt.timeout = 300; 
        apkt.no_hangup = 1;
        BfvRcvProcessPkt(lp, &apkt);
        if (apkt.res.status == BT_STATUS_TIMEOUT)
        {
            if (need_display)  /* see comment at display() */
                display("timeout");
            continue;
        }
        else if (apkt.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&apkt.res,&bterr);
            printf ("BfvRcvProcessPkt: %s\n", bterr.long_msg);
            exit(1);
        }
        
    }  /* end while (run_flag) */
    
    display_mode = 0;  /* disable display */
    
    setup_notify(lp, 0);  /* disable callback functions */
}


/* setup or tear down signal bit, alarm and error notification */
/* mode = 0 for tear down */
/* mode = 1 for setup */
void
setup_notify(BTLINE *lp, int mode)
{
    int span, slot, start_span, end_span;
    struct args_signaling asig;
    struct args_board_notify anotify;
    BTERR bterr;
    
    /* Check to see if the user entered a specific span on the command line */
    /* User specifies span numbers starting at 1 */
    if (user_span == 0)  /* all spans */
    {
        start_span = 0;
        end_span = MAX_SPANS;
    }
    else  /* specific span */
    {
        start_span = (user_span-1);  /* zero based */
        end_span = user_span;
    }
    
    /* Start monitoring on all valid spans */
    for (span = start_span; span < end_span; span++)
    {
        /* only watch spans with active timeslots */
        if (any_timeslot[span] == 0)
            continue;

        for (slot = 0; slot < MAX_SLOTS; slot++)
        {
            if (timeslots[span][slot] == 1)
            {
                if (debug)
                {
                    if (mode)
                        printf("=== watching span %d  timeslot %d\n",
                               span + 1, slot + 1);
                    else
                        printf("=== not watching span %d  timeslot %d\n",
                               span + 1, slot + 1);
                }
                
                BT_ZERO(asig);
                asig.unit = span + 1;  
                asig.stream = 0;    
                asig.time_slot = slot;
                if (mode)
                    asig.func = sigmonFunc;
                else
                    asig.func = NULL;
                asig.mon_mode = BT_MON_MODE_ENABLE;
                BfvCallSignalingStateMonitor(lp, &asig);
                if (asig.res.status != BT_STATUS_OK)
                {
                    BfvErrorMessage(lp, &asig.res, &bterr);
                    fprintf(stderr, "BfvCallSignalingStateMonitor: %s\n",
                            bterr.long_msg);
                    exit(1);
                }
            }
        }
    }
    
    /* Set up notification of alarm events */
    BT_ZERO(anotify);
    anotify.enable = 1;
    anotify.notify_type = ADMIN_NOTIFY_NETWORK_ALARM_DEF;
    if (mode)
        anotify.func = notifyFunc;
    else
        anotify.func = NULL;
    BfvBoardNotify(lp, &anotify);
    if (anotify.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp, &anotify.res, &bterr);
        fprintf(stderr, "BfvBoardNotify (alarms): %s\n", bterr.long_msg);
        exit(1);
    }
    
    /* Set up notification of error events */
    BT_ZERO(anotify);
    anotify.enable = 1;
    anotify.notify_type = ADMIN_NOTIFY_NETWORK_ERROR_DEF;
    if (mode)
        anotify.func = notifyFunc;
    else
        anotify.func = NULL;
    BfvBoardNotify(lp, &anotify);
    if (anotify.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp, &anotify.res, &bterr);
        fprintf(stderr, "BfvBoardNotify (errors): %s\n", bterr.long_msg);
        exit(1);
    }
    
    /* Set up notification of Ethernet link status events */
    BT_ZERO(anotify);
    anotify.enable = 1;
    anotify.notify_type = ADMIN_NOTIFY_ENET_LINK_DEF;
    if (mode)
        anotify.func = notifyFunc;
    else
        anotify.func = NULL;
    BfvBoardNotify(lp, &anotify);
    if (anotify.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp, &anotify.res, &bterr);
        fprintf(stderr, "BfvBoardNotify (Enet links): %s\n", bterr.long_msg);
        exit(1);
    }
}


void
display(char *caller)
{
    int max_slots;  /* for the display */
    int span, bit, slot, n, start_span, end_span;
    static unsigned long last_time = 0L;
    unsigned long now;
    
    if (!display_mode)  /* skip display while initializing */
        return;


    /* only update the display if at least 2 seconds have elapsed */
    now = millitime();
    if ((now - last_time) < 2000)
    {
        /* The need_display flag is used to indicate the need to update the
           display one more time after the display has been supressed due
           to timing considerations.  This can happen if there is a flurry of
           of messages, the display is skipped, and then the messages stop.
           This creates a condition where there is some more recent data
           that has not yet been displayed. */
        need_display = 1;  /* display was skipped */
        return;
    }
    
    if (!debug && !api_debug)
        console_clear();

    last_time = now;
    
    if (debug)
        printf("  display(%s):\n", caller);

    printf("Board Temperature: %3.1fF  %3.1fC\n",
           temperatureF,
           temperatureC);
    
    /* display Ethernet span status */
    printf("Ethernet link 1: %s\n", link_text());
    
    /* Check to see if the user entered a specific span on the command line */
    /* User specifies span numbers starting at 1 */
    if (user_span == 0)  /* all spans */
    {
        start_span = 0;
        end_span = MAX_SPANS;
    }
    else  /* specific span */
    {
        start_span = (user_span-1);  /* zero based */
        end_span = user_span;
    }
    
    /* Index by spans starting at 0 */
    for (span = start_span; span < end_span; span++)  /* span */
    {
        /* Make sure there are valid timeslots on this span */
        if (any_timeslot[span] == 0)
            continue;

        /* Print out the header */
        printf("Mod 0x%02x Span %i  %s  FRM %06u  CRC %06u  BPV %06u  SLIP %06u\n", 
               user_module, 
               span + 1,
               alarm_text(network_alarm[span]),
               network_error[span][ERR_FRM],
               network_error[span][ERR_CRC],
               network_error[span][ERR_BPV],
               network_error[span][ERR_SLIP]);
        
        
        /* test for T1 or E1, skip 25, 29 if T1 */
        if (span_type[span] == CALL_SW_PORT_E1_DEF)
        {
            printf(" Ch: 1    5    9   13   17   21   25   29   Last reset: %s    \n",
                   time_str[span]);
            max_slots = 30;
        }
        else
        {
            printf(" Ch: 1    5    9   13   17   21             Last reset: %s    \n",
                   time_str[span]);
            max_slots = 24;
        }
        
        
        /* print input bits */
        printf(" In:");
        n = 0;
        
        for (bit = 0; bit < MAX_BITS; bit++)  /* Bits A, B, C, D */
        {
            /* timeslot 0-23 or 0-29 */
            for (slot = 0; slot < max_slots; slot++)
            {
                if ((n++ % 4) == 0)
                    printf(" ");

                printf("%c", bits_in[span][bit][slot]);
            }
            printf("\n    ");
            n = 0;
        }
        
        
        /* print output bits */
        printf("\nOut:");
        n = 0;
        
        for (bit = 0; bit < MAX_BITS; bit++)  /* Bits A, B, C, D */
        {
            /* timeslot 0-29 or 0-23 */
            for (slot = 0; slot < max_slots; slot++)
            {
                if ((n++ % 4) == 0)
                    printf(" ");

                printf("%c", bits_out[span][bit][slot]);
            }
            printf("\n    ");
            n = 0;
        }
        printf("\n");
    }  /* end if (span) */
    
    need_display = 0;  /* display was updated */
}


char *
alarm_text(int code)
{
    static char alm_txt[50];
    
    switch(code)
    {
      case ADMIN_UNIT_ALARM_NONE_DEF:
        return("No Alarms");
        break;
        
      case ADMIN_UNIT_ALARM_LOS_DEF:
        return("LOS      ");
        break;
        
      case ADMIN_UNIT_ALARM_RAI_DEF:
        return("RAI      ");
        break;
        
      case ADMIN_UNIT_ALARM_AIS_DEF:
        return("AIS      ");
        break;
        
      default:
        sprintf(alm_txt, "AL 0x%04X", code);  /* unknown alarm type */
        return(alm_txt);
        break;
    }
}


char *
link_text(void)
{
    static char lnk_txt[50];
    
    switch(link_status)
    {
      case ADMIN_ENET_LINK_UP_DEF:
        return("UP                   ");
        break;
        
      case ADMIN_ENET_LINK_DOWN_DEF:
        return("DOWN                 ");
        break;
        
      case ADMIN_ENET_LINK_FAILOVER_DEF:
      case ADMIN_ENET_LINK_FAILOVER_TRUNK_DEF:
        sprintf(lnk_txt, "FAILOVER TO %d", link_failover_unit);
        return(lnk_txt);
        break;
        
      case ADMIN_ENET_LINK_FAILOVER_UNSUCC_DEF:
        return("FAILOVER UNSUCCESSFUL");
        break;
        
      default:
        /* unknown link status type */
        sprintf(lnk_txt, "status 0x%04X", link_status);
        return(lnk_txt);
        break;
    }
}


/* reset the error counts for a span (one based) */
void
reset_error_counts (int span)
{
    int err, prt;
    time_t now;
    struct tm *tm;
    
    prt = span - 1;  /* zero based */
    
    for (err = 0; err < MAX_ERRORS; err++)
        network_error[prt][err] = 0;
    
    /* update reset timestamp */
    now = time(NULL);
    tm = localtime (&now);
    strftime (time_str[prt], sizeof(time_str[prt]), "%m/%d %H:%M:%S", tm);
}


int
read_temperature(BTLINE *lp)
{
    struct args_temperature args;
    BTERR bterr;
    
    BT_ZERO(args);
    BfvBoardTemperatureGet(lp, &args);
    
    if (args.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(NULL, &args.res,&bterr);
        fprintf(stderr, "BfvBoardTemperatureGet: %s\n",bterr.long_msg);
        return(1);
    }
    else
    {
        temperatureC = ((double) args.current_temp) / 2.0;
        temperatureF = (temperatureC * 1.8) + 32;
        return(0);
    }
}


unsigned long
millitime(void)
{
    time_x tb;
    
    osiGet_Time(&tb);
    return (tb.time * 1000 + tb.millitm);
}


void
console_clear(void)
{
#ifdef  WINNT
    system("cls");
#else
    system("clear");
#endif
}


int
key_test(int *key)
{
    char inkey;
    
    if (kbhit() == 0)
        return (0);    /* no key */

    inkey = osiWait_For_Keyboard_Entry();
    if (key != NULL)
        *key = (int) inkey;

    return (1);        /* have a key */
}
