/* btver.c */
/* Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved. */

/**************************************************************************
The btver program displays version information for the driver, Bfv API,
boot ROM firmware, control processor firmware, and DSP firmware.
***************************************************************************/

#include <stdio.h>
#include <string.h>

#include "btlib.h"
#include "millver.h"

void ReportVersionInfo(void);
int GetNumberOfBostonBoards(int mod_num);
void ReportBoardInfo(void);
void GetFirmwareInfo(int mod);
BTLINE *AttachSession(unsigned mod_num);
void DetachSession(BTLINE *lp);


/* usage */
void
usage ()
{
    fprintf (stderr,"usage: btver [-m mod] [-v]\n" 
             "      -m mod  displaying the btver information for the specified module.\n"
             "      -v      verbose mode.\n\n");
    fprintf (stderr,"      NOTE: It is required that the firmware be loaded on the board in\n");
    fprintf (stderr,"            order to obtain ctrl processor and DSP information.\n");
    exit(1);
}


/* global vars */
int g_nNumberOfBoards;
unsigned nModID[16];
int driver_not_running;
int display_auto;


/* main */
int
main(int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    int display_module = -1;
    
    BT_API_SET_VER();
    
    while ((c = getopt (argc, argv, "m:va")) != EOF)
    {
        switch (c)
        {
          case 'm':
            if (sscanf(optarg,"%x",&display_module) != 1)
                usage();
            break;
          case 'v':
            BfvDebugModeSet(DEBUG_ALL);
            break;
          case 'a':
            display_auto = 1;
            break;
          default:
            usage ();
        }
    }
    
    ReportVersionInfo();
    g_nNumberOfBoards = GetNumberOfBostonBoards(display_module);
    ReportBoardInfo();
    
    return(0);
}

/* Search system for the number of boards installed */
int
GetNumberOfBostonBoards(int mod_num)
{
    BTERR bterr;
    struct args_addr_info args_info;
    int bos_cnt = 0;
    unsigned mod;
    
    for (mod = 0; mod <= M_ADDR_MAX_COMPONENT_VALUE; mod++)
    {
        BT_ZERO(args_info);
        /* if user selected module */
        if (mod_num >= 0)
        {
            mod = M_ADDR_COMPONENT_UPPER_LIM;
            args_info.mod_num = mod_num;
        }
        else
            args_info.mod_num = mod;
        BfvModuleInfo(NULL,&args_info);
        if (args_info.res.status != BT_STATUS_OK)
        {
            if (args_info.res.status == BT_STATUS_ERROR &&
                args_info.res.line_status == APIERR_DRV_OPEN_ERROR)
            {
                driver_not_running = 1;
                return(0);
            }
            
            BfvErrorMessage(NULL,&args_info.res,&bterr);
            fprintf (stderr,"%s\n",bterr.long_msg);
            exit(1);
        }
        
        if (args_info.exists)
        {
            if (args_info.hardware)
                nModID[bos_cnt++] = args_info.mod_num;
            else if (mod_num >= 0)
            {
                fprintf(stderr,
                        "Specified module (%X) is not a hardware module.\n",
                        mod_num);
                exit(1);
            }
        }
        else if (mod_num >= 0)
        {
            fprintf(stderr,"Specified module (%X) doesn't exist.\n",
                    mod_num);
            exit(1);
        }
    }
    return(bos_cnt);
}


/* Report version info */
void
ReportVersionInfo(void)
{
    unsigned a_ver = LINE_API_VER_LOADED();
    unsigned d_ver = LINE_DRIVER_VER_LOADED(NULL);
    unsigned s_ver = LINE_SDK_VER_LOADED();
    char *tmp, buf[20];

    printf ("- Boston Bfv API v%u.%u.%02u (Build %u), compiled with v%s (Build %u)\n",
            (a_ver % 100000) / 1000,
            (a_ver % 1000) / 100,
            a_ver % 100,
            a_ver / 100000,
            API_VERSION,MILL_API_BUILD_NUM);

    printf ("- Boston Driver  v%u.%u.%02u (Build %u)\n",
            (d_ver / 10000000),
            (d_ver / 1000000) % 10,
            (d_ver / 10000) % 100,
            (d_ver % 10000));

    if (display_auto)
    {
        sprintf(buf,", A%u",LINE_SDK_AUTO_LOADED());
        tmp = buf;
    }
    else
        tmp = "";
    printf ("- Boston SDK     v%u.%u.%02u (Build %u%s), ",
            (s_ver / 10000000) % 100,
            (s_ver / 100000) % 100,
            (s_ver / 1000) % 100,
            s_ver % 1000,
            tmp);
    if (display_auto)
    {
        sprintf(buf,", A%u",SDK_AUTO_NUMBER);
        tmp = buf;
    }
    else
        tmp = "";
    printf("compiled with v%u.%u.%02u (Build %u%s)\n",
           SDK_MAJOR_VERSION,
           SDK_MIDDLE_VERSION,
           SDK_MINOR_VERSION,
           SDK_BUILD_NUMBER,
           tmp);
}

/* Report Board Info */
void
ReportBoardInfo(void)
{
    int c; 
    
    if (driver_not_running)
        printf ("\n***** Driver not running *****\n");

    if (g_nNumberOfBoards < 1)
        printf ("\n***** No boards were found in the system *****\n");

    for (c = 0; c < g_nNumberOfBoards; c++)
        GetFirmwareInfo(c);

    printf ("\n******************************************************************************\n");
}

/* Get firmware info loaded on board */
void
GetFirmwareInfo(int mod)
{
    BTLINE *lp;
    int n;
    char *tmp;
    char tmp_buf[100];
    
    printf ("\n\t-------< Firmware Info for module 0x%02X >--------\n",
            nModID[mod]);
    printf ("FirmType     Version%s     Date                Comment\n",
            ((display_auto) ? "      " : ""));
    printf ("==============================================================================\n");
    
    if ((lp = AttachSession(nModID[mod])) != NULL)
    {
        /* report flash firmware loaded */
        printf(" Boot Rom:   ");
        if (LINE_FIRM_BOOT_ROM_MAJOR(lp))
        {
            sprintf(tmp_buf,"%u.%u.%uB%u",
                    LINE_FIRM_BOOT_ROM_MAJOR(lp),
                    LINE_FIRM_BOOT_ROM_MIDDLE(lp),
                    LINE_FIRM_BOOT_ROM_MINOR(lp),
                    LINE_FIRM_BOOT_ROM_BUILD(lp));
            if (display_auto)
            {
                tmp = tmp_buf + strlen(tmp_buf);
                sprintf(tmp,"A%u",LINE_FIRM_BOOT_ROM_AUTO_NUM(lp));
                printf("%-18s",tmp_buf);
            }
            else
                printf("%-12s",tmp_buf);
            printf("%-11s  %s\n",
                   LINE_FIRM_BOOT_ROM_DATE(lp),
                   LINE_FIRM_BOOT_ROM_COMMENT(lp));
        }
        else
            printf("Not available\n");

        /* report control processor firmware loaded */
        printf("Ctrl Proc:   ");
        if (LINE_FIRM_CTRL_PROC_MAJOR(lp))
        {
            sprintf(tmp_buf,"%u.%u.%uB%u",
                    LINE_FIRM_CTRL_PROC_MAJOR(lp),
                    LINE_FIRM_CTRL_PROC_MIDDLE(lp),
                    LINE_FIRM_CTRL_PROC_MINOR(lp),
                    LINE_FIRM_CTRL_PROC_BUILD(lp));
            if (display_auto)
            {
                tmp = tmp_buf + strlen(tmp_buf);
                sprintf(tmp,"A%u",LINE_FIRM_CTRL_PROC_AUTO_NUM(lp));
                printf("%-18s",tmp_buf);
            }
            else
                printf("%-12s",tmp_buf);
            printf ("%-11s  %s\n",
                    LINE_FIRM_CTRL_PROC_DATE(lp),
                    LINE_FIRM_CTRL_PROC_COMMENT(lp));
        }
        else
            printf("Not available\n");

        /* report DSP firmware loaded */
        for (n = 0; n < LINE_FIRM_NUM_DSPS(lp); n++)
        {
            if (LINE_FIRM_DSP_MAJOR(lp,n))
            {
                printf("  DSP[%2d]:   ",n);
                sprintf(tmp_buf,"%u.%u.%uB%u",
                        LINE_FIRM_DSP_MAJOR(lp,n),
                        LINE_FIRM_DSP_MIDDLE(lp,n),
                        LINE_FIRM_DSP_MINOR(lp,n),
                        LINE_FIRM_DSP_BUILD(lp,n));
                if (display_auto)
                {
                    tmp = tmp_buf + strlen(tmp_buf);
                    sprintf(tmp,"A%u",LINE_FIRM_DSP_AUTO_NUM(lp,n));
                    printf("%-18s",tmp_buf);
                }
                else
                    printf("%-12s",tmp_buf);
                printf ("%-11s  %s\n",
                        LINE_FIRM_DSP_DATE(lp,n),
                        LINE_FIRM_DSP_COMMENT(lp,n));
            }
        }
        if (LINE_FIRM_NUM_DSPS(lp) == 0)
            printf("  DSP[ 0]:   Not available\n");
        /* release module */
        DetachSession(lp);
    }
    else
        printf("Attach error.\n");
}


/* Attach Session for module */
BTLINE *
AttachSession(unsigned mod_num)
{
    struct args_line_admin args;
    BTLINE *lp;
    
    BT_ZERO(args);
    args.dest_addr.mm_bModule = mod_num;
    args.dest_addr.mm_bChannel = 2;
    args.present = 1;
    args.unique = 1;
    if ((lp = BfvSessionAttach(&args)) == NULL)
    {
        /* If channel 2 not available, try channel 1 */
        BT_ZERO(args);
        args.dest_addr.mm_bModule = mod_num;
        args.dest_addr.mm_bChannel = 1;
        args.present = 1;
        args.unique = 0;
        if ((lp = BfvSessionAttach(&args)) == NULL)
            return(NULL);
    }
    
    BT_ZERO(args);
    args.config_file_name = "btcall.cfg";
    args.skip_teleph = 1;
    if (BfvLineReset (lp, &args) < 0)
        return(NULL);
    return(lp);
}

/* detach Session */
void
DetachSession(BTLINE *lp)
{
    struct args_line_admin args;
    
    if (lp != NULL)
    {
        BT_ZERO(args);
        BfvSessionDetach(lp,&args);      
    }
}
