/* debug_control.c */
/* Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved. */

/**************************************************************************
The debug_control program is used to control the debug mode output
of running programs.
***************************************************************************/

#include "btlib.h"
#include <string.h>
#include <ctype.h>

void
usage ()
{
    fprintf (stderr, "usage: debug_control [-v] [-u chan]\n");
    fprintf (stderr, "   [-d debug_mode [-f file1 [-f file2 [-m max_size]]]\n");
    fprintf (stderr, "   [-F 0|1 [-a]]\n");
    fprintf (stderr, "   [-c debug_type[,...] [-l level] [-L file]]\n");
    fprintf (stderr, "-d = Set API debug mode, optionally with file or files with limit.\n");
    fprintf (stderr, "-F = Function entry/exit debugging, optionally with arguments.\n");
    fprintf (stderr, "-c = Call control debugging, optionally with level or file.\n");
    fprintf (stderr, "   debug_type = api | l3l4 | l4l3 | int | host | ip (one or more)\n");
    fprintf (stderr, "   level = none | error | warning | basic | verbose(dflt)\n");
    fprintf (stderr, "-v = Enable local API debug mode.\n");
    fprintf (stderr, "At least one of -d, -F, or -c is required.\n");
    fprintf (stderr, "The -u option is required for -d and -F.\n");
    exit(1);
}

int
main(int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    BTERR bterr;
    struct args_debug_mode args;
    int mode_set = 0;
    char *fnames[2];
    int num_files = 0;
    int unit_set = 0;
    char *ch1,*ch2;
    int cc_set = 0;
    int func_dbg_set = 0;
    int tmp;

    BT_API_SET_VER();
    
    BT_ZERO(args);

    while ((c = getopt (argc, argv, "u:vd:f:m:c:l:L:F:a")) != EOF)
    {
        switch (c)
        {
          case 'u':
            args.unit = atoi (optarg);
            unit_set = 1;
            break;
          case 'v':
            /* Turn on local debug mode */
            BfvDebugModeSet(DEBUG_ALL);
            break;
          case 'd':
            if (mode_set)
                usage();
            args.mode = atoi (optarg);
            fnames[0] = fnames[1] = NULL;
            mode_set = 1;
            args.options |= API_DBG_MODE_SET;
            break;
          case 'f':
            if (!mode_set || num_files >= 2)
                usage();
            args.options |= API_DBG_FILE_SET;
            fnames[num_files++] = optarg;
            break;
          case 'm':
            if (num_files != 2)
                usage();
            args.file_limit = atoi (optarg);
            break;
          case 'c':
            if (cc_set)
                usage();
            for (ch1 = optarg; *ch1; ch1 = ch2)
            {
                ch2 = ch1;
                while (isalnum(*ch2))
                    ch2++;
                if (*ch2 != ',' && *ch2 != '\0')
                    usage();
                if (*ch2 != '\0')
                {
                    *ch2 = 0;
                    ch2++;
                }
                if (!stricmp(ch1,"api"))
                    args.options |= API_DBG_CC_API_SET;
                else if (!stricmp(ch1,"l3l4"))
                    args.options |= API_DBG_CC_L3L4_SET;
                else if (!stricmp(ch1,"l4l3"))
                    args.options |= API_DBG_CC_L4L3_SET;
                else if (!stricmp(ch1,"int"))
                    args.options |= API_DBG_CC_INT_SET;
                else if (!stricmp(ch1,"host"))
                    args.options |= API_DBG_CC_HOST_MOD_SET;
                else if (!stricmp(ch1,"ip"))
                    args.options |= API_DBG_CC_IP_STACK_SET;
                else
                    usage();
            }
            cc_set = 1;
            args.cc_trace_level = ecc_trace_verbose;
            break;
          case 'l':
            if (cc_set++ != 1)
                usage();
            if (!stricmp(optarg,"none"))
                args.cc_trace_level = ecc_trace_none;
            else if (!stricmp(optarg,"error"))
                args.cc_trace_level = ecc_trace_error;
            else if (!stricmp(optarg,"warning"))
                args.cc_trace_level = ecc_trace_warning;
            else if (!stricmp(optarg,"basic"))
                args.cc_trace_level = ecc_trace_basic;
            else if (!stricmp(optarg,"verbose"))
                args.cc_trace_level = ecc_trace_verbose;
            else
                usage();
            break;
          case 'L':
            if (!cc_set || args.cc_trace_file_name)
                usage();
            args.cc_trace_file_name = optarg;
            args.options |= API_DBG_CC_LOG_FILE_SET;
            break;
          case 'F':
            if (func_dbg_set++)
                usage();
            args.options |= API_DBG_FUNC_ENTRY_EXIT;
            tmp = atoi(optarg);
            if (tmp)
                args.func_dbg_opts |= API_DBG_FUNC_OPT_ENA;
            break;
          case 'a':
            if (func_dbg_set++ != 1)
                usage();
            args.func_dbg_opts |= API_DBG_FUNC_OPT_ARGS;
            break;
          default:
            usage ();
        }
    }
    
    if ((args.options & (API_DBG_MODE_SET|API_DBG_FUNC_ENTRY_EXIT)) &&
        !unit_set)
        usage();

    if (optind != argc)
        usage ();

    if (num_files == 2 && args.file_limit == 0)
        usage();

    if (args.options == 0)
        usage();

    if (mode_set)
    {
        args.fname1 = fnames[0];
        args.fname2 = fnames[1];
    }
    args.options |= API_DBG_EXT_CTRL;
    BfvDebugModeSetAdv(&args);
    if (args.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(NULL,&args.res,&bterr);
        fprintf (stderr,"BfvDebugModeSetAdv: %s\n",bterr.long_msg);
        exit(1);
    }
}
