/* firmload.c */
/* Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved. */

/**************************************************************************
The firmload program is used to download a complete set of
standard firmware files to all boards in a system. Any error
encountered will abort the download procedure.
***************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <malloc.h>
#include <memory.h>
#if defined(WINNT) || defined(__OS2__) || defined(__QNX__) || defined(WIN95)
#include <stdlib.h>
#endif
#include <errno.h>
#include <time.h>
#ifdef WINNT
#include <io.h>
#endif
#include "btlib.h"

unsigned firm_dest = 0xffffffff;
unsigned config_spec = 0;
char* firm_dir;
int exit_code = 0;
int err_ct = 0;

char *names_list_2[] = { "cp.bin",
                         "cp1000.bin",
                         "ram.bin" };
char *names_list_0[] = { "bootdsp.hex" };
char *names_list_1[] = { "dsp1000.hex",
                         "dsp1000_5.hex",
                         "dsp1000_v34.hex",
                         "dsp2020.hex",
                         "dsp2020_5.hex",
                         "dsp_bos.hex" };
char *names_list_1_cab[] = { "dsp1000_ld.hex" };
char *names_list_1_esch[] = { "dsp1000_ud.hex",
                              "dsp1034_ud.hex" };

#define LIST2_SIZE (sizeof(names_list_2)/sizeof(*names_list_2))
#define LIST0_SIZE (sizeof(names_list_0)/sizeof(*names_list_0))
#define LIST1_SIZE (sizeof(names_list_1)/sizeof(*names_list_1))
#define LIST1_CAB_SIZE (sizeof(names_list_1_cab)/sizeof(*names_list_1_cab))
#define LIST1_ESCH_SIZE (sizeof(names_list_1_esch)/sizeof(*names_list_1_esch))


unsigned module_list[255];
unsigned num_modules = 0;
char fname0[512], fname1[512], fname1_cab[512], fname1_esch[512], fname2[512];
enum { SOMBRERO,
       CABERNET,
       ESCHER }
board_category;
int sombrero_used, cabernet_used, escher_used;

int firm_dl(BTLINE *lp,int firm_type,char *fname);
void pick_file_names(char **name_list,int name_list_size,char *fname);
int determine_dsp_loader(void);
void check_board_type(void);
int get_board_type(struct args_addr_info *args);

void usage ()
{
    fprintf (stderr, "Usage: firmload [options] firmware_dir [module_num ...]\n");
    fprintf (stderr, "options: [-c conf_spec] [-d] [-b 0|1] [-q] [-e]\n");
    fprintf (stderr, "For the current directory, firmware_dir should be '.'\n");
    fprintf (stderr, "  -c conf_spec = Configuration specification value.\n");
    fprintf (stderr, "  -d = Do not download DSP firmware.\n");
    fprintf (stderr, "  -b = Specify whether to download DSP boot loader (default = auto).\n");
    fprintf (stderr, "  -q = Quiet mode (API debug mode off).\n");
    fprintf (stderr, "  -e = Stop on first error.\n");
    exit (1);
}

int
main(int argc, char **argv)
{
    int c;
    extern int optind;
    extern char *optarg;
    BTLINE *lp;
    BTERR bterr;
    struct args_line_admin args_admin;
    struct args_addr_info args_info;
    unsigned mod_index;
    int list_given;
    int num_downloaded = 0;
    int quiet_mode = 0;
    int do_boot_loader = -1;
    int do_dsp = 1;
    int err_stop = 0;
    char *fname1_cur;

    BT_API_SET_VER();

    while ((c = getopt (argc, argv, "c:b:dqe")) != EOF)
    {
        switch (c)
        {
          case 'c':
            config_spec = atoi(optarg);
            break;
          case 'b':
            do_boot_loader = atoi(optarg);
            if (do_boot_loader != 0 && do_boot_loader != 1)
                usage();
            break;
          case 'd':
            do_dsp = 0;
            break;
          case 'q':
            quiet_mode = 1;
            break;
          case 'e':
            err_stop = 1;
            break;
          default:
            usage ();
        }
    }
    
    if (optind >= argc)
        usage ();
    
    firm_dir = argv[optind++];
    
    /* Use just the module numbers specified. */
    if (list_given = (optind < argc))
        while (optind < argc)
        {
            if (sscanf(argv[optind++],"%x",&module_list[num_modules++]) != 1)
                usage();
        }
    else
    {
        void pkt_ver_kludge(int q);

        pkt_ver_kludge(quiet_mode);
        /* No module numbers specified. Use full list. */
        while (num_modules + 2 < M_ADDR_MAX_COMPONENT_VALUE)
        {           
            module_list[num_modules] = num_modules + 2;
            num_modules++;
        }
    }
    
    pick_file_names(names_list_2,LIST2_SIZE,fname2);
    if (do_dsp)
    {
        if (do_boot_loader == -1)
            do_boot_loader = determine_dsp_loader();
        if (do_boot_loader)
            pick_file_names(names_list_0,LIST0_SIZE,fname0);

        BfvDebugModeSet(DEBUG_NONE);
        check_board_type();
        if (!quiet_mode)
            BfvDebugModeSet(DEBUG_ALL);

        if (sombrero_used)
            pick_file_names(names_list_1,LIST1_SIZE,fname1);
        if (cabernet_used)
            pick_file_names(names_list_1_cab,LIST1_CAB_SIZE,fname1_cab);
        if (escher_used)
            pick_file_names(names_list_1_esch,LIST1_ESCH_SIZE,fname1_esch);
    }

    for (mod_index = 0; mod_index < num_modules; mod_index++)
    {
        /* When checking for module existence, do not produce debug output */
        BfvDebugModeSet(DEBUG_NONE);

        BT_ZERO(args_info);
        args_info.mod_num = module_list[mod_index];
        BfvModuleInfo(NULL,&args_info);
        if (args_info.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(NULL,&args_info.res,&bterr);
            fprintf (stderr,"BfvModuleInfo: %s\n",bterr.long_msg);
            exit (1);
        }
        if (!args_info.exists)
        {
            /* Only report this error if list specified. */
            if (list_given)
            {
                fprintf(stderr,"Module %X doesn't exist.\n",
                        module_list[mod_index]);
                exit_code = 1;
            }
            continue;
        }
        if (!args_info.hardware)
        {
            /* Only report this error if list specified. */
            if (list_given)
            {
                fprintf(stderr,"Module %X is not a hardware module.\n",
                        module_list[mod_index]);
                exit_code = 1;
            }
            continue;
        }
        switch (get_board_type(&args_info))
        {
          case SOMBRERO:
          default:
            fname1_cur = fname1;
            break;
          case CABERNET:
            fname1_cur = fname1_cab;
            break;
          case ESCHER:
            fname1_cur = fname1_esch;
            break;
        }

        num_downloaded++;

        /* When doing the attach, reset, and download, use debug mode, unless
           quiet mode is specified. */
        if (!quiet_mode)
            BfvDebugModeSet(DEBUG_ALL);

        BT_ZERO(args_admin);
        args_admin.dest_addr.mm_bModule = module_list[mod_index];
        args_admin.dest_addr.mm_bChannel = 1;
        args_admin.present = 1;
        args_admin.unique = 1;
        if ((lp = BfvSessionAttach(&args_admin)) == NULL)
        {
            BfvErrorMessage(lp,&args_admin.res,&bterr);
            fprintf (stderr,"BfvSessionAttach for module %X: %s\n",
                     module_list[mod_index],bterr.long_msg);
            err_ct++;
            exit_code = 1;
            if (err_stop)
                goto end_it;
            continue;
        }
           
        BT_ZERO(args_admin);
        args_admin.config_file_name = "btcall.cfg";
        args_admin.no_init = 2;
        args_admin.skip_teleph = 1;
        if (BfvLineReset (lp, &args_admin) < 0)
        {
            BfvErrorMessage(lp,&args_admin.res,&bterr);
            fprintf (stderr, "BfvLineReset for module %X: %s: status %lX.\n",
                     module_list[mod_index],bterr.long_msg,
                     args_admin.reset_status);
            /* Even if BfvLineReset gets an error, keep going. */
            fprintf (stderr, "Continuing regardless of reset error.\n");
        }

        if (!firm_dl(lp,2,fname2))
            if (!(do_dsp && do_boot_loader && firm_dl(lp,0,fname0)))
                if (!(do_dsp && firm_dl(lp,1,fname1_cur)))
                    if (quiet_mode)
                        printf("Download to module %X done.\n",
                               module_list[mod_index]);

        if (err_stop && err_ct > 0)
            goto end_it;

        BT_ZERO(args_admin);
        BfvSessionDetach(lp,&args_admin);
        if (args_admin.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(lp,&args_admin.res,&bterr);
            fprintf (stderr,"BfvSessionDetach for module %X: %s\n",
                     module_list[mod_index],bterr.long_msg);
            err_ct++;
            exit_code = 1;
            if (err_stop)
                goto end_it;
        }
    }

end_it:
    printf("Downloads attempted on %d module",num_downloaded);
    if (num_downloaded != 1)
        putchar('s');
    if (list_given)
        printf(" out of %d specified",num_modules);
    printf(" with %d error",err_ct);
    if (err_ct != 1)
        putchar('s');
    printf(".\n");

    if (!list_given && num_downloaded == 0)
        exit_code = 1;

    return(exit_code);
}

int
firm_dl(BTLINE *lp,int firm_type,char *fname)
{
    struct args_download args_download;
    BTERR bterr;
    int ret;

    printf("Downloading file \"%s\"\n",fname);
    BT_ZERO(args_download);
    args_download.fname = fname;
    args_download.specify_type = 1;
    args_download.dl_type = firm_type;
    args_download.dl_dest = firm_dest;
    if (firm_type == 2)
        args_download.config_spec_value = config_spec;
    ret = BfvFirmwareDownload(lp,&args_download);
    if (args_download.res.status != BT_STATUS_OK)
    {
        BfvErrorMessage(lp,&args_download.res,&bterr);
        fprintf (stderr,"BfvFirmwareDownload for module %X: %s\n",
                 LINE_DEST_ADDR(lp).mm_bModule,bterr.long_msg);
        err_ct++;
        exit_code = 1;
    }

#if 0
    printf("Done, result value %d.\n",ret);
#endif

    return(ret);
}

void pick_file_names(char **name_list,int name_list_size,char *fname)
{
    int i;

    for (i = 0; i < name_list_size; i++)
    {
        sprintf(fname,"%s/%s",firm_dir,name_list[i]);
        if (access(fname,F_OK) == 0)
            return;
    }

    if (name_list_size == 1)
        fprintf(stderr,"%s/%s does not exist.\n",firm_dir,name_list[0]);
    else
    {
        fprintf(stderr,"Neither ");
        for (i = 0; i < name_list_size; i++)
        {
            fprintf(stderr,"%s/%s",firm_dir,name_list[i]);
            if (i < name_list_size - 2)
                fprintf(stderr,", ");
            else if (i == name_list_size - 2)
                fprintf(stderr," nor ");
        }
        fprintf(stderr," exists.\n");
    }

    exit(1);
}

#define SEARCH_STRING "@[#] NOBOOT"
#define SEARCH_STRING_LEN (sizeof(SEARCH_STRING) - 1)

int determine_dsp_loader(void)
{
    FILE *fp;
    int ch;
    int ret = 1;
    int cur_index = 0;

#ifdef USES_FAT_FILESYSTEM
    fp = fopen(fname2,"rb");
#else
    fp = fopen(fname2,"r");
#endif
    if (fp == NULL)
    {
        fprintf(stderr,"Error opening %s for scanning!\n",fname2);
        exit(1);
    }

    while ((ch = getc(fp)) != EOF)
    {
        if (ch == SEARCH_STRING[cur_index])
        {
            if (++cur_index == SEARCH_STRING_LEN)
            {
                ret = 0;
                break;
            }
        }
        else
            cur_index = 0;
    }

    fclose(fp);
    return(ret);
}

#define PCI_BROOKTROUT_VENDOR           0x12e4
#define PCI_BROOKTROUT_DEVICE_BAROSSA   0xB005
#define PCI_BROOKTROUT_DEVICE_APPLEGATE 0xB006
#define PCI_BROOKTROUT_DEVICE_ESCHER    0xB007
#define PCI_BROOKTROUT_VENDOR_8241      0x1057
#define PCI_BROOKTROUT_DEVICE_8241      0x0006
int get_board_type(struct args_addr_info *args)
{
    if (args->pci_sub_vend_id == PCI_BROOKTROUT_VENDOR &&
        (args->pci_sub_id == PCI_BROOKTROUT_DEVICE_BAROSSA ||
         args->pci_sub_id == PCI_BROOKTROUT_DEVICE_APPLEGATE) ||
        /* temporary */
        args->pci_vend_id == PCI_BROOKTROUT_VENDOR_8241 &&
        args->pci_dev_id == PCI_BROOKTROUT_DEVICE_8241)
        return(CABERNET);
    if (args->pci_sub_vend_id == PCI_BROOKTROUT_VENDOR &&
        args->pci_sub_id == PCI_BROOKTROUT_DEVICE_ESCHER)
        return(ESCHER);
    return(SOMBRERO);
}

void check_board_type(void)
{
    unsigned mod_index;
    struct args_addr_info args_info;
    BTERR bterr;

    for (mod_index = 0; mod_index < num_modules; mod_index++)
    {
        BT_ZERO(args_info);
        args_info.mod_num = module_list[mod_index];
        BfvModuleInfo(NULL,&args_info);
        if (args_info.res.status != BT_STATUS_OK)
        {
            BfvErrorMessage(NULL,&args_info.res,&bterr);
            fprintf (stderr,"BfvModuleInfo: %s\n",bterr.long_msg);
            exit (1);
        }
        if (!args_info.exists)
            continue;
        if (!args_info.hardware)
            continue;

        switch (get_board_type(&args_info))
        {
          case SOMBRERO:
          default:
            sombrero_used++;
            break;
          case CABERNET:
            cabernet_used++;
            break;
          case ESCHER:
            escher_used++;
            break;
        }
    }
}

/* Do something to access the driver so that M_ADDR_MAX_COMPONENT_VALUE
   will be valid before we need to use it. */
void pkt_ver_kludge(int q)
{
    struct args_addr_info args_info;

    BfvDebugModeSet(DEBUG_NONE);
    BT_ZERO(args_info);
    BfvModuleInfo(NULL,&args_info);
    if (!q)
        BfvDebugModeSet(DEBUG_ALL);
}
