/*
 * infopkt.h -- fax and voice api info packet definitions
 * Copyright (c) 1989-2004 Brooktrout Technology Inc.  All rights reserved.
 */

/*
 * an infopkt is a chunk of data preceeded by a header.  The common
 * part of a header contains a type code and a length.  Some type
 * codes imply a longer header before the data.  The length field
 * includes the total size of the header, plus optional data.
 * A file containing infopkts is just a bunch of infopkts concatenated
 * together.  At the infopkt level, you can concatenate two files
 * a get a valid infopkt file, although this may not make sense for
 * some applications.
 */

#ifndef _infopkt_h_
#define _infopkt_h_

/* stdio.h needed for infopkt.h */
#ifndef EOF
#include <stdio.h>
#endif

#pragma pack(1)

/*
 * Headers are in Intel byte order, other machines must swap.
 */  

/* No infopkt will be bigger than this, so that it is easy to deal
 * with them in normal arrays on 16-bit platforms.
 */
#define INFOPKT_MAX_LENGTH 30000
/* maximum nesting depth */
#define MAX_INDIR_FILES 3

enum infopkt_types {
    INFOPKT_G3=32490,                   /* Data Types */
    INFOPKT_BTG3,
    INFOPKT_ASCII,
    INFOPKT_CVSD,                       /* obsolete */
    INFOPKT_G3_STRIP_PARAMETERS,        /* Tag Types */
    INFOPKT_ASCII_STRIP_PARAMETERS,
    INFOPKT_ASCII_PAGE_PARAMETERS,
    INFOPKT_CVSD_PARAMETERS,            /* obsolete */
    INFOPKT_T30_PARAMETERS,
    INFOPKT_BEGINNING_OF_PAGE,
    INFOPKT_DOCUMENT_PARAMETERS,
    INFOPKT_INDIR_G3,                   /* Indirect data file types */
    INFOPKT_INDIR_BTG3,
    INFOPKT_INDIR_ASCII,
    INFOPKT_INDIR_CVSD,                 /* obsolete */
    INFOPKT_INDIR_INFOPKT,              /* indirect infopkt stream file */
    INFOPKT_USER0,
    INFOPKT_USER1,
    INFOPKT_USER2,
    INFOPKT_USER3,
    INFOPKT_USER4,
    INFOPKT_USER5,
    INFOPKT_USER6,
    INFOPKT_USER7,
    INFOPKT_USER8,
    INFOPKT_USER9,
    INFOPKT_INDIR_TIFF,
    INFOPKT_END_OF_SPEECH,
    INFOPKT_SPEECH_PARAMETERS,
    INFOPKT_SPEECH,
    INFOPKT_INDIR_SPEECH,
    INFOPKT_PROMPT_MAP,
    INFOPKT_ANNOTATION,
    INFOPKT_FAX_HDR,
    INFOPKT_PAGE_PARAMETERS,
    INFOPKT_BFT,
    INFOPKT_INDIR_DCX,
    INFOPKT_OVERLAY,
    INFOPKT_SPACE,
    INFOPKT_EFF_PAGE_PARAMETERS,
    INFOPKT_INDIR_WAVE
};

enum indir_mode {
    INDIR_MODE_FOLLOW,
    INDIR_MODE_RETURN_AND_FOLLOW,
    INDIR_MODE_RETURN,
    INDIR_MODE_FOLLOW_NOUSER /* follow packets and toss user pkts */
};

#define IPS_FLAG_UNGET  0x01
#define IPS_FLAG_MEM    0x02
#define IPS_FLAG_FP     0x04

struct infopkt_stream {
    FILE *fpstack[MAX_INDIR_FILES];
    int next_fp;
    FILE *fp;
    char *buf;
    unsigned short bufsize;
    unsigned short flags;
    struct infopkt *unget_ip;
    M_CB1 void (* M_CB2 user_func)
         (struct infopkt *ip,struct infopkt_stream *ips,char *arg);
    char *user_arg;
};

struct mem_infopkt_info {
    unsigned char *bufptr;
    long bufsize;
    long max_valid_data;
    long offset_within_buf;
    long offset_of_buf;
    unsigned char *cur_ptr;
    M_CB1 void (* M_CB2 buf_func)
         (struct infopkt_stream *ips,long old_offset,long new_offset,
          unsigned char **buf,long *bufsize,char *arg);
    char *buf_arg;
};

/* All structures written to files must have their multi-byte integers in
   little-endian format. */

struct infopkt {
    unsigned short type;
    unsigned short length; /* including the header */
    /* more header and data follows */
};

struct g3pkt {
    struct infopkt hdr;
    /* group 3 data follows */
};

struct btg3pkt {
    struct infopkt hdr;
    /* group 3 data follows */
};
struct asciipkt {
    struct infopkt hdr;
    /* ascii data follows */
};

/* obsolete */
struct cvsdpkt {
    struct infopkt hdr;
    unsigned short flags;               /* these two are obsolete */
    unsigned short rate;                /* not used */
};

struct g3strippkt {                     /* cmd 57 */
    struct infopkt hdr;
    unsigned short resolution;
    unsigned short width;
    unsigned short data_fmt;
};

struct asciistrippkt {                  /* cmd 57 */
    struct infopkt hdr;
    unsigned short resolution;
    unsigned short width;
    unsigned short eof_char;
    unsigned short font_no;
    unsigned short left_margin;
    unsigned short right_margin;
    unsigned short line_spacing;
};

/* obsolete */
struct asciipagepkt {                   /* cmd 5a */
    struct infopkt hdr;
    unsigned short top_margin;
    unsigned short bottom_margin;
    unsigned short length;
    unsigned short pad_flag;
};

struct pageparampkt {                   /* cmd 5a */
    struct infopkt hdr;
    unsigned short top_margin;
    unsigned short bottom_margin;
    unsigned short length;
    unsigned short ascii_pad;
    unsigned short image_pad;
    unsigned short image_break;
    unsigned short image_margin;
};

/* obsolete */
struct cvsdparampkt {
    struct infopkt hdr;
    unsigned short flags;
    unsigned short rate;
};

struct t30parampkt {
    struct infopkt hdr;
    unsigned char bit_rate;     /* cmd 44 */
    unsigned char scan_time;    /* cmd 42 */
    unsigned char coding_type;  /* not implemented */
    unsigned char modulation_type;
    unsigned char line_compression;
};

/* send: signal EOPAGE w/ new params; receive: insert into infopkt stream */
struct docpkt {
    struct infopkt hdr;
    unsigned char resolution;   /* cmd 41 */
    unsigned char hor_width;    /* cmd 41 */
    unsigned char vert_length;  /* not implemented */
};

/* send: only to signal EOPAGE */
struct boppkt {
    struct infopkt hdr;
};

struct indirg3pkt {
    struct infopkt hdr;
};

struct indirbtg3pkt {
    struct infopkt hdr;
};

struct indirasciipkt {
    struct infopkt hdr;
};

/* obsolete */
struct indircvsdpkt {
    struct infopkt hdr;
};

struct indirinfopkt {
    struct infopkt hdr;
};

struct userpkt {
    struct infopkt hdr;
    /* user defined data follows */
};

struct indirtiffpkt {
    struct infopkt hdr;
};

struct indirdcxpkt {
    struct infopkt hdr;
};

/* Speech parameter infopkt */
struct SPI {
    struct infopkt hdr;
    unsigned short sample_rate;
    unsigned char coding_format;
    unsigned char bits_per_sample;
    unsigned char afe_rate;
    unsigned char data_fmt;
    unsigned char agc;
};

struct speechpkt {
    struct infopkt hdr;
};

struct indirspeechpkt {
    struct infopkt hdr;
};

struct eospkt {
    struct infopkt hdr;
    unsigned short mode;
};

struct promptpkt {
    struct infopkt hdr;
    unsigned num_phrases;
    /* map data follows */
};

struct prompt_phrase {
    unsigned phrase_size;
    unsigned offset;
};

#define MAX_PHRASES_PER_MAP \
   ((INFOPKT_MAX_LENGTH - sizeof(struct promptpkt)) / \
    sizeof(struct prompt_phrase))

typedef struct prompt_map {
    struct infopkt_stream *ips;
    struct promptpkt promptpkt;
    struct prompt_phrase *phrases;
} PROMPT_MAP;

struct annotpkt {
    struct infopkt hdr;
    /* ascii data follows */
};

struct faxhdrpkt {
    struct infopkt hdr;
    unsigned char placement;
    unsigned char insert_mode;
    /* ascii data for label follows */
};

struct bftpkt {                 /* cmd 41 */
    struct infopkt hdr;
    unsigned char bft_type;
};

struct overlaypkt {
    struct infopkt hdr;
    char overlay_number;
};

struct spacepkt {
    struct infopkt hdr;
    short spacing;
    unsigned char units;
};

struct effpagepkt {
    struct infopkt hdr;
    unsigned eff_page_type;
};

struct indirwavepkt {
    struct infopkt hdr;
};




/* All new fields for an argument structure must be added to the end of
   the structure. Compatibility support must also be added to appropriate
   internal source files, as well.
*/

struct args_infopkt {
    RES res;
    struct infopkt_stream *ips;
    MCONST char *fname;
    MCONST char *fmode;
    unsigned char *buf;
    long buf_size;
    struct infopkt *ip;
    enum indir_mode i_mode;
    M_CB1 void (* M_CB1 user_func)(MCONST struct infopkt *ip,
                                   struct infopkt_stream *ips,
                                   char *arg);
    char *arg;
    int use_open_file;
    FILE *fp;
    long offset;
    int origin;
    PROMPT_MAP *prompt_map;
    int limited_read;
    long valid_data;
    M_CB1 void (* M_CB2 buf_func)
         (struct infopkt_stream *ips,long old_offset,long new_offset,
          unsigned char **buf,long *bufsize,char *arg);
};

/* All new fields for an argument structure must be added to the end of
   the structure. Compatibility support must also be added to appropriate
   internal source files, as well.
*/


#pragma pack()

#endif /* _infopkt_h_ */


#ifdef MILL_DO_PROTOS

B_APIE1 struct infopkt_stream * B_APIE2 MillInfopktOpen(struct args_infopkt
                                                        *args);
B_APIE1 struct infopkt_stream * B_APIE2 MillInfopktOpenMem (struct args_infopkt
                                                            *args);
B_APIE1 long B_APIE2 MillInfopktFtell (struct args_infopkt *args);
B_APIE1 int B_APIE2 MillInfopktFseek (struct args_infopkt *args);
B_APIE1 int B_APIE2 MillInfopktClose (struct args_infopkt *args);
B_APIE1 MCONST struct infopkt * B_APIE2 MillInfopktGet (struct args_infopkt *args);
B_APIE1 int B_APIE2 MillInfopktUnget (struct args_infopkt *args);
B_APIE1 int B_APIE2 MillInfopktPut (struct args_infopkt *args);
B_APIE1 void B_APIE2 MillInfopktUser (struct args_infopkt *args);

#endif
