/* osi_SL.c */
/* Solaris version of OS independent function calls */
/* Copyright (c) 2001 Brooktrout Technology Inc.  All rights reserved. */

#include "osilib.h"

// insert component version string into library
#define COMPONENT_NAME	OSI_API
#include "../../../bfv.api/inc/component_version.h"

/* Solaris */
/* Thread Functions */
/* Using Solaris threads */
/* requires library switch -lthread */
/* requires #include <thread.h> */

/******************************************************************************
* Function:    
*	int osiThread_Create ( void*             stack_start,
*						   int               stack_size,
*						   void*             (*start_routine)(void *),
*						   void*             arguments,
*						   int               flags, 
*						   unsigned long*    thr_id)
*
* Inputs:
*	stack_start		pointer to base address of new threads stack
*	stack_size		size of new threads stack
*	start_routine	pointer to function new thread will execute
*	arguments		pointer to the argument to start_routine
*	flags			attributes for the new thread
*					limited to START_SUSPENDED or 0
*	thr_id			pointer to identifier of new thread. Set by Thread_Create
*
*	returns 0 on success or error code on failure
*
*******************************************************************************/
int osiThread_Create ( void*             stack_start,
				       int               stack_size, 
					   void*             (*start_routine)(void *),  
					   void*             arguments, 
					   int               flags, 
					   unsigned long*    thr_id)
{
	int status;
	thread_t * thrID_ptr;
	thread_t thr_id_tmp;


	if (thr_id == NULL)
	{
		thrID_ptr = &thr_id_tmp;
	}
	else
	{
		thrID_ptr = (thread_t*)thr_id;
	}

	status = thr_create(
				stack_start,				/* stack_base */
				(size_t)stack_size,			/* stack_size */
				start_routine,				/* start_func */
				arguments,					/* arg to start function*/
				(long)flags,				/* flags */
				thrID_ptr);					/* new_thread_id */
	
	return status;
	
}

/******************************************************************************
* Function:  
*	int osiThread_Wait( bt_thread thr_id,  int *status)
*
* Inputs:
*	thr_id	thread to wait for
*	status	pointer to exit status of thr_id. Set by Thread_Wait
*
*	returns 0 on success or error code on failure
*
*******************************************************************************/
int osiThread_Wait( bt_thread thr_id,  int *status)
{
	int result;
	int status_tmp;

	result = thr_join( (thread_t)thr_id, NULL, (void **)&status_tmp);

	if (status != NULL)
		*status = status_tmp;

	return result;
}

/******************************************************************************
* Function:  
*	int osiThread_Destroy( bt_thread thr_id)
*
* Inputs:
*	thr_id	thread to wait for
*
*	returns 0 on success or error code on failure
*
*******************************************************************************/
int osiThread_Destroy( bt_thread thr_id)
{
	return 0;	/* this function not needed for Solaris */
}

/* Solaris */
/* Using the posix version of semaphore */
/* requires compile switch -lposix4 */
/* requires <semaphore.h> */

/******************************************************************************
* Function:
*	int osiSemaphore_Init (bt_sem * Sem_ID_ptr, int resources, int type)
*
* Inputs:
*	Sem_ID		pointer to semaphore ID. Set by Semaphore_Init
*	resources	number of resources protected by this semaphore
*	type		type of semaphore to create. Limited to S_THREAD or S_PROCESS
*
*	returns 0 on success or -1 on failure
*
*******************************************************************************/
int osiSemaphore_Init (bt_sem * Sem_ID_ptr, int resources, int type)
{
	int result;

	result = sem_init((sem_t*)Sem_ID_ptr, type, resources);

	return result;
}

/******************************************************************************
* Function:
*	int osiSemaphore_Destroy (bt_sem * Sem_ID)
*
* Inputs:
*	Sem_ID		pointer to semaphore to destroy
*
*	returns 0 on success or -1 on failure
*
*******************************************************************************/
int osiSemaphore_Destroy(bt_sem* Sem_ID)
{
	int result;

	result = sem_destroy((sem_t*)Sem_ID);

	return result;
}

/******************************************************************************
* Function:
*	int osiSemaphore_Increment (bt_sem * Sem_ID)
*
* Inputs:
*	Sem_ID		pointer to semaphore to release
*
*	returns 0 on success or -1 on failure
*
*******************************************************************************/
int osiSemaphore_Increment (bt_sem* Sem_ID)
{
	int result;

	result = sem_post((sem_t*)Sem_ID);

	return result;
}

/******************************************************************************
* Function:
*	int osiSemaphore_Decrement (bt_sem * Sem_ID_ptr, long timeout)
*
* Inputs:
*
*	Sem_ID		pointer to semaphore to acquire
*	timeout		timeout in case semaphore cannot be acquired, in milliseconds
*                    or INFINITE if no timeout
*
*	returns  0 on success
*	        -1 on timeout
*	        -2 on failure	(NT only)
*
*******************************************************************************/
int osiSemaphore_Decrement (bt_sem *Sem_ID, long timeout)
{
	int result;

	if (timeout == 0)
	{
		result = sem_trywait(Sem_ID);
		if (result != 0)
			return (-1);	// timeout
		else
			return (0);		// success
	}
	else if (timeout != INFINITE)
	{
		while (timeout > 0)
		{
			result = sem_trywait(Sem_ID);
			if ((result != 0) )
			{
				osiWait(TIMEOUT_INCREMENT);
				timeout = timeout - TIMEOUT_INCREMENT;
				if (timeout <= 0)
				{
					// last chance
					result = sem_trywait(Sem_ID);
					if (result != 0)
					{
						return(-1);	// timeout
					}
					else
						return(0);	// success
				}
			}
			else
				return (0);		// success
		}
	}
	sem_wait((sem_t*)Sem_ID);

	return (0);		// success
}

/* Solaris */
/* Mutex Functions */
/* Solaris version of mutex */
/* requires compile switch -lthread */
/* requires #include <thread.h> */
/* requires #include <synch.h> */

/******************************************************************************
* Function:
*	int osiMutex_Init (mutex_t * Mutex_ID, int type)
*
* Inputs:
*
*	Mutex		pointer to a mutex
*	type		either M_THREAD or M_PROCESS
*
*	returns 0 on success or error code on failure
*
*******************************************************************************/  
int osiMutex_Init (bt_mutex * Mutex_ID, int type)
{
	return mutex_init(Mutex_ID, type, NULL);
}

/******************************************************************************
* Function:
*	int osiMutex_Acquire(mutex_t * Mutex_ID, long timeout )
*
* Inputs:
*
*	Mutex_ID		pointer to a mutex
*	timeout			timeout in case mutex cannot be acquired, in milliseconds
*                    or INFINITE if no timeout
*
*	returns  0 on success
*	        -1 on timeout
*	        -2 on failure	(NT only)
*
*******************************************************************************/
int osiMutex_Acquire(bt_mutex * Mutex_ID , long timeout)
{
	int result;

	if (timeout == 0)
	{
		result = mutex_trylock(Mutex_ID);
		if (result != 0)
			return (-1);	// timeout
		else
			return (0);		// success
	}
	else if (timeout != INFINITE)
	{
		while (timeout > 0)
		{
			result = mutex_trylock(Mutex_ID);
			if ((result != 0) )
			{
				osiWait(TIMEOUT_INCREMENT);
				timeout = timeout - TIMEOUT_INCREMENT;
				if (timeout <= 0)
				{
					// last chance
					result = mutex_trylock(Mutex_ID);
					if (result != 0)
					{
						return(-1);	// timeout
					}
					else
						return(0);	// success
				}
			}
			else
				return (0);		// success
		}
	}
	mutex_lock(Mutex_ID);

	return (0);		// success
}

/******************************************************************************
* Function:
*	int osiMutex_Release(mutex_t * Mutex_ID)
*
* Inputs:
*
*	Mutex_ID		pointer to a mutex
*
*	returns 0 on success or error code on failure
*
*******************************************************************************/
int osiMutex_Release(bt_mutex * Mutex_ID)
{
	return mutex_unlock(Mutex_ID);
}

/******************************************************************************
* Function:
*	int osiMutex_Destroy(mutex_t * Mutex_ID)
*
* Inputs:
*
*	Mutex_ID		pointer to a mutex
*
*	returns 0 on success or error code on failure
*
*******************************************************************************/
int osiMutex_Destroy(bt_mutex * Mutex_ID)
{
	return 0;	/* this function not needed for Solaris */
}

/* Solaris */
/* Time and Keyboard related functions */

/******************************************************************************
* Function:
*	void osiWait (unsigned long MilliTime)
*
* Inputs:
*
*	MilliTime	time in milliseconds to wait before proceeding
*
* requires compile switch -lposix4
* requires #include <time.h>
*
*******************************************************************************/
void osiWait(unsigned long MilliTime)
{
	struct timespec rqt;

	rqt.tv_sec = MilliTime / 1000;
	rqt.tv_nsec = (MilliTime % 1000) * 1000000;
	nanosleep(&rqt, NULL);
}

/******************************************************************************
* Function:
*	int kbhit ()
*
* Inputs:
*    none
*
*	return is number of characters waiting at keyboard
*
* requires #include <termios.h>
* requires #include <unistd.h>
* requires #include <stropts.h>
*
*******************************************************************************/
int kbhit(void)
{
#ifdef _LP64
	return (1);
#else
	long count = 0;
	static struct termios stored_settings;
	struct termios new_settings;

	/* save original settings */
    tcgetattr(0,&stored_settings);

    new_settings = stored_settings;

	/* set to single key sroke input */
    /* Disable canonical mode, and set buffer size to 1 byte */
    new_settings.c_lflag &= (~ICANON);
    new_settings.c_cc[VTIME] = 0;
    new_settings.c_cc[VMIN] = 1;

    tcsetattr(0,TCSANOW,&new_settings);

    /*  Add  up the count of characters in the file buffer          */
    /*  and the count of characters in the device buffer.           */
    if( stdin != NULL )
	{
		/* count is number of characters waiting since last read  */
		ioctl( stdin->_file, FIONREAD, &count );
		count += stdin->_cnt;
	}

	/* restore original keypress settings*/
    tcsetattr(0,TCSANOW,&stored_settings);

    return (int)count; 
#endif
}

/******************************************************************************
* Function:
*	char osiWait_For_Keyboard_Entry ()
*
* Inputs:
*		None
*
*	return is first character of input
*
* requires #include <termios.h>
* requires #include <unistd.h>
* requires #include <stropts.h>
* requires #include <stdio.h>
*
*******************************************************************************/
char osiWait_For_Keyboard_Entry ()
{
    long count;
	int c2;
	char c;
	static struct termios stored_settings;
	struct termios new_settings;

	/* save original settings */
    tcgetattr(0,&stored_settings);

    new_settings = stored_settings;

	/* set to single key sroke input */
    /* Disable canonical mode, and set buffer size to 1 byte */
    new_settings.c_lflag &= (~ICANON);
    new_settings.c_cc[VTIME] = 0;
    new_settings.c_cc[VMIN] = 1;

    tcsetattr(0,TCSANOW,&new_settings);

    /*  Add  up the count of characters in the file buffer          */
    /*  and the count of characters in the device buffer.           */
    count = 0;
    if( stdin != NULL )
	{
#ifdef _LP64
		count = 0;
		while (count == 0)
		{
			c2 = getchar();
			if (c2 == EOF)
			{
				osiWait(100);
			}
			else
			{
				count = 1;
				c = (char)c2;
			}
		}
#else
		while (count == 0)
		{
			/* kbhit replacement */
			/* count is number of characters waiting since last read  */
			ioctl( stdin->_file, FIONREAD, &count );
			count += stdin->_cnt;
			osiWait(100);
		}

		c = getchar();
#endif
    }

	/* restore original keypress settings*/
    tcsetattr(0,TCSANOW,&stored_settings);

    return c; 
}


/******************************************************************************
* Function:
*	void osiGet_Time ( time_x Time)
*
* Inputs:
*
*	Time	structure with current system time
*
* requires #include <sys/timeb.h>
*
*******************************************************************************/
void osiGet_Time ( time_x * timebuffer)
{
	ftime(timebuffer);
}

/******************************************************************************
* Function:
*	void osiGet_Millitime (void)
*
*   Returns time in milliseconds
*
* requires #include <sys/timeb.h>
*
*******************************************************************************/
unsigned long osiGet_Millitime(void)
{
	time_x tb;

	ftime(&tb);
	return (tb.time * 1000 + tb.millitm);
}

/******************************************************************************
* Function:
*	void osiGet_Time_Stamp(char * Time_Stamp)
*
* Inputs:
*
*	Time_Stamp	char string with current system time
*
* requires #include <sys/timeb.h>
* requires #include <time.h>
* requires #include <stdio.h>
*
*******************************************************************************/
void osiGet_Time_Stamp(char * Time_Stamp)
{
	struct	timeb timebuffer;
	char*  timeline;

	ftime(&timebuffer);

	timeline = ctime(&(timebuffer.time));

	/* kill new line character */
	timeline[24] = 0;

	sprintf(Time_Stamp,"%.19s.%hu %s", timeline, timebuffer.millitm, &timeline[20]);

}

/* Solaris */
/* Socket Functions */
/* requires compile switch -lsocket -lnsl */
/* #include <sys/socket.h>
/* #include <unistd.h>     /* For gethostname() */
/* #include <netdb.h>      /* For gethostbyname() */
/* #include <netinet/in.h> /* For htons() */
/******************************************************************************
* Function:
*	int osiSockets_Init (void)
*
* Inputs: None
*
*	
* Always returns 0 for success.
*
*******************************************************************************/
int osiSockets_Init (void)
{
	/* Do nothing. */  
	/* This function is for Windows NT compatibilty. */
	return (0);
}

/******************************************************************************
* Function:
*	int osiSockets_Destroy (void)
*
* Inputs: None
*
*	
* Always returns 0 for success.
*
*******************************************************************************/
int osiSockets_Destroy (void)
{
	/* Do nothing. */  
	/* This function is for Windows NT compatibilty. */
	return (0);
}

/******************************************************************************
* Function:
*	bt_socket osiSocket_Create (int iAddressFamily, 
*                               int iSocketType,
*                               int iProtocol)
*
* Inputs:
*
*	
* Return a socket descriptor on success and INVALID_SOCKET on error.
*
*******************************************************************************/
bt_socket osiSocket_Create (int iAddressFamily, int iSocketType, int iProtocol)
{	
	return ( socket (iAddressFamily, iSocketType, iProtocol) );
}

/******************************************************************************
* Function:
*	int	osiSocket_Bind (bt_socket btsSocket,
*                       struct sockaddr *saAddress,
*                       int iAddrLength)
*
* Inputs:
*
*	
* Return 0 on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int	osiSocket_Bind (bt_socket btsSocket, struct sockaddr *saAddress, int iAddrLength)
{
	return ( bind (btsSocket, saAddress, (socklen_t)iAddrLength) );
}

/******************************************************************************
* Function:
*	osiSocket_Connect (bt_socket btsSocket,
*                      struct sockaddr *saAddress,
*                      int iAddrLength)
*
* Inputs:
*
*	
* Return 0 on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int	osiSocket_Connect (bt_socket btsSocket, const struct sockaddr *saAddress, int iAddrLength)
{
	return ( connect (btsSocket, saAddress, (socklen_t)iAddrLength) );
}

/******************************************************************************
* Function:
*	int	osiSocket_Listen (bt_socket btsSocket, int iBackLog)
*
* Inputs:
*
*	
* Return 0 on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int	osiSocket_Listen (bt_socket btsSocket, int iBackLog)
{
	return ( listen (btsSocket, iBackLog) );
}

/******************************************************************************
* Function:
*	bt_socket osiSocket_Accept (bt_socket btsSocket,
*                               struct sockaddr *saAddress,
*                               int *iAddrLength)
*
* Inputs:
*
*	
* Return a valid socket descriptor on success or INVALID_SOCKET on error.
*
*******************************************************************************/
bt_socket osiSocket_Accept (bt_socket btsSocket, struct sockaddr *saAddress, int *iAddrLength)
{
	return ( accept (btsSocket, saAddress, (socklen_t *)iAddrLength) ); 
}

/******************************************************************************
* Function:
*	int	osiSocket_Close (bt_socket btsSocket)
*
* Inputs:
*
*	
* Return 0 on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int	osiSocket_Close (bt_socket btsSocket)
{
	return ( close (btsSocket) );
}

/******************************************************************************
* Function:
*	int osiSocket_Recv (bt_socket btsSocket,
*                       void *pvBuffer,
*                       int iBufferLength,
*                       int iFlags)
*
* Inputs:
*
*	
* Return the number of bytes received on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int osiSocket_Recv (bt_socket btsSocket, void *pvBuffer, int iBufferLength, int iFlags)
{
	return ( recv (btsSocket, pvBuffer, (size_t)iBufferLength, iFlags) );
}

/******************************************************************************
* Function:
*	int osiSocket_Send (bt_socket btsSocket,
*                       const void *pvBuffer,
*                       int iBufferLength,
*                       int iFlags)
*
* Inputs:
*
*	
* Return the number of bytes sent on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int osiSocket_Send (bt_socket btsSocket, const void *pvBuffer, int iBufferLength, int iFlags)
{
	return ( send (btsSocket, pvBuffer, (size_t)iBufferLength, iFlags) );
}

/******************************************************************************
* Function:
*	int osiSocket_SetOption (bt_socket btsSocket,
*                            int iLevel,
*                            int iOptName,
*                            const void *pvOptVal,
*                            int iOptLength)
*
* Inputs:
*
*	
* Return 0 on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int osiSocket_SetOption (bt_socket btsSocket, int iLevel, int iOptName, const void *pvOptVal, int iOptLength)
{
	return ( setsockopt (btsSocket, iLevel, iOptName, pvOptVal, (socklen_t)iOptLength) );
}

/******************************************************************************
* Function:
*	int	osiSocket_GetHostName (char *pchHostName, int iNameLength)
*
* Inputs:
*
*	
* Return 0 on success or SOCKET_ERROR on error.
*
*******************************************************************************/
int	osiSocket_GetHostName (char *pchHostName, int iNameLength)
{
	return (gethostname (pchHostName, (size_t)iNameLength) );
}

/******************************************************************************
* Function:
*	struct hostent * osiSocket_GetHostByName (const char *pchHostName)
*
* Inputs:
*
*	
* Return a pointer to a hostent structure on success or NULL on error.
*
*******************************************************************************/
struct hostent * osiSocket_GetHostByName (const char *pchHostName)
{
	return ( gethostbyname (pchHostName) );
}

/******************************************************************************
* Function:
*	unsigned short int osiSocket_htons (unsigned short int HostByte)
*
* Inputs:
*
*	
* Converts a short int from host byte order to network byte order
* The i386 host byte order is LSB first.  The network byte order is MSB first.
*
*******************************************************************************/
unsigned short int osiSocket_htons (unsigned short int HostByte)
{
	return ( htons (HostByte) );
}

/******************************************************************************
* Function:
*	unsigned short int osiLastError
*
* Inputs:
*
*	
* Retrieves last function error.
*
*******************************************************************************/
unsigned long osiLastError (void)
{
	return ( errno );
}
