#include <stdio.h>
#include <string.h>
#include "osilib.h"

/** Globals for osiDelayf *****************************************************/

/* This is the granularity of the delay funtion. */
#define OSI_DELAYF_MIN_DELAY 100

static int delayfInitialized=0;
static bt_mutex listProtector;
static bt_sem stopThreadSem,
    threadIsNowStoppedSem;
static bt_delayf *flistHead=NULL,
    *flistTail=NULL;

/** Functions **************************************************************/

/** List maintenence functions **/
/* These should /not/ be accessible outside delayf.c, so all declared
 * as static.
 */

static int SetTriggerTime(time_x *triggerTime, unsigned long delay)
{
    triggerTime->time += delay / 1000;
    triggerTime->millitm += (unsigned short) delay % 1000;

    /* Check for overflow on millisecond part */
    if (triggerTime->millitm > 999)
    {
	/* Overflow of 1000 milliseconds found; subtract it from millisec
	 * value and increment one second to second value to compensate.
	 */
	triggerTime->millitm -= 1000;
	(triggerTime->time) += 1;
    }

    return(0);
}


static int TimeCompare(const time_x *first, const time_x *second)
{
    /* Compare seconds (time); if they're the same, then compare
     * milliseconds (millitm).
     */
    if (first->time > second->time)
	return (1);

    if (first->time < second->time)
	return (-1);

    /* first->time must be == second->time */
    if (first->millitm > second->millitm)
	return(1);
    if (first->millitm < second->millitm)
	return(-1);

    /* They must be the same */
    return(0);
}


static int ExamineNextInList(bt_delayf *head, const time_x *currTime,
			     bt_delayf **todo)
{
    int result=0;

    if (NULL == head)
    {  /* The list is empty. Buh-bye. */
	return(0);
    }

    result = TimeCompare(currTime, &(head->triggerTime));

    if (result > 0)
    {
	*todo = head;
	return(1);
    }
    else
    {
	*todo = NULL;
	return(0);
    }
}


static int CreateNewItemInList(bt_delayf **head, bt_delayf **tail, bt_delayf **id,
			       void *fn, void *args, time_x *triggerTime)
{
    int compare=0;
    bt_delayf *newList=NULL,
	tmpList,
	*prevList=NULL,
	*currList=NULL;

    /* Initialize this just in case we bomb out */
    *id = NULL;

    /* Create the new list structure item */
    newList = malloc(sizeof(bt_delayf));
    if (NULL == newList)
	return(OSI_DELAYF_MEM_ERR);

    /* Now fill out values in the structures */
    newList->fn = (void (*) (struct bt_delayfStr *, void *))fn;
    newList->args = args;
    memcpy(&(newList->triggerTime), triggerTime, sizeof(time_x));
    newList->next = NULL;
    newList->prev = NULL;

    /* Attach tmp node to head of list */
    tmpList.next = *head;
    tmpList.prev = NULL;
    if (NULL != tmpList.next)
	tmpList.next->prev = &tmpList;

    prevList = &tmpList;
    currList = tmpList.next;
    /* Now find a place in the ordered list for this event */
    for ( ; currList != NULL; prevList = currList, currList = currList->next)
    {
	compare = TimeCompare(&(newList->triggerTime),
			      &(currList->triggerTime));
	if (compare <= 0)
	{  /* New node belongs where curr is or before */
	    break;
	}
    }

    /* We allow more than one item in the list to have the same time, so we
     * don't have to see if these are duplicate events.
     */

    /* Insert the new item into the list */
    prevList->next = newList;
    newList->prev = prevList;
    newList->next = currList;

    /* See if this is the new tail of the list */
    if (NULL != currList)
    {
	currList->prev = newList;
    }
    else
    {
	*tail = newList;
    }

    *id = newList;

    /* Unhook the tmp node */
    *head = tmpList.next;
    (*head)->prev = NULL;

    return(0);
}


static int FindNode(bt_delayf *head, bt_delayf **listItem, bt_delayf *id)
{
    bt_delayf *curr=NULL;

    if (NULL == head)
    {	/* Empty list */
	*listItem = NULL;
	return(-1);
    }

    /* Search through list till we find the id */
    for (curr = head; NULL != curr; curr = curr->next)
    {
	if (curr == id)
	{
	    *listItem = curr;
	    return(0);
	}
    }

    *listItem = NULL;
    return(1);
}


static int DelItemFromList(bt_delayf **head, bt_delayf **tail, bt_delayf **id)
{
    int retVal=0;
    bt_delayf *toDelete=NULL,
	*tmp=NULL;

    retVal = FindNode(*head, &toDelete, *id);

    if ((-1 == retVal) || (1 == retVal))
    {   /* Node not found (list empty, or node does not exist)
	 * Not an error condition; just return.
	 */
	return(OSI_DELAYF_NOT_FOUND);
    }

    /* Double check that it isn't null, though FindNode should have detected
     * this anyway. Counts as "not found".
     */
    if (NULL == toDelete)
	return(OSI_DELAYF_NOT_FOUND);

    /* Is this the head member? */
    if (*head == toDelete)
    {
	//printf("It's the head.\n");
	(*head) = toDelete->next;
	if (NULL != *head)
	    (*head)->prev = NULL;
    }
    else if (NULL == toDelete->next)
    {  /* This is the tail */
	//printf("It's the tail.\n");
	tmp = toDelete->prev;
	tmp->next = NULL;

	/* Update tail to point to last current member */
	(*tail) = tmp;
    }
    else
    {  /* Somewhere in the list */
	//printf("It's in the list, neither head nor tail.\n");
	/* Get the previous node and update it to point to next node */
	tmp = toDelete->prev;
	tmp->next = toDelete->next;

	/* Get the next node and update it to point to previous one */
	tmp = toDelete->next;
	tmp->prev = toDelete->prev;
    }

    /* Free the data and the pointer */
    free(toDelete);
    toDelete = NULL;

    /* NULL out the id so any other reference to it is caught */
    *id = NULL;

    return(0);
}


static void FreeEntireList(bt_delayf **head, bt_delayf **tail)
{
    bt_delayf *tmp=NULL,
	dummy;

    dummy.next = *head;

    tmp = dummy.next;
    /* Go through entire list, freeing anything that may be left. */
    while(NULL != tmp)
    {
	dummy.next = tmp->next;
	free(tmp);
	tmp = dummy.next;
    }

    *head = NULL;
    *tail = NULL;
}


static void *delayf_thr(void *arg)
{
    int doNow=0,
	retVal=0;
    time_x currTime;
    bt_delayf *oneThingToDo=NULL;

    while (1)
    {
	/* See if we should end the thread. 99.97% of the time, we should
	 * not stop the thread, so we can use the timeout value on the
	 * semaphore as a Sleep() / usleep() replacement.
	 */
	retVal = osiSemaphore_Decrement(&stopThreadSem, OSI_DELAYF_MIN_DELAY);
	if ((0 == retVal) || (-2 == retVal))
	{ /* Yes, stop this thread! */
	    break;
	}

	/* Get the current time */
	osiGet_Time(&currTime);

	/* Grab mutex for list */
	if (0 != osiMutex_Acquire(&listProtector, INFINITE))
	{
	    /* This should never fail. We might wait indefinitely, but it
	     * should never return any kind of error.
	     */
	    break;
	}

	/* See if there's anything to do now in the list */
	doNow = ExamineNextInList(flistHead, &currTime, &oneThingToDo);

	while (doNow)
	{
	    /* Run the function specified by oneThingToDo.fn with arg */
	    oneThingToDo->fn(oneThingToDo, oneThingToDo->args);

	    /* Delete the item we just ran */
	    if (0 != DelItemFromList(&flistHead, &flistTail, &oneThingToDo))
	    {  /* A return value other than zero would mean that &oneThingToDo
		* was NULL... this would be extremely strange, if we ran
		* oneThingToDo->fn() above just a moment ago... In practice,
		* this should never happen, so let's get out of here!
		*/
		break;
	    }

	    /* See if there's anything else to do at this moment */
	    doNow = ExamineNextInList(flistHead, &currTime, &oneThingToDo);
	}

	/* Release mutex for list */
	retVal = osiMutex_Release(&listProtector);
	if (0 != retVal)
	{  /* osiMutex_Release should return 0 for success. If not,
	    * something's messed up. So get out!
	    */
	    break;
	}
    }

    /* Notify delayf we're ending, and then end. */
    retVal = osiSemaphore_Increment(&threadIsNowStoppedSem);
    if (0 != retVal)
    {  /* Anything other than 0 for return value is an error. Do nothing,
	* since we're breaking out of here next anyway. 
	*/
	;
    }
    osiThread_End(0);

    /* This shuts up the compiler: */
    return(0);
}

/** User fuctions **/

/******************************************************************************
* Function:
*       int osiDelayf_Init( void )
*
* osiDelayf_Init() must be called only from the main execution thread, and
* may not be called again until osiDelayf_Terminate() has been called. There
* may be only one osiDelayf queue per process.
*
* Inputs: none
* Output: none
*
* returns  0 on success
*******************************************************************************/
int osiDelayf_Init(void)
{
    int retVal=0;
    unsigned long threadID=0;

    /* Has delayf already been initialized? */
    if (delayfInitialized)
	return(OSI_DELAYF_ALREADY_INITIALIZED);

    /* Create mutex to protect delay list */
    if (0 != osiMutex_Init(&listProtector, M_THREAD))
	return(OSI_DELAYF_ERR);

    /* Protect delay list immediatly during setup. */
    if (0 != osiMutex_Acquire(&listProtector, INFINITE))
	return(OSI_DELAYF_ERR);

    /* Init delay list with NULL, as there are no functions to run yet. */
    flistHead = NULL;
    flistTail = NULL;

    /* Create thread stop semaphore */
    if (0 != osiSemaphore_Init(&stopThreadSem, 1, 0))
	return(OSI_DELAYF_ERR);
    if (0 != osiSemaphore_Decrement(&stopThreadSem, 1))
	return(OSI_DELAYF_ERR);
    /* Create thread has stopped notification semaphore */
    if (0 != osiSemaphore_Init(&threadIsNowStoppedSem, 1, 0))
	return(OSI_DELAYF_ERR);
    if (0 != osiSemaphore_Decrement(&threadIsNowStoppedSem, 1))
	return(OSI_DELAYF_ERR);

    /* Create thread */
    if (0 != osiThread_Create(NULL, 0, delayf_thr, NULL, 0, &threadID))
	return(OSI_DELAYF_ERR);

    /* Release relevent mutexes */
    if (0 != osiMutex_Release(&listProtector))
	return(OSI_DELAYF_ERR);

    /* Allow other delayf fuctions to operate since we've successfully
     * started.
     */
    delayfInitialized = ~0;

    return(0);
}

/******************************************************************************
* Function:
*       osiDelayf_Create(void *func,
*                        void *arg,
*                        unsigned long delay,
*                        bt_delayf **id)
*
* Schedules function *func to be run with arguments *arg in delay milliseconds
* from now. Time quantum for execution is 100 milliseconds. If a delay of less
* than 100 milliseconds is specified, the delay will automatically be set to
* 100 milliseconds.
*                        
* Inputs:
*       func             Pointer to a user supplied function
*       arg              Pointer to a user supplied argument
*       delay            Time delay in milliseconds (min 100)
* Output:
*       id               Handle to delayed function
*
* returns 0 for success
*         OSI_DELAYF_NOT_INITIALIZED if osiDelayf_Init has not yet been called
*         OSI_DELAYF_ERR on failure
*
*******************************************************************************/
int osiDelayf_Create(void *func, void *arg, unsigned long delay, bt_delayf **id)
{
    time_x triggerTime;

    if (! delayfInitialized)
	return(OSI_DELAYF_NOT_INITIALIZED);

    /* Calculate time that this event should fire */
    /* Silently guarantee min granularity to the time */
    if (delay < OSI_DELAYF_MIN_DELAY)
	delay = OSI_DELAYF_MIN_DELAY;

    /* Get the current time, and set the trigger time for this time, plus
     * whatever the specified delay was.
     */
    osiGet_Time(&triggerTime);
    SetTriggerTime(&triggerTime, delay);

    /* Grab mutex for list */
    if (0 != osiMutex_Acquire(&listProtector, INFINITE))
	return(OSI_DELAYF_ERR);

    /* Add new delayed function to the list */
    if (0 != CreateNewItemInList(&flistHead, &flistTail, id,
				 func, arg, &triggerTime))
	return(OSI_DELAYF_ERR);

    /* Release mutex for list */
    if (0 != osiMutex_Release(&listProtector))
	return(OSI_DELAYF_ERR);

    return(0);
}


/******************************************************************************
* Function:
*       osiDelayf_Cancel(bt_delayf **id)
*
* Cancels execution of function specified by id, if it hasn't been run yet.
* id is then set to NULL. If the function specified by id is not found, id is
* still set to NULL, and OSI_DELAYF_NOT_FOUND is returned. This is not
* necessarily an error; it is up to the user to decide whether to handle it as
* an error condition or not.
*
* Inputs:
*       id               Handle to delayed function
* Output:
*       id               Reset to NULL during osiDelayf_Cancel.
*
* returns 0 for success
*         OSI_DELAYF_NOT_INITIALIZED if osiDelayf_Init has not yet been called
*         OSI_DELAYF_ERR on failure
*         OSI_DELAYF_NOT_FOUND if the function specified by id can't be found.
*                              Not necessarily an error.
*
*******************************************************************************/
/* Input: pointer to handle of delayed function
 * Output: pointer to handle of delayed function is set to NULL.
 * Return value: 0 for success
 */
int osiDelayf_Cancel(bt_delayf **id)
{
    int retVal=0;

    /* Make sure delayf has been initialized */
    if (! delayfInitialized)
	return(OSI_DELAYF_NOT_INITIALIZED);

    /* Grab mutex for list */
    if (0 != osiMutex_Acquire(&listProtector, INFINITE))
	return(OSI_DELAYF_ERR);

    /* Find item in list. If found, remove it, otherwise just leave. */
    retVal = DelItemFromList(&flistHead, &flistTail, id);
    /* retVal is checked after listProtector is released! */

    /* Release mutex for list */
    if (0 != osiMutex_Release(&listProtector))
	return(OSI_DELAYF_ERR);

    /* This check for retVal is from DelItemFromList() above. */
    switch (retVal)
    {
      case 0: /* Success! */
	return(0);
	break;
      case 1: /* Not found -- not really an error, but nice to report back. */
	return(OSI_DELAYF_NOT_FOUND);
	break;
      default:
	return(OSI_DELAYF_ERR);
    }

    //return(0);
}


/******************************************************************************
* Function:
*       osiDelayf_Terminate(void)
*
* Stops osiDelayf execution, and frees all resources associated with osiDelayf
* system. Any scheduled tasks to be run that haven't been run yet will be
* canceled.
*                        
* Inputs: none
*
* Output: none
*
* returns 0 for success
*         OSI_DELAYF_NOT_INITIALIZED if osiDelayf_Init has not yet been called
*         OSI_DELAYF_ERR on failure
*
*******************************************************************************/
int osiDelayf_Terminate(void)
{
    int retVal=0;

    /* Make sure delayf has been initialized */
    if (! delayfInitialized)
	return(OSI_DELAYF_NOT_INITIALIZED);

    /* This must happen early on, to prevent other threads from getting to
     * list mutex during delayf shutdown.
     */
    delayfInitialized = 0;

    /* Now tell thread to stop */
    if (0 != osiSemaphore_Increment(&stopThreadSem))
	return(OSI_DELAYF_ERR);

    /* Wait up to 500 milliseconds for thread to stop */
    retVal = osiSemaphore_Decrement(&threadIsNowStoppedSem, 500);
    if (retVal == -2)
	return(OSI_DELAYF_ERR);

    /* Free thread stop semaphore */
    if (0 != osiSemaphore_Destroy(&stopThreadSem))
	return(OSI_DELAYF_ERR);

    /* Free thread has stopped notice semaphore */
    if (0 != osiSemaphore_Destroy(&threadIsNowStoppedSem))
	return(OSI_DELAYF_ERR);

    /* Grab mutex for list before destroying anything in list */
    if (0 != osiMutex_Acquire(&listProtector, INFINITE))
	return(OSI_DELAYF_ERR);

    /* Free all items in list */
    FreeEntireList(&flistHead, &flistTail);

    /* Release mutex for list */
    if (0 != osiMutex_Release(&listProtector))
	return(OSI_DELAYF_ERR);

    /* Free list mutex */
    if (0 != osiMutex_Destroy(&listProtector))
	return(OSI_DELAYF_ERR);

    return(0);
}
