/* @(#)parse_did.c	1.14	02 May 1995
**
** Copyright notice:  Copyright 1995, The Bristol Group, Ltd.
** The source of this program is provided only to IsoFax customers who
** need to modify it for non-standard DID hardware.  No copies beyond
** normal backups may be made, and further distribution of listings or
** copies is strictly prohibited.  Use of the binaries and/or derivative
** binaries is covered by the Bristol Softare License Agreement.  Use of
** DID for fax routing requires a specific IsoFax extra charge license.
**
** If you are not absolutely certain about a permitted usage, contact the main
** Bristol Group office for clarification.
*******************************************************************************/
/*
** parse_did - parse did information
**
** SYNOPSIS
**	#include "parse_did.h"
**
**	int parse_did_mess(int did_number,tty_buf_t *ptbuf);
**
** DESCRIPTION
**	parse_did_mess() parses information generated by DID hardware,
**	and creates DID status files.
**
** NOTES
**	Using other hardware than the "Nicollet" for DID fax routing:
**
**	It is assumed that these other devices are connected via RS232 and
**	produce some kind of report at the beginning of each incoming call.
**	If the device cannot be configured for 9600 baud, then modify the
**	#define below.  If the baud rate is not 4800 or 9600, contact Bristol.
**
**	Most, if not all of the modification will be to the function
**	parse_did_mess(). Please look at this function for details.
**
** RETURN VALUES
**	parse_did_mess() returns 1 on success, 0 otherwise.
**
** SEE ALSO
**	parse_did.h, map_trunk_to_tty(), write_num_to_status_file()
**	(the functions are documented below).
*******************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#if defined(SUNOS)
	int printf(char *, ...);
	int fprintf(FILE *,char *, ...);
	int sscanf(char *,char *, ...);
	void *memmove(void *,void *,unsigned int);
#endif

#include "parse_did.h"

extern int Trace_on;
void trprintf(char *format, ...);

int parse_did_mess(int did_num,tty_buf_t *ptbuf)
{
	int trunk = 0;
	int called_number = 0;
	int card_number;
	int jack_number;
	int i;
	char ttyname[PARSE_DID_BUFSIZ];
	char tmp[PARSE_DID_BUFSIZ];
	char *pc;
	char *pend;
	char *pstart;
	char *tmppc;

/* The following parses "Ascom" DID records;  modify it as needed for other
 * types of equipment.  The important thing is to get the trunk number right.
 *
 * A typical Ascom report would be:  1 1 5555F (trunk 1, extension 5555)
 * there maybe an option to remove spaces in the future so support 1155F
 * the extension number maybe 2 to 6 digits
 */
	if(Trace_on) {
		memcpy(tmp,ptbuf->buf,ptbuf->count);
		*(tmp+ptbuf->count) = (char)0;
		trprintf("parse_did: DID box %d: [%s]\n", did_num,tmp);
	};

	pend = ptbuf->buf + ptbuf->count-1;
	pstart = ptbuf->buf;
	
	for(;;) {
		/*
		 * step over any CR and LF characters
		 */
		while(pstart <= pend && *pstart < ' ') {
			++pstart;
		}

		pc = pstart;

		/*
		 * find the next record terminator..
		 */
		while(pc <= pend && *pc != 'F') {
			++pc;
		}
		if(pc > pend) break;

		/*
		 * off-hook messages are at least 5 characters long (including
		 * the terminal 'F'..  if we got less 
		 * something's garbled.
		 */
		i = pc - pstart + 1;
		if( i < 5) {
			if(Trace_on) {
				memcpy(tmp,pstart,i);
				*(tmp+i) = (char)0;
				trprintf("parse_did: unexpected input [%s]\n",
					tmp);
			}
			pstart += i;
			continue;
		}

		/*
		 * the first character should be the router number
		 * (1 through E).
		 */
		pc = pstart;
		if(isdigit(*pc)) {
			card_number = *pc - '0';
			if (card_number == 0) { 
				trprintf("parse_did: unexpected card number [%c]\n",*pc);
				pstart += i;
				continue;
			}
		}
		else {
			switch(*pc) {
			case 'A':
				card_number = 10;
				break;
			case 'B':
				card_number = 11;
				break;
			case 'C':
				card_number = 12;
				break;
			case 'D':
				card_number = 13;
				break;
			case 'E':
				card_number = 14;
				break;
			default:
			trprintf("parse_did: unexpected card number [%c]\n",
				*pc);
				pstart += i;
				continue;
			}
		}
		pc++;
		if(*pc == ' ') pc++;

		/*
		 * the second character should be the jack number on the
		 * card (1 or 2);
		 */
		jack_number = *(pc) - '0';
		if(jack_number != 1 && jack_number != 2) {
			trprintf("parse_did: unexpected jack number [%c]\n",
				*(pc));
			pstart += i;
			continue;
		}
 
                pc++; 
 		if(sscanf(pc,"%d",&called_number) != 1) {
			if(Trace_on) {
				tmppc=tmp;
				while(pc <= pend && *pc != 'F') {
					*tmppc=*pc;
					pc++;
					tmppc++;
				}
				*tmppc = (char)0;
			trprintf("parse_did: unexpected extension [%s]\n",tmp);
			}
			pstart += i;
			continue;
		}
		pstart += i;



		/* 
		 * Check the did_num matches the router number
		 */
		if(did_num!=card_number) {
                      if(Trace_on) {
                        trprintf("parse_did: router#=%d is on did line#=%d ignored report\n",card_number,did_num);
                        }
                        continue;
                }

		/*
		 * Ascom supports up to 2 trunks per DID box
		 */
		trunk = (did_num - 1) * 2 + jack_number;

		/*
		 * we've got a complete record..
		 * generate a status file..
		 */
		trprintf("parse_did: trunk#=%d called#=%d\n",trunk,
			called_number);

		/*
		 * Find the trunk in MODEMS, and return its device name.
		 */
		if(map_trunk_to_tty(trunk,ttyname) <= 0) {
		trprintf("parse_did: couldn't find tty for trunk #%d\n",
				trunk);
			continue;
		}

		trprintf("faxdid: found tty '%s' for trunk #%d\n",ttyname,
			trunk);

		if(write_num_to_status_file(ttyname,called_number) < 0) {
			trprintf("parse_did: can't write to status file\n");
		}
	}

	/*
	 * ok.. shift the stuff that we've parsed..
	 */
	if(pstart != ptbuf->buf) {
		ptbuf->count -= (pstart - ptbuf->buf);
		memmove(ptbuf->buf,pstart,ptbuf->count);
	}
	return 1;
}

/*
** open_tty - open tty from did
**
** SYNOPSIS
**	int open_tty(char *ttyname);
**
** RETURN VALUES
**	oprn_tty() returns a file descriptor on success, -1 otherwise.
*******************************************************************************/

int OPEN_TTY_9600_BAUD(char *);
int OPEN_TTY_4800_BAUD(char *);

/*
 * use this baud rate (4800 is currently the only other option).
 *	#define BAUD_RATE 4800
 */
#define BAUD_RATE 9600

int open_tty(char *ttyname)
{
	switch (BAUD_RATE) {
	case 4800:
		return OPEN_TTY_4800_BAUD(ttyname);
	case 9600:
		return OPEN_TTY_9600_BAUD(ttyname);
	default:
		return -1;
	}
}

/*
** map_trunk_to_tty - get the tty name that took this call.
**
** SYNOPSIS
**	int map_trunk_to_tty(int trunk,char *ttyname);
**
** DESCRIPTION
**	map_trunk_to_tty() retrieves the tty name of the modem
**	corresponding to trunk line that took this call.  On
**	success, the tty name (minus the /dev/ prefix) is
**	placed in ttyname.
**
** RETURN VALUES
**	map_trunk_to_tty() returns 1 on success, 0 if no match was
**	found, and -1 on failure.
*******************************************************************************/

/*
** write_num_to_status_file - generate status file entry for call
**
** SYNOPSIS
**	int write_num_to_status_file(char *ttyname,int number)
**
** DESCRIPTION
**
** RETURN VALUES
**	write_num_to_status_file() returns 1 on success, -1
**	otherwise.
*******************************************************************************/
