
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>

#include "Pplatform.h"
#include "mlog.h"

#define BUFINC 256

static char SCCS_ID[] = "@(#)|mlog.c|9|01|3|0|0|07/26/99|14:26:42";

int
mlog_get_record(FILE *fp,				/* where to read from  */
				mlog_rec_t *recp,		/* pointer to buffer   */
				int rec_buf_size,		/* how big the buffer is */
				int *field_count)		/* device count        */
{
	static int dirty = 0;
	static int mfields = 0;
	static int bufsize = 0;
	static char *buf = (char *)0;

	int i;
	int off_pos;
	int fld_len;
	mlog_field_t *fldp;
	char *p;
	char *bufp;
	char mask_buf[24];

	if(!fp || !recp || !field_count)
	{
		fputs("mlog_get_record:  not enough arguments.\n", stderr);
		return MLOG_ERROR_BAD_ARG;
	}

	if(!dirty)		/* if dirty, last call had insufficient buffer space */
	{
		if(!buf)	/* first time through */
		{
			bufsize = sizeof(mlog_rec_t)
						+ sizeof(mlog_field_t) * 10
						+ 2;							/* rev + \n */

			if((buf = (char *)malloc(bufsize)) == (char *)0)
			{
				bufsize = 0;
				fprintf(stderr, "mlog_get_record:  Failed to allocate memory, "									"errno=%d.\n", errno);
				return MLOG_ERROR_MEMORY;
			}
		}

		/*
		 * Get the whole line.  Reallocate memory
		 * if necessary.
		 */
		for(i = 0;;)
		{
			p = (char *)0;
			if(!fgets(buf + i, bufsize - i, fp) || (p = strchr(buf + i, '\n')))
				break;

			i += strlen(buf + i);
			if((p = (char *)realloc(buf, bufsize + BUFINC)) == (char *)0)
			{
				fprintf(stderr, "mlog_get_record:  Failed to resize memory, "
								"errno=%d.\n", errno);
				return MLOG_ERROR_MEMORY;
			}
			buf = p;
			bufsize += BUFINC;
		}

		if(p == (char *)0)					/* EOF or error */
		{
			if(feof(fp))
				return MLOG_ERROR_EOF;

			/*
			 * This really is an error:  either there was a race
			 * condition or the end of the log is corrupt.  We
			 * may want to forgo the error message though...
			 */
			fprintf(stderr,
					"mlog_get_record:  1) incomplete record, errno=%d.\n",
					errno);
			return MLOG_ERROR_BAD_REC;
		}
		*p = '\0';
	}

	/* One additional surface check for a complete record */
	if((i = strlen(buf)) < (sizeof(mlog_rec_t) - sizeof(mlog_field_t)))
	{
		fprintf(stderr, "mlog_get_record:  2) incomplete record, errno=%d.\n",
				errno);
		/*
		*/
		fprintf(stderr, "mlog_get_record:  i=%d, sizeof(mlog_rec_t)=%d\n",
						i, sizeof(mlog_rec_t));
		return MLOG_ERROR_BAD_REC;
	}

	dirty = 1;

	/*
	 * Count the devices before filling in the buf.
	 * Use the 'dirty' variable to store the count.
	 */
	switch(*buf)
	{
		case MLOG_REV_1:
		default:
			off_pos = MLOG_REV_LEN + MLOG_TIME_LEN;
			fld_len = MLOG_DEVICE_LEN + MLOG_MASK_LEN;
	}

	if((i - off_pos) % fld_len)			/* shouldn't happen, odd length */
	{
		dirty = 0;						/* we want to read from file again */
		fputs("mlog_get_record:  assert(off_pos+fld_len <= i)\n", stderr);
		fprintf(stderr, "\t[%s]\n", buf);
		return MLOG_ERROR_BAD_REC;
	}
	mfields = (i - off_pos) / fld_len;	/* fields are fixed in size */

	/*
	 * Now we need at least sizeof(mlog_rec_t) of buffer space.
	 * This must follow the 'mfields' assignment.  Otherwise we may
	 * be complaining about memory when we don't even have a valid
	 * record.
	 * Calculation:  there are mfields of sizeof(mlog_field_t)
	 *               and 1 sizeof(mlog_rec_t) which includes 1 field.
	 *               Therefore one header minus the size of field plus
	 *               the number of fields times the size of a field.
	 */
	if((unsigned)rec_buf_size
			< (sizeof(mlog_rec_t)+sizeof(mlog_field_t)*(mfields - 1)))
		return MLOG_ERROR_BUF_TOO_SMALL;

	/*  It's clear sailing from here.  Reset the dirty flag */
	dirty = 0;

	/* Initialize some helper variables */
	p = mask_buf;
	*p++ = '0';
	*p++ = 'x';
	bufp = buf;
	memset(recp, 0, rec_buf_size);

	/* Fill in the mlog_rec_t and other "out" data */
	recp->rev = *bufp++;
	memcpy(p, bufp, MLOG_TIME_LEN);
	bufp += MLOG_TIME_LEN;
	p[MLOG_TIME_LEN] = '\0';
	recp->log_time = strtoul(mask_buf, (char **)0, 16);
	*field_count = mfields;

	/*
	 * Fill in the array of mlog_field_t's.  We have determined at
	 * this point that we have enough memory, so don't bother checking
	 * size restraints.
	 */
	for(fldp = recp->field; mfields; mfields--, fldp++)
	{
		memcpy(fldp->device, bufp, MLOG_DEVICE_LEN);
		bufp += MLOG_DEVICE_LEN;

		/*
		 * mask_buf - includes a leading "0x"
		 * p        - points to the char past mask_buf's "0x"
		 */
		memcpy(p, bufp, MLOG_MASK_LEN);
		bufp += MLOG_MASK_LEN;
		p[MLOG_MASK_LEN] = '\0';
		
		/* force a short */
		fldp->mask = (short)(strtoul(mask_buf, (char **)0, 16) & 0x0000ffff);
	}

	return MLOG_ERROR_OK;
}

int
mlog_put_record(FILE *fp,				/* where to write to   */
				mlog_rec_t *recp,		/* pointer to the buff */
				int field_count)		/* how many devices    */

{
	static int first_time = 1;
	mlog_field_t *p;

	if(!fp || !recp)
		return 0;

	/* prh 2/10/99 Indicate very first record on a new server executiion.
       If rev '1' write 'a' */
	if (first_time)
	{
		fputc(recp->rev - '1' + 'a', fp);
		first_time = 0;
	}
	else
		fputc(recp->rev, fp);

	fprintf(fp, "%08x", recp->log_time);
	p = recp->field;
	while(field_count > 0)
	{
		fwrite(p->device, 1, MLOG_DEVICE_LEN, fp);
		fprintf(fp, "%04x", p->mask);
		++p;
		--field_count;
	}
	fputc('\n', fp);

	return 1;
}




