
#ifndef _INCLUDE_MLOG_H
#define _INCLUDE_MLOG_H

/* SCCS_ID = "@(#)mlog.h	9.01.2.0.0	07/26/99	14:35:55" */

/*
 *  Format of a modem log record:
 *   Rev        - ascii char 0-9
 *   Timestamp  - ascii UNIX timestamp in hexidecimal
 *   DeviceN    - 10 byte ascii string with ascii 0x20(' ') fill
 *   Use/Status - ascii hexidecimal bitmapping of the modem_log_status
 *                and modem_log_usage enumerations below.
 *   EOR        - end of record is delimited by an ascii 0x0a('\n')
 *
 *  A record is variable length and gives a statistical representation
 *  of the modem usage across all currently declared modems.  Modem
 *  slots that are available but not declared will NOT be printed.
 *
 *  record (in bytes):
 *    1       8        10         2      ...   10         2        1
 *  <rev><timestamp><device1><use/status>...<device1><use/status><\n>
 *
 *  use/status mask (in bits): type 15 - 7, status 6 - 0, U - Unused
 *   15                                                     7
 *  | U | U | U |Send/Voice|Voice|DID info|Receive|Send|Send/Receive|
 *     6                                    0
 *     U | U |Voice|Receiving|Sending|Down|Free|
 */

#define MLOG_VAR_SIZE_ARRAY 1
#define MLOG_REV_LEN 1
#define MLOG_DEVICE_LEN 10
#define MLOG_TIME_LEN 8
#define MLOG_MASK_LEN 4

/*
 * The modem_log_status enumeration corresponds to bits
 * 0 -> 6 of the use/status mask.  2 of which are for
 * future expansion.
 */
enum modem_log_status
{
    MLOG_STATUS_FREE        = 0x00000001,
    MLOG_STATUS_DOWN        = 0x00000002,
    MLOG_STATUS_SENDING     = 0x00000004,
    MLOG_STATUS_RECEIVE     = 0x00000008,
    MLOG_STATUS_VOICE       = 0x00000010,
    MLOG_STATUS_UNUSED_1    = 0x00000020,
    MLOG_STATUS_UNUSED_2    = 0x00000040,
};

/*
 * The modem_log_status enumeration corresponds to bits
 * 7 -> 15 of the use/status mask.  3 of which are for
 * future expansion.
 */
enum modem_log_usage
{
    MLOG_USE_SEND_REC       = 0x00000080,
    MLOG_USE_SEND           = 0x00000100,
    MLOG_USE_RECEIVE        = 0x00000200,
    MLOG_USE_DID_INFO       = 0x00000400,
    MLOG_USE_VOICE          = 0x00000800,
    MLOG_USE_SEND_VOICE     = 0x00001000,
    MLOG_USE_UNUSED_1       = 0x00002000,
    MLOG_USE_UNUSED_2       = 0x00004000,
    MLOG_USE_UNUSED_3       = 0x00008000,
};

/*
 * The modem_error_values enumeration describes the return values
 * from mlog_get_record().
 */
enum modem_error_values
{
	MLOG_ERROR_OK,              /* Got record successfully */
	MLOG_ERROR_MEMORY,          /* Could not alloc/realloc memory */
	MLOG_ERROR_BAD_ARG,         /* Not enough data passed as arguments */
	MLOG_ERROR_BAD_REC,         /* Record was determined to be badly formed */
	MLOG_ERROR_BUF_TOO_SMALL,   /* The buffer passes as an arg is to small */
	MLOG_ERROR_EOF              /* No more records */
};

enum mlog_rev
{
    MLOG_REV_1              = '1',  /* Current rev, 6-24-96 */
    MLOG_REV_2              = '2',
    MLOG_REV_3              = '3',
    MLOG_REV_4              = '4',
    MLOG_REV_5              = '5',
    MLOG_REV_6              = '6',
    MLOG_REV_7              = '7',
    MLOG_REV_8              = '8',
    MLOG_REV_9              = '9'
};

/*
 *  The idea here is that you can dynamically (or statically)
 *  allocate a variable size array to store the values in.  Thus,
 *    {
 *      rev
 *      log_time
 *      field[0]
 *      field[1]
 *        ...
 *      field[N]
 *    }
 */
typedef struct 
{
    unsigned short mask;                     /* use/status mask */
    char device[MLOG_DEVICE_LEN];            /* device name, ws filled */
    char unused[16 - sizeof(unsigned short) - MLOG_DEVICE_LEN]; /* align 4 */
} mlog_field_t;

typedef struct 
{
    char rev;                                /* version of record   */
    char unused[3];                          /* align on 4 byte boundary */
    time_t log_time;
    mlog_field_t field[MLOG_VAR_SIZE_ARRAY]; /* must be the last field */
} mlog_rec_t;

/*
 * Macro for allocating a record buffer with space
 * for _n_modems modems.
 */
#define MLOG_BUF_ALLOC(_n_modems) \
	((mlog_rec_t *)malloc(sizeof(mlog_rec_t) + \
	(sizeof(mlog_field_t) * (_n_modems - MLOG_VAR_SIZE_ARRAY))))

/*
 * Macro for re-allocating a record buffer with space
 * for _n_modems modems.
 */
#define MLOG_BUF_REALLOC(_obuf,_n_modems) \
	((mlog_rec_t *)realloc(_obuf, sizeof(mlog_rec_t) + \
	(sizeof(mlog_field_t) * (_n_modems - MLOG_VAR_SIZE_ARRAY))))

int mlog_get_record(FILE *fp,                /* where to read from  */
                    mlog_rec_t *recp,        /* pointer to buffer pointer  */
					int buf_size,            /* how big the buffer is */
                    int *field_count);       /* device count        */

int mlog_put_record(FILE *fp,                /* where to write to   */
                    mlog_rec_t *recp,        /* pointer to the buff */
                    int field_count);        /* how many devices    */


#ifdef __cplusplus
    extern "C" {
#endif

#ifdef __cplusplus
    }
#endif

#endif
